
	NOLIST  ;turn listing off, then turn it back on as needed
;******************************************************************************
; Helper macros for Microchip PIC16Fxxx
; Copyright (c) 2008, 2009, 2010 Don Julien.  djulien@thejuliens.net
; May be freely distributed and used for non-commerical purposes (non-commercial share-alike).  
; See the license agreement terms below.
;
; This software was written and is released in the hope that it will help honor and celebrate 
; the birth of Christ, who is *the* Light of the World.  In recognition and appreciation of His 
; love and sacrifice for us, and by His inspiration and enablement, I offer up this work to Him.
; If you feel inclined to support this project financially, please make a donation to a needy 
; family or to a local (Biblical) church.
; For comments, bugs, or other inquiries, please contact the author at djulien@thejuliens.net.
;
; This file contains "helper macros" to make writing code for the PIC easier (imho).
; Some of these macros define pseudo-opcodes to extend the instruction set of mid-range PICs.
; No code is generated or RAM allocated directly from within this file, so unused macros will
; not waste any processor resources; macros expand within the caller as needed.
; This file grew fairly large, but the individual macros and pseudo-opcodes themselves are fairly simple.
; The .LST file is extremely helpful for checking if these macros are working correctly.
; References to page#s in any of the comments refer to the Microchip PIC16F688 datasheet DS41203C.pdf
; unless otherwise noted.
;
; This code was compiled at tested with MPASM 5.20, which is BUGGY.
; I had to work around the following MPASM bugs:
; 1. Macro LOCALs are sometimes not unique (probably due to something like hash table collisions)
; 2. Pass 1 vs. pass 2 addresses do not resolve to the same values if there is conditional code expansion based on addresses (error 116)
; 3. Macro parameters get trashed in recursive macros (parameters are acting like globals, not local to the macro)
; 4. Assembler can't figure out which parameters to use when multiple in-line macros are used on the same line
; 5. Forward references don't resolve sometimes (they are left with a value of 0)
;
; Revision history:
; version  date    who  why
; 0.3     3/2008    DJ  first version; wrote some of the basic functions
; 0.4     4/2008    DJ  add threads, arithmetic helpers, etc
; 1.0     8/2008    DJ  added misc timing and pin I/O helpers for MRR controller
; 1.0a    2/2009    DJ  add goto/call/return via bank1; misc fixes + additions
; 1.0b    3/23/09   DJ  added "else" statements to ifbit, incdec, and if# macros
; 1.0c    4/7/09    DJ  finally solved the error 116 problem! backed out of Placeholder and EntryPoint workarounds
; 1.0d    4/27/09   DJ  looks like the assembler gives up on forward addresses occassionally, so we still need the PlaceHolder macros :(
; 1.0e    11/09/09  DJ  added hints to arith and conditional macros for a little better optimization; misc tweaks
; 1.0f    2/15/10   DJ  misc clean up; tighten up bank affinity on calls/gotos; reduce IF nesting level (hit the limit); allow non-nested placeholders; added definitions for 16F689/690
;         4/17/10   DJ  finish off readability clean-up in preparation for first source code publication
;
; SOFTWARE LICENSE AGREEMENT:
; This work is licensed under the Creative Commons Attribution-Noncommercial-Share Alike 3.0 
; Unported License.  To view a copy of this license, visit 
; http://creativecommons.org/licenses/by-nc-sa/3.0/ or send a letter to 
; Creative Commons, 171 Second Street, Suite 300, San Francisco, California, 94105, USA.
;
; Basically, this software is supplied to you solely for your own personal, non-commerical use,
; and I ask that if you make changes, adaptations, fixes, etc that you freely share those.
; THIS SOFTWARE IS PROVIDED "AS IS".  NO EXPRESS, IMPLIED, STATUTORY OR OTHER WARRANTIES 
; OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.  AUTHOR 
; SHALL NOT BE LIABLE FOR SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, FOR ANY REASON OR
; UNDER ANY CIRCUMSTANCES WHATSOEVER.  THIS SOFTWARE MUST NOT BE USED FOR CRITICAL APPLICATIONS
; SUCH AS HEAVY MACHINERY, LIFE SUPPORT, ETC.  YOU USE THIS SOFTWARE AT YOUR OWN RISK.
; But if you have questions or problems with it, let me know and I'll try to help if I can.
;******************************************************************************

	subtitle  "Fun With Macros - PIC16Fxxx helper macros and pseudo-opcodes"
	errorlevel -302  ;this is a useless/annoying message because the assembler doesn't handle it well (always generates warning when accessing registers in bank 1, even if you've set the bank select bits correctly)
	errorlevel -306  ;this is a useless/annoying message because the assembler doesn't handle it well (always generates warning when accessing page 1, even if you've set the page select bits correctly)
	NOEXPAND  ;turn macro expansion off, then turn it back on selectively as needed


	CONSTANT TRUE = -1
	CONSTANT FALSE = 0
	CONSTANT MAYBE = 2  ;not FALSE and not TRUE
	CONSTANT DONT_CARE = FALSE  ;arbitrary, but it's safer to turn off a don't-care feature than to leave it on

	CONSTANT COMMA = 0x2c ;A',' ;commas cause assembler syntax error, so use a symbol to wrap it

#define CONTINUE  ;dummy placeholder (for readability only)
;#define THEN  ;dummy keyword for IF readability
;#define ELSE  ;dummy keyword for IF readability

;these constants require more than 16 bits:
ONE_SECOND set d'1000000'  ;#usec/sec
ONE_MSEC set d'1000'  ;#usec/msec
_1SEC set ONE_SECOND  ;less verbose
_1MSEC set ONE_MSEC  ;less verbose

	CONSTANT PIC16Fxxx_version = 0x10f  ;bump this whenever functionality in this file changes!


;******************************************************************************
;macros to control expansion of other macros:
;this allows some clutter to be removed from the .LST file
;******************************************************************************

	VARIABLE MEXPAND_STACKLO = 1, MEXPAND_STACKHI = 0  ;default is ON (for caller, set at eof in this file)
	VARIABLE MEXPAND_MAXNEST = 0, MEXPAND_CURNEST = 0

;push current macro expand directive, then set new one (max 31 nested levels):
;params:
; onoff = whether to turn macro expansion on or off
EXPAND_PUSH MACRO onoff
	NOEXPAND  ;first turn it off to hide clutter in here
MEXPAND_CURNEST += 1  ;keep track of current nesting level
MEXPAND_STACKLO *= 2  ;push current value (shift left)
	if MEXPAND_CURNEST > MEXPAND_MAXNEST
MEXPAND_MAXNEST = MEXPAND_CURNEST  ;keep track of high-water mark
	endif
;use this if more than 32 levels needed:
;MEXPAND_STACKHI *= 2
;	if MEXPAND_STACKLO & ASM_MSB
;MEXPAND_STACKHI += 1
;MEXPAND_STACKLO &= ~ASM_MSB
;	endif
	if !(onoff)
		EXITM
	endif
MEXPAND_STACKLO += 1
																				EXPAND  ;turn listing on
	ENDM


;restore previous macro expand directive:
;params:
; dummy = dummy parameter (needed in order to force MPASM to allow optional params for this macro)
; expected_nesting = expected nesting level after pop (optional); must have a trailing "&" if passed
EXPAND_POP MACRO dummy, nesting
	NOEXPAND  ;first turn it off to hide clutter in here
	LOCAL EXP_NEST = nesting -1  ;optional param; defaults to -1 if not passed
MEXPAND_CURNEST -= 1  ;keep track of current nesting level
MEXPAND_STACKLO /= 2  ;pop previous value (shift right)
;only needed if reach 16 levels:
;	if MEXPAND_STACKLO & ASM_MSB  ;< 0
;MEXPAND_STACKLO &= ~ASM_MSB  ;1-MEXPAND_STACKLO  ;make correction for assembler sign-extend
;	endif
;use this if more than 32 levels needed:
;	if MEXPAND_STACKHI & 1
;MEXPAND_STACKLO += ASM_MSB
;	endif
;MEXPAND_STACKHI /= 2
;errif does this:
;	if !(MEXPAND_STACKLO & 1)  ;pop, leave off
;		EXITM
;	endif
;																				EXPAND  ;listing back on again
	ERRIF (EXP_NEST != -1) && (MEXPAND_CURNEST != EXP_NEST), [ERROR] Mismatched macro expanders: nesting is #v(MEXPAND_CURNEST) but should be #v(EXP_NEST)
	ENDM


;repeat a statement the specified number of times:
;stmt can refer to "repeater" for iteration-specific behavior
;stmt cannot use more than 1 parameter (MPASM gets confused by commas; doesn't know which macro gets the params)
;params:
; count = #times to repeat stmt
; stmt = statement to be repeated
REPEAT MACRO count, stmt
	NOEXPAND  ;hide clutter
	EXPAND_PUSH FALSE
	LOCAL repeater = 0
	while repeater < (count)  ;0, 1, ..., count-1
																				EXPAND  ;show generated code
	stmt
		NOEXPAND  ;hide clutter
repeater += 1
		if repeater > d'1000'  ;paranoid; prevent run-away code expansion
			ERRIF TRUE, [ERROR] Code generator loop too big: #v(repeater) of #v(count)
			EXITM
		endif
	endw
	EXPAND_POP,
	ENDM


;******************************************************************************
;error/debug assertion message macros:
;******************************************************************************

;show error message if condition is true:
;params:
; assert = condition that must (not) be true
; message = message to display if condition is true (values can be embedded using #v)
ERRIF MACRO assert, message
	NOEXPAND  ;hide clutter
	if assert
		error message
	endif
	if !(MEXPAND_STACKLO & 1)
		EXITM
	endif
																				EXPAND  ;listing back on again
	ENDM


;alternate version to allow commas to be placed within the message:
ERRIF2 MACRO assert, msg1, msg2
	NOEXPAND  ;hide clutter
	if assert
		error msg1, msg2
	endif
	if !(MEXPAND_STACKLO & 1)
		EXITM
	endif
																				EXPAND  ;listing back on again
	ENDM


;show warning message if condition is true:
;params:
; assert = condition that must (not) be true
; message = message to display if condition is true (values can be embedded using #v)
WARNIF MACRO assert, message
	NOEXPAND  ;hide clutter
	if assert
		messg message
	endif
	if !(MEXPAND_STACKLO & 1)
		EXITM
	endif
																				EXPAND  ;listing back on again
	ENDM


;alternate version to allow commas to be placed within the message:
WARNIF2 MACRO assert, msg1, msg2
	NOEXPAND  ;hide clutter
	if assert
		messg msg1, msg2
	endif
	if !(MEXPAND_STACKLO & 1)
		EXITM
	endif
																				EXPAND  ;listing back on again
	ENDM


;******************************************************************************
;misc. assembler-related helper macros:
;******************************************************************************

	VARIABLE NEXT_UNIQID = 0  ;unique value to help assember generate unique symbols (sometimes it generates duplicates, which leads to error 116)

ASM_MSB set 0x80000000  ;assembler uses 32-bit values
	ERRIF (!ASM_MSB) || (2*ASM_MSB), [ERROR] ASM_MSB is not set correctly: msb == #v(ASM_MSB) and 2*msb == #v(2*ASM_MSB)  ;paranoid self-check

#define ASM_NOP  VARIABLE adummy = 0  ;KLUDGE; assembler needs something, so give it a benign statement


;ternary operator (like C/C++ "?:" operator):
#define IIF(TF, tval, fval)  (((TF)!=0)*(tval) + ((TF)==0)*(fval))
	ERRIF ((3 == 3) != 1) || ((3 != 3) != 0), "[ERROR] IIF is broken:" true == #v(3 == 3) and false == #v(3 != 3)  ;paranoid self-check


;misc arithmetic helpers:
#define rdiv(num, den)  (((num)+(den)/2)/(den))  ;rounded divide (at compile-time)
#define divup(num, den)  (((num)+(den)-1)/(den))  ;round-up divide
#define err_rate(ideal, actual)  ((d'100'*(ideal)-d'100'*(actual))/(ideal))  ;%error
#define mhz(freq)  #v((freq)/ONE_SECOND)MHz  ;used for text messages
#define kbaud(freq)  #v((freq)/1000)kb  ;used for text messages
#define sgn(x)  IIF((x) < 0, -1, (x) != 0)  ;-1/0/+1
#define abs(x)  IIF((x) < 0, -(x), x)  ;absolute value
#define MIN(x, y)  IIF((x) < (y), x, y)  ;use upper case so it won't match text within ERRIF/WARNIF messages
#define MAX(x, y)  IIF((x) > (y), x, y)  ;use upper case so it won't match text within ERRIF/WARNIF messages
	ERRIF abs(-d'42') != d'42', "[ERROR] abs macro is broken: abs(-42)" == #v(abs(-d'42'))  ;paranoid self-check
	ERRIF sgn(-d'42') != -1, "[ERROR] sgn macro is broken: sgn(-42)" == #v(sgn(-d'42'))  ;paranoid self-check
	ERRIF MIN(-4, -3) != -4, "[ERROR] MIN macro is broken: min(-4, -3)" == #v(MIN(-4, -3))  ;paranoid self-check
	ERRIF MAX(4, 3) != 4, "[ERROR] MAX macro is broken: max(4, 3)" == #v(MAX(4, 3))  ;paranoid self-check


;turn a tri-state boolean value into an int:
#define BOOL2INT(bool)  abs(bool)
	CONSTANT b2i_FALSE = BOOL2INT(FALSE)  ;line too long
	CONSTANT b2i_TRUE = BOOL2INT(TRUE)  ;line too long
	CONSTANT b2i_MAYBE = BOOL2INT(MAYBE)  ;line too long
	ERRIF b2i_FALSE == b2i_TRUE, "[ERROR] BOOL2INT broken: FALSE" == #v(b2i_FALSE) and TRUE == #v(b2i_TRUE)  ;paranoid self-check
	ERRIF b2i_FALSE == b2i_MAYBE, "[ERROR] BOOL2INT broken: FALSE" == #v(b2i_FALSE) and MAYBE == #v(b2i_MAYBE)  ;paranoid self-check
	ERRIF b2i_TRUE == b2i_MAYBE, "[ERROR] BOOL2INT broken: TRUE" == #v(b2i_TRUE) and MAYBE == #v(b2i_MAYBE)  ;paranoid self-check
	ERRIF BOOL2INT(TRUE) != 1, [ERROR] BOOL2INT(TRUE) broken: TRUE == #v(BOOL2INT(TRUE))  ;paranoid self-check
	ERRIF BOOL2INT(FALSE) != 0, [ERROR] BOOL2INT(FALSE) broken: FALSE == #v(BOOL2INT(FALSE))  ;paranoid self-check


;******************************************************************************
;processor-specific definitions:
;I think it's better to add more definitions in here in order to support additional PIC16Fxxx family members,
;rather than using conditional logic, definitions or compiles elsewhere in the code
;******************************************************************************

;instruction timing:
;NOTE: these require 32-bit arithmetic
#define InstrPerSec(clock)  ((clock)/INSTR_CYCLES)  ;#instructions/sec at specified clock speed
#define uSec2Instr(usec, clock)  rdiv((usec)*InstrPerSec(clock), ONE_SECOND)  ;#instructions that can be executed in specified time at clock speed
#define Instr2uSec(numinst, clock)  rdiv((numinst)*ONE_SECOND+1, InstrPerSec(clock)+1)  ;time to execute specified #instructions at clock speed; KLUDGE: "+1" avoids divide-by-0 errors but doesn't signficantly affect the result
#define LoopIter(usec, each)  rdiv(usec, each)  ;#loop iterations within specified time


;partial opcodes:
;use named consts so they show in LST file
	CONSTANT ADDWFdf = 0x700
	CONSTANT SUBWFdf = 0x200
	CONSTANT XORWFdf = 0x600
	CONSTANT BSFbf = 0x1400
	CONSTANT BCFbf = 0x1000
	CONSTANT BTFSCbf = 0x1800
	CONSTANT BTFSSbf = 0x1C00
	CONSTANT CALLk = 0x2000
	CONSTANT GOTOk = 0x2800
	CONSTANT RETSUB = 0x0008
	CONSTANT ADDLWk = 0x3E00


;comma-less instruction wrappers:
;these are useful for instructions used as parameters to macros, since MPASM can't match up the parameters correctly
;#define ADDLIT(reg, val)  da 0x3000+ (val)%0x100, 0x780+ (reg)%0x80
;#define ADDREG(reg, val)  da 0x800+ (val)%0x80, 0x780+ (reg)%0x80
#define ADDWF_wrapper(reg, dest)  OPCODE ADDWFdf+ ((dest)!=0)*0x80+ ((reg)%0x80)
#define SUBWF_wrapper(reg, dest)  OPCODE SUBWFdf+ ((dest)!=0)*0x80+ ((reg)%0x80)
;#define XORWF_wrapper(reg, dest)  OPCODE XORWFdf+ ((dest)!=0)*0x80+ ((reg)%0x80)
;#define SETBIT(reg, bit, val)  OPCODE IIF(val, 0x1400, 0x1000)+ (bit)*0x80+ ((reg)%0x80)  ;line too long
#define SETBIT(reg, bit, val)  OPCODE BCFbf+ ((val)!=0)*0x400+ (bit)*0x80+ ((reg)%0x80)  ;bcf/bsf
#define TESTBIT(reg, bit, val)  OPCODE BTFSCbf+ ((val)!=0)*0x400+ (bit)*0x80+ ((reg)%0x80)  ;btfsc/btfss
#define JUMPRET(adrs, ret)  OPCODE GOTOk- ((ret)!=0)*0x800+ ((adrs)%0x800)  ;call/goto
#define RETURNIE(ei)  OPCODE RETSUB+ ((ei)!=0)  ;return/retfie


;wrapper to avoid optional comma handling problems for DA when used as nested macro:
;this also allows instructions to be generated dynamically from parts (ie, a conditional expression can use "btfsc" vs. "btfss", etc)
;params:
;  opc = opcode (can be generated dynamically)
OPCODE MACRO opc
	NOEXPAND  ;hide clutter in LST file
#if 0  ;readability
;	LOCAL OPC = opc
;	EXPAND_PUSH FALSE
;	if (OPC & 0xfc00) == BCFbf
#endif  ;0
																				EXPAND  ;show generated code
	da opc  ;instruction generated from comma-less stmt
	NOEXPAND
;	EXPAND_POP,
	ENDM


;set some generic definitions to control addressing, etc:
;values that are approprirate for most of the target PICs are set here, then overridden later as needed
#define GENERAL_RAM_START  0x20  ;general-purpose RAM starts at 0x20 on PIC16F688
#define NONBANKED_RAM_START  0x70  ;RAM above this address can be accessed via any bank (page 124 + 10)
#define BANK_SIZE  0x80  ;RAM at this address or above is banked and needs bank-select bits (7 bit register address)
#define MAX_RAM  0x1ff  ;general-purpose RAM ends at 0x1ff on PIC16F688, but a note of page 13 says that RP1 + IRP are not used, which would mean that only 0..0xff is addressable?
#define UPPER_BANK_START  0x120  ;start of general-purpose RAM in highest bank
#define UPPER_BANK_SIZE  d'80'  ;length of general-purpose RAM in highest bank
#define RESET_VECTOR  0  ;processor reset vector address
#define INTERRUPT_VECTOR  4  ;processor interrupt vector address
#define MAX_ROM  0x0fff  ;0x1fff  ;page 9 says PC is 13 bits and program memory ends at 0x1fff, but there's only 4K words, so it actually ends at 0x0fff and PCLATH<4> is useless
#define EE_START  0x2100  ;EEPROM starts at 0x2100 on PIC16F688
;for immediate (static) addresses, 11 bits come from the operand, and 2 bits from PCLATH
;for indirect (dynamic) addresses, 8 bits come from the register operand and 5 from PCLATH
;hence, the effective page size for immediate addressing is 0x800, but for indirect/register addressing is 0x100
;see page 21 for a description of PCL/PCLATH and immediate addresses
#define REG_PAGE_SIZE  0x100  ;code at this address or above is paged and needs page select bits (8 bit address)
#define LIT_PAGE_SIZE  0x800  ;code at this address or above is paged and needs page select bits (11 bit address)
#define INSTR_CYCLES  4  ;each instruction takes 4 clock cycles
#define FIRST_PORT  0xA  ;lowest possible port# (using hex in lieu of alpha)
#define LAST_PORT  0xF  ;highest possible port# (using hex in lieu of alpha)
#define SELF_WRITE  FALSE  ;allows self-write?
#define SERIN_PIN  0xC5  ;serial input pin, RC5
#define SEROUT_PIN  0xC4  ;serial output pin, RC4
#define CLKIN_PIN  0xA5  ;ext clock pin, RA5
#define RESET_PIN  0xA3  ;this pin can only be used for input, RA3
#define MAX_SFR  0xA0  ;max address of special function registers


;for compatibility with other processors:
#ifdef GPIO  ;12F6xx
 #define PORTA  GPIO
 #define TRISA  TRISIO
 #define CMCON0  CMCON
 #define _FCMEN_OFF  -1  ;undefined?
 #define _BOR_OFF  _BODEN_OFF  ;missing symbol in p16f688.inc
 #define _IESO_OFF  -1  ;undefined?
#endif  ;def GPIO

#ifndef NOT_RAPU
 #define NOT_RAPU  NOT_RABPU
#endif  ;NOT_RAPU

#ifndef RAIE
 #define RAIE  RABIE
#endif  ;RAIE

#ifndef RAIF
 #define RAIF  BABIF
#endif  ;RAIF

#ifndef SENDB
 #define SENDB  SENB
#endif  ;SENDB

#ifndef SBODEN
 #define SBODEN  SBOREN
#endif  ;SBODEN

;fix up inconsistent naming:
#ifndef _BOR_OFF
 #define _BOR_OFF  _BOD_OFF
#endif  ;ndef _BOR_OFF

#ifndef NOT_BOR
 #define NOT_BOR  NOT_BOD
#endif  ;ndef NOT_BOR


;set processor-specific definitions:
;use Configure -> Select Device in MPLAB to set the processor type, then the code below will use the correct .INC for the selected processor
;only need to set processor-specific values here; generic 16Fxxx defaults were set above
#ifdef __16F688  ;this is the "standard" PIC used in Renard controllers
	INCLUDE "p16f688.inc"  ;be sure to fix syntax error in this include file before compiling (with older versions of MPLAB)
	NOLIST  ;reduce clutter
;	messg "Compiled for 16F688: 8 I/O pins gives 8x7 = 56 SSR channels (or 112 when SSR-doubled)"
	CONSTANT PORT#v(0xA)_BITS = b'00111111'  ;16F688 has 6-bit ports; change as needed for port sizes on other processors
	CONSTANT PORT#v(0xC)_BITS = b'00111111'  ;16F688 has 6-bit ports; change as needed for port sizes on other processors
 #undefine LAST_PORT
 #define LAST_PORT  0xC
 #define FSRH  0x190  ;use an unused register location as alias for STATUS.IRP
#endif  ;__16F688

#ifdef __16F675  ;this one can't support Renard (no serial I/O), but can be used as a small, stand-alone controller
	INCLUDE "p16f675.inc"
	NOLIST  ;reduce clutter
;	messg "Compiled for 16F675: 4 I/O pins gives 4x3 = 12 SSR channels (or 24 when SSR-doubled)"
	CONSTANT PORT#v(0xA)_BITS = b'00111111'  ;16F675 has one 6-bit port; change as needed for port sizes on other processors
 #undefine MAX_RAM  ;override
 #define MAX_RAM  0xdf  ;general-purpose RAM ends at 0xff on PIC16F675, but only 0..0xdf is addressable
 #undefine MAX_ROM  ;override
 #define MAX_ROM  0x07ff  ;PIC16F675 has 1K words (2K bytes); TODO: CHECK THIS
 #undefine LAST_PORT
 #define LAST_PORT  FIRST_PORT
	messg "NOT TESTED!"
#endif  ;__16F675

#ifdef __16F87
	INCLUDE "p16f87.inc"
	NOLIST  ;reduce clutter
;	messg "compiled for 16F87: 12 I/O pins gives 12x11 = 132 SSR channels (or 164 when SSR-doubled); max 57% on at a time"
 #undefine UPPER_BANK_START
 #define UPPER_BANK_START  0x190  ;start of general-purpose RAM in highest bank
 #undefine UPPER_BANK_SIZE
 #define UPPER_BANK_SIZE  d'96'  ;length of general-purpose RAM in highest bank
	CONSTANT PORT#v(0xA)_BITS = b'11111111'  ;PORTA is 8 bits
	CONSTANT PORT#v(0xB)_BITS = b'11111111'  ;PORTB is 8 bits
 #undefine LAST_PORT
 #define LAST_PORT  0xB
 #define FSRH  0x190  ;use an unused register location as alias for STATUS.IRP
	messg "NOT TESTED!"
#endif  ;__16F87

#ifdef __16F726
	INCLUDE "p16f726.inc"
	NOLIST  ;reduce clutter
;	messg "compiled for 16F726: 22 I/O pins gives 22x21 = 462 SSR channels (924 when SSR-doubled); max 16% on at a time"
 #undefine MAX_ROM
 #define MAX_ROM  0x1fff  ;8K words, so it actually ends at 0x1fff and PCLATH<4> is used
 #undefine UPPER_BANK_START
 #define UPPER_BANK_START  0x190  ;start of general-purpose RAM in highest bank
 #undefine UPPER_BANK_SIZE
 #define UPPER_BANK_SIZE  d'96'  ;length of general-purpose RAM in highest bank
	CONSTANT PORT#v(0xA)_BITS = b'11111111'  ;PORTA is 8 bits
	CONSTANT PORT#v(0xB)_BITS = b'11111111'  ;PORTB is 8 bits
	CONSTANT PORT#v(0xC)_BITS = b'11111111'  ;PORTC is 8 bits
;	CONSTANT PORT#v(0xE)_BITS = b'00001000'  ;PORTE is 1 bit; why bother?
 #undefine LAST_PORT
 #define LAST_PORT  0xC
 #define FSRH  0x190  ;use an unused register location as alias for STATUS.IRP
	messg "NOT TESTED!"
#endif  ;__16F726

#ifdef __16F689 
 #define __16F689or690
	INCLUDE "p16f689.inc"
	NOLIST  ;reduce clutter
	CONSTANT PORT#v(0xA)_BITS = b'00111111'  ;PORTA is 6 bits
	CONSTANT PORT#v(0xB)_BITS = b'11110000'  ;PORTB is 4 bits, but they are left-justified!
	CONSTANT PORT#v(0xC)_BITS = b'11111111'  ;PORTC is 8 bits
 #undefine SERIN_PIN
 #define SERIN_PIN  0xB5  ;serial input pin, RB5
 #undefine SEROUT_PIN
 #define SEROUT_PIN  0xB7  ;serial output pin, RB7
 #undefine LAST_PORT
 #define LAST_PORT  0xC
 #undefine MAX_SFR
 #define MAX_SFR  0x1A0  ;max special function register
 #define FSRH  0x1FF  ;use an unused register location as alias for STATUS.IRP
	messg "NOT TESTED!"
#endif  ;def __16F689

#ifdef __16F690
 #define __16F689or690
	INCLUDE "p16f690.inc"
	NOLIST  ;reduce clutter
	CONSTANT PORT#v(0xA)_BITS = b'00111111'  ;PORTA is 6 bits
	CONSTANT PORT#v(0xB)_BITS = b'11110000'  ;PORTB is 4 bits, but they are left-justified!
	CONSTANT PORT#v(0xC)_BITS = b'11111111'  ;PORTC is 8 bits
 #undefine SERIN_PIN
 #define SERIN_PIN  0xB5  ;serial input pin, RB5
 #undefine SEROUT_PIN
 #define SEROUT_PIN  0xB7  ;serial output pin, RB7
 #undefine LAST_PORT
 #define LAST_PORT  0xC
 #undefine MAX_SFR
 #define MAX_SFR  0x1A0  ;max special function register
 #define FSRH  0x1FF  ;use an unused register location as alias for STATUS.IRP
	messg "NOT TESTED!"
#endif  ;def __16F690

#ifndef STATUS
	error "[ERROR] PIC16Fxxx.inc file not found; processor type might be undefined or unsupported!"
#endif


;get page# of a code address:
#define LITPAGEOF(dest)  ((dest)/LIT_PAGE_SIZE)  ;used for direct addressing (thru opcode)
#define REGPAGEOF(dest)  ((dest)/REG_PAGE_SIZE)  ;used for indirect addressing (thru register)

;address of last code page:
#define LAST_LIT_PAGE_START  (LIT_PAGE_SIZE*LITPAGEOF(MAX_ROM))  ;start of last literal page (immediate addressing)
#define LAST_REG_PAGE_START  (REG_PAGE_SIZE*REGPAGEOF(MAX_ROM))  ;start of last register page (indirect/register addressing)


;decide if RP0 is needed if caller did not specify:
;the 16F688 uses RP0 for Bank 0/1 select
#ifndef USE_RP0
 #define USE_RP0  (MAX_RAM >= BANK_SIZE)  ;RP0 needed for Bank 0/2 vs. 1/3 select
#endif  ;MAX_RAM >= BANK_SIZE


;decide if RP1 is needed if caller did not specify:
;the note on page 15 says RP1 is not used, so we really only need one bsf/bcf instruction anyway, not 2
;but the memory map on page 8 seems to indicate that it *is* used?
#ifndef USE_RP1
 #define USE_RP1  (MAX_RAM >= 0x100)  ;RP1 needed for Bank 0/1 vs. 2/3 select
#endif  ;MAX_RAM >= 0x100


;set a safe default:
#ifndef USE_INTS
 #define USE_INTS  FALSE
#endif  ;USE_INTS


;decide if RP0 is needed if caller did not specify:
#ifndef USE_IRP
 #define USE_IRP  (MAX_RAM >= 0x100)  ;IRP needed for Bank 0/1 vs. 2/3 select
#endif  ;MAX_RAM >= 0x100


;set this if using indirect addressing on banks 2 or 3:
#if USE_IRP
 #define BANKSEL_INDIRECT(reg)  setbit STATUS, IRP, BANKOF(reg) & 2  ;need to set IRP to address bank 2 + 3 indirectly
#else  ;USE_IRP
 #define BANKSEL_INDIRECT(reg)  ASM_NOP
#endif  ;USE_IRP


;check if caller wants to override PCLATH addressing:
#ifndef USE_PCLATH3
 #define USE_PCLATH3  (LITPAGEOF(MAX_ROM) & 1)
#else  ;def USE_PCLATH3
 #define EXPLICIT_PCLATH  TRUE
#endif  ;def USE_PCLATH3

#ifndef USE_PCLATH4
 #define USE_PCLATH4  (LITPAGEOF(MAX_ROM) & 2)
#else  ;def USE_PCLATH4
 #define EXPLICIT_PCLATH  TRUE
#endif  ;def USE_PCLATH4


;addressing sanity checks and info:
	WARNIF2 USE_RP0, [INFO] Using RP0: max RAM #v(MAX_RAM), bank size #v(BANK_SIZE)
	WARNIF2 !USE_RP0, [INFO] Not using RP0: max RAM #v(MAX_RAM), bank size #v(BANK_SIZE)
	WARNIF USE_RP0 && (MAX_RAM < BANK_SIZE), [CAUTION] Max RAM #v(MAX_RAM) < Bank size #v(BANK_SIZE) but RP0 is enabled: useless extra instructions will be used
	WARNIF !USE_RP0 && (MAX_RAM >= BANK_SIZE), [CAUTION] Max RAM #v(MAX_RAM) >= Bank size #v(BANK_SIZE) but RP0 is disabled: this prevents direct addressing of upper banks

	WARNIF2 USE_RP1, [INFO] Using RP1: max RAM #v(MAX_RAM), bank size #v(BANK_SIZE)
	WARNIF2 !USE_RP1, [INFO] Not using RP1: max RAM #v(MAX_RAM), bank size #v(BANK_SIZE)
	WARNIF USE_RP1 && (MAX_RAM < 0x100), [CAUTION] Max RAM #v(MAX_RAM) < 0x100 but RP1 is enabled: useless extra instructions will be used
	WARNIF !USE_RP1 && (MAX_RAM >= 0x100), [CAUTION] Max RAM #v(MAX_RAM) >= 0x100 but RP1 is disabled: this prevents direct addressing of upper banks

	WARNIF2 USE_IRP, [INFO] Using IRP: max RAM #v(MAX_RAM), bank size #v(BANK_SIZE)
	WARNIF2 !USE_IRP, [INFO] Not using IRP: max RAM #v(MAX_RAM), bank size #v(BANK_SIZE)
	WARNIF USE_IRP && (MAX_RAM < 0x100), [CAUTION] Max RAM #v(MAX_RAM) < 0x100 but IRP is enabled: useless extra instructions will be used
	WARNIF !USE_IRP && (MAX_RAM >= 0x100), [CAUTION] Max RAM #v(MAX_RAM) >= 0x100 but IRP is disabled: this prevents indirect addressing of those banks

	WARNIF2 USE_PCLATH3, [INFO] Using PCLATH<3>: max ROM #v(MAX_ROM), page size #v(LIT_PAGE_SIZE)
	WARNIF2 !USE_PCLATH3, [INFO] Not using PCLATH<3>: max ROM #v(MAX_ROM), page size #v(LIT_PAGE_SIZE)
	WARNIF2 USE_PCLATH4, [INFO] Using PCLATH<4>: max ROM #v(MAX_ROM), page size #v(LIT_PAGE_SIZE)
	WARNIF2 !USE_PCLATH4, [INFO] Not using PCLATH<4>: max ROM #v(MAX_ROM), page size #v(LIT_PAGE_SIZE)


;define some extra symbols for easier port references:
#ifdef PORTA
	CONSTANT IOC#v(0xA) = IOCA
	CONSTANT PORT#v(0xA) = PORTA, TRIS#v(0xA) = TRISA
	CONSTANT PORTOF#v(PORTA) = 0xA, PORTOF#v(TRISA) = 0xA
#else
	CONSTANT PORT#v(0xA) = -1, TRIS#v(0xA) = -1, PORT#v(0xA)_BITS = 0  ;helps caller to know port not present
#endif  ;def PORTA
#ifdef PORTB
;	CONSTANT IOC#v(0xB) = IOCB
	CONSTANT PORT#v(0xB) = PORTB, TRIS#v(0xB) = TRISB
	CONSTANT PORTOF#v(PORTB) = 0xB, PORTOF#v(TRISB) = 0xB
#else
	CONSTANT PORT#v(0xB) = -1, TRIS#v(0xB) = -1, PORT#v(0xB)_BITS = 0  ;helps caller to know port not present
#endif  ;def PORTB
#ifdef PORTC
;	CONSTANT IOC#v(0xC) = IOCC
	CONSTANT PORT#v(0xC) = PORTC, TRIS#v(0xC) = TRISC
	CONSTANT PORTOF#v(PORTC) = 0xC, PORTOF#v(TRISC) = 0xC
#else
	CONSTANT PORT#v(0xC) = -1, TRIS#v(0xC) = -1, PORT#v(0xC)_BITS = 0  ;helps caller to know port not present
#endif  ;def PORTC
#ifdef PORTD
;	CONSTANT IOC#v(0xD) = IOCD
	CONSTANT PORT#v(0xD) = PORTD, TRIS#v(0xD) = TRISD
	CONSTANT PORTOF#v(PORTD) = 0xD, PORTOF#v(TRISD) = 0xD
#else
	CONSTANT PORT#v(0xD) = -1, TRIS#v(0xD) = -1, PORT#v(0xD)_BITS = 0  ;helps caller to know port not present
#endif  ;def PORTD
#ifdef PORTE
;	CONSTANT IOC#v(0xE) = IOCE
	CONSTANT PORT#v(0xE) = PORTE, TRIS#v(0xE) = TRISE
	CONSTANT PORTOF#v(PORTE) = 0xE, PORTOF#v(TRISE) = 0xE
#else
	CONSTANT PORT#v(0xE) = -1, TRIS#v(0xE) = -1, PORT#v(0xE)_BITS = 0  ;helps caller to know port not present
#endif  ;def PORTE
#ifdef PORTF
	CONSTANT IOC#v(0xF) = IOCF
	CONSTANT PORT#v(0xF) = PORTF, TRIS#v(0xF) = TRISF
	CONSTANT PORTOF#v(PORTF) = 0xF, PORTOF#v(TRISF) = 0xF
#else
	CONSTANT PORT#v(0xF) = -1, TRIS#v(0xF) = -1, PORT#v(0xF)_BITS = 0  ;helps caller to know port not present
#endif  ;def PORTF


;******************************************************************************
;banking/paging/memory management macros:
;these correct or at least identify addressing problems for RAM banks and code pages
;helps to reduce late-night errors, but LOOK AT THE LISTING FILE to make sure they are doing what they are supposed to
;******************************************************************************

	CONSTANT TMR1 = TMR1L  ;alias for Timer 1 lo+high bytes (little endian)
;#if TMR1H != (TMR1L+1)
;	ERRIF TRUE, "Timer1 is NOT little-endian", TMR1L, TMR1H
;;#else
;;	messg "Timer1 is little-endian (okay)"
;#endif


	CONSTANT BT_NONBANKED = (-1)>>16  ;non-banked RAM location
	CONSTANT BT_UNKNOWN = -1  ;invalid RAM location


;the following macros expand past MPASM's 200 column limit:
;#define BANKOF(reg)  IIF(((reg) == INDF) || ((reg) == PCL) || ((reg) == STATUS) || ((reg) == FSR) || ((reg) == PCLATH) || ((reg) == INTCON) || ((reg)%BANK_SIZE >= NONBANKED_RAM_START), -1, (reg)/BANK_SIZE)
;#define ISPORT(reg)  ((((reg) >= PORTA) && ((reg) <= PORTF)) || (((reg) >= TRISA) && ((reg) <= TRISF)))
;#define ISREG(adrs, size)  (((adrs)%BANK_SIZE >= GENERAL_RAM_START) || ((size) < 2) || (HIBYTE(adrs) == (adrs)+(size)-1))
;KLUDGE: to shorten these lines, we need to use a symbolic lookup table to handle this instead, so that's what the code below does
;KLUDGE: MPASM will only iterate 256 times or generate 100 lines within a while, so a reusable macro is used here:
	VARIABLE BTREG = 0


;this macro will generate other constants
;params:
; limit = #iterations; allows continuation since MPASM limits #iterations to ~ 200
GENERATE_REGINFO MACRO limit
	NOEXPAND  ;reduce clutter
	EXPAND_PUSH FALSE
	LOCAL LIMIT = MIN(MAX_RAM, limit)
	while BTREG <= LIMIT
		VARIABLE NOTBANKED = (BTREG == INDF) || (BTREG == PCL) || (BTREG == STATUS) || (BTREG == FSR) || (BTREG == PCLATH) || (BTREG == INTCON) || ((BTREG%BANK_SIZE) >= NONBANKED_RAM_START)
		VARIABLE ISBADRAM = FALSE  ; (BTREG == 0x06) || ((BTREG >= 0x08) && (BTREG <= 0x09)) || etc. TODO: check for bad RAM locations?
		VARIABLE BANKNUM = IIF(NOTBANKED, BT_NONBANKED, IIF(ISBADRAM, BT_UNKNOWN, BTREG/BANK_SIZE))  ;bank# of this register
		VARIABLE REGSIZES = IIF(ISBADRAM, 0, 1<<1)  ;all valid registers are at least 1 byte
		VARIABLE UPPER_BYTE = IIF((BTREG%BANK_SIZE) >= GENERAL_RAM_START, BTREG+1, 0)  ;default address of upper byte for 16-bit registers (assumed to be contiguous unless overridden below)
;choose the correct upper byte for non-contiguous 16-bit registers:
;the only valid 16-bit regs are TMR1L+H, ANSEL+ANSELH, PCL+PCLATH, EEADR+EEADRH, EEDAT+EEDATH, SPBRG+SPBRGH
;FSR is actually a 9-bit register, so a pseudo-reg is used so it can be semantically handled like a 16-bit reg
		if BTREG == TMR1L  ;TMR1/TMR1H are already little-endian, so this is redundant
UPPER_BYTE = TMR1H
REGSIZES |= 1<<2  ;this is a valid 16-bit register
		endif
#ifdef ANSELH
		if BTREG == ANSEL  ;ANSEL/ANSELH are already little-endian, so this is redundant
UPPER_BYTE = ANSELH
REGSIZES |= 1<<2  ;this is a valid 16-bit register
		endif
#endif  ;def ANSELH
		if BTREG == PCL  ;PCL/PCLATH are disjoint
UPPER_BYTE = PCLATH
REGSIZES |= 1<<2  ;this is a valid 16-bit register
		endif
		if BTREG == EEADR  ;EEADR/EEADRH are disjoint
UPPER_BYTE = EEADRH
REGSIZES |= 1<<2  ;this is a valid 16-bit register
		endif
		if BTREG == EEDAT  ;EEDAT/EEDATH are disjoint
UPPER_BYTE = EEDATH
REGSIZES |= 1<<2  ;this is a valid 16-bit register
		endif
		if BTREG == SPBRG  ;SPBRG/SPBRGH are big-endian
UPPER_BYTE = SPBRGH
REGSIZES |= 1<<2  ;this is a valid 16-bit register
		endif
		if BTREG == FSR  ;high byte of FSR is only 1 bit (IRP bit in STATUS)
UPPER_BYTE = FSRH  ;pseudo-reg
REGSIZES |= 1<<2  ;this is a valid 16-bit pseudo-register
		endif
		VARIABLE ISPORT = (BTREG == PORT#v(0xA)) || (BTREG == PORT#v(0xB)) || (BTREG == PORT#v(0xC)) || (BTREG == PORT#v(0xD)) || (BTREG == PORT#v(0xE)) || (BTREG == PORT#v(0xF))
		VARIABLE ISTRIS = (BTREG == TRIS#v(0xA)) || (BTREG == TRIS#v(0xB)) || (BTREG == TRIS#v(0xC)) || (BTREG == TRIS#v(0xD)) || (BTREG == TRIS#v(0xE)) || (BTREG == TRIS#v(0xF))
		if (BTREG%BANK_SIZE) >= GENERAL_RAM_START
REGSIZES |= (1<<2) | (1<<3) | (1<<4)  ;user-defined registers can be any size; set them as valid 16-, 24- or 32-bit register
		endif
		CONSTANT REGINFO_#v(BTREG) = BANKNUM<<16 | UPPER_BYTE<<6 | BOOL2INT(ISTRIS | ISPORT) | REGSIZES
		if !(BTREG%BANK_SIZE)  ;define one extra const at the start of each bank
			CONSTANT OFBANK_#v(BTREG/BANK_SIZE) = GENERAL_RAM_START + BTREG
		endif
BTREG += 1
		if BTREG > d'1000'
			ERRIF TRUE, [ERROR] GENERATE_REGINFO broken macro: reg #v(BTREG) exceeded limit #v(limit)
			EXITM
		endif
	endw
	EXPAND_POP,
	ENDM
	CONSTANT REGINFO_ = (BT_NONBANKED<<16) + 1  ;KLUDGE: this will make BANKOF(-1) == not banked
;	CONSTANT REGINFO_#v(BT_NONBANKED) = BT_NONBANKED<<16  ;KLUDGE: this will make BANKOF(not banked) == not banked
;	CONSTANT REGINFO_#v(BT_UNKNOWN) = BT_NONBANKED<<16  ;KLUDGE: this will make BANKOF(-1) == not banked
;	CONSTANT REGINFO_#v(0x7fffa0) = 
	CONSTANT OFBANK_#v(BT_NONBANKED) = BT_UNKNOWN
	CONSTANT OFBANK_ = BT_UNKNOWN + 1  ;KLUDGE: this will make OFBANK(-1) == -1

	REPEAT divup(MAX_RAM, d'250'), GENERATE_REGINFO BTREG + d'250'  ;KLUDGE: MPASM only handles 256 iterations, so do it in chunks


;now define macros to get info about a register using the constants generated above:
#define BANKOF(reg)  ((REGINFO_#v(reg))>>d'16')  ;-1 means not banked
#define HIBYTE(reg)  ((REGINFO_#v(reg)>>6) & 0x3ff)  ;NOTE: might be > 0xff on some PICs
#define ISPORT(reg)  (REGINFO_#v(reg)&1)  ;PORTx or TRISx
#define ISREG(adrs, size)  (REGINFO_#v(adrs)&(1<<(size)))
;line too long:
;#define OFBANK(bank)  IIF((bank) != BT_NONBANKED, GENERAL_RAM_START + BANK_SIZE*(bank), BT_UNKNOWN)  ;a bank-selected address within the specified bank
#define OFBANK(bank)  OFBANK_#v(bank)  ;a bank-selected address within the specified bank

;BANKOF tests:
	ERRIF BANKOF(INDF) != BT_NONBANKED, [ERROR] "BANKOF(INDF)" == #v(BANKOF(INDF)) but should be #v(BT_NONBANKED)  ;paranoid self-check
	ERRIF BANKOF(PCL) != BT_NONBANKED, [ERROR] "BANKOF(PCL)" == #v(BANKOF(PCL)) but should be #v(BT_NONBANKED)  ;paranoid self-check
	ERRIF BANKOF(STATUS) != BT_NONBANKED, [ERROR] "BANKOF(STATUS)" == #v(BANKOF(STATUS)) but should be #v(BT_NONBANKED)  ;paranoid self-check
	ERRIF BANKOF(FSR) != BT_NONBANKED, [ERROR] "BANKOF(FSR)" == #v(BANKOF(FSR)) but should be #v(BT_NONBANKED)  ;paranoid self-check
	ERRIF BANKOF(PCLATH) != BT_NONBANKED, [ERROR] "BANKOF(PCLATH)" == #v(BANKOF(PCLATH)) but should be #v(BT_NONBANKED)  ;paranoid self-check
	ERRIF BANKOF(INTCON) != BT_NONBANKED, [ERROR] "BANKOF(INTCON)" == #v(BANKOF(INTCON)) but should be #v(BT_NONBANKED)  ;paranoid self-check
	ERRIF BANKOF(d'20') != 0, [ERROR] "BANKOF(20)" == #v(BANKOF(d'20')) but should be 0  ;paranoid self-check
	ERRIF BANKOF(0x7f) != BT_NONBANKED, [ERROR] "BANKOF(0x7f)" == #v(BANKOF(0x7f)) but should be #v(BT_NONBANKED)  ;paranoid self-check
	ERRIF BANKOF(0x84) != 1, [ERROR] "BANKOF(0x84)" == #v(BANKOF(0x84)) but should be 1  ;paranoid self-check
	ERRIF BANKOF(0x184) != 3, [ERROR] "BANKOF(0x184)" == #v(BANKOF(0x184)) but should be 3  ;paranoid self-check
	ERRIF BANKOF(MAX_RAM) != BT_NONBANKED, [ERROR] "BANKOF(MAX_RAM)" == #v(BANKOF(MAX_RAM)) but should be #v(BT_NONBANKED)  ;paranoid self-check
	ERRIF BANKOF(BT_UNKNOWN) != BT_NONBANKED, [ERROR] "BANKOF(-1)" == #v(BANKOF(BT_UNKNOWN)) but should be #v(BT_NONBANKED)  ;paranoid self-check
	ERRIF BANKOF(OFBANK(0)) != 0, [ERROR] "BANKOF(BANK 0)" == #v(BANKOF(OFBANK(0))) but should be 0  ;paranoid self-check
	ERRIF BANKOF(OFBANK(1)) != 1, [ERROR] "BANKOF(BANK 1)" == #v(BANKOF(OFBANK(1))) but should be 1  ;paranoid self-check
	ERRIF BANKOF(OFBANK(BT_NONBANKED)) != BT_NONBANKED, [ERROR] "BANKOF(BANK x)" == #v(BANKOF(OFBANK(BT_NONBANKED))) but should be #v(BT_NONBANKED)  ;paranoid self-check
;HIBYTE tests:
	ERRIF HIBYTE(TMR1L) != TMR1H, [ERROR] "HIBYTE(TMR1L)" == #v(HIBYTE(TMR1L)) but should be #v(TMR1H)  ;paranoid self-check
	ERRIF HIBYTE(PCL) != PCLATH, [ERROR] "HIBYTE(PCL)" == #v(HIBYTE(PCL)) but should be #v(PCLATH)  ;paranoid self-check
	ERRIF HIBYTE(EEADR) != EEADRH, [ERROR] "HIBYTE(EEADR)" == #v(HIBYTE(EEADR)) but should be #v(EEADRH)  ;paranoid self-check
	ERRIF HIBYTE(EEDAT) != EEDATH, [ERROR] "HIBYTE(EEDAT)" == #v(HIBYTE(EEDAT)) but should be #v(EEDATH)  ;paranoid self-check
	ERRIF HIBYTE(SPBRG) != SPBRGH, [ERROR] "HIBYTE(SPBRG)" == #v(HIBYTE(SPBRG)) but should be #v(SPBRGH)  ;paranoid self-check
	ERRIF HIBYTE(FSR) != FSRH, [ERROR] "HIBYTE(FSR)" == #v(HIBYTE(FSR)) but should be #v(FSRH)  ;paranoid self-check
	ERRIF HIBYTE(1), [ERROR] "HIBYTE(1)" == #v(HIBYTE(1)) but should be 0  ;paranoid self-check
;ISPORT tests:
	ERRIF ISPORT(0x7f), [ERROR] "ISPORT(0x7f)" == #v(ISPORT(0x7f)) but should be 0  ;paranoide self-check
	ERRIF !ISPORT(PORT#v(FIRST_PORT)), [ERROR] "ISPORT(FIRST_PORT)" == #v(ISPORT(PORT#v(FIRST_PORT))) but should be !0  ;paranoide self-check
	ERRIF !ISPORT(TRIS#v(LAST_PORT)), [ERROR] "ISPORT(LAST_TRIS)" == #v(ISPORT(TRIS#v(LAST_PORT))) but should be !0  ;paranoide self-check
;ISREG tests:
	ERRIF !ISREG(0x7f, 2), [ERROR] "ISREG(0x7f,2)" == #v(ISREG(0x7f, 2)) but should be !0  ;paranoide self-check
	ERRIF !ISREG(TMR1L, 2), [ERROR] "ISREG(TMR1L,2)" == #v(ISREG(TMR1L, 2)) but should be !0  ;paranoide self-check
	ERRIF ISREG(PCL, 3), [ERROR] "ISREG(PCL,3)" == #v(ISREG(PCL, 3)) but should be 0  ;paranoide self-check
	ERRIF !ISREG(0x20, 4), [ERROR] "ISREG(0x20,4)" == #v(ISREG(0x20, 4)) but should be !0  ;paranoide self-check


;set non/banked RAM allocation limits:
	CONSTANT RAM_START#v(BOOL2INT(FALSE)) = NONBANKED_RAM_START, RAM_END#v(BOOL2INT(FALSE)) = BANK_SIZE
	CONSTANT RAM_START#v(BOOL2INT(TRUE)) = GENERAL_RAM_START, RAM_END#v(BOOL2INT(TRUE)) = NONBANKED_RAM_START
	VARIABLE NEXT_RAM#v(BOOL2INT(MAYBE)) = MAX_RAM  ;this one can be overridden by caller


;convenience wrappers for SAFE_ALLOC macro:

;banked alloc:
BDCL MACRO name
	NOEXPAND
	SAFE_ALLOC name, TRUE
	ENDM

;non-banked alloc:
NBDCL MACRO name
	NOEXPAND
	SAFE_ALLOC name, FALSE
	ENDM

;reallocate:
;CAUTION: caller must ensure that usage is mutually exclusive
REDCL MACRO name
	NOEXPAND
	SAFE_ALLOC name, MAYBE
	ENDM


;allow space to be reallocated:
;TODO: maybe use a stack frame allocation mechanism?
;params:
; ramstart = start address of reallocatable area
; ramend = end address of reallocatable area
REALLOC_SPACE MACRO ramstart, ramend
	NOEXPAND  ;reduce clutter
	CONSTANT RAM_START#v(BOOL2INT(MAYBE)) = ramstart, RAM_END#v(BOOL2INT(MAYBE)) = ramend
NEXT_RAM#v(BOOL2INT(MAYBE)) = ramstart
	WARNIF FULL_WARN, [CAUTION] RAM #v(ramstart)..#v(ramend) being reused: caller must ensure mutually exclusive usage
	ENDM


	VARIABLE RAM_BLOCK = 0, NEXT_RAM#v(BOOL2INT(TRUE)) = GENERAL_RAM_START, NEXT_RAM#v(BOOL2INT(FALSE)) = NONBANKED_RAM_START

;allocate a banked/non-banked/reallocated variable:
;checks for address overflow on allocated variables
;also saves banked or non-banked RAM address for continuation in a later CBLOCK
;params:
; name = variable name to allocate
; banked = flag controlling where it is allocated; TRUE/FALSE == yes/no, MAYBE == reallocate from caller-specified pool of reusable space
SAFE_ALLOC MACRO name, banked
	NOEXPAND  ;reduce clutter
	EXPAND_PUSH TRUE  ;show RAM allocations in LST
	CBLOCK NEXT_RAM#v(BOOL2INT(banked))  ;continue where we left off last time
	name
	ENDC  ;can't span macros
	EXPAND_PUSH FALSE
RAM_BLOCK += 1  ;need a unique symbol name so assembler doesn't complain; LOCAL won't work inside CBLOCK
	CBLOCK
	LAST_RAM_ADDRESS_#v(RAM_BLOCK):0  ;get address of last alloc; need additional CBLOCK because macros cannot span CBLOCKS
	ENDC
NEXT_RAM#v(BOOL2INT(banked)) = LAST_RAM_ADDRESS_#v(RAM_BLOCK)  ;update pointer to next available RAM location
	ERRIF LAST_RAM_ADDRESS_#v(RAM_BLOCK) > MAX_RAM, [ERROR] SAFE_ALLOC: RAM overflow #v(LAST_RAM_ADDRESS_#v(RAM_BLOCK)) > max #v(MAX_RAM)
	ERRIF LAST_RAM_ADDRESS_#v(RAM_BLOCK) > RAM_END#v(BOOL2INT(banked)), [ERROR] SAFE_ALLOC: RAM overflow #v(LAST_RAM_ADDRESS_#v(RAM_BLOCK)) > end #v(RAM_END#v(BOOL2INT(banked)))
	ERRIF LAST_RAM_ADDRESS_#v(RAM_BLOCK) <= RAM_START#v(BOOL2INT(banked)), [ERROR] SAFE_ALLOC: RAM overflow #v(LAST_RAM_ADDRESS_#v(RAM_BLOCK)) <= start #v(RAM_START#v(BOOL2INT(banked)))
	EXPAND_POP,
	EXPAND_POP,
	ENDM


	VARIABLE PREVBANK = BT_UNKNOWN  ;track currently selected bank; starts unknown
	VARIABLE NUMBANKSEL = 0, NONBANKSEL = 0  ;count #instructions used/saved for bank selects; used to track macro perf stats
;	VARIABLE NESTED_BANKCHK = 0  ;detect recursion (mainly for debug)

;optimized bank select:
;this keeps track of the previous bank selected and then decides whether RP0/RP1 need to be set
;assembler always generates 2 bsf/bcf instructions for BANKSEL (for RP0 + RP1), even if only one is needed
;this macro only sets them *if needed*, so it will be more efficient than BANKSEL (saves 1 - 2 instructions over BANKSEL)
;for now, this is only implemented for immediate (literal) addresses; doesn't handle indirect addresses and IRP yet (which are only used in a few places anyway)
;when fixit param == MAYBE, conditional code is generated to set RP0 and/or RP1 as needed
;HOWEVER, this can cause problems within a btf, because there is no room for the extra bank instruction(s)
;in those cases, set fixit = FALSE to just check that RP0/RP1 are set correctly, and explicitly set/reset RP1+RP0 *before* the btf instruction with fixit = TRUE/MAYBE
;set PREVBANK to -1 to "flush" the remembered bank select when unknown; this will force RP0/RP1 to be set the next time

;params:
; reg = register whose bank is to be selected
; fixit = whether or not to generate bank select instructions: TRUE/FALSE = yes/no, MAYBE = only if bank is different from previously active bank
; undef_ok = flag telling whether to silently ignore undefined bank (optional); defaults to FALSE; must end with "|" if passed
BANKCHK MACRO reg, fixit, undef_ok
	NOEXPAND  ;hide clutter in LST file
	if (undef_ok 0) && ((reg) == BT_UNKNOWN)
		EXITM
	endif
	EXPAND_PUSH FALSE
;	ERRIF NESTED_BANKCHK, "[ERROR] Recursive BANKCHK", $, NESTED_BANKCHK
;NESTED_BANKCHK += 1
	LOCAL REG = reg  ;assember bug; save as LOCAL to force eval (else "missing operand" and "missing argument" errors); also helps with "line too long" messages
	LOCAL FIXIT = fixit
	ERRIF (REG < 0) || (REG > MAX_RAM), [ERROR] BANKCHK: bad register address #v(REG) out of range 0..#v(MAX_RAM)
	if BANKOF(REG) == BT_NONBANKED  ;not banked; don't need to set RP0/RP1
NONBANKSEL += 1
		EXPAND_POP,
;--NESTED_BANKCHK
		EXITM
	endif
	if BANKOF(REG) == BANKOF(PREVBANK)  ;don't need to set RP0/RP1
;PREVBANK = REG  ;update last-used reg anyway (helpful for debug)
NONBANKSEL += 1  ;count saved instructions (macro perf)
;;--NESTED_BANKCHK
	endif
	if BANKOF(PREVBANK) == BT_NONBANKED  ;force a bank select
;PREVBANK = ~REG  ;this will force both RP0+RP1 to be set below
PREVBANK = OFBANK(BANKOF(REG) ^ 3)  ;this will force both RP0+RP1 to need updates below
	endif
;	BANKSEL reg  ;BANKSEL always generates bsf/bcf on RP0 and RP1, even if not needed!
#if USE_RP1  ;the note on page 15 says RP1 is not used, so we really only need one bsf/bcf instruction anyway, not 2?
;only set/reset RP1 if needed:
	if ((BANKOF(REG) & 2) != (BANKOF(PREVBANK) & 2)) || (FIXIT == TRUE)
		ERRIF !FIXIT, [ERROR] BANKCHK: you need to set/reset RP1 for reg #v(REG) here
		WARNIF (BANKOF(REG) & 2) == (BANKOF(PREVBANK) & 2), [INFO] BANKCHK: you don't need to set/reset RP1 for reg #v(REG) here
	endif
	if (((BANKOF(REG) & 2) != (BANKOF(PREVBANK) & 2)) || (FIXIT == TRUE)) && FIXIT  ;IFs too deep; repeated here to reduce nesting level
		setbit2 STATUS, RP1, BANKOF(REG) & 2  ;select bank 0+1 or 2+3
NUMBANKSEL += 1  ;count saved instructions (macro perf)
	endif
#else  ;USE_RP1
	ERRIF BANKOF(REG) & 2, [ERROR] BANKCHK: you need to set RP1 for reg #v(REG) here
NONBANKSEL += 1
#endif  ;USE_RP1
#if USE_RP0
;only set/reset RP0 if needed:
	if ((BANKOF(REG) & 1) != (BANKOF(PREVBANK) & 1)) || (FIXIT == TRUE)
		ERRIF !FIXIT, [ERROR] BANKCHK: you need to set/reset RP0 for reg #v(REG) here
		WARNIF (BANKOF(REG) & 1) == (BANKOF(PREVBANK) & 1), [INFO] BANKCHK: you don't need to set/reset RP0 for reg #v(REG) here
	endif
	if (((BANKOF(REG) & 1) != (BANKOF(PREVBANK) & 1)) || (FIXIT == TRUE)) && FIXIT  ;IFs too deep; repeated here to reduce nesting level
		setbit2 STATUS, RP0, BANKOF(REG) & 1  ;select bank 0/2 or 1/3
NUMBANKSEL += 1  ;count saved instructions (macro perf)
	endif
#else  ;USE_RP0
	ERRIF BANKOF(REG) & 1, [ERROR] BANKCHK: you need to set RP0 for reg #v(REG) here
NONBANKSEL += 1
#endif  ;USE_RP0
	if FIXIT
PREVBANK = REG  ;remember current bank select for next time
	endif
	EXPAND_POP,
;--NESTED_BANKCHK
	ENDM


;force bank select on a register (unconditional):
;params:
; reg = register whose bank is to become active
; force = (unused)
SETBANK MACRO reg, force
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
;no longer needed:
;	if force
;		if (force) == TRUE
;PREVBANK = -1  ;flush previous bank
;		endif
;		BANKCHK $, adrs, force  ;force bank select bits to be set
;	else
;	endif
PREVBANK = reg  ;override previous bank
	EXPAND_POP,
	ENDM


;generate a "shadow address" (same lower address, but different upper page):
;this is used to defeat PAGECHK when it is known for *certain* that PCLATH is set correctly
#define SHADOW_ADRS(thing)  (((thing)&(LIT_PAGE_SIZE-1)) | ($&~(LIT_PAGE_SIZE-1)))


	VARIABLE PREVPAGE = -1  ;track currently selected bank; starts unknown
	VARIABLE NUMPAGESEL = 0, NONPAGESEL = 0  ;count #instructions used/saved for page selects; used to track macro perf stats

;optimized page select:
;this tracks the previous page selected and then decides whether PCLATH needs to be set
;assembler always generates 2 bsf/bcf instructions for PAGESEL (for PCLATH<4> + PCLATH<3>), even if only one is needed
;this macro only sets them *if needed*, so it will be more efficient than PAGESEL (saves 1 - 2 instructions over PAGESEL)
;for now, this is only implemented for immediate (literal) addresses; doesn't handle indirect addresses and PCLATH yet (which are only used in a few places, anyway)
;when fixit == MAYBE, conditional code is generated to set PCLATH<4> and/or PCLATH<3> as needed
;HOWEVER, this can cause problems within a btf, because there is no room for the extra bank instruction(s)
;AND the assembler doesn't handle variable code addresses correctly between pass 1 and pass (which can lead to error 116 "Address label duplicated or different in second pass"); forward references are not resolved yet in pass1
;in those cases, set fixit = FALSE to just check that PCLATH is set correctly, and explicitly set/reset PCLATH *before* the btf instruction with fixit = TRUE/MAYBE
;set PREVPAGE to -1 to "flush" the remembered page select when unknown; this will force PCLATH to be set the next time

;params:
; dest = address whose page is to be selected
; fixit = whether or not to generate page select instructions: TRUE/FALSE = yes/no, MAYBE = only if page is different from previously active page
; undef_ok = flag telling whether to silently ignore undefined page (optional); defaults to FALSE; must end with "|" if passed
PAGECHK MACRO dest, fixit, undef_ok
	NOEXPAND  ;reduce clutter
	if (undef_ok 0) && ((dest) == -1)
		EXITM
	endif
	EXPAND_PUSH FALSE
	LOCAL SRC = $
	LOCAL DEST = dest  ;assember bug; save as LOCAL to force eval (else "missing operand" and "missing argument" errors); also helps with "line too long" messages
	LOCAL FIXIT = fixit
#if USE_INTS  ;instruction at INTERRUPT_VECTOR should be saving registers
	ERRIF DEST == INTERRUPT_VECTOR, [ERROR] PAGECHK: int vector #v(DEST) should be saving registers
#endif  ;USE_INTS
	ERRIF2 (DEST < INTERRUPT_VECTOR) || (DEST > MAX_ROM) || (SRC < RESET_VECTOR) || (SRC > MAX_ROM), [ERROR] PAGECHK: bad address #v(DEST), you'll need to verify the address yourself
	WARNIF LITPAGEOF(DEST - 2) != LITPAGEOF(DEST + 2), [INFO] WARNING: destination #v(DEST) is close to page boundary: carefully check address arithmetic here!
	if LITPAGEOF(DEST) == LITPAGEOF(PREVPAGE)  ;don't need to set PCLATH<4:3>
NONPAGESEL += 1  ;count saved instructions (macro perf)
	endif
	if PREVPAGE < 0  ;force a page select
;PREVPAGE = ~DEST  ;this will force both PCLATH<4> and PCLATH<3> to be set below
PREVPAGE = LIT_PAGE_SIZE * (LITPAGEOF(DEST) ^ 3)  ;this will force both PCLATH<4> and <3> to need updates below
	endif
;	PAGESEL dest  ;PAGESEL always generates bsf/bcf on PCLATH<4> and PCLATH<3>, even if not needed!
#if USE_PCLATH4  ;page 9 says PC is 13 bits, but there's only 4K words, so it wraps after 0x1000 and there is really only 0x1000 words so PCLATH<4> is useless?
;only set/reset PCLATH<4> if needed:
	if ((LITPAGEOF(DEST) & 2) != (LITPAGEOF(PREVPAGE) & 2)) || (FIXIT == TRUE)
		ERRIF !FIXIT, [ERROR] PAGECHK: you need to set/reset PCLATH<4> for adrs #v(DEST) here
		WARNIF (LITPAGEOF(DEST) & 2) == (LITPAGEOF(PREVPAGE) & 2), [INFO] PAGECHK: you don't need to set/reset PCLATH<4> for adrs #v(DEST) here
		if FIXIT
			setbit PCLATH, 4, LITPAGEOF(DEST) & 2  ;select page 0+1 or 2+3
NUMPAGESEL += 1  ;count saved instructions (macro perf)
		endif
	endif
#else  ;USE_PCLATH4
	ERRIF LITPAGEOF(DEST) & 2, [ERROR] PAGECHK: you need to set PCLATH<4> for adrs #v(DEST) here
NONPAGESEL += 1
#endif  ;USE_PCLATH4
#if USE_PCLATH3
;only set/reset PCLATH<3> if needed:
	if ((LITPAGEOF(DEST) & 1) != (LITPAGEOF(PREVPAGE) & 1)) || (FIXIT == TRUE)
		ERRIF !FIXIT, [ERROR] PAGECHK: you need to set/reset PCLATH<3> for adrs #v(DEST) here
		WARNIF (LITPAGEOF(DEST) & 1) == (LITPAGEOF(PREVPAGE) & 1), [INFO] PAGECHK: you don't need to set/reset PCLATH<3> for reg #v(DEST) here
		if FIXIT
			setbit PCLATH, 3, LITPAGEOF(DEST) & 1  ;select page 0/2 or 1/3
NUMPAGESEL += 1  ;count saved instructions (macro perf)
		endif
	endif
#else  ;USE_PCLATH3
	ERRIF LITPAGEOF(DEST) & 1, [ERROR] PAGECHK: you need to set PCLATH<3> for reg #v(DEST) here
NONPAGESEL += 1
#endif  ;USE_PCLATH3
	if FIXIT
PREVPAGE = dest  ;remember where we went to for next time
	endif
	EXPAND_POP,
	ENDM


;force a code page select (unconditional):
;params:
; adrs = address whose page is to become active
; force = (unused)
SETPAGE MACRO adrs, force
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
;no longer needed:
;	if force
;		if (force) == TRUE
;PREVPAGE = -1  ;flush previous page
;		endif
;		PAGECHK $, adrs, force  ;force page select bits to be set
;	else
;	endif
PREVPAGE = adrs  ;override previous page
	EXPAND_POP,
	ENDM


;conditional convenience wrappers for specific processors:
UGLY_PASS12FIX688 MACRO ofs
#ifdef  __16F688
	UGLY_PASS12FIX ofs
#endif
	ENDM

UGLY_PASS12FIX690 MACRO ofs
#ifdef  __16F689or690
	UGLY_PASS12FIX ofs
#endif
	ENDM


	VARIABLE PASS1_FIXUPS = 0, PASS2_FIXUPS = 0  ;used to track macro perf stats

;VERY ugly kludge to help the assembler get back on track:
;this macro is needed because of SETPAGE/PCLATH optimization and the assembler's inability to handle that
;use this macro ahead of a label that gets an error 116 "Address label duplicated or different in second pass"
;there appear to be 2 main causes for this: pass 1 vs. pass 2 addresses out of sync, and LOCAL identifier name clashes
;the NEXT_UNIQID variable addresses the second of those causes, while this macro addresses the first one
;this macro can be used to pad out the address during pass 1, but be skipped during pass 2 (or vice versa), so that the address is consistent between pass 1 + 2
;it's best to use it within dead code chunks (ie, AFTER a goto), where the extra instructions will NEVER be executed; this avoids any run-time overhead
;I'm not sure why symbolic addresses occassionally get out of alignment between pass 1 and pass 2; it's inconsistent - sometimes the assembler recovers correctly and sometimes not
;usage of this macro is trial and error; only add it in one place at a time, and adjust it until the error 116 goes away (use the .LST file to check the addresses in pass 1 vs. 2)
;if a pass 1 address is higher than a pass 2 address, use a +ve offset; this will put nop's in the final executable code
;if a pass 1 address is less than a pass 2, use a -ve offset; this will only generate nop's in pass 1, and won't actually take up any code space in pass 2

;params:
; pass2ofs = amount to adjust; if pass 2 address > pass 1 address, use pass2ofs > 0; else use pass2ofs < 0; there can be errors in either direction
UGLY_PASS12FIX MACRO pass2ofs
	NOEXPAND  ;reduce clutter
	EXPAND_PUSH FALSE
	if (pass2ofs) < 0  ;slide pass 2 addresses down (pad pass 1 address up, actually)
		if !eof  ;this is true only during pass 1, because assembler hasn't resolved the address yet (eof label MUST be at end, so it isn't resolved yet)
			REPEAT -(pass2ofs), nop  ;insert dummy instructions to move address during pass 1; these won't be present during pass 2
		endif
;		WARNIF eof, "[WARNING] Unneeded pass 1 fixup", pass2ofs, eof  ;won't ever see this message (output discarded during pass 1)
PASS1_FIXUPS += 0x10000-(pass2ofs)  ;lower word = #prog words; upper word = #times called
	endif
	if (pass2ofs) > 0  ;slide pass 2 addresses up
		if eof  ;this is true only during pass 2, when address has been resolved (eof label MUST be at end, so it isn't resolved yet in pass 1)
			REPEAT pass2ofs, nop
		endif
		WARNIF !eof, [WARNING] Unneeded #v(pass2ofs) pass 2 fixup
PASS2_FIXUPS += 0x10000+(pass2ofs)  ;lower word = #prog words; upper word = #times called
	endif
	EXPAND_POP,
	ENDM


;this is a kludgey way to help assembler resolve addresses
;often the assembler isn't smart enough to figure out forward references
;these macros are used for forward references and other symbols that are not known until later
;this will leave an n-word gap in the prog area, which we can go back and fill in later after the unknown symbol has finally been resolved
;individual macros are defined here to avoid the use of commas within operands to other macros

;leave a 1-word placeholder:
;this is a work-around for when assember can't resolve forward addresses (intermittent problem with assembler)
;params:
; stmt = dummy param for readability
LEAVE_PLACEHOLDER1 MACRO stmt
	NOEXPAND
	LEAVE_PLACEHOLDER_ 1, stmt
	ENDM

;make sure correct bank is selected first, then leave placeholder (allows fixed length placeholder when bank is uncertain):
LEAVE_PLACEHOLDER1BANK MACRO reg, stmt
	NOEXPAND
	BANKCHK reg, MAYBE,  ;allow neargoto when placeholder is filled in
	LEAVE_PLACEHOLDER_ 1, stmt
	ENDM

;make sure PCLATH is set first, then leave placeholder (allows fixed length placeholder when page is uncertain):
LEAVE_PLACEHOLDER1JUMP MACRO stmt
	NOEXPAND
	PAGECHK $, MAYBE,  ;allow neargoto when placeholder is filled in
	LEAVE_PLACEHOLDER_ 1, stmt
	ENDM

;leave a 2-word placeholder:
;this is a work-around for when assember can't resolve forward addresses (intermittent problem with assembler)
;params:
; stmt = dummy param for readability
LEAVE_PLACEHOLDER2 MACRO stmt
	NOEXPAND
	LEAVE_PLACEHOLDER_ 2, stmt
	ENDM


	VARIABLE PLACEHOLDER_STACK = 0, PLACEHOLDER_HIWATER = 0  ;stack control

;leave an n-word placeholder:
;this is a work-around for when assember can't resolve forward addresses (intermittent problem with assembler)
;params:
; len = #words to leave open
; stmt = dummy param for readability
LEAVE_PLACEHOLDER_ MACRO len, stmt
	NOEXPAND  ;reduce clutter
	EXPAND_PUSH FALSE
PLACEHOLDER_STACK += 1  ;stack; don't overwrite previous if still active
	if PLACEHOLDER_STACK > PLACEHOLDER_HIWATER  ;allocate another stack frame (defines more symbols in MPASM symbol table)
		VARIABLE PLACEHOLDER#v(PLACEHOLDER_STACK)_STADRS
		VARIABLE PLACEHOLDER#v(PLACEHOLDER_STACK)_ENADRS
		VARIABLE PLACEHOLDER#v(PLACEHOLDER_STACK)_PREVBANK
		VARIABLE PLACEHOLDER#v(PLACEHOLDER_STACK)_PREVPAGE
		VARIABLE PLACEHOLDER#v(PLACEHOLDER_STACK)_RESUME
		VARIABLE PLACEHOLDER#v(PLACEHOLDER_STACK)_SVBANK
		VARIABLE PLACEHOLDER#v(PLACEHOLDER_STACK)_SVPAGE
PLACEHOLDER_HIWATER = PLACEHOLDER_STACK
	endif
PLACEHOLDER#v(PLACEHOLDER_STACK)_STADRS = $
PLACEHOLDER#v(PLACEHOLDER_STACK)_ENADRS = $ + (len)
																				EXPAND  ;show generated code
PLACEHOLDER#v(PLACEHOLDER_STACK)_PREVBANK = PREVBANK
PLACEHOLDER#v(PLACEHOLDER_STACK)_PREVPAGE = PREVPAGE
	ORG PLACEHOLDER#v(PLACEHOLDER_STACK)_STADRS + (len)  ;leave a gap; will come back and fill it in later (after symbols resolve or other decisions are made)
	NOEXPAND
	EXPAND_POP,
	ENDM


#define STRICT_OR_EMPTY(which)  IIF($ == PLACEHOLDER#v(which)_STADRS, MAYBE, TRUE)

;go back and fill in most recent placeholder:
;a stack is used for nested placeholders, so they must be filled in reverse order
;I guess you really can write relocatable absolute code afterall  :)
;params:
; which = which placeholder to fill (optional); if specified, overrides stack behavior; NOTE: this MUST be used if nested stmt blocks may also contain placeholders
; stmt = statement(s) to fill in; empty stmt to abandon (undo) the placeholder
FILL_PLACEHOLDER MACRO which, stmt
	NOEXPAND  ;reduce clutter
	LOCAL WHICH = which -1  ;optional param; must end with "&" if passed
	if WHICH == -1  ;param not passed
WHICH = PLACEHOLDER_STACK  ;default to top of stack
	endif
	EXPAND_PUSH FALSE
	START_FILL_PLACEHOLDER WHICH&,
																				EXPAND  ;show generated code
	stmt
	NOEXPAND
	END_FILL_PLACEHOLDER WHICH&, STRICT_OR_EMPTY(WHICH)
	EXPAND_POP,
	ENDM


;start placeholder fill-in:
;params:
; which = which placeholder to fill (optional); if specified, overrides stack behavior; NOTE: this MUST be used if nested stmt blocks may also contain placeholders
; dummy = dummy parameter (needed in order to force MPASM to accept params for this macro)
;	VARIABLE PHFILL_STARTED = 0  ;tolerate partial recursion
START_FILL_PLACEHOLDER MACRO which, dummy
	NOEXPAND  ;hide clutter in LST file
	LOCAL WHICH = which -1  ;optional param; must end with "&" if passed
	if WHICH == -1  ;param not passed
WHICH = PLACEHOLDER_STACK  ;default to top of stack
	endif
	EXPAND_PUSH FALSE
PLACEHOLDER#v(WHICH)_SVBANK = PREVBANK  ;save current bank
PLACEHOLDER#v(WHICH)_SVPAGE = PREVPAGE  ;save current page
PREVBANK = PLACEHOLDER#v(WHICH)_PREVBANK
PREVPAGE = PLACEHOLDER#v(WHICH)_PREVPAGE
PLACEHOLDER#v(WHICH)_RESUME = $  ;save current place
																				EXPAND  ;show generated code
	ORG PLACEHOLDER#v(WHICH)_STADRS  ;go back and fill in placeholder
	NOEXPAND
	EXPAND_POP,
	ENDM


;finish placeholder fill-in:
;params:
; which = which placeholder to fill (optional); if specified, overrides top-of-stack behavior; NOTE: MUST match corresponding START_FILL_PLACEHOLDER
; strict = determines whether to verify that address space is completely filled: TRUE/FALSE = yes/no, MAYBE = abandon placeholder without filling it
END_FILL_PLACEHOLDER MACRO which, strict
	NOEXPAND  ;hide clutter in LST file
	LOCAL WHICH = which -1  ;optional param; must end with "&" if passed
	if WHICH == -1  ;param not passed
WHICH = PLACEHOLDER_STACK  ;default to top of stack
	endif
	EXPAND_PUSH FALSE
;	if ($ < PLACEHOLDER#v(n)_ENADRS) && (PLACEHOLDER#v(n)_ENADRS > PLACEHOLDER#v(n)_STADRS+1)
;		WARNIF "whoops, overestimated placeholder space", $, PLACEHOLDER#v(n)_ENADRS
;	endif
	if (strict) != MAYBE
		if strict
			REPEAT PLACEHOLDER#v(WHICH)_ENADRS - $, nop  ;didn't use it all; pad it out
			ADRSCHK PLACEHOLDER#v(WHICH)_ENADRS
		endif
																				EXPAND  ;show generated code
		ORG PLACEHOLDER#v(WHICH)_RESUME  ;resume where we left off
PREVBANK = PLACEHOLDER#v(WHICH)_SVBANK  ;restore remembered bank
		NOEXPAND
	endif
PREVBANK = PLACEHOLDER#v(WHICH)_SVBANK  ;restore remembered bank
PREVPAGE = PLACEHOLDER#v(WHICH)_SVPAGE  ;restore remembered page
	while WHICH < PLACEHOLDER_STACK  ;slide top of stack down
PLACEHOLDER#v(WHICH)_STADRS = PLACEHOLDER#v(WHICH + 1)_STADRS
PLACEHOLDER#v(WHICH)_ENADRS = PLACEHOLDER#v(WHICH + 1)_ENADRS
PLACEHOLDER#v(WHICH)_PREVBANK = PLACEHOLDER#v(WHICH + 1)_PREVBANK
PLACEHOLDER#v(WHICH)_PREVPAGE = PLACEHOLDER#v(WHICH + 1)_PREVPAGE
PLACEHOLDER#v(WHICH)_RESUME = PLACEHOLDER#v(WHICH + 1)_RESUME
PLACEHOLDER#v(WHICH)_SVBANK = PLACEHOLDER#v(WHICH + 1)_SVBANK
PLACEHOLDER#v(WHICH)_SVPAGE = PLACEHOLDER#v(WHICH + 1)_SVPAGE
WHICH += 1
	endw
PLACEHOLDER_STACK -= 1  ;pop placeholder stack
	ERRIF PLACEHOLDER_STACK < 0, [ERROR] Placeholder stack underflow: #v(PLACEHOLDER_STACK)
	EXPAND_POP,
	ENDM


;verify that address calculations are correct:
;the most likely cause that they are incorrect is bank or page select optimization, so check BANKCHK/PAGECHK usage if this generates an error
;params:
; adrs = what the current address should be if address calculations were correct
ADRSCHK MACRO adrs
	NOEXPAND
	ERRIF $ != (adrs), [ERROR] ADRSCHK: guessed wrong on target address (expected #v(adrs) at #v($)). check for correct bank and page select
	ENDM


;******************************************************************************
;flow control pseudo-opcodes:
;these opcodes are all bank- and page- safe
;every label (goto or call/return) can have a "bank affinity"; this determines which bank is active upon entry or exit
;check the .LST file to see the resulting opcodes and addresses
;******************************************************************************

;enable/disbale interrupts:
;silently ignored if ints not used
;params:
; onoff = flag whether to enable or disable interrupts
intenable_ MACRO onoff
	NOEXPAND  ;reduce clutter
	EXPAND_PUSH FALSE
#if USE_INTS
	if onoff
		setbit INTCON, GIE, onoff
	else
;http://www.pic101.com/mcgahee/picuart.zip includes a note about a Microchip recommendation for a loop to ensure that GIE is cleared
;MicroChip says: If an IRQ happens when the "bcf intcon,gie" line is executing, the GIE bit will be cleared, but the interrupt will still occur.
		whilebit INTCON, GIE, !(onoff), SETBIT(INTCON, GIE, onoff),
	endif
#endif  ;USE_INTS
	EXPAND_POP,
	ENDM


;comma-less wrappers for goto/call/return:
;bank affinity is determined by destination address

;call a function and return later (dest can be anywhere):
;CAUTION: generated code size can vary, which might interfere with assember operation
;params:
; dest = destination address (can be near or far)
call_ MACRO dest
	NOEXPAND
	jump dest, MAYBE, TRUE  ;near/far tolerant; do bank select if necessary
	ENDM

;this one does not return (dest can be anywhere):
;CAUTION: generated code size can vary, which might interfere with assember operation
;params:
; dest = destination address (can be near or far)
goto_ MACRO dest
	NOEXPAND
	jump dest, MAYBE, FALSE  ;near/far tolerant; never returns
	ENDM


;near call (destination must be on current page):
;this must be used for instructions immediately following a btf conditional skip instruction (only room for one instruction)
;params:
; dest = destination address (must be in same page)
nearcall MACRO dest
	NOEXPAND
	jump dest, FALSE, TRUE  ;destination must be in same page
	ENDM

;near goto (destination must be on current page):
;this must be used for instructions immediately following a btf conditional skip instruction (only room for one instruction)
;params:
; dest = destination address (must be in same page)
neargoto MACRO dest
	NOEXPAND
	jump dest, FALSE, FALSE  ;destination must be in same page
	ENDM


;far call (destination address not in same page):
;this cannot be used for instructions immediately following a btf conditional skip instruction (only room for one instruction)
;params:
; dest = destination address (assumed to NOT be in same page)
farcall MACRO dest
	NOEXPAND
	jump dest, TRUE, TRUE  ;destination not in same page
	ENDM

;far goto (destination address not in same page):
;this cannot be used for instructions immediately following a btf conditional skip instruction (only room for one instruction)
;params:
; dest = destination address (assumed to NOT be in same page)
fargoto MACRO dest
	NOEXPAND
	jump dest, TRUE, FALSE  ;destination not in same page
	ENDM


;return and do not re-enable interrupts:
;src label controls bank affinity upon return
;params:
; src = address where we came in thru (determines bank affinity upon return)
return_ MACRO src
	NOEXPAND
	returnFROM src, FALSE
	ENDM

;return and re-enable interrupts:
;src label controls bank affinity upon return
;params:
; src = address where we came in thru (determines bank affinity upon return)
retfie_ MACRO src
	NOEXPAND
	returnFROM src, TRUE
	ENDM


;jump to a destination address (call or goto):
;first checks if dest address is in range
;can be used for near/far/any jump addresses
;destination address must use jumptarget macro to specify bank affinity to help avoid save/restore overhead of RP0+RP1
;params:
; dest = destination address; can be near or far
; fixpage = flag telling whether to set PCLATH: TRUE/FALSE for far/near addresses, MAYBE for unknown (PCLATH will be set only if needed)
; want_ret = flag telling whether or not to return: TRUE/FALSE = yes/no
jump MACRO dest, fixpage, want_ret
	NOEXPAND  ;hide clutter in LST file
	LOCAL DestBankAffinity = InBankOf(dest)  ;must be declared before usage
	LOCAL RetBankAffinity = OutBankOf(dest)  ;must be declared before usage
	LOCAL DEST = dest, FIXPAGE = fixpage, WANT_RET  ;DEBUG
	if TRUE
																				EXPAND  ;show CALL/GOTO params
PREVPAGE = PREVPAGE  ;show last known page select (shown for debug)
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
DestBankAffinity = DestBankAffinity  ;show destination bank affinity
RetBankAffinity = RetBankAffinity  ;show returning bank affinity
DEST = dest  ;show CALL/GOTO params
FIXPAGE = fixpage ^ (IsPCLHSafe(dest)<<16)  ;show CALL/GOTO params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	WARNIF (($ < noop12500) || ($ > noop4)) && ((dest) == ($+1)), [INFO] Warning: useless goto#v(dest)
	if BANKOF(DestBankAffinity) != BT_NONBANKED
		BANKCHK DestBankAffinity, MAYBE,  ;set RP0+RP1 to known state refore jumping
	endif
;this doesn't really save anything; skip it:
;	if ((dest) == $+2) && (REGPAGEOF($) == REGPAGEOF(dest))  ;no savings if PCLATH is not already set
;		WARNIF TRUE, "[CAUTION] Check for correct jump here", dest, $
;																				EXPAND  ;show GOTO params + generated code
;svPREVPAGE = svPREVPAGE  ;show last known page select (shown for debug)
;DEST = dest  ;show GOTO params
;FIXPAGE = fixpage  ;show GOTO params
;	incf PCL, F  ;short-cut; faster exec than goto?

#ifdef EXPLICIT_PCLATH  ;skip all the fancy optimization; just set it
;	PAGECHK dest, TRUE,  ;force PCLATH
	if LITPAGEOF(MAX_ROM) & 2
		setbit PCLATH, 4, LITPAGEOF(dest) & 2
	endif
	if LITPAGEOF(MAX_ROM) & 1
		setbit PCLATH, 3, LITPAGEOF(dest) & 1
	endif
	SETPAGE $, FALSE  ;we've already set PCLATH; defeat auto-checking
#else  ;def EXPLICIT_PCLATH
	PAGECHK dest, fixpage,  ;does not alter W; this is important for params passed thru W
#endif  ;def EXPLICIT_PCLATH
;assembler always seems to generate page select during first pass, so second pass gives error 116 "Address label duplicated or different in second pass"
#if 0  ;this is more concise, but less readable in LST file:
	JUMPRET(dest, want_ret)  ;CAUTION: PCLATH<4:3> must be correct for this to jump to the right place
#else  ;0
	if !(want_ret)
																				EXPAND  ;show generated code
	goto dest  ;go and don't come back
		NOEXPAND
	else
																				EXPAND  ;show generated code
	call dest  ;go and come back later
		NOEXPAND
	endif
#endif  ;0
	if !(want_ret)
		EXPAND_POP,
		EXITM
	endif
#ifdef EXPLICIT_PCLATH  ;restore PCLATH for caller
	if LITPAGEOF(MAX_ROM) & 2
		setbit PCLATH, 4, LITPAGEOF($) & 2
	endif
	if LITPAGEOF(MAX_ROM) & 1
		setbit PCLATH, 3, LITPAGEOF($) & 1
	endif
	SETPAGE $, FALSE  ;we've already set PCLATH; defeat auto-checking
#else  ;def EXPLICIT_PCLATH
	if !IsPCLHSafe(dest)  ;dest did not preserve PCLATH
		SETPAGE -1, TRUE  ;PCLATH is unknown after return
	endif
#endif  ;def EXPLICIT_PCLATH
	SETBANK RetBankAffinity, TRUE  ;if caller had unspecified bank affinity, this will set current bank to unknown after return
	EXPAND_POP,
	ENDM


;return from a function:
;to avoid save/restore overhead of RP0+RP1, caller/callee can use a bank select convention (affinity)
;bank affinity was specified on entry point, so that must be passed as a parameter here
;params:
;  src = jumpdest being returned from (used to determine bank select convention when returning to caller)
; intenable = flag telling whether to enable interrupts: TRUE/FALSE
returnFROM MACRO src, intenable
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
	LOCAL SrcBankAffinity = OutBankOf(src)  ;must be declared before usage; was offset by 1
	if BANKOF(SrcBankAffinity) != BT_NONBANKED
		BANKCHK SrcBankAffinity, MAYBE,  ;set RP0+RP1 to known state refore jumping
	endif
	if IsPCLHSafe(src)
		PAGECHK $, MAYBE,  ;caller expects PCLATH to be unchanged, so restore it here
	endif
#if 0  ;this is more concise, but less readable in LST file:
	RETURNIE(intenable)  ;retfie/return
#else  ;0
	if intenable
																				EXPAND  ;show generated code
	retfie  ;return and enable ints
		NOEXPAND
	else
																				EXPAND  ;show generated code
	return  ;return without enabling ints
		NOEXPAND
	endif
#endif  ;0
	EXPAND_POP,
	ENDM


#define InBankOf_(name)  IBA#v(0)name
#define OutBankOf_(name)  OBA#v(0)name
#define IsPCLHSafe(name)  PCLH#v(0)name
#define InBankOf(name)  (IBA#v(0)name-1)
#define OutBankOf(name)  (OBA#v(0)name-1)

;declare bank affinity for a named destination address:
;this just sets a const which can be checked later to control conditional bank selects; value is offset by 1 to avoid -1; #v is used just for symbol concatenation
;bank affinity for forward references must be declared before used, since it affects whether bank selects need to be done prior to goto/call
;params:
;  name = symbolic name of label to set bank affinity for
;  inbankreg = bank select convention to use when jumping to the named address (optional)
;  outbankreg = bank select convention to use when returning from the named address (optional); defaults to inbankreg
;  pclh_safe = flag telling whether PCLATH is altered by dest (optional); defaults to FALSE
;  fwd = flag telling if this is a forware reference (optional, defaults to TRUE); TRUE/FALSE == yes/no, MAYBE == yes but use a LOCAL rather than CONSTANT
SetBankAffinity MACRO name, inbankreg, outbankreg, pclh_safe, fwd
	NOEXPAND  ;reduce clutter
	LOCAL InBank = inbankreg 0  ;optiona param; must end with "| if passed
	if !InBank
InBank = BT_UNKNOWN
	endif
	LOCAL OutBank = outbankreg 0  ;optiona param; must end with "| if passed
	if !OutBank
OutBank = InBank
	endif
	LOCAL unsafe = (pclh_safe 0) & (LONGJUMP_HINT 0)  ;line too long
	if fwd -1  ;this is a forward reference (no ":" after name)
		if (fwd -1) != TRUE
			VARIABLE InBankOf_(name) = (InBank) + 1  ;BANKOF(inbankreg 0) + 1  ;offset by 1 for consistency with other case below
			VARIABLE OutBankOf_(name) = (OutBank) + 1  ;BANKOF(OutBank) + 1  ;offset by 1 for consistency with other case below
			VARIABLE IsPCLHSafe(name) = BOOL2INT(unsafe == 0)  ;remember whether PCLATH needs to be restored
		else
			CONSTANT InBankOf_(name) = (InBank) + 1  ;BANKOF(inbankreg 0) + 1  ;offset by 1 for consistency with other case below
			CONSTANT OutBankOf_(name) = (OutBank) + 1  ;BANKOF(OutBank) + 1  ;offset by 1 for consistency with other case below
			CONSTANT IsPCLHSafe(name) = BOOL2INT(unsafe == 0)  ;remember whether PCLATH needs to be restored
		endif
	else  ;in-place declaration (name ends with a ":")
		CBLOCK (InBank) + 1  ;BANKOF(inbankreg 0) + 1  ;offset by 1 to avoid invalid address
		InBankOf_(name) 0  ;name ends with ":"; specify length of 0 to avoid allocating memory
		ENDC
		CBLOCK OutBank + 1  ;BANKOF(OutBank) + 1  ;offset by 1 to avoid invalid address
		OutBankOf_(name) 0  ;name ends with ":"; specify length of 0 to avoid allocating memory
		ENDC
		CBLOCK BOOL2INT(unsafe == 0)  ;remember whether PCLATH needs to be restored
		IsPCLHSafe(name) 0  ;name ends with ":"; specify length of 0 to avoid allocating memory
		ENDC
	endif
	ENDM


	VARIABLE PREVJUMP = 0  ;remember location of previous jump

;define a call or goto destination with consistent bank switching:
;this allows bank select housekeeping to be automatic even with conditional execution paths
;to get the smallest code space, bank affinity for the majority of callers should be used
;to get the fastest execution, bank affinity of the most frequent caller should be used
;fall-thrus count as a caller with respect to frequency or majority

;params:
;  name = symbolic label to define; caller should put ":" after the name parameter; this allows easier searching for labels within the source code
;  inbankreg1 = bank select convention to use when jumping to this address (optional if already declared); caller is responsible for following this
;  inbankreg2 = bank select double-check (optional); if both reg1 and reg2 are passed, they must be in the same bank or an error will be flagged; this allows for easy checking of multiple paths for consistency
;  outbankreg = bank select convention upon return (optional); defaults to inbound bank affinity
;  flags = controls whether execution can fall thru from above this label, whether bank affinity was forward-declared, or whether dest alters PCLATH (optional, all default to TRUE)
;optional params must have a trailing "|" if they are passed; otherwise MPASM will report a syntax error
jumpdest MACRO name, inbankreg1, inbankreg2, outbankreg, flags
	NOEXPAND  ;hide clutter in LST file
	LOCAL InBankAffinity1 = inbankreg1 0, InBankAffinity2 = inbankreg2 0, FLAGS = flags 0  ;optional params; must end with "|" if passed; 0 is used because 0 is not a useful register address
	LOCAL OutBankAffinity = outbankreg 0  ;optional param; must end with "|" if passed
	EXPAND_PUSH FALSE
	if !InBankAffinity1 || ISLIT(InBankAffinity1)
InBankAffinity1 = IIF(FLAGS & (FALLTHRU_BANKSEL 0), PREVBANK, BT_UNKNOWN)  ;use currently selected bank for fall-thru unless caller specified otherwise
	endif
	if !InBankAffinity2 || ISLIT(InBankAffinity2)
InBankAffinity2 = IIF(FLAGS & (FALLTHRU_BANKSEL 0), PREVBANK, BT_UNKNOWN)  ;use currently selected bank for fall-thru unless caller specified otherwise
	endif
	if !OutBankAffinity  ;not specified; default to inbound bank affinity
OutBankAffinity = InBankAffinity1
	endif
	LOCAL BA1 = BANKOF(InBankAffinity1), BA2 = BANKOF(InBankAffinity2)  ;line too long
	ERRIF (BA1 != BT_NONBANKED) && (BA2 != BT_NONBANKED) && (BA1 != BA2), [ERROR] Conflicting bank affinity: #v(BA1):#v(InBankAffinity1) vs. #v(BA2):#v(InBankAffinity2)  ;must match if both params passed
	LOCAL BankMismatch = (PREVJUMP == $) && (BANKOF(PREVBANK) != BA1)
	WARNIF BankMismatch, [CAUTION] WARNING: Conflicting bank selects #v(BA1):#v(InBankAffinity1) vs. #v(BANKOF(PREVBANK)):#v(PREVBANK) at name #v($)
;	if !(FLAGS & (FWDREF 0))  ;caller did not specify FWDREF
;looks like MPASM doesn't complain about CONSTANT/VARIABLE and CBLOCK symbols with the same name, so just do this all the time:
		SetBankAffinity name, InBankAffinity1|, OutBankAffinity|, (FLAGS & (LONGJUMP_HINT 0))|, FALSE&  ;set it in case not set yet (so callers know how to get here)
;	endif
	if FLAGS & (FALLTHRU_PAGESEL 0)
		WARNIF LITPAGEOF($) != LITPAGEOF($ + 4), [INFO] Label pre-padded to avoid 1K-word page boundary at #v($)..#v($ + 4)  ;make sure PAGECHK uses actual destination page
		while LITPAGEOF($) != LITPAGEOF($ + 4)
			nop  ;pad up to next 1K-word page to avoid crossing page boundary (simplifies address arithmetic)
		endw
		PAGECHK $, MAYBE,  ;make page select consistent when falling thru
	else
		SETPAGE $, FALSE  ;remember PCLATH; PCLATH<4:3> must have been set to current address since we jumped here successfully, but PCLATH<2:0> might be incorrect
	endif
	if (FLAGS & (FALLTHRU_BANKSEL 0)) && (BANKOF(InBankAffinity1) != BT_NONBANKED)
		BANKCHK InBankAffinity1, MAYBE,  ;make bank select consistent when falling thru
	else  ;no fall-thru; assume caller-specified bank is correct
		SETBANK InBankAffinity1, FALSE  ;override cached bank
	endif
																				EXPAND  ;show generated code
InBankAffinity1 = InBankAffinity1  ;show inbound bank affinity
OutBankAffinity = OutBankAffinity  ;show returning bank affinity
FLAGS = FLAGS & (LONGJUMP_HINT FWDREF FALLTHRU_PAGESEL FALLTHRU_BANKSEL 0)  ;show LONGJUMP_HINT=8 FWDREF=4 FALLTHRU_PAGESEL=2 FALLTHRU_BANKSEL=1 flags
name  ;can call or goto this label, or maybe fall thru
PREVBANK = PREVBANK  ;show currently selected bank
PREVPAGE = PREVPAGE  ;show currently selected page
	NOEXPAND  ;reduce clutter
PREVJUMP = $  ;remember prev jump location; this helps detect redundant or conflicting settings
	EXPAND_POP,
	ENDM


;call a function, but store return address in register instead of hardware stack:
;this is NOT recursize; local variables are used for return address and parameters
;params:
; dest = destination address; can be near or far
; stack = register to hold return address (16 bits)
;stackcall MACRO dest, stack
;	NOEXPAND
;	EXPAND_PUSH FALSE
;	LOCAL retadrs  ;RETADRS = $+5  ;assember losses track of addresses, so we have to help it here
;	mov16 _LITERAL(retadrs), stack  ;save return address in register rather than on the stack
;	goto_ dest
;retadrs:
;	EXPAND_POP,
;	ENDM


;return from function:
;params:
; stack = register holding return address (16 bits)
;stackreturn MACRO stack
;	NOEXPAND
;	EXPAND_PUSH FALSE
;	mov16 stack, PCL
;	SETBANK -1, FALSE  ;PCLATH is probably wrong for any code following this, so just forget it
;	EXPAND_POP,
;	ENDM


	VARIABLE IFSTMT, ELSESTMT  ;addresses of parts of if/else block, in case caller wants to jump to them from outside

;conditional if/else block execution:
;set up 2 (optional) instruction blocks, and jump around them as needed
;MPASM doesn't provide good support for optional macro params, which made this more complicated than it otherwise would be; it was a fun exercise, though
;this could probably be optimized more
;params:
;  ifstmt = statement to execute if preceding conditional instruction was true (fall thru)
;  elsestmt = statement to execute if preceding conditional instruction was false (jump over)
;  hints = execution hints; zero or more of:
;          BIGSTMT_HINT ==> ifstmt is more than 1 instr and must be jumped over (not needed if elsestmt used)
;          BOTHOK_HINT ==> okay to execute elsestmt first in order to avoid jump over (speculative exec, reduces code size); in this case ifstmt still needs to be 1 instruction (but BIGSTMT_HINT can override this)
;          NOJUMP_HINT ==> PCLATH does not need to be valid to execute ifstmt; BIGSTMT_HINT will override this
;          NORET_HINT ==> execution does not return or fall thru, so bank/page select does not need to be preserved
ifelseblock1 MACRO ifstmt, elsestmt, hints
	NOEXPAND  ;reduce clutter
	EXPAND_PUSH FALSE
	LOCAL HINTS = hints 0  ;optional param; must end with "|" if passed
;KLUDGE: assembler occassionally generates a duplicate, non-unique suffix on LOCALs (probably a bug in hash algorithm or symbol table manager)
;this causes annoying error 116 "Address label duplicated or different in second pass", and also messes up forward references (leaves symbolic addresses with values of 0, which will cause a jump to 0 so PIC will restart)
;to work around this MPASM problem, we FORCE the suffixes to be unique by explicitly generating unique symbols for LOCALs
	LOCAL UNIQID = NEXT_UNIQID 
NEXT_UNIQID += 1
	LOCAL istrue#v(UNIQID), isfalse#v(UNIQID), isdone#v(UNIQID)  ;KLUDGE: force LOCALs to be unique, in case assembler doesn't generate a unique hash
;execute next instr if condition was TRUE:
;there is no (documented) way to determine whether an optional parameter is passed to a macro, so we have to figure it out by expanding the parameter and then checking if it generated any instructions
	LOCAL SAVE_BANK = PREVBANK, START_IFELSE = $  ;save current bank and page so it can be restored if optional execution paths change it (needs to be in a consistent state upon macro exit)
	LOCAL NOTMY_HINTS = HINTS & ~(NOJUMP_HINT BIGSTMT_HINT BOTHOK_HINT NORET_HINT 0);FWDREF LONGJUMP_HINT 0)  ;line too long
	WARNIF NOTMY_HINTS, [WARNING] Ignoring unrecognized IFELSEBLOCK hint #v(NOTMY_HINTS)
	LEAVE_PLACEHOLDER1 jump_over_elsestmt_OR_jump_over_ifstmt_OR_ifstmt  ;this instruction will vary according to size of if/else stmts
	LOCAL MYPH1 = PLACEHOLDER_STACK  ;remember top of placeholder stack in case elsestmt adds more
 jumpdest isfalse#v(UNIQID):,,,, FALLTHRU  ;jump or fall-thru to here if condition is false
	if !(HINTS & (BOTHOK_HINT 0))  ;elsestmt was not already executed (speculatively)
																				EXPAND  ;show generated code
	elsestmt
		NOEXPAND
	endif
	LOCAL HAS_ELSESTMT = ($ != isfalse#v(UNIQID))
	if HAS_ELSESTMT
		LOCAL ELSE_BANK = PREVBANK  ;remember selected bank after elsestmt
		LOCAL FALLTHRU_STYLE = IIF(HINTS & (NOJUMP_HINT 0), FALLTHRU 0, FALLTHRU_PAGESEL 0)  ;line too long; doesn't really fall thru if there is a jump or no stmt there
		if !(HINTS & (NORET_HINT 0))
			LEAVE_PLACEHOLDER1JUMP neargoto isdone#v(UNIQID)  ;jump over ifstmt
			LOCAL MYPH2 = PLACEHOLDER_STACK  ;remember top of placeholder stack in case ifstmt adds more
		else
FALLTHRU_STYLE = 0
		endif
 jumpdest istrue#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;(FALLTHRU_STYLE & ~(FALLBOK_HINT 0))|  ;& ~FALLTHRU  ;jump to here if condition is true; use same bank as before elsestmt so jump over won't require bank select (only reserved 1 word for jump)
																				EXPAND  ;show generated code
	ifstmt
		NOEXPAND
		LOCAL HAS_IFSTMT = ($ != istrue#v(UNIQID))
		if !(HINTS & (NORET_HINT 0))
FALLTHRU_STYLE = IIF(HAS_IFSTMT && (HINTS & (NOJUMP_HINT 0)), FALLTHRU 0, FALLTHRU_PAGESEL 0)  ;line too long; doesn't really fall thru if there is a jump or no stmt there
		endif
		if HAS_IFSTMT
 jumpdest isdone#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;11 jump or fall-thru to here when done; use same bank as before elsestmt so jump over won't require bank select (only reserved 1 word for jump)
			if !(HINTS & (NORET_HINT 0))
				FILL_PLACEHOLDER MYPH2&, neargoto isdone#v(UNIQID)  ;jump over ifstmt
			endif
			FILL_PLACEHOLDER MYPH1&, neargoto istrue#v(UNIQID)  ;jump over elsestmt
		else  ;don't need the trailing goto around ifstmt
			if !(HINTS & (NORET_HINT 0))
				FILL_PLACEHOLDER MYPH2&, ASM_NOP  ;didn't need this one; reclaim unused instruction space
			endif
 jumpdest isdone#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;12 jump or fall-thru to here when done; use same bank as before elsestmt so jump over won't require bank select (only reserved 1 word for jump)
			FILL_PLACEHOLDER MYPH1&, neargoto isdone#v(UNIQID)  ;jump over elsestmt
		endif
	endif
	if (!HAS_ELSESTMT) && (HINTS & (BIGSTMT_HINT 0))  ;IFs nested too deep; repeat IF conditions
		LEAVE_PLACEHOLDER1 neargoto isdone#v(UNIQID)  ;jump over ifstmt
		LOCAL MYPH2 = PLACEHOLDER_STACK  ;remember top of placeholder stack in case ifstmt adds more
 jumpdest istrue#v(UNIQID):, SAVE_BANK|,,,  ;jump to here if condition is true; use same bank as before elsestmt so jump over won't require bank select (only reserved 1 word for jump)
																				EXPAND  ;show generated code
	ifstmt
		NOEXPAND
		LOCAL HAS_IFSTMT = ($ != istrue#v(UNIQID))
		WARNIF !HAS_IFSTMT, [WARNING] Don't need BIGSTMT_HINT if there is no IF stmt
		LOCAL FALLTHRU_STYLE = IIF(HINTS & (NOJUMP_HINT 0), FALLTHRU 0, FALLTHRU_PAGESEL 0)  ;line too long; doesn't really fall thru if there is a jump or no stmt there
 jumpdest isdone#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;13 jump or fall-thru to here when done; use same bank as before ifstmt so jump over won't require bank select (only reserved 1 word for jump)
		FILL_PLACEHOLDER MYPH2&, neargoto isdone#v(UNIQID)  ;jump around if condition is false
		FILL_PLACEHOLDER MYPH1&, neargoto istrue#v(UNIQID)  ;jump around if condition is false
	endif
	if (!HAS_ELSESTMT) && !(HINTS & (BIGSTMT_HINT 0))  ;IFs nested too deep; repeat IF conditions
istrue#v(UNIQID) = isfalse#v(UNIQID) - 1
		START_FILL_PLACEHOLDER MYPH1&,
																				EXPAND  ;show generated code
	ifstmt  ;ifstmt must fit in 1 instruction unless BIGSTMT_HINT or elsestmt specified
		NOEXPAND
		LOCAL HAS_IFSTMT = ($ != istrue#v(UNIQID))
		LOCAL FALLTHRU_STYLE = IIF(HAS_IFSTMT && (HINTS & (NOJUMP_HINT 0)), FALLTHRU 0, 0)  ;line too long; doesn't really fall thru if there is a jump or no stmt there
 jumpdest isdone#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;14 jump or fall-thru to here when done; use same bank as before ifstmt so jump over won't require bank select (only reserved 1 word for jump)
		END_FILL_PLACEHOLDER MYPH1&, STRICT_OR_EMPTY(MYPH1)
	endif
	ERRIF LITPAGEOF($) != LITPAGEOF(START_IFELSE), [ERROR] if/else block spans page boundary: starts #v(START_IFELSE) and ends #v($)
;CAUTION: these are globals so caller can use them, but that means they must be set at macro exit (else a recursive macro call would undo previously set values)
	SetBankAffinity IFSTMT, SAVE_BANK|,,, MAYBE&  ;allow caller to jump to ifstmt
	SetBankAffinity ELSESTMT, SAVE_BANK|,,, MAYBE&  ;allow caller to jump to elsestmt
IFSTMT = IIF(HAS_IFSTMT, istrue#v(UNIQID), isdone#v(UNIQID))  ;allow caller to jump in from elsewhere
ELSESTMT = IIF(HAS_ELSESTMT, isfalse#v(UNIQID), isdone#v(UNIQID))  ;allow caller to jump in from elsewhere
	EXPAND_POP,
	ENDM
#if 1  ;more ugliness; KLUDGE: params get trashed in recursve macros, so we define a couple more copies here to try to avoid that problem
ifelseblock2 MACRO ifstmt, elsestmt, hints
	NOEXPAND  ;reduce clutter
	EXPAND_PUSH FALSE
	LOCAL HINTS = hints 0
;KLUDGE: assembler occassionally generates a duplicate, non-unique suffix on LOCALs (probably a bug in hash algorithm or symbol table manager)
;this causes annoying error 116 "Address label duplicated or different in second pass", and also messes up forward references (leaves symbolic addresses with values of 0, which will cause a jump to 0 so PIC will restart)
;to work around this MPASM problem, we FORCE the suffixes to be unique by explicitly generating unique symbols for LOCALs
	LOCAL UNIQID = NEXT_UNIQID 
NEXT_UNIQID += 1
	LOCAL istrue#v(UNIQID), isfalse#v(UNIQID), isdone#v(UNIQID)  ;KLUDGE: force LOCALs to be unique, in case assembler doesn't generate a unique hash
;execute next instr if condition was TRUE:
;there is no (documented) way to determine whether an optional parameter is passed to a macro, so we have to figure it out by expanding the parameter and then checking if it generated any instructions
	LOCAL SAVE_BANK = PREVBANK, START_IFELSE = $  ;save current bank and page so it can be restored if optional execution paths change it (needs to be in a consistent state upon macro exit)
	LOCAL NOTMY_HINTS = HINTS & ~(NOJUMP_HINT BIGSTMT_HINT BOTHOK_HINT NORET_HINT 0);FWDREF LONGJUMP_HINT 0)  ;line too long
	WARNIF NOTMY_HINTS, [WARNING] Ignoring unrecognized IFELSEBLOCK hint #v(NOTMY_HINTS)
	LEAVE_PLACEHOLDER1 jump_over_elsestmt_OR_jump_over_ifstmt_OR_ifstmt  ;this instruction will vary according to size of if/else stmts
	LOCAL MYPH1 = PLACEHOLDER_STACK  ;remember top of placeholder stack in case elsestmt adds more
 jumpdest isfalse#v(UNIQID):,,,, FALLTHRU  ;jump or fall-thru to here if condition is false
	if !(HINTS & (BOTHOK_HINT 0))  ;elsestmt was not already executed (speculatively)
																				EXPAND  ;show generated code
	elsestmt
		NOEXPAND
	endif
	LOCAL HAS_ELSESTMT = ($ != isfalse#v(UNIQID))
	if HAS_ELSESTMT
		LOCAL ELSE_BANK = PREVBANK  ;remember selected bank after elsestmt
		LOCAL FALLTHRU_STYLE = IIF(HINTS & (NOJUMP_HINT 0), FALLTHRU 0, FALLTHRU_PAGESEL 0)  ;line too long; doesn't really fall thru if there is a jump or no stmt there
		if !(HINTS & (NORET_HINT 0))
			LEAVE_PLACEHOLDER1JUMP neargoto isdone#v(UNIQID)  ;jump over ifstmt
			LOCAL MYPH2 = PLACEHOLDER_STACK  ;remember top of placeholder stack in case ifstmt adds more
		else
FALLTHRU_STYLE = 0
		endif
 jumpdest istrue#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;(FALLTHRU_STYLE & ~(FALLBOK_HINT 0))|  ;& ~FALLTHRU  ;jump to here if condition is true; use same bank as before elsestmt so jump over won't require bank select (only reserved 1 word for jump)
																				EXPAND  ;show generated code
	ifstmt
		NOEXPAND
		LOCAL HAS_IFSTMT = ($ != istrue#v(UNIQID))
		if !(HINTS & (NORET_HINT 0))
FALLTHRU_STYLE = IIF(HAS_IFSTMT && (HINTS & (NOJUMP_HINT 0)), FALLTHRU 0, FALLTHRU_PAGESEL 0)  ;line too long; doesn't really fall thru if there is a jump or no stmt there
		endif
		if HAS_IFSTMT
 jumpdest isdone#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;11 jump or fall-thru to here when done; use same bank as before elsestmt so jump over won't require bank select (only reserved 1 word for jump)
			if !(HINTS & (NORET_HINT 0))
				FILL_PLACEHOLDER MYPH2&, neargoto isdone#v(UNIQID)  ;jump over ifstmt
			endif
			FILL_PLACEHOLDER MYPH1&, neargoto istrue#v(UNIQID)  ;jump over elsestmt
		else  ;don't need the trailing goto around ifstmt
			if !(HINTS & (NORET_HINT 0))
				FILL_PLACEHOLDER MYPH2&, ASM_NOP  ;didn't need this one; reclaim unused instruction space
			endif
 jumpdest isdone#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;12 jump or fall-thru to here when done; use same bank as before elsestmt so jump over won't require bank select (only reserved 1 word for jump)
			FILL_PLACEHOLDER MYPH1&, neargoto isdone#v(UNIQID)  ;jump over elsestmt
		endif
	endif
	if (!HAS_ELSESTMT) && (HINTS & (BIGSTMT_HINT 0))  ;IFs nested too deep; repeat IF conditions
		LEAVE_PLACEHOLDER1 neargoto isdone#v(UNIQID)  ;jump over ifstmt
		LOCAL MYPH2 = PLACEHOLDER_STACK  ;remember top of placeholder stack in case ifstmt adds more
 jumpdest istrue#v(UNIQID):, SAVE_BANK|,,,  ;jump to here if condition is true; use same bank as before elsestmt so jump over won't require bank select (only reserved 1 word for jump)
																				EXPAND  ;show generated code
	ifstmt
		NOEXPAND
		LOCAL HAS_IFSTMT = ($ != istrue#v(UNIQID))
		WARNIF !HAS_IFSTMT, [WARNING] Don't need BIGSTMT_HINT if there is no IF stmt
		LOCAL FALLTHRU_STYLE = IIF(HINTS & (NOJUMP_HINT 0), FALLTHRU 0, FALLTHRU_PAGESEL 0)  ;line too long; doesn't really fall thru if there is a jump or no stmt there
 jumpdest isdone#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;13 jump or fall-thru to here when done; use same bank as before ifstmt so jump over won't require bank select (only reserved 1 word for jump)
		FILL_PLACEHOLDER MYPH2&, neargoto isdone#v(UNIQID)  ;jump around if condition is false
		FILL_PLACEHOLDER MYPH1&, neargoto istrue#v(UNIQID)  ;jump around if condition is false
	endif
	if (!HAS_ELSESTMT) && !(HINTS & (BIGSTMT_HINT 0))  ;IFs nested too deep; repeat IF conditions
istrue#v(UNIQID) = isfalse#v(UNIQID) - 1
		START_FILL_PLACEHOLDER MYPH1&,
																				EXPAND  ;show generated code
	ifstmt  ;ifstmt must fit in 1 instruction unless BIGSTMT_HINT or elsestmt specified
		NOEXPAND
		LOCAL HAS_IFSTMT = ($ != istrue#v(UNIQID))
		LOCAL FALLTHRU_STYLE = IIF(HAS_IFSTMT && (HINTS & (NOJUMP_HINT 0)), FALLTHRU 0, 0)  ;line too long; doesn't really fall thru if there is a jump or no stmt there
 jumpdest isdone#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;14 jump or fall-thru to here when done; use same bank as before ifstmt so jump over won't require bank select (only reserved 1 word for jump)
		END_FILL_PLACEHOLDER MYPH1&, STRICT_OR_EMPTY(MYPH1)
	endif
	ERRIF LITPAGEOF($) != LITPAGEOF(START_IFELSE), [ERROR] if/else block spans page boundary: starts #v(START_IFELSE) and ends #v($)
;CAUTION: these are globals so caller can use them, but that means they must be set at macro exit (else a recursive macro call would undo previously set values)
	SetBankAffinity IFSTMT, SAVE_BANK|,,, MAYBE&  ;allow caller to jump to ifstmt
	SetBankAffinity ELSESTMT, SAVE_BANK|,,, MAYBE&  ;allow caller to jump to elsestmt
IFSTMT = IIF(HAS_IFSTMT, istrue#v(UNIQID), isdone#v(UNIQID))  ;allow caller to jump in from elsewhere
ELSESTMT = IIF(HAS_ELSESTMT, isfalse#v(UNIQID), isdone#v(UNIQID))  ;allow caller to jump in from elsewhere
	EXPAND_POP,
	ENDM
ifelseblock3 MACRO ifstmt, elsestmt, hints
	NOEXPAND  ;reduce clutter
	EXPAND_PUSH FALSE
	LOCAL HINTS = hints 0
;KLUDGE: assembler occassionally generates a duplicate, non-unique suffix on LOCALs (probably a bug in hash algorithm or symbol table manager)
;this causes annoying error 116 "Address label duplicated or different in second pass", and also messes up forward references (leaves symbolic addresses with values of 0, which will cause a jump to 0 so PIC will restart)
;to work around this MPASM problem, we FORCE the suffixes to be unique by explicitly generating unique symbols for LOCALs
	LOCAL UNIQID = NEXT_UNIQID 
NEXT_UNIQID += 1
	LOCAL istrue#v(UNIQID), isfalse#v(UNIQID), isdone#v(UNIQID)  ;KLUDGE: force LOCALs to be unique, in case assembler doesn't generate a unique hash
;execute next instr if condition was TRUE:
;there is no (documented) way to determine whether an optional parameter is passed to a macro, so we have to figure it out by expanding the parameter and then checking if it generated any instructions
	LOCAL SAVE_BANK = PREVBANK, START_IFELSE = $  ;save current bank and page so it can be restored if optional execution paths change it (needs to be in a consistent state upon macro exit)
	LOCAL NOTMY_HINTS = HINTS & ~(NOJUMP_HINT BIGSTMT_HINT BOTHOK_HINT NORET_HINT 0);FWDREF LONGJUMP_HINT 0)  ;line too long
	WARNIF NOTMY_HINTS, [WARNING] Ignoring unrecognized IFELSEBLOCK hint #v(NOTMY_HINTS)
	LEAVE_PLACEHOLDER1 jump_over_elsestmt_OR_jump_over_ifstmt_OR_ifstmt  ;this instruction will vary according to size of if/else stmts
	LOCAL MYPH1 = PLACEHOLDER_STACK  ;remember top of placeholder stack in case elsestmt adds more
 jumpdest isfalse#v(UNIQID):,,,, FALLTHRU  ;jump or fall-thru to here if condition is false
	if !(HINTS & (BOTHOK_HINT 0))  ;elsestmt was not already executed (speculatively)
																				EXPAND  ;show generated code
	elsestmt
		NOEXPAND
	endif
	LOCAL HAS_ELSESTMT = ($ != isfalse#v(UNIQID))
	if HAS_ELSESTMT
		LOCAL ELSE_BANK = PREVBANK  ;remember selected bank after elsestmt
		LOCAL FALLTHRU_STYLE = IIF(HINTS & (NOJUMP_HINT 0), FALLTHRU 0, FALLTHRU_PAGESEL 0)  ;line too long; doesn't really fall thru if there is a jump or no stmt there
		if !(HINTS & (NORET_HINT 0))
			LEAVE_PLACEHOLDER1JUMP neargoto isdone#v(UNIQID)  ;jump over ifstmt
			LOCAL MYPH2 = PLACEHOLDER_STACK  ;remember top of placeholder stack in case ifstmt adds more
		else
FALLTHRU_STYLE = 0
		endif
 jumpdest istrue#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;(FALLTHRU_STYLE & ~(FALLBOK_HINT 0))|  ;& ~FALLTHRU  ;jump to here if condition is true; use same bank as before elsestmt so jump over won't require bank select (only reserved 1 word for jump)
																				EXPAND  ;show generated code
	ifstmt
		NOEXPAND
		LOCAL HAS_IFSTMT = ($ != istrue#v(UNIQID))
		if !(HINTS & (NORET_HINT 0))
FALLTHRU_STYLE = IIF(HAS_IFSTMT && (HINTS & (NOJUMP_HINT 0)), FALLTHRU 0, FALLTHRU_PAGESEL 0)  ;line too long; doesn't really fall thru if there is a jump or no stmt there
		endif
		if HAS_IFSTMT
 jumpdest isdone#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;11 jump or fall-thru to here when done; use same bank as before elsestmt so jump over won't require bank select (only reserved 1 word for jump)
			if !(HINTS & (NORET_HINT 0))
				FILL_PLACEHOLDER MYPH2&, neargoto isdone#v(UNIQID)  ;jump over ifstmt
			endif
			FILL_PLACEHOLDER MYPH1&, neargoto istrue#v(UNIQID)  ;jump over elsestmt
		else  ;don't need the trailing goto around ifstmt
			if !(HINTS & (NORET_HINT 0))
				FILL_PLACEHOLDER MYPH2&, ASM_NOP  ;didn't need this one; reclaim unused instruction space
			endif
 jumpdest isdone#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;12 jump or fall-thru to here when done; use same bank as before elsestmt so jump over won't require bank select (only reserved 1 word for jump)
			FILL_PLACEHOLDER MYPH1&, neargoto isdone#v(UNIQID)  ;jump over elsestmt
		endif
	endif
	if (!HAS_ELSESTMT) && (HINTS & (BIGSTMT_HINT 0))  ;IFs nested too deep; repeat IF conditions
		LEAVE_PLACEHOLDER1 neargoto isdone#v(UNIQID)  ;jump over ifstmt
		LOCAL MYPH2 = PLACEHOLDER_STACK  ;remember top of placeholder stack in case ifstmt adds more
 jumpdest istrue#v(UNIQID):, SAVE_BANK|,,,  ;jump to here if condition is true; use same bank as before elsestmt so jump over won't require bank select (only reserved 1 word for jump)
																				EXPAND  ;show generated code
	ifstmt
		NOEXPAND
		LOCAL HAS_IFSTMT = ($ != istrue#v(UNIQID))
		WARNIF !HAS_IFSTMT, [WARNING] Don't need BIGSTMT_HINT if there is no IF stmt
		LOCAL FALLTHRU_STYLE = IIF(HINTS & (NOJUMP_HINT 0), FALLTHRU 0, FALLTHRU_PAGESEL 0)  ;line too long; doesn't really fall thru if there is a jump or no stmt there
 jumpdest isdone#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;13 jump or fall-thru to here when done; use same bank as before ifstmt so jump over won't require bank select (only reserved 1 word for jump)
		FILL_PLACEHOLDER MYPH2&, neargoto isdone#v(UNIQID)  ;jump around if condition is false
		FILL_PLACEHOLDER MYPH1&, neargoto istrue#v(UNIQID)  ;jump around if condition is false
	endif
	if (!HAS_ELSESTMT) && !(HINTS & (BIGSTMT_HINT 0))  ;IFs nested too deep; repeat IF conditions
istrue#v(UNIQID) = isfalse#v(UNIQID) - 1
		START_FILL_PLACEHOLDER MYPH1&,
																				EXPAND  ;show generated code
	ifstmt  ;ifstmt must fit in 1 instruction unless BIGSTMT_HINT or elsestmt specified
		NOEXPAND
		LOCAL HAS_IFSTMT = ($ != istrue#v(UNIQID))
		LOCAL FALLTHRU_STYLE = IIF(HAS_IFSTMT && (HINTS & (NOJUMP_HINT 0)), FALLTHRU 0, 0)  ;line too long; doesn't really fall thru if there is a jump or no stmt there
 jumpdest isdone#v(UNIQID):, SAVE_BANK|,,, FALLTHRU_STYLE|  ;14 jump or fall-thru to here when done; use same bank as before ifstmt so jump over won't require bank select (only reserved 1 word for jump)
		END_FILL_PLACEHOLDER MYPH1&, STRICT_OR_EMPTY(MYPH1)
	endif
	ERRIF LITPAGEOF($) != LITPAGEOF(START_IFELSE), [ERROR] if/else block spans page boundary: starts #v(START_IFELSE) and ends #v($)
;CAUTION: these are globals so caller can use them, but that means they must be set at macro exit (else a recursive macro call would undo previously set values)
	SetBankAffinity IFSTMT, SAVE_BANK|,,, MAYBE&  ;allow caller to jump to ifstmt
	SetBankAffinity ELSESTMT, SAVE_BANK|,,, MAYBE&  ;allow caller to jump to elsestmt
IFSTMT = IIF(HAS_IFSTMT, istrue#v(UNIQID), isdone#v(UNIQID))  ;allow caller to jump in from elsewhere
ELSESTMT = IIF(HAS_ELSESTMT, isfalse#v(UNIQID), isdone#v(UNIQID))  ;allow caller to jump in from elsewhere
	EXPAND_POP,
	ENDM
#endif


;computed goto setup:
;all bits of PCLATH must be set when modifying PCL via arithmetic statements
;params:
; gotoc_adrs = start address of computer goto
GotocSetup MACRO gotoc_adrs
	NOEXPAND
	mov8 _LITERAL(HIGH(gotoc_adrs)), PCLATH
	SETPAGE gotoc_adrs, FALSE  ;update page tracker with new PCLATH contents
	ENDM


	VARIABLE CASE_CTL, PREV_CASE, CASE_COUNT = 0

;case statement setup:
;params:
; ctlvalue = value used in case evaluation
START_CASE MACRO ctlvalue
	NOEXPAND  ;hide clutter in LST file
	loadw_ ctlvalue
CASE_CTL = ctlvalue
CASE_COUNT += 0x100-1  ;macro perf
PREV_CASE = 0
	ENDM


;evaluate a case:
;optimized for literal values to avoid extra W loads
;params:
; op = operand
; cmpvalue = value to compare to
; stmt = statement to execute if case is true
; hint = execution hints (optional); pass-thru to ifelse_
CASE MACRO op, cmpvalue, stmt, hints
	NOEXPAND  ;reduce clutter
	LOCAL CMPVAL = cmpvalue
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show CASE params
CMPVAL = cmpvalue  ;show CASE params
PREV_CASE = PREV_CASE  ;show CASE params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	ERRIF !ISLIT(cmpvalue), [ERROR] Case value must be literal: #v(cmpvalue)
	LOCAL CASE_DIFF = _LITERAL(LOW(UNLIT(cmpvalue) - PREV_CASE))  ;line too long
	if8 WREG, op, CASE_DIFF, stmt,, hints
CASE_COUNT += 1  ;measure macro usage/perf
PREV_CASE = UNLIT(cmpvalue)  ;keep track of previous value, to avoid reloading W each time
	EXPAND_POP,
	ENDM


;dummy macros for readability:

DEFAULT_CASE MACRO stmt  ;dummy macro, just for readability
	stmt
	ENDM

END_CASE MACRO  ;dummy macro, just for readability
	NOEXPAND
	ENDM


;define a function prototype:
;also allocates RAM for local/parameter variables
;DEFFUNC MACRO name, numparams
;	LOCAL n = 0
;	CBLOCK NEXT_BRAM
;	while n < (numparams)
;		name_#v(n)
;n += 1
;	endw
;	ENDC
;	SAFE_ENDC TRUE
;name:
;	ENDM


;call a function:
;CALLFUNC MACRO name, first, second, third, fourth
;#ifdef name_3
;	mov8 fourth, name_3
;#endif
;#ifdef name_2
;	mov8 third, name_2
;#endif
;#ifdef name_1
;	mov8 second, name_1
;#endif
;#ifdef name_0
;	mov8 first, name_0
;#endif
;	call_ name
;	ENDM


;******************************************************************************
;bit manipulation macros and pseudo-opcodes:
;these opcodes are all bank- and page- safe, and show resulting opcodes in .LST file
;******************************************************************************

;value -> bit map (provides a log2 function):
;usage: BIT_#v(bit#) = power of 2
;equivalent to the following definitions:
;#define BIT_65536  d'16'
;#define BIT_32768  d'15'
;#define BIT_16384  d'14'
;#define BIT_8192  d'13'
;#define BIT_4096  d'12'
;#define BIT_2048  d'11'
;#define BIT_1024  d'10'
;#define BIT_512  9
;#define BIT_256  8
;#define BIT_128  7
;#define BIT_64  6
;#define BIT_32  5
;#define BIT_16  4
;#define BIT_8  3
;#define BIT_4  2
;#define BIT_2  1
;#define BIT_1  0
;#define BIT_0  0
	VARIABLE bit = -1
	while bit <= d'16'  ;-1, 0, 1, 2, ..., 16
		CONSTANT BIT_#v(IIF(bit < 0, 0, 1<<bit)) = IIF(bit < 0, 0, bit)
bit += 1
	endw
	ERRIF BIT_1024 != d'10', [ERROR] BIT_ constants are bad: BIT_1024 == #v(BIT_1024) but should be #v(1<<d'10')  ;paranoid self-check
	ERRIF BIT_1 | BIT_0, [ERROR] BIT_ constants are bad: BIT_1 == #v(BIT_1) and BIT_0 == #v(BIT_0) but should be 0  ;paranoid self-check


;count number of bits and highest bit in a byte value:
;MPASM chokes on line lengths > 255, so we need to define a bunch of constants here
;usage: NUMBITS_#v(value) = #bits on in value, HIGHBIT_#v(value) = highest bit# on (8 == msb, 1 == lsb)
	VARIABLE bitvals = 0, numbits, lastbit
	while bitvals < 256  ;0, 1, 2, ..., 255
lastbit = 0
numbits = 0
bit = 0
		while bit < 8  ;0, 1, ..., 7
			if bitvals & (1<<bit)
lastbit = bit + 1
numbits += 1
			endif
bit += 1
		endw
		CONSTANT NUMBITS_#v(bitvals) = numbits
		CONSTANT HIGHBIT_#v(bitvals) = lastbit
bitvals += 1
	endw
	ERRIF NUMBITS_0, [ERROR] NUMBITS_ constants are bad: NUMBITS_0 == #v(NUMBITS_0) but should be 0  ;paranoid self-check
	ERRIF NUMBITS_4 != 1, [ERROR] NUMBITS_ constants are bad: NUMBITS_4 == #v(NUMBITS_4) but should be 1  ;paranoid self-check
	ERRIF NUMBITS_42 != 3, [ERROR] NUMBITS_ constants are bad: NUMBITS_42 == #v(NUMBITS_42) but should be 3  ;paranoid self-check
	ERRIF NUMBITS_255 != 8, [ERROR] NUMBITS_ constants are bad: NUMBITS_255 == #v(NUMBITS_255) but should be 8  ;paranoid self-check
	ERRIF HIGHBIT_0, [ERROR] HIGHBIT_ constants are bad: HIGHBIT_0 == #v(HIGHBIT_0) but should be 0  ;paranoid self-check
	ERRIF HIGHBIT_4 != 3, [ERROR] HIGHBIT_ constants are bad: HIGHBIT_4 == #v(HIGHBIT_4) bu should be 1  ;paranoid self-check
	ERRIF HIGHBIT_42 != 6, [ERROR] HIGHBIT_ constants are bad: HIGHBIT_42 == #v(HIGHBIT_42) but should be 6  ;paranoid self-check
	ERRIF HIGHBIT_255 != 8, [ERROR] HIGHBIT_ constants are bad: HIGHBIT_255 == #v(HIGHBIT_255) but should be 8  ;paranoid self-check


	VARIABLE BITDCL_COUNT_#v(BOOL2INT(TRUE)) = 0, BITDCL_COUNT_#v(BOOL2INT(FALSE)) = 0, BITVAR_INIT = 0

;allocate a 1-bit variable (cuts down on memory usage):
;currently bit variables are allocated in non-banked RAM, to cut down on bank selects
;caller is responsible for initializing bit vbls
;params:
; name = symbolic name to allocate
BITDCL MACRO name  ;, banked
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
	LOCAL banked = FALSE  ;don't need this param; hard-code to OFF
	if !(BITDCL_COUNT_#v(banked) % 8)  ;allocate more storage space
		if banked
			BDCL BITDCL#v(BITDCL_COUNT_#v(banked) / 8)
		else
			NBDCL BITDCL#v(BITDCL_COUNT_#v(banked) / 8)
		endif
		WARNIF2 BITVAR_INIT, [CAUTION] Bit vars already initialized, this one #v(BITDCL_COUNT_#v(banked)) won't be
	endif
	EXPAND_PUSH TRUE  ;show bit variable definition
	CONSTANT name = BITDCL_COUNT_#v(banked)  ;remember where the bit is
	EXPAND_POP,
BITDCL_COUNT_#v(banked) += 1
	EXPAND_POP,
	ENDM


;setbit parameters for a variable allocated via BITDCL:
#define BITVAR(name)  BITDCL#v((name)/8), (name)%8
#define BITVADRS(name)  BITDCL#v((name)/8)


;clear all BITDCL variables:
CLEAR_BITDCLS MACRO
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
;	mov#v(8*divup(BITDCL_COUNT, 8)) _LITERAL(0), BITDCL0  ;this will only work if they are consecutive
	LOCAL bit = 0
	while bit < BITDCL_COUNT_#v(BOOL2INT(FALSE))
		mov8 _LITERAL(0), BITVADRS(bit)  ;clear bit var storage space
bit += 8
	endw
BITVAR_INIT = $  ;remember when it was initialized
	EXPAND_POP,
	ENDM


;copy a bit:
;CAUTION: uses speculative copy, so dest bit may be set twice
;based on a tip from Microchip to reset bit first, then set it conditionally
;this results in fewer instructions (3 instead of 4)
;params:
; srcreg = register containing bit to copy
; srcbit = which bit in srcreg to copy
; destreg = register where bit is to be copied to
; destbit = which bit in destreg to copy to
bitcopy MACRO srcreg, srcbit, destreg, destbit
	NOEXPAND  ;hide clutter in LST file
#if 0  ;line too long
	ifbit srcreg, srcbit, TRUE, SETBIT(destreg, destbit, TRUE), SETBIT(destreg, destbit, FALSE), BOTHOK_HINT
#else
	setbit destreg, destbit, FALSE  ;line too long
	ifbit srcreg, srcbit, TRUE, SETBIT(destreg, destbit, TRUE),,
#endif
	ENDM


;reverse a bit:
;params:
; reg = register containing bit to reverse
; bit = which bit in reg to reverse
; hints = execution hints; zero or more of:
;          ZPRESERVE_HINT ==> caller wants to preserve STATUS.Z; affects how efficiently bit can be reversed
revbit MACRO reg, bit, hints
	NOEXPAND  ;hide clutter in LST file
	if (hints 0) & (ZPRESERVE_HINT 0)  ;need to preserve STATUS.Z
		ifbit reg, bit, TRUE, SETBIT(reg, bit, FALSE), SETBIT(reg, bit, TRUE), NOJUMP_HINT  ;4 instr (btf + jump + bsf/bcf); can't use BOTHOK_HINT here
	else
		xor8 reg, _LITERAL(1<<(bit)), reg,  ;this is faster than testing and setting the bit (2 instr instead of 4), but alters WREG and STATUS.Z
	endif
	ENDM


;set/reset a bit:
;params:
; reg = register containing bit to set/reset
; bit = which bit to set/reset in reg
; val = whether to set or reset bit; 0 = reset, !0 = set
setbit MACRO reg, bit, val
	NOEXPAND  ;reduce clutter
	LOCAL REG = reg
	LOCAL BIT = bit
	LOCAL VALUE = val
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show SETBIT params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
REG = reg  ;show SETBITS params
BIT = bit  ;show SETBIT params
VALUE = val  ;show SETBIT params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	if REG == WREG  ;special case
		if VALUE
																				EXPAND  ;show generated code
	iorlw 1<<(bit)  ;turn the bit ON
			NOEXPAND
		else
																				EXPAND  ;show generated code
	andlw ~(1<<(bit))  ;turn the bit OFF
			NOEXPAND
		endif
		EXPAND_POP,
		EXITM
	endif
	BANKCHK REG, MAYBE,
#if 0  ;this is more concise, but less readable in LST file:
	SETBIT(REG, BIT, VALUE)  ;set bit on/off
#else  ;0
;use macro params rather than locals here, so they show in LST file; this causes recursion problems, though
	if !VALUE
																				EXPAND  ;show generated code
	bcf reg, bit  ;turn bit OFF
		NOEXPAND
	else
																				EXPAND  ;show generated code
	bsf reg, bit  ;turn bit ON
		NOEXPAND
	endif
#endif  ;0
	EXPAND_POP,
	ENDM
#if 1  ;more ugliness; KLUDGE: params get trashed in recursive macros, so we define a couple more copies here to try to avoid that
setbit2 MACRO reg, bit, val
	NOEXPAND ;reduce clutter
	LOCAL REG = reg
	LOCAL BIT = bit
	LOCAL VALUE = val
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show SETBIT params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
REG = reg  ;show SETBITS params
BIT = bit  ;show SETBIT params
VALUE = val  ;show SETBIT params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	if REG == WREG  ;special case
		if VALUE
																				EXPAND  ;show generated code
	iorlw 1<<(bit)  ;turn the bit ON
			NOEXPAND
		else
																				EXPAND  ;show generated code
	andlw ~(1<<(bit))  ;turn the bit OFF
			NOEXPAND
		endif
		EXPAND_POP,
		EXITM
	endif
	BANKCHK REG, MAYBE,
#if 0  ;this is more concise, but less readable in LST file:
	SETBIT(REG, BIT, VALUE)  ;set bit on/off
#else  ;0
;use macro params rather than locals here, so they show in LST file; this causes recursion problems, though
	if !VALUE
																				EXPAND  ;show generated code
	bcf reg, bit  ;turn bit OFF
		NOEXPAND
	else
																				EXPAND  ;show generated code
	bsf reg, bit  ;turn bit ON
		NOEXPAND
	endif
#endif  ;0
	EXPAND_POP,
	ENDM
#endif


;values for STATUS.C after add/sub:
	CONSTANT C_CARRY = TRUE
	CONSTANT C_NOCARRY = FALSE
	CONSTANT C_BORROW = FALSE
	CONSTANT C_NOBORROW = TRUE


;ifbit params to use for common tests:
#define EQZ  STATUS, Z, TRUE
#define NEZ  STATUS, Z, FALSE
#define EQUALS  STATUS, Z, TRUE  ;used after a subtract/compare
#define NOT_EQUAL  STATUS, Z, FALSE  ;used after a subtract/compare

#define LTZ  STATUS, C, C_BORROW  ;used after a subtract/compare
#define GEZ  STATUS, C, C_NOBORROW  ;used after a subtract/compare
#define LESS_THAN  STATUS, C, C_BORROW  ;used after a subtract/compare
#define GREATER_EQUAL  STATUS, C, C_NOBORROW  ;used after a subtract/compare
#define CARRY  STATUS, C, C_CARRY
#define NOBORROW  STATUS, C, C_NOBORROW
#define BORROW  STATUS, C, C_BORROW
#define NOCARRY  STATUS, C, C_NOCARRY


;comma-less wrappers for status bit re/sets:
setz_ MACRO
	NOEXPAND
	setbit EQUALS
	ENDM

clrz_ MACRO
	NOEXPAND
	setbit NOT_EQUAL
	ENDM

;skpc MACRO
;	btfss STATUS, C
;	ENDM

;skpnc MACRO
;	btfsc STATUS, C
;	ENDM

;complement W:
;comw MACRO
;	sublw   0
;;	xorlw   0xff  ;first invert W
;;	addlw   1  ;then add 1 to get 2's complement
;	ENDM

clrc_ MACRO
	NOEXPAND
	setbit NOCARRY
	ENDM

setc_ MACRO
	NOEXPAND
	setbit CARRY
	ENDM


;do something until a bit value changes:
;params:
; reg = register containing bit to check
; bit = which bit to check
; val = bit value to check
; whilestmt = statements to execute while bit is at specified value (optional)
; hints = execution hints; zero or more of:
;          EQUAL_HINT ==> caller wants STATUS.Z to be correct upon return
;          COMPARE_HINT ==> caller wants STATUS.C to be correct upon return
;          UNTIL_HINT ==> execute stmt UNTIL condition is true, rather than WHILE
whilebit MACRO reg, bit, val, whilestmt, hints
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
;KLUDGE: assembler occassionally generates a duplicate, non-unique suffix on LOCALs (probably a bug in hash algorithm or symbol table manager)
;this causes annoying error 116 "Address label duplicated or different in second pass", and also messes up forward references (leaves symbolic addresses with values of 0, which will cause a jump to 0 so PIC will restart)
;to work around this ASSEMBLER problem, we FORCE the suffixes to be unique by explicitly generating unique symbols for LOCALs
	LOCAL UNIQID = NEXT_UNIQID 
NEXT_UNIQID += 1
	LOCAL beforestmt#v(UNIQID), whileloop#v(UNIQID), HINTS = hints 0  ;optional param; must end with "|" if passed
	WARNIF HINTS & ~(EQUAL_HINT COMPARE_HINT UNTIL_HINT 0), [WARNING] Unrecognized whilebit hints: #v(HINTS & ~(EQUAL_HINT COMPARE_HINT 0))
	if !(HINTS & (UNTIL_HINT 0))  ;check condition before executing whilestmt; this uses an extra jump at start, but avoids it at the end and also avoids inverting the relational compare
		LEAVE_PLACEHOLDER1BANK reg, neargoto whileloop#v(UNIQID)
		LOCAL MYPH = PLACEHOLDER_STACK  ;remember top of placeholder stack in case whilestmt adds more
	endif
 jumpdest beforestmt#v(UNIQID):, reg|,,, IIF(HINTS & (UNTIL_HINT 0), FALLTHRU 0, 0)|  ;(FALLTHRU_STYLE & (LONGJUMP_HINT 0))|
																				EXPAND  ;show generated code
	whilestmt
	NOEXPAND
;now go back and fill in around it:
	if $ == beforestmt#v(UNIQID)  ;no while stmt
		if !(HINTS & (UNTIL_HINT 0))  ;check condition before executing whilestmt
			FILL_PLACEHOLDER MYPH&, ASM_NOP  ;placeholder not needed; reclaim space
		endif
 jumpdest whileloop#v(UNIQID):,,,, FALLTHRU  ;bank already selected
		ifbit reg, bit, val, neargoto whileloop#v(UNIQID),, (HINTS & ~(LONGJUMP_HINT 0))|
	else  ;while stmt
 jumpdest whileloop#v(UNIQID):,,,, FALLTHRU  ;bank already selected
		ifbit reg, bit, val, neargoto beforestmt#v(UNIQID),, (HINTS & ~(LONGJUMP_HINT 0))|
		if !(HINTS & (UNTIL_HINT 0))  ;check condition before executing whilestmt
			FILL_PLACEHOLDER MYPH&, neargoto whileloop#v(UNIQID)
		endif
	endif
	EXPAND_POP,
	ENDM


	VARIABLE NESTED_IFBIT = 0  ;track nested macros; MPASM trashes params, so warn about this

;test a bit value:
;ifstmt must be 1 word if there is no elsestmt; otherwise if/else stmt can be any size
;PAGECHK ifstmt/elsestmt needs to call PAGECHK at the end if PCLATH has been altered; this macro can only set it at the start
;this macro can be recursive, so params MUST be copied to local vbls (otherwise MPASM bug trashes params as stack unwinds)
;CAUTION: THOROUGHLY check any changes made here; this macro is used extensively; always check .LST file to verify it does what you think it's doing
;params:
; reg = register containing bit to check
; bit = which bit to check
; val = bit value to check
; ifstmt = statements to execute if bit is at specified value (optional)
; elsestmt = statements to execute if bit is *not* at specified value (optional)
; hints = execution hints; zero or more of:
;          BIGSTMT_HINT ==> ifstmt is more than 1 instr and must be jumped over (not needed if elsestmt used)
;          BOTHOK_HINT ==> okay to execute elsestmt first in order to avoid jump over (reduces code size); in this case ifstmt still needs to be 1 instruction (but BIGSTMT_HINT can override this)
;          NOJUMP_HINT ==> PCLATH does not need to be valid to execute ifstmt; BIGSTMT_HINT will override this
ifbit MACRO reg, bit, val, ifstmt, elsestmt, hints
	NOEXPAND  ;reduce LST clutter
	LOCAL REG = reg
	LOCAL BIT = bit
	LOCAL VALUE = val
	LOCAL HINTS = hints 0  ;optional param; must end with "|" if passed
	WARNIF NESTED_IFBIT && FULL_WARN, [CAUTION] Recursive #v(NESTED_IFBIT) usage of ifbit: params may be trashed  ;MPASM trashes params, so warn about this
NESTED_IFBIT += 1
	LOCAL NOTMY_HINTS = HINTS & ~(NOJUMP_HINT BOTHOK_HINT BIGSTMT_HINT 0)  ;line too long
	WARNIF NOTMY_HINTS, [WARNING] Ignoring unrecognized IFBIT hint(s): #v(NOTMY_HINTS)
	if TRUE  ;MEXPAND_STACKLO & 1
																				EXPAND  ;show IFBIT params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
REG = reg  ;show IFBIT params
BIT = bit  ;show IFBIT params
VALUE = val  ;show IFBIT params
HINTS = hints 0  ;show IFBIT params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	if HINTS & (BOTHOK_HINT 0)  ;speculative IF; execute else stmt first to cut down on instructions (assumed to be benign)
																				EXPAND  ;show generated code
	elsestmt  ;do this one first to avoid the need to jump around it later if condition is false
		NOEXPAND
	endif
	if (HINTS & (NOJUMP_HINT BIGSTMT_HINT 0)) != (NOJUMP_HINT 0)  ;need to jump over
;NOTE: it appears that PCLATH does NOT need to be correct when executing an BTFSC/BTFSS instruction, so only set it if conditional instruction needs it
		PAGECHK $, MAYBE,  ;allow neargoto around if/else stmt
	endif
	if (HINTS & (BIGSTMT_HINT BOTHOK_HINT 0)) == (BIGSTMT_HINT 0)
VALUE = !VALUE  ;statement block is too big to follow btf immediately; reverse the compare and jump over if false (this will avoid a jump-over if there is no else stmt)
	endif
	BANKCHK REG, MAYBE,
#if 0  ;this is more concise, but less readable in LST file:
	TESTBIT(REG, BIT, !VALUE)  ;test bit and execute next instruction only if bit is NOT the specified value
#else  ;0
	if VALUE
																				EXPAND  ;show generated code
	btfsc reg, bit  ;do next instr only if bit is ON
		NOEXPAND
	else
																				EXPAND  ;show generated code
	btfss reg, bit  ;do next instr only if bit is OFF
		NOEXPAND
	endif
#endif  ;0
	if (HINTS & (BIGSTMT_HINT BOTHOK_HINT 0)) == (BIGSTMT_HINT 0)  ;swap stmts
		ifelseblock#v(NESTED_IFBIT + NESTED_INCDEC) elsestmt, ifstmt, (HINTS & ~(BIGSTMT_HINT 0))|
	else  ;stmts as-is
		ifelseblock#v(NESTED_IFBIT + NESTED_INCDEC) ifstmt, elsestmt, HINTS|
	endif
	EXPAND_POP,
NESTED_IFBIT -= 1
	ENDM


;convenience wrapper:
bitblt8 MACRO src, offbits, onbits, dest
	NOEXPAND
	bitblt_ src, offbits, onbits, dest, 1
	ENDM

;turn bits on/off (somewhat optimized):
;uses bcf/bsf for single bits
;uses and/ior for multiple bits
;TODO: need to check this for read-modify-write safety
;params:
; src = original value
; offbits = bitmask of bits to turn off
; onbits = bitmask of bits to turn on
; dest = destination register
; count = #bytes to set
bitblt_ MACRO src, offbits, onbits, dest, count
	NOEXPAND  ;hide clutter in LST file
	LOCAL SRC = src
	LOCAL OFFBITS = offbits
	LOCAL ONBITS = onbits
	LOCAL DEST = dest
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show BITBLT params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
SRC = src  ;show BITBLT params
OFFBITS = offbits  ;show BITBLT params
ONBITS = onbits  ;show BITBLT params
DEST = dest  ;show BITBLT params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	WARNIF (!ONBITS) && (!OFFBITS), [CAUTION] BITBLT: no bits to turn on or off
	ERRIF ONBITS & OFFBITS, [ERROR] BITBLT: bit on/off conflict #v(ONBITS) vs. #v(OFFBITS)
	ERRIF (ONBITS | OFFBITS) & ~0xff, [ERROR] BITBLT: too many bits to turn on or off in #v(ONBITS) and #v(OFFBITS)
	LOCAL DESTBITS = IIF(ISPORT(DEST), PORT_BITS#v(PORTOF#v(DEST)), 0xff)  ;some ports don't have all bits
	if (ONBITS | OFFBITS) == DEST_BITS  ;set the entire register
		mov8 _LITERAL(ONBITS), DEST
		EXPAND_POP,
		EXITM
	endif
	if (NUMBITS_#v(OFFBITS) == 1) && (SRC == DEST)
		setbit DEST, BIT_#v(offbits), FALSE
	else
		if OFFBITS  ;2 or more bits to turn off; at least one bit will remain on, since the OFFBITS == 0xff case was handled earrlier
			and8 SRC, _LITERAL((offbits) ^ 0xff), DEST  ;turn off the bits
SRC = DEST
		endif
	endif
	if (NUMBITS_#v(ONBITS) == 1) && (SRC == DEST)
		setbit DEST, BIT_#v(onbits), TRUE
	else
		if ONBITS  ;2 or more bits to turn on; at least one bit will remain off, since the ONBITS == 0xff case was handled earlier
			ior8 SRC, _LITERAL(onbits), DEST  ;turn on the bits
SRC = DEST
		endif
	endif
	EXPAND_POP,
	ENDM


;******************************************************************************
;conditional compare and branching macros/functions:
;helps to reduce late-night errors, but LOOK AT THE LISTING FILE to make sure they are doing what they are supposed to
;******************************************************************************

;assembler doesn't like special chars as macro params, so use #defines:
#define LT  <
#define EQ  ==
#define GT  >
#define LE  <=
#define NE  !=
#define GE  >=

;there's no way to test directly for a relational operator (due to MPASM syntax limitations), but it can be deduced through specially chosen test cases:
#define isLEorEQorGE(relop)  (0 relop 0)  ;relop is one of: <=, ==, >=
#define isLTorLEorNE(relop)  (0 relop 1)  ;relop is one of: <, <=, =
#define isGTorGEorNE(relop)  (1 relop 0)  ;relop is one of: >, >=, =


;hints for conditional statements:
;these are optional parameters, so they are constructed as ORs
#define NO_HINTS  0|
#define FALLTHRU_BANKSEL  1|  ;execution can fall thru from above; make bank select consistent
#define FALLTHRU_PAGESEL  2|  ;execution can fall thru from above; make PCLATH consistent
#define FALLTHRU  (FALLTHRU_BANKSEL FALLTHRU_PAGESEL 0)|  ;make both bank select + PCLATH consistent when falling thru
#define FWDREF  4|  ;bank affinity was already declared
#define LONGJUMP_HINT  8|  ;PCLATH is changed
#define NORET_HINT  0x10|  ;execution does not return/fall thru
#define NOJUMP_HINT  0x100|  ;neither condition jumps, so PCLATH does not need to be set
#define BOTHOK_HINT  0x200|  ;it's safe to execute both branches (avoids extra jump around true stmt)
#define EQUAL_HINT  0x400|  ;only need to test equality; need to set STATUS.Z but don't need to set STATUS.C
#define COMPARE_HINT  0x800|  ;need to set both STATUS.Z and STATUS.C
#define BIGSTMT_HINT  0x1000|  ;first conditional stmt is more than 1 instruction; need to jump around it; USE ONLY IF NOT SUPPLYING AN ELSE STMT AND NOT BOTHOK, otherwise generates incorrectly reversed jump
#define MORESHIFT_HINT  0x2000|  ;continuation from previous shift; no need to clear STATUS.C
#define UNTIL_HINT 0x4000|  ;execute whilestmt first before checking exit condition
;#define IFNORET_HINT  0x4000|  ;if stmt never falls thru
;#define ELSENORET_HINT  0x8000|  ;else stmt never falls thru
;aliases:
#define ZPRESERVE_HINT  EQUAL_HINT  ;need to preserve STATUS.Z
#define CPRESERVE_HINT  COMPARE_HINT  ;need to preserve STATUS.C


;convenience wrappers:
while8 MACRO op1, relop, op2, whilestmt, hints
	while_ op1, 1, relop, op2, 1, whilestmt, hints
	ENDM

while16 MACRO op1, relop, op2, whilestmt, hints
	while_ op1, 2, relop, op2, 2, whilestmt, hints
	ENDM

while24 MACRO op1, relop, op2, whilestmt, hints
	while_ op1, 3, relop, op2, 3, whilestmt, hints
	ENDM

while32 MACRO op1, relop, op2, whilestmt, hints
	while_ op1, 4, relop, op2, 4, whilestmt, hints
	ENDM


;do something while a condition is true:
;params:
; op1 = first operand
; op1bytes = size of first operand
; relop = relational operator: <, =, >, <=, !=, >=
; op2 = second operand
; op2bytes = size of second operand
; whilestmt = statement(s) to execute while condition is true
; hints = execution hints; zero or more of:
;          EQUAL_HINT ==> caller wants STATUS.Z to be correct upon return
;          COMPARE_HINT ==> caller wants STATUS.C to be correct upon return
;          UNTIL_HINT ==> execute stmt UNTIL condition is true, rather than WHILE
while_ MACRO op1, op1bytes, relop, op2, op2bytes, whilestmt, hints
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
	LOCAL beforestmt, whileloop, HINTS = hints 0  ;optional param; must end with "|" if passed
	WARNIF HINTS & ~(EQUAL_HINT COMPARE_HINT 0), [WARNING] Unrecognized While hints: #v(HINTS & ~(EQUAL_HINT COMPARE_HINT 0))
	if !(HINTS & (UNTIL_HINT 0))  ;check condition before executing whilestmt; this uses an extra jump at start, but avoids it at the end and also avoids inverting the relational compare
		LEAVE_PLACEHOLDER1BANK op1, neargoto whileloop
		LOCAL MYPH = PLACEHOLDER_STACK  ;remember top of placeholder stack in case whilestmt adds more
	endif
 jumpdest beforestmt:, op1|, op2|,, IIF(HINTS & (UNTIL_HINT 0), FALLTHRU 0, 0)|  ;(FALLTHRU_STYLE & (LONGJUMP_HINT 0))|
																				EXPAND  ;show generated code
	whilestmt
	NOEXPAND
;now go back and fill in around it:
	if $ == beforestmt  ;no while stmt
		if !(HINTS & (UNTIL_HINT 0))  ;check condition before executing whilestmt
			FILL_PLACEHOLDER MYPH&, ASM_NOP  ;placeholder not needed; reclaim space
		endif
 jumpdest whileloop:,,,, FALLTHRU  ;bank already selected
		ifelse_ op1, op1bytes, relop, op2, op2bytes, neargoto whileloop,, (HINTS & ~(LONGJUMP_HINT 0))|
	else  ;while stmt
 jumpdest whileloop:,,,, FALLTHRU  ;bank already selected
		ifelse_ op1, op1bytes, relop, op2, op2bytes, neargoto beforestmt,, (HINTS & ~(LONGJUMP_HINT 0))|
		if !(HINTS & (UNTIL_HINT 0))  ;check condition before executing whilestmt
			FILL_PLACEHOLDER MYPH&, neargoto whileloop
		endif
	endif
	EXPAND_POP,
	ENDM


;convenience wrappers:

if8 MACRO op1, relop, op2, ifstmt, elsestmt, hints
	NOEXPAND
	ifelse_ op1, 1, relop, op2, 1, ifstmt, elsestmt, hints
	ENDM

if16 MACRO op1, relop, op2, ifstmt, elsestmt, hints
	NOEXPAND
	ifelse_ op1, 2, relop, op2, 2, ifstmt, elsestmt, hints
	ENDM

if24 MACRO op1, relop, op2, ifstmt, elsestmt, hints
	NOEXPAND
	ifelse_ op1, 3, relop, op2, 3, ifstmt, elsestmt, hints
	ENDM

if32 MACRO op1, relop, op2, ifstmt, elsestmt, hints
;	NOEXPAND
	EXPAND
	ifelse_ op1, 4, relop, op2, 4, ifstmt, elsestmt, hints
	ENDM


	VARIABLE NESTED_IFELSE = 0  ;track nested macros; MPASM trashes params, so warn about this

;in-line if statement (conditional statement execution):
;NOTE: ifstmt must be 1 word if there is no elsestmt; otherwise if/else can be any size
;minor optimation is done: ifelse_ X > lit is transformed to ifelse_ X >= lit+1
;NOTE: ifstmt and elsestmt will be swapped for X > non-lit
;CAUTION: THOROUGHLY check any changes made here; this macro is used extensively; always check .LST file to verify it does what you think it's doing
;relational operators must be passed as symbols, or else MPASM will complain about syntax
;params:
; op1 = first operand
; op1bytes = size of first operand
; relop = relational operator: <, =, >, <=, !=, >= (must use symbols to pass this)
; op2 = second operand
; op2bytes = size of second operand
; ifstmt = statement(s) to execute if condition is true
; elsestmt = statement(s) to execute if condition is false
; hints = execution hints; zero or more of:
;          EQUAL_HINT ==> caller wants STATUS.Z to be correct upon return
;          COMPARE_HINT ==> caller wants STATUS.C to be correct upon return
;          BIGSTMT_HINT ==> ifstmt is more than 1 instr and must be jumped over (not needed if elsestmt used)
;          BOTHOK_HINT ==> okay to execute elsestmt first in order to avoid jump over (speculative exec, reduces code size); in this case ifstmt still needs to be 1 instruction (but BIGSTMT_HINT can override this)
;          NOJUMP_HINT ==> PCLATH does not need to be valid to execute ifstmt; BIGSTMT_HINT will override this
ifelse_ MACRO op1, op1bytes, relop, op2, op2bytes, ifstmt, elsestmt, hints
	NOEXPAND  ;hide clutter in LST file
	WARNIF NESTED_IFELSE, [CAUTION] Recursive #v(NESTED_IFELSE) usage of ifelse: params may be trashed  ;MPASM trashes params, so warn about this
NESTED_IFELSE += 1
	EXPAND_PUSH FALSE
	LOCAL NOTMY_HINTS = (hints 0) & ~(COMPARE_HINT EQUAL_HINT NOJUMP_HINT BIGSTMT_HINT BOTHOK_HINT 0)  ;line too long
	WARNIF NOTMY_HINTS, [WARNING] Ignoring unrecognized IF hint #v(NOTMY_HINTS)
	LOCAL CMP_HINTS = ((hints 0) & (COMPARE_HINT EQUAL_HINT 0))  ;line too long
;relational operator can be inferred by checking how it handles 0 and 1;
;I saw this trick somewhere on the web; sorry, i didn't write it down at the time; if you know, please let me know so i can add a refernce to it
#if 1  ;IFs nested too deep; this flattens them out, but is harder to read
	LOCAL islit2 = ISLIT(op2)  ;line too long
	if UNLIT(op2) >= (1<<(8*op2bytes)) - 1  ;line too long
islit2 = FALSE  ;const will wrap so +1 won't work
	endif
	if (0 relop 0) && (0 relop 1) && islit2  ; <= lit; convert to < +1 to avoid testing both Z and C
		EXPAND_PUSH TRUE  ;show what this does
		cmp_ op1, op1bytes, _LITERAL(UNLIT(op2) + 1), op2bytes, COMPARE_HINT CMP_HINTS|
		ifbit BORROW, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
		EXPAND_POP,
		EXPAND_POP,
NESTED_IFELSE -= 1
		EXITM
	endif
	if (0 relop 0) && (0 relop 1)  ; <= reg; must check Z and C
		EXPAND_PUSH TRUE  ;show what this does
		cmp_ op1, op1bytes, op2, op2bytes, EQUAL_HINT COMPARE_HINT CMP_HINTS|
;CAUTION: 2 btf instructions are "chained" together here
		TESTBIT(STATUS, C, FALSE)  ;skip next instruction if STATUS.C is OFF (BORROW on); was < so skip the = check
		ifbit EQUALS, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
		EXPAND_POP,
		EXPAND_POP,
NESTED_IFELSE -= 1
		EXITM
	endif
	if (0 relop 0) && (1 relop 0)  ; >=
		EXPAND_PUSH TRUE  ;show what this does
		cmp_ op1, op1bytes, op2, op2bytes, COMPARE_HINT CMP_HINTS|
		ifbit NOBORROW, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
		EXPAND_POP,
		EXPAND_POP,
NESTED_IFELSE -= 1
		EXITM
	endif
	if (0 relop 0) && ((op2) == LIT0) && ((op1bytes) == 1)  ;special case: == 0
		if (op1) == WREG  ;KLUDGE: update STATUS.Z so bit test will be correct
																				EXPAND  ;show generated code
	addlw 0
			NOEXPAND
		else
			EXPAND_PUSH TRUE  ;show what this does
			loadw_ op1  ;just load W to compare to 0
			EXPAND_POP,
		endif
		ifbit EQUALS, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
		EXPAND_POP,
NESTED_IFELSE -= 1
		EXITM
	endif
	if (0 relop 0)  ; ==
		EXPAND_PUSH TRUE  ;show what this does
		cmp_ op1, op1bytes, op2, op2bytes, EQUAL_HINT CMP_HINTS|
		ifbit EQUALS, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
		EXPAND_POP,
		EXPAND_POP,
NESTED_IFELSE -= 1
		EXITM
	endif
	if (0 relop 1) && (1 relop 0) && ((op2) == LIT0) && ((op1bytes) == 1)  ;special case: != 0
		if (op1) == WREG  ;KLUDGE: update STATUS.Z so bit test will be correct
																				EXPAND  ;show generated code
	addlw 0
			NOEXPAND
		else
			EXPAND_PUSH TRUE  ;show what this does
			loadw_ op1  ;just load W to compare to 0
			EXPAND_POP,
		endif
		ifbit NOT_EQUAL, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
		EXPAND_POP,
NESTED_IFELSE -= 1
		EXITM
	endif
	if (0 relop 1) && (1 relop 0)  ; !=
		EXPAND_PUSH TRUE  ;show what this does
		cmp_ op1, op1bytes, op2, op2bytes, EQUAL_HINT CMP_HINTS|
		ifbit NOT_EQUAL, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
		EXPAND_POP,
		EXPAND_POP,
NESTED_IFELSE -= 1
		EXITM
	endif
	if (1 relop 0) && ISLIT(op2)  ; >; convert to >= +1 to avoid testing both Z and C
		EXPAND_PUSH TRUE  ;show what this does
		cmp_ op1, op1bytes, _LITERAL(UNLIT(op2) + 1), op2bytes, COMPARE_HINT CMP_HINTS|
		ifbit NOBORROW, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
		EXPAND_POP,
		EXPAND_POP,
NESTED_IFELSE -= 1
		EXITM
	endif
	if 1 relop 0  ; > must check Z and C
		WARNIF TRUE, [CAUTION] if/else stmt swapped in order to evaluate >
		EXPAND_PUSH TRUE  ;show what this does
		cmp_ op1, op1bytes, op2, op2bytes, EQUAL_HINT COMPARE_HINT CMP_HINTS|
;CAUTION: 2 btf instructions are "chained" together here
		TESTBIT(STATUS, C, FALSE)  ;skip next instruction if STATUS.C is OFF (BORROW on); was < so skip the = check
		ifbit EQUALS, elsestmt, ifstmt, ((hints 0) & ~CMP_HINTS)|
		EXPAND_POP,
		EXPAND_POP,
NESTED_IFELSE -= 1
		EXITM
	endif
;< is the only case left:
	EXPAND_PUSH TRUE  ;show what this does
	cmp_ op1, op1bytes, op2, op2bytes, COMPARE_HINT CMP_HINTS|
	ifbit BORROW, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
	EXPAND_POP,
#else
	if 0 relop 0  ; <=, ==, >=
		if 0 relop 1  ; <=
			if ISLIT(op2)  ;convert to < +1 to avoid testing both Z and C
				EXPAND_PUSH TRUE  ;show what this does
				cmp_ op1, op1bytes, _LITERAL(UNLIT(op2) + 1), op2bytes, COMPARE_HINT CMP_HINTS|
				ifbit BORROW, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
				EXPAND_POP,
			else  ;must check Z and C
				EXPAND_PUSH TRUE  ;show what this does
				cmp_ op1, op1bytes, op2, op2bytes, EQUAL_HINT COMPARE_HINT CMP_HINTS|
;CAUTION: 2 btf instructions are "chained" together here
				TESTBIT(STATUS, C, FALSE)  ;skip next instruction if STATUS.C is OFF (BORROW on); was < so skip the = check
				ifbit EQUALS, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
				EXPAND_POP,
			endif
		else
			if 1 relop 0  ; >=
				EXPAND_PUSH TRUE  ;show what this does
				cmp_ op1, op1bytes, op2, op2bytes, COMPARE_HINT CMP_HINTS|
				EXPAND_POP,
				ifbit NOBORROW, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
			else  ; ==
				if ((op2) == LIT0) && ((op1bytes) == 1)  ;special case: compare to 0
					if (op1) == WREG  ;KLUDGE: update STATUS.Z so bit test will be correct
																				EXPAND  ;show generated code
	addlw 0
						NOEXPAND
					else
						EXPAND_PUSH TRUE  ;show what this does
						loadw_ op1  ;just load W to compare to 0
						EXPAND_POP,
					endif
				else
					EXPAND_PUSH TRUE  ;show what this does
					cmp_ op1, op1bytes, op2, op2bytes, EQUAL_HINT CMP_HINTS|
					EXPAND_POP,
				endif
				ifbit EQUALS, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
			endif
		endif
	else  ; <, !=, >
		if (0 relop 1) && (1 relop 0)  ; !=
			if ((op2) == LIT0) && ((op1bytes) == 1)  ;special case: compare to 0
				if (op1) == WREG  ;KLUDGE: update STATUS.Z so bit test will be correct
																				EXPAND  ;show generated code
	addlw 0
					NOEXPAND
				else
					EXPAND_PUSH TRUE  ;show what this does
					loadw_ op1  ;just load W to compare to 0
					EXPAND_POP,
				endif
			else
				EXPAND_PUSH TRUE  ;show what this does
				cmp_ op1, op1bytes, op2, op2bytes, EQUAL_HINT CMP_HINTS|
				EXPAND_POP,
			endif
			ifbit NOT_EQUAL, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
		else
			if 1 relop 0  ; >
				if ISLIT(op2)  ;convert to >= +1 to avoid testing both Z and C
					EXPAND_PUSH TRUE  ;show what this does
					cmp_ op1, op1bytes, _LITERAL(UNLIT(op2) + 1), op2bytes, COMPARE_HINT CMP_HINTS|
					EXPAND_POP,
					ifbit NOBORROW, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
				else  ;must check Z and C
					EXPAND_PUSH TRUE  ;show what this does
					cmp_ op1, op1bytes, op2, op2bytes, EQUAL_HINT COMPARE_HINT CMP_HINTS|
;CAUTION: 2 btf instructions are "chained" together here
					TESTBIT(STATUS, C, FALSE)  ;skip next instruction if STATUS.C is OFF (BORROW on); was < so skip the = check
					ifbit EQUALS, elsestmt, ifstmt, ((hints 0) & ~CMP_HINTS)|
					EXPAND_POP,
				endif
			else  ;<
				EXPAND_PUSH TRUE  ;show what this does
				cmp_ op1, op1bytes, op2, op2bytes, COMPARE_HINT CMP_HINTS|
				EXPAND_POP,
				ifbit BORROW, ifstmt, elsestmt, ((hints 0) & ~CMP_HINTS)|
			endif
		endif
	endif
#endif
	EXPAND_POP,
NESTED_IFELSE -= 1
	ENDM


;******************************************************************************
;register/arithmetic and stack handling macros:
;somewhat optimized; generates reasonably tight code
;helps to reduce late-night errors, but LOOK AT THE LISTING FILE to make sure they are doing what they are supposed to
;PIC16xxx stack is only 8 deep and there are no instructions to access it
;these macros implement a custom stack using RAM and indirect addressing
;multi-byte macros are little-endian
;******************************************************************************

;NOTE: assembler uses 32-bit constants internally; use the top bit to mark literal values, since it is not usually needed for code generation (which is only 8 or 14 bits)
#define _LITERAL(n)  (ASM_MSB | (n))  ;prepend this to any 8-, 16- or 24-bit literal values used as pseudo-opcode parameters, to distinguish them from register addresses (which are only 1 byte)
#define UNLIT(thing)  ((thing) & ~ASM_MSB)  ;convert from literal to number (strips _LITERAL tag)
#define ISLIT(thing)  (((thing) & ASM_MSB) && !ISPSEUDO(thing))  ;check for a literal

	CONSTANT LIT0 = _LITERAL(0)

;pseudo-registers:
;#define ISPSEUDO(reg)  (((reg) == WREG) || ((reg) == STACK) || ((reg) == UNSTACK))  ;line too long
#define ISPSEUDO(reg)  (((reg) <= STACK) && ((reg) >= UNSTACK))
	CONSTANT STACK = -5  ;use this instead of "INDF" (W == INDF == 0, but we need to distinguish between them); this will inc/dec stack pointer also
	CONSTANT WREG = -9  ;use this instead of "W" (W == INDF == 0, but we need to distinguish between them)
	CONSTANT UNSTACK = -13  ;use this for a stack that fills downward rather than upward


;figure out the largest digit in a 32-bit decimal value:
	VARIABLE LARGEST_DECDIG = 1, NUMDIG_DEC32 = 1
	while ((d'10' * LARGEST_DECDIG) > 0) && (((d'10' * LARGEST_DECDIG) / d'10') == LARGEST_DECDIG)  ;1, 10, 100, ..., (until wrap)
LARGEST_DECDIG *= d'10'
NUMDIG_DEC32 += 1
	endw


;comma-less wrapper for value and reg parameters:
;this is a kludgey way to pass a register and literal value as one param
;can handle a literal value up to 23 bits, then use the 8 remaining bits to hold register address
;unfortunately MPASM doesn't handle nested macro parameters well, so we need this kludge in order to keep code concise
	CONSTANT LIT23bit = _LITERAL(0x807fffff)  ;line too long
#define RVpack(val, reg)  (((val) & LIT23bit) | ((reg)<<23))
#define RV2reg(ValAndReg)  (((ValAndReg)>>23) & 0xff)
#define RV2val(ValAndReg)  ((ValAndReg) & 0x807fffff)


;load register with a literal value:
;this does not alter anything except the target register
;STATUS and W are not affected, and no temps are used
;however, it takes several instructions
;params:
; value = value to be loaded; can be literal or register
; reg = destination register
movlf_ MACRO value, reg
	NOEXPAND  ;hide clutter in LST file
	LOCAL VALUE = value
	LOCAL REG = reg
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show MOVLF params
VALUE = value  ;show MOVLF params
REG = reg  ;show MOVLF params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
;copy a bit at a time, in order to not affect STATUS or W:
	clear_ reg
	LOCAL bit = 0
	while bit < 8  ;0, 1, 2, ... 7
		LOCAL want_bit = TRUE
		if (reg) == PCLATH  ;only need to copy 2 bits
want_bit = FALSE
#if USE_PCLATH3
			if bit == 3
want_bit = TRUE
			endif
#endif
#if USE_PCLATH4
			if bit == 4
want_bit = TRUE
			endif
#endif
		endif
		if ISPORT(reg)  ;might not need all bits
want_bit = PORT#v(PORTOF#v(reg))_BITS & (1<<bit)
		endif
		if ISLIT(value)
			setbit reg, bit, UNLIT(value) & (1<<bit)
		else
			ifbit value, bit, SETBIT(reg, bit, TRUE),,
		endif
bit += 1
	endw
	EXPAND_POP,
	ENDM


;load accumulater with a register or value:
;FSR contains stack pointer; decremented before pop (stack empties downward)
;stack is assumed never to cross a bank boundary
;params:
; src = literal value or register to load into W
loadw_ MACRO src
	NOEXPAND  ;hide clutter in LST file
	LOCAL SRC = src
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show LOADW params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
SRC = src  ;show LOADW params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	if SRC == WREG  ;noop; already there
		EXPAND_POP,
		EXITM
	endif
	if SRC == FSRH  ;KLUDGE: upper "byte" is only IRP bit in STATUS
		ifbit STATUS, IRP, TRUE, movlw 1, movlw 0, BOTHOK_HINT  ;turn STATUS.IRP into a whole byte
		EXPAND_POP,
		EXITM
	endif
	if SRC == STACK  ;pop
																				EXPAND  ;show generated code
	decf FSR, F
	movf INDF, W
		NOEXPAND
		EXPAND_POP,
		EXITM
	endif
	if SRC == UNSTACK  ;pop upward
																				EXPAND  ;show generated code
	movf INDF, W
	incf FSR, F
		NOEXPAND
		EXPAND_POP,
		EXITM
	endif
	if ISLIT(SRC)
;a clrw is the same size as movlw 0, so there's no point in checking for that special case here:
		WARNIF UNLIT(SRC) & ~0xff, [CAUTION] LOADW: dropping upper bytes #v(UNLIT(SRC) & ~0xff) from literal #v(SRC)
																				EXPAND  ;show generated code
	movlw UNLIT(src) 
		NOEXPAND
	else  ;register
		BANKCHK SRC, MAYBE,
																				EXPAND  ;show generated code
	movf src, W
		NOEXPAND
	endif
	EXPAND_POP,
	ENDM


;save accumulator to a register:
;FSR contains stack pointer; incremented after push (stack fills upward)
;stack is assumed to never cross a bank boundary
;params:
; dest = destination register where to copy W to
storew_ MACRO dest
	NOEXPAND  ;hide clutter in LST file
	LOCAL DEST = dest
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show STOREW params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
DEST = dest  ;show STOREW params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	if DEST == WREG  ;noop; already there
		EXPAND_POP,
		EXITM
	endif
	if DEST == FSRH  ;KLUDGE: upper "byte" is only 1 bit: STATUS.IRP
		ifbit NOT_EQUAL, SETBIT(STATUS, IRP, TRUE), SETBIT(STATUS, IRP, FALSE), BOTHOK_HINT  ;set STATUS.IRP if upper byte was non-0
		EXPAND_POP,
		EXITM
	endif
	if DEST == STACK  ;push
																				EXPAND  ;show generated code
	movwf INDF
	incf FSR, F
		NOEXPAND
		EXPAND_POP,
		EXITM
	endif
	if DEST == UNSTACK  ;push downward
																				EXPAND  ;show generated code
	decf FSR, F
	movwf INDF
		NOEXPAND
		EXPAND_POP,
		EXITM
	endif
	BANKCHK dest, MAYBE,
																				EXPAND  ;show generated code
	movwf dest
	NOEXPAND
	EXPAND_POP,
	ENDM


;clear a register:
;FSR contains stack pointer; incremented after push (stack fills upward)
;stack is assumed to never cross a bank boundary
;does not alter W (unless dest == WREG)
;params:
; dest = destination register to clear
clear_ MACRO dest
	NOEXPAND  ;hide clutter in LST file
	LOCAL DEST = dest
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show CLEAR params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
DEST = dest  ;show CLEAR params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	if DEST == WREG
																				EXPAND  ;show generated code
	clrw  ;a clrw is the same size as movlw 0, so this doesn't save any space
		NOEXPAND
		EXPAND_POP,
		EXITM
	endif
	if DEST == STACK  ;push
																				EXPAND  ;show generated code
	clrf INDF
	incf FSR, F
		NOEXPAND
		EXPAND_POP,
		EXITM
	endif
	if DEST == UNSTACK  ;push downward
																				EXPAND  ;show generated code
	decf FSR, F
	clrf INDF
		NOEXPAND
		EXPAND_POP,
		EXITM
	endif
	BANKCHK dest, MAYBE,
																				EXPAND  ;show generated code
	clrf dest
	NOEXPAND
	EXPAND_POP,
	ENDM


;convenience wrappers:

gettop4 MACRO src, dest
	NOEXPAND
	swapcopy_ src, dest, TRUE, TRUE
	ENDM

swap_ MACRO src, dest
	NOEXPAND
	swapcopy_ src, dest, FALSE, TRUE
	ENDM

copy_ MACRO src, dest
	NOEXPAND
	swapcopy_ src, dest, FALSE, FALSE
	ENDM


;copy a register or literal value to register:
;STATUS bits are *not* affected unless stripping the upper or lower 4 bits
;params:
; src = source register or literal value
; dest = destination register
; strip = flag telling whether to remove upper 4 bits
; preserve_src = flag telling whether src must be preserved (copied value will be swapped)
swapcopy_ MACRO src, dest, strip, preserve_src
	NOEXPAND  ;hide clutter in LST file
	LOCAL SRC = src, DEST = dest, STRIP = strip
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show SWAPCOPY params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
SRC = src  ;show SWAPCOPY params
DEST = dest  ;show SWAPCOPY params
STRIP = strip  ;show SWAPCOPY params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	if ISLIT(SRC)
		WARNIF UNLIT(SRC) & ~0xff, [CAUTION] SWAP/COPY: dropping upper bytes #v(UNLIT(SRC) & ~0xff) from literal #v(SRC)
																				EXPAND  ;show generated code
	movlw UNLIT(src)  ;a clrw is the same size as movlw 0, so there's no point in checking for that special case
		NOEXPAND
	else  ;register
		ERRIF SRC == WREG, [ERROR] SWAP/COPY: not supported on W
		if (SRC == STACK) || (SRC == UNSTACK)
SRC = INDF
		endif
		BANKCHK SRC, MAYBE,
		if preserve_src  ;preserves source, but copied value is nibble-swapped and it must be done twice
																				EXPAND  ;show generated code
	swapf SRC, W
			NOEXPAND
		else  ;source is nibble-swapped
																				EXPAND  ;show generated code
	swapf SRC, F
	swapf SRC, W
			NOEXPAND
		endif
	endif
	if strip
																				EXPAND  ;show generated code
	andlw 0xf
		NOEXPAND
	endif
	storew_ DEST
	EXPAND_POP,
	ENDM


;get bottom nibble:
;params:
; src = source register or literal value
; dest = destination register
getbottom4 MACRO src, dest
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
	loadw_ src
																				EXPAND  ;show generated code
	andlw 0xf
	NOEXPAND
	storew_ dest
	EXPAND_POP,
	ENDM


;inc or dec a register, with optional conditional stmt following:
;params:
; reg = source register
; dest = destination register
; updown = flag telling whether to increment or decrement
; skip = flag telling whether to conditionally skip next instruction
bumpskipz MACRO reg, dest, updown, skip
	NOEXPAND  ;hide clutter in LST file
	LOCAL REG = reg, DEST = dest, UPDOWN = updown, SKIP = skip
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show BUMPSKIP params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
REG = reg  ;show BUMPSKIP params
DEST = dest  ;show BUMPSKIP params
UPDOWN = updown  ;show BUMPSKIP params
SKIP = skip  ;show BUMPSKIP params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	if (reg) == WREG  ;special case
																				EXPAND  ;show generated code
	addlw IIF(updown, 1, LOW -1)
		NOEXPAND
		if (dest) != WREG
			storew_ dest
		endif
		if skip
																				EXPAND  ;show generated code
	btfss STATUS, Z
			NOEXPAND
		endif
	else
		BANKCHK reg, MAYBE,
		if skip
			if updown
																				EXPAND  ;show generated code
	incfsz reg, IIF(dest == reg, F, W)
				NOEXPAND
			else
																				EXPAND  ;show generated code
	decfsz reg, IIF(dest == reg, F, W)
				NOEXPAND
			endif
			ERRIF dest != reg, [ERROR] Inc/dec destination #v(dest) vs. reg #v(reg) not implemented (movwf won't fit)
		else
			if updown
																				EXPAND  ;show generated code
	incf reg, IIF(dest == reg, F, W)
				NOEXPAND
			else
																				EXPAND  ;show generated code
	decf reg, IIF(dest == reg, F, W)
				NOEXPAND
			endif
		endif
		if DEST != REG
			storew_ dest
		endif
	endif
	EXPAND_POP,
	ENDM


;comma-less wrappers:
;(kludge to allow nested macro calls with multiple parameters)
;NOTE: this assumes dest is a register without checking it
;#define MOV8(src, dest)  MOV8_WRAPPER IIF(ISLIT(src), LIT0, 0) | UNLIT(src)*0x100+(dest)
;MOV8_WRAPPER MACRO src_dest
;	mov8 IIF(ISLIT(src_dest), LIT0, 0) | (UNLIT(src_dest)/0x100), (src_dest)%0x100
;	ENDM

clear8 MACRO reg
	mov8 _LITERAL(0), reg
	ENDM

clear16 MACRO reg
	mov16 _LITERAL(0), reg
	ENDM

clear24 MACRO reg
	mov24 _LITERAL(0), reg
	ENDM

clear32 MACRO reg
	mov32 _LITERAL(0), reg
	ENDM


;comma-less wrappers:
MOV8RV MACRO reg_and_val
	MOVRV_ reg_and_val, 1
	ENDM

MOV16RV MACRO reg_and_val
	MOVRV_ reg_and_val, 2
	ENDM

MOV24RV MACRO reg_and_val
	MOVRV_ reg_and_val, 3
	ENDM

MOV32RV MACRO reg_and_val
	MOVRV_ reg_and_val, 4
	ENDM


;copy a register or literal value to a register:
;params:
; reg_and_val = packed destination register and value; value can be literal or register
; bytes = #bytes to copy
MOVRV_ MACRO reg_and_val, bytes
	NOEXPAND  ;hide clutter in LST file
	LOCAL REGVAL = reg_and_val  ;line too long
	if ISLIT(REGVAL)
		mov#v(8*bytes) _LITERAL(RV2val(REGVAL)), RV2reg(REGVAL)
	else
		mov#v(8*bytes) RV2val(REGVAL), RV2reg(REGVAL)
	endif
	ENDM


;convenience wrappers for mov_:
push8 MACRO reg
	NOEXPAND
	push_(reg, 1)
	ENDM

pop8 MACRO reg
	NOEXPAND
	pop_(reg, 1)
	ENDM

;empty stack in reverse direction (up instead of down):
unpush8 MACRO reg
	mov_ UNSTACK, 1, reg, 1, FALSE, TRUE
	ENDM

mov8 MACRO src, dest
	NOEXPAND
	mov_ src, 1, dest, 1, FALSE, TRUE
	ENDM

;don't warn if move is not needed:
mov8nop MACRO src, dest
	NOEXPAND
	movnop_ src, 1, dest, 1, FALSE, TRUE  ;FALSE
	ENDM

not8 MACRO src, dest
	NOEXPAND
	mov_ src, 1, dest, 1, MAYBE, TRUE
	ENDM

;cast up 8 to 16 bits (unsigned):
mov8_16 MACRO src, dest
	NOEXPAND
	mov_ src, 1, dest, 2, FALSE, TRUE
	ENDM

;cast up 8 to 32 bits (unsigned):
mov8_32 MACRO src, dest
	NOEXPAND
	mov_ src, 1, dest, 4, FALSE, TRUE
	ENDM

mov16 MACRO src, dest
	NOEXPAND
	mov_ src, 2, dest, 2, FALSE, TRUE
	ENDM

mov16nop MACRO src, dest
	NOEXPAND
	movnop_ src, 2, dest, 2, FALSE, TRUE
	ENDM

not16 MACRO src, dest
	NOEXPAND
	mov_ src, 2, dest, 2, MAYBE, TRUE
	ENDM

push16 MACRO reg
	NOEXPAND
	push_(reg, 2)
	ENDM

pop16 MACRO reg
	NOEXPAND
	pop_(reg, 2)  ;NOTE: pulls bytes from stack in reverse order
	ENDM

mov24 MACRO src, dest
	NOEXPAND
	mov_ src, 3, dest, 3, FALSE, TRUE
	ENDM

mov24nop MACRO src, dest
	NOEXPAND
	movnop_ src, 3, dest, 3, FALSE, TRUE
	ENDM

not24 MACRO src, dest
	NOEXPAND
	mov_ src, 3, dest, 3, MAYBE, TRUE
	ENDM

push24 MACRO reg
	NOEXPAND
	push_(reg, 3)
	ENDM

pop24 MACRO reg
	NOEXPAND
	pop_(reg, 3)  ;NOTE: pulls bytes from stack in reverse order
	ENDM

mov32 MACRO src, dest
	NOEXPAND
	mov_ src, 4, dest, 4, FALSE, TRUE
	ENDM

mov32nop MACRO src, dest
	NOEXPAND
	movnop_ src, 4, dest, 4, FALSE, TRUE
	ENDM

not32 MACRO src, dest
	NOEXPAND
	mov_ src, 4, dest, 4, MAYBE, TRUE
	ENDM

push32 MACRO reg
	NOEXPAND
	push_(reg, 4)
	ENDM

pop32 MACRO reg
	NOEXPAND
	pop_(reg, 4)  ;NOTE: pulls bytes from stack in reverse order
	ENDM

;reduce macro nesting a little:
#define push_(reg, bytes)  mov_ reg, bytes, STACK, bytes, FALSE, TRUE
#define pop_(reg, bytes)  mov_ STACK, bytes, reg, bytes, TRUE, TRUE  ;NOTE: pulls bytes from stack in reverse order

;push_ MACRO reg, bytes
;	NOEXPAND
;	mov_ reg, bytes, STACK, bytes, FALSE, TRUE
;	ENDM

;pop_ MACRO reg, bytes
;	NOEXPAND
;	mov_ STACK, bytes, reg, bytes, TRUE, TRUE  ;NOTE: pulls bytes from stack in reverse order
;	ENDM

;silently remove redundant reg copy:
movnop_ MACRO src, srcbytes, dest, destbytes, reverse, want_uniq
	NOEXPAND
	if src == dest  ;silently ignore
		EXITM
	endif
	mov_ src, srcbytes, dest, destbytes, reverse, want_uniq
	ENDM


;multi-byte move (it's actually a "copy", not "move") 8/16/24/32-bit register/literal to register:
;somewhat optimized
;NOTE: the PIC16xxx MOVF F,W instruction alters the Z STATUS bit; this macro can be used to avoid that in some cases
;does a cast if srcbytes != destbytes (no sign extend)
;stack is assumed to never cross a bank boundary
;NOTE: stack fills up, empties down, so that POP PCL will do the last stack pointer update (decrement) *before* setting PCL
;NOTE: PUSH puts bytes onto the stack in order, while POP takes them off in REVERSE order (so that POP PCLATH/PCL will do PCL last)
;PUSH PCL and POP PCL can be used to implement function calls/returns using the custom stack
;no stack frame is used for local variables (too much address arithmetic), but a thread-local stack can be used if FSR is saved/restored appropriately
;to make a recursive function, pass parameters via stack instead of statically allocating them (use PUSH/POP for recursion)
;CAUTION: THOROUGHLY check any changes made here; this macro is used extensively; always check .LST file to verify it does what you think it's doing
;params:
; src = source register or literal value; if src == STACK, a POP is performed (FSR decremented before each byte)
; srcbytes = #bytes to copy from src
; dest = destination register
; destbytes = #bytes to copy into dest; if dest == STACK, a PUSH is performed (FSR increments after each byte)
; reverse = flag telling whether to copy in reverse order; must be used to control PCLATH vs. PCL order; MAYBE => complement
; want_uniq = flag telling whether src and dest should be different
mov_ MACRO src, srcbytes, dest, destbytes, reverse, want_uniq
	NOEXPAND  ;hide clutter in LST file
	LOCAL SRC = src, SRCBYTES = srcbytes
	LOCAL DEST = dest, DESTBYTES = destbytes
	if TRUE  ;MEXPAND_STACKLO & 1
																				EXPAND  ;show MOV params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
SRC = src  ;show MOV params
SRCBYTES = srcbytes  ;show MOV params
DEST = dest  ;show MOV params
DESTBYTES = destbytes  ;show MOV params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
;assembler bug; save as LOCALs to force eval (else "missing operand" and "missing argument" errors)
	LOCAL SRC_0 = SRC, SRC_1 = (SRC+1), SRC_2 = (SRC+2), SRC_3 = (SRC+3)
	LOCAL DEST_0 = DEST, DEST_1 = (DEST+1), DEST_2 = (DEST+2), DEST_3 = (DEST+3)
	LOCAL byte = 0, REV = reverse
	WARNIF (want_uniq) && (SRC == DEST), [CAUTION] MOV: useless move #v(SRC)  ;) && ((SRC != WREG) || (DEST != WREG))
	if (SRC == DEST) && ((srcbytes) == (destbytes)) && !(reverse)  ;nothing to do
		EXPAND_POP,
		EXITM
	endif
	if ISLIT(SRC)  ;unpack SRC bytes
SRC_1 = _LITERAL((UNLIT(SRC)/0x100) & 0xff)
SRC_2 = _LITERAL((UNLIT(SRC)/0x10000) & 0xff)
SRC_3 = _LITERAL((UNLIT(SRC)/0x1000000) & 0xff)
SRC_0 = _LITERAL(SRC & 0xff)  ;trim off extra bits after they are copied to SRC_1+2+3
		if (SRC_3 == _LITERAL(0x7f)) && (SRC_2 == _LITERAL(0xff))  ;KLUDGE: caller probably wanted all the bits
SRC_3 = _LITERAL(0xff)
		endif
		WARNIF (srcbytes < 2) && (SRC_1 != LIT0), [CAUTION] MOV: dropping upper bytes1 #v(SRC_1) from source
		WARNIF (srcbytes < 3) && (SRC_2 != LIT0), [CAUTION] MOV: dropping upper bytes2 #v(SRC_2) from source
		WARNIF (srcbytes < 4) && (SRC_3 != LIT0), [CAUTION] MOV: dropping upper bytes3 #v(SRC_3) from source
	else
		if ISPSEUDO(SRC)  ;preserve pseudo-registers
SRC_1 = SRC
SRC_2 = SRC
SRC_3 = SRC
		else
SRC_1 = HIBYTE(SRC)  ;make correction for non-contiguous upper byte 
			ERRIF !ISREG(SRC, srcbytes), [ERROR] MOV: invalid multi-byte source register #v(srcbytes):#v(SRC)
		endif
	endif
	if srcbytes < 2
SRC_1 = LIT0
	endif
	if srcbytes < 3
SRC_2 = LIT0
	endif
	if srcbytes < 4
SRC_3 = LIT0
	endif

	ERRIF ISLIT(DEST), [ERROR] MOV: destination must be register not literal #v(DEST)
	if ISPSEUDO(DEST)  ;preserve pseudo-registers
DEST_1 = DEST
DEST_2 = DEST
DEST_3 = DEST
	else
DEST_1 = HIBYTE(DEST)  ;make correction for non-contiguous upper byte 
		ERRIF !ISREG(DEST, destbytes), [ERROR] MOV: invalid multi-byte destination register #v(destbytes):#v(DEST)
	endif
;assembler bug: getting "symbol undefined" on destbytes within WHILE loop below, so use local vbls instead
	if destbytes < 2
DEST_1 = LIT0
	endif
	if destbytes < 3
DEST_2 = LIT0
	endif
	if destbytes < 4
DEST_3 = LIT0
	endif

	if (reverse) == TRUE  ;move bytes in reverse order (pop from stack)
byte = -4
REV = -1
		if (SRC_0 != LIT0) && (SRC_0 != STACK) && (SRC_0 != UNSTACK) && (SRC_0 == SRC_1)  ;W already loaded; reuse it
SRC_0 = WREG
		endif
		if (SRC_1 != LIT0) && (SRC_1 != STACK) && (SRC_1 != UNSTACK) && (SRC_1 == SRC_2)  ;W already loaded; reuse it
SRC_1 = WREG
		endif
		if (SRC_2 != LIT0) && (SRC_2 != STACK) && (SRC_2 != UNSTACK) && (SRC_2 == SRC_1)  ;W already loaded; reuse it
SRC_2 = WREG
		endif
	else  ;bytes in order (push onto stack or normal register copy)
byte = 1
REV = 1
		if (SRC_3 != LIT0) && (SRC_3 != STACK) && (SRC_3 != UNSTACK) && (SRC_3 == SRC_2)  ;W already loaded; reuse it
SRC_3 = WREG
		endif
		if (SRC_2 != LIT0) && (SRC_2 != STACK) && (SRC_2 != UNSTACK) && (SRC_2 == SRC_1)  ;W already loaded; reuse it
SRC_2 = WREG
		endif
		if (SRC_1 != LIT0) && (SRC_1 != STACK) && (SRC_1 != UNSTACK) && (SRC_1 == SRC_0)  ;W already loaded; reuse it
SRC_1 = WREG
		endif
	endif
	WHILE (byte * REV >= 1) && (byte * REV <= 4)  ;4, 3, 2, 1 or 1, 2, 3, 4; copy bytes in MSB-to-LSB or LSB-to-MSB order
		if DEST_#v(byte * REV - 1) != LIT0  ;start with MSB (required if loading PCL/PCLATH)
			if SRC_#v(byte * REV - 1) == LIT0  ;special case: source is 0
				clear_ DEST_#v(byte * REV - 1)
			else  ;non-0 source
				if ((reverse) != FALSE) && ((reverse) != TRUE)  ;complement
					if SRC == WREG
																				EXPAND  ;show generated code
	xorlw 0xff  ;complement W
						NOEXPAND
					else
						if ISLIT(SRC_#v(byte * REV - 1))
							loadw_ _LITERAL(LOW ~UNLIT(SRC_#v(byte * REV - 1)))
						else
							BANKCHK SRC_#v(byte * REV - 1), MAYBE,
																				EXPAND  ;show generated code
	comf SRC_#v(byte * REV - 1), IIF(SRC == DEST, F, W)
							NOEXPAND
						endif
					endif
					if SRC != DEST
						storew_ DEST_#v(byte * REV - 1)
					endif
				else
					loadw_ SRC_#v(byte * REV - 1)
					storew_ DEST_#v(byte * REV - 1)
				endif
			endif
		endif
byte += 1
	ENDW
	EXPAND_POP,
	ENDM


;to test or understand how the MOV/PUSH/POP macros above work, use the following code:
;call this function from somewhere,
;then view Disassembly Listing and File Registers in MPLAB IDE and step through the code using F7
#if FALSE  ;test/debug
;	LIST
	CBLOCK 0x20
	DATA8  ;equ 0x20
	DATA32:4  ;equ 0x21
	STACK_AREA  ;equ 0x25
	ENDC
MOV_test: jumptarget,
;	BANKCHK 0x20, MAYBE,  ;set RP0+RP1
	mov8 _LITERAL(-1), DATA8
	mov32 _LITERAL(-1), DATA32
	nop
	mov8 _LITERAL(0), DATA8
	mov32 _LITERAL(0), DATA32
	nop
	mov8 _LITERAL(0x55), DATA8
	mov32 _LITERAL(0x11223344), DATA32
	nop
	mov32 _LITERAL(-1), STACK_AREA
	mov8 0x21, STACK_AREA
	nop
 	mov32 _LITERAL(-1), STACK_AREA
	mov32 0x21, STACK_AREA
	nop
	mov32 _LITERAL(-1), STACK_AREA
	mov8 _LITERAL(STACK_AREA), FSR  ;stack pointer
	push8 _LITERAL(1)
	nop
	mov32 _LITERAL(-1), STACK_AREA
	mov8 _LITERAL(STACK_AREA), FSR  ;stack pointer
	push32 DATA32
	nop
	mov32 _LITERAL(-1), DATA32
	pop32 DATA32
	return
;these ones should generate a compile error:
	mov8 0x50, _LITERAL(2)
	mov32 0x51, _LITERAL(3)
	pop8 _LITERAL(1)
;	NOLIST
#endif


;kludgey wrappers to avoid commas on nested macros:
PresetTimer1 MACRO preset
	add16 TMR1, preset, TMR1,  ;make it tick again at the right time; add/sub automatically corrects for timing latency
	ENDM

;kludgey wrappers to avoid commas on nested macros:
PresetTimer0 MACRO preset
	add8 TMR0, preset, TMR0,  ;make it tick again at the right time; add/sub automatically corrects for timing latency
	ENDM


;comma-less wrappers:
ADD16RV MACRO reg_and_val
	NOEXPAND  ;hide clutter in LST file
	LOCAL dest = RV2reg(reg_and_val)  ;line too long
	LOCAL value = RV2val(reg_and_val)  ;line too long
	if ISLIT(reg_and_val)
		add16 dest, _LITERAL(value), dest,
	else
		add16 dest, value, dest,
	endif
	ENDM

;IOR8RV MACRO reg_and_val
;	NOEXPAND
;	LOCAL dest = RV2reg(reg_and_val)  ;line too long
;	LOCAL value = RV2val(reg_and_val)  ;line too long
;	if ISLIT(reg_and_val)
;		ior8 dest, _LITERAL(value), dest,
;	else
;		ior8 dest, value, dest,
;	endif
;	ENDM


;convenience wrappers for arith_:
add8 MACRO src, value, dest, hints
	NOEXPAND
	add_(src, 1, value, 1, dest, hints)
	ENDM

sub8 MACRO src, value, dest, hints
	NOEXPAND
	sub_(src, 1, value, 1, dest, hints)
	ENDM

cmp8 MACRO src, value, hints
	NOEXPAND
	cmp_(src, 1, value, 1, hints)
	ENDM

and8 MACRO src, value, dest, hints
	NOEXPAND
	and_(src, 1, value, 1, dest, hints)
	ENDM

ior8 MACRO src, value, dest, hints
	NOEXPAND
	ior_(src, 1, value, 1, dest, hints)
	ENDM

xor8 MACRO src, value, dest, hints
	NOEXPAND
	xor_(src, 1, value, 1, dest, hints)
	ENDM

add16 MACRO src, value, dest, hints
	NOEXPAND
	add_(src, 2, value, 2, dest, hints)
	ENDM

add16_8 MACRO src, value, dest, hints
	NOEXPAND
	add_(src, 2, value, 1, dest, hints)
	ENDM

sub16 MACRO src, value, dest, hints
	NOEXPAND
	sub_ src, 2, value, 2, dest, hints
	ENDM

cmp16 MACRO src, value, hints
	NOEXPAND
	cmp_(src, 2, value, 2, hints)
	ENDM

and16 MACRO src, value, dest, hints
	NOEXPAND
	and_(src, 2, value, 2, dest, hints)
	ENDM

ior16 MACRO src, value, dest, hints
	NOEXPAND
	ior_(src, 2, value, 2, dest, hints)
	ENDM

xor16 MACRO src, value, dest, hints
	NOEXPAND
	xor_(src, 2, value, 2, dest, hints)
	ENDM

and24 MACRO src, value, dest, hints
	NOEXPAND
	and_(src, 3, value, 3, dest, hints)
	ENDM

add24 MACRO src, value, dest, hints
	NOEXPAND
	add_(src, 3, value, 3, dest, hints)
	ENDM

sub24 MACRO src, value, dest, hints
	NOEXPAND
	sub_(src, 3, value, 3, dest, hints)
	ENDM

cmp24 MACRO src, value, hints
	NOEXPAND
	cmp_(src, 3, value, 3, hints)
	ENDM

ior24 MACRO src, value, dest, hints
	NOEXPAND
	ior_(src, 3, value, 3, dest, hints)
	ENDM

xor24 MACRO src, value, dest, hints
	NOEXPAND
	xor_(src, 3, value, 3, dest, hints)
	ENDM

add32 MACRO src, value, dest, hints
	NOEXPAND
	add_(src, 4, value, 4, dest, hints)
	ENDM

sub32 MACRO src, value, dest, hints
	NOEXPAND
	sub_(src, 4, value, 4, dest, hints)
	ENDM

cmp32 MACRO src, value, hints
	NOEXPAND
	cmp_(src, 4, value, 4, hints)
	ENDM

;NOTE: this one generates redundant add/sub 0's which should be optimized out
add32_8 MACRO src, value, dest, hints
	NOEXPAND
	add_(src, 4, value, 1, dest, hints)
	ENDM

add32_16 MACRO src, value, dest, hints
	NOEXPAND
	add_(src, 4, value, 2, dest, hints)
	ENDM

;NOTE: this one generates redundant add/sub 0's which should be optimized out
sub32_8 MACRO src, value, dest, hints
	NOEXPAND
	sub_(src, 4, value, 1, dest, hints)
	ENDM

and32 MACRO src, value, dest, hints
	NOEXPAND
	and_(src, 4, value, 4, dest, hints)
	ENDM

ior32 MACRO src, value, dest, hints
	NOEXPAND
	ior_(src, 4, value, 4, dest, hints)
	ENDM

xor32 MACRO src, value, dest, hints
	NOEXPAND
	xor_(src, 4, value, 4, dest, hints)
	ENDM

;reduce macro nesting a little:
#define add_(src, srcbytes, value, valbytes, dest, hints)  arith_  A'+', src, srcbytes, value, valbytes, dest, hints
#define sub_(src, srcbytes, value, valbytes, dest, hints)  arith_  A'-', src, srcbytes, value, valbytes, dest, hints
#define cmp_(src, srcbytes, value, valbytes, hints)  arith_  A'?', src, srcbytes, value, valbytes, WREG, hints
#define and_(src, srcbytes, value, valbytes, dest, hints)  arith_  A'A', src, srcbytes, value, valbytes, dest, hints
#define ior_(src, srcbytes, value, valbytes, dest, hints)  arith_  A'I', src, srcbytes, value, valbytes, dest, hints
#define xor_(src, srcbytes, value, valbytes, dest, hints)  arith_  A'X', src, srcbytes, value, valbytes, dest, hints

;add_ MACRO src, srcbytes, value, valbytes, dest, hints
;	NOEXPAND
;	arith_  A'+', src, srcbytes, value, valbytes, dest, hints
;	ENDM

;sub_ MACRO src, srcbytes, value, valbytes, dest, hints
;	NOEXPAND
;	arith_  A'-', src, srcbytes, value, valbytes, dest, hints
;	ENDM

;cmp_ MACRO src, srcbytes, value, valbytes, hints
;	NOEXPAND
;	arith_  A'?', src, srcbytes, value, valbytes, WREG, hints
;	ENDM

;and_ MACRO src, srcbytes, value, valbytes, dest, hints
;	NOEXPAND
;	arith_  A'A', src, srcbytes, value, valbytes, dest, hints
;	ENDM

;ior_ MACRO src, srcbytes, value, valbytes, dest, hints
;	NOEXPAND
;	arith_  A'I', src, srcbytes, value, valbytes, dest, hints
;	ENDM

;xor_ MACRO src, srcbytes, value, valbytes, dest, hints
;	NOEXPAND
;	arith_  A'X', src, srcbytes, value, valbytes, dest, hints
;	ENDM


;boolean and algebraic arithmetic 8/16/24/32-bit register/literal to/from register, place result in register:
;source and destination registers are assumed to be the same size
;value to be added or subtracted can be a different size (smaller)
;and/ior/xor algoritms are identical except for and vs. ior vs. xor instruction
;for subtract, set dest to W for a compare (throws away result, just sets Z and C)
;add algorithm is based on http://www.piclist.com/techref/microchip/math/add/32brb.htm
;subtract algorithm is based on http://www.piclist.com/techref/microchip/math/sub/index.htm
;algoritms are identical except for add vs. sub instruction
;params:
; func = function to perform: +, -, And, Ior, Xor, compare ("?")
; src = source register or literal value
; srcbytes = #bytes in src
; value = operand
; valbytes = #bytes in value
; dest = destination register; #bytes assumed to be the same as src
; hints = execution hints; zero or more of:
;          EQUAL_HINT ==> caller wants STATUS.Z to be correct upon return
;          COMPARE_HINT ==> caller wants STATUS.C to be correct upon return
arith_ MACRO func, src, srcbytes, value, valbytes, dest, hints
	NOEXPAND  ;hide clutter in LST file
	LOCAL SRC = src
	LOCAL VAL = value
	LOCAL DEST = dest
	LOCAL HINTS = hints 0  ;optional param; must end with "|" if passed
	LOCAL destbytes = srcbytes  ;assume dest and src are same size
	if TRUE  ;MEXPAND_STACKLO & 1
																				EXPAND  ;show ARITH params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
SRC = src  ;show ARITH params
VAL = value  ;show ARITH params
DEST = dest  ;show ARITH params
HINTS = hints 0  ;show ARITH params
destbytes = srcbytes  ;show ARITH params; assume dest and src are same size
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	LOCAL NOTMY_HINTS = (hints 0) & ~(EQUAL_HINT COMPARE_HINT 0)  ;line too long
	WARNIF NOTMY_HINTS, [WARNING] Ignoring unrecognized ARITH hint #v(NOTMY_HINTS)
	if (VAL == LIT0) && !((hints 0) & (EQUAL_HINT COMPARE_HINT 0))
		EXPAND_PUSH TRUE  ;show what this does
		mov_ IIF(func == A'A', value, src), srcbytes, dest, srcbytes, FALSE, TRUE
		EXPAND_POP,
		EXPAND_POP,
		EXITM
	endif
;misc optimization, mainly for eeread/eewrite addressing:
	if ((func) == A'+') && !((hints 0) & (EQUAL_HINT COMPARE_HINT 0))
		if VAL == LIT0
			EXPAND_PUSH TRUE  ;show what this does
			mov#v(8*destbytes)nop SRC, DEST
			EXPAND_POP,
			EXPAND_POP,
			EXITM
		endif
		if (VAL == _LITERAL(1)) && (SRC == DEST)
			EXPAND_PUSH TRUE  ;show what this does
			inc#v(8*destbytes) DEST
			EXPAND_POP,
			EXPAND_POP,
			EXITM
		endif
	endif
;needless optimization:
;	LOCAL FUNC = func, cmpdone
;sub8 WREG, _LITERAL(10), WREG
;	if (SRC == WREG) || (ISLIT(SRC) && !ISLIT(VAL))  ;swap operands
;SRC = value
;VAL = src
;		if (SRC == WREG) || (FUNC == A'?')
;FUNC = A'#'  ;need to complement STATUS.C after compare
;		else
;			if (FUNC == A'-') && ISLIT(VAL)  ;complement operand to relax operator ordering (helpful in case SRC == WREG)
;FUNC = A'+'
;VAL = _LITERAL(-UNLIT(VAL))  ;complement and add so operands are cummutative
;			endif
;		endif
;		WARNIF TRUE, "ARITH: swapped operands to make value loadable", SRC, VAL
;	endif
;substitute for more efficient functions:
;don't do these ones; caller might want STATUS.C
;	if (src == dest) && (value == LITERAL 1)
;		if func == A'+'
;			inc_ src, dest, srcbytes
;			EXITM
;		endif
;		if func == A'-'
;			dec_ src, dest, srcbytes
;			EXITM
;		endif
;	endif
;complement operand to preserve W (useful for CASE stmts):
	if (SRC == WREG) && ISLIT(VAL) && ((func == A'-') || (func == A'?'))
		arith_ A'+', SRC, srcbytes, _LITERAL((-1>>(32 - (8*valbytes))) & -UNLIT(VAL)), valbytes, dest, hints  ;sublw gives lit-W, but we want W-lit; complement value and use addlw 
		EXPAND_POP,
		EXITM
	endif

;KLUDGE: assembler occassionally generates a duplicate, non-unique suffix on LOCALs (probably a bug in hash algorithm or symbol table manager)
;this causes annoying error 116 "Address label duplicated or different in second pass", and also messes up forward references (leaves symbolic addresses with values of 0, which will cause a jump to 0 so PIC will restart)
;to work around this ASSEMBLER problem, we FORCE the suffixes to be unique by explicitly generating unique symbols for LOCALs
	LOCAL UNIQID = NEXT_UNIQID 
NEXT_UNIQID += 1
	LOCAL NEED_COMP = FALSE, cmpdone_#v(UNIQID)
	if (SRC == WREG)  ;|| ((VAL == DEST) && (DEST != WREG))  ;need to swap operands
SRC = value
VAL = src
NEED_COMP = (func == A'-') || (func == A'?')  ;these operators are not cummutative, so result needs to be complemented
		if HINTS & (COMPARE_HINT 0)
			WARNIF2 NEED_COMP, [CAUTION] ARITH: swapped operands #v(SRC) and #v(VAL) to preserve WREG, STATUS.C will be incorrect after -
		endif
	endif
;assembler bug; save as LOCALs to force eval (else "missing operand" and "missing argument" errors)
	LOCAL SRC_0 = SRC, SRC_1 = (SRC+1), SRC_2 = (SRC+2), SRC_3 = (SRC+3)
	LOCAL VAL_0 = VAL, VAL_1 = (VAL+1), VAL_2 = (VAL+2), VAL_3 = (VAL+3)
	LOCAL DEST_0 = DEST, DEST_1 = (DEST+1), DEST_2 = (DEST+2), DEST_3 = (DEST+3)

	if ISLIT(SRC)  ;unpack SRC bytes
SRC_1 = _LITERAL((UNLIT(SRC)/0x100) & 0xff)
SRC_2 = _LITERAL((UNLIT(SRC)/0x10000) & 0xff)
SRC_3 = _LITERAL((UNLIT(SRC)/0x1000000) & 0xff)
SRC_0 = _LITERAL(SRC & 0xff)  ;trim off extra bits after they are copied to SRC_1+2+3
		if (SRC_3 == _LITERAL(0x7f)) && (SRC_2 == _LITERAL(0xff))  ;KLUDGE: caller probably wanted all the bits
SRC_3 = _LITERAL(0xff)
		endif
		WARNIF (srcbytes < 2) && (SRC_1 != LIT0), [CAUTION] ARITH: dropping upper bytes1 #v(SRC_1) from source #v(SRC)
		WARNIF (srcbytes < 3) && (SRC_2 != LIT0), [CAUTION] ARITH: dropping upper bytes2 #v(SRC_2) from source #v(SRC)
		WARNIF (srcbytes < 4) && (SRC_3 != LIT0), [CAUTION] ARITH: dropping upper bytes3 #v(SRC_3) from source #v(SRC)
	else
		if ISPSEUDO(SRC)  ;preserve pseudo-registers
SRC_1 = SRC
SRC_2 = SRC
SRC_3 = SRC
			ERRIF (SRC == WREG) && (srcbytes > 1), [ERROR] ARITH: WREG will be lost during multi-byte operation
		else
SRC_1 = HIBYTE(SRC)  ;make correction for non-contiguous upper byte 
			ERRIF !ISREG(SRC, srcbytes), [ERROR] ARITH: invalid multi-byte source register #v(srcbytes):#v(SRC)
		endif
	endif
	if (srcbytes) < 2
SRC_1 = LIT0
	endif
	if (srcbytes) < 3
SRC_2 = LIT0
	endif
	if (srcbytes) < 4
SRC_3 = LIT0
	endif

	ERRIF ISLIT(DEST), [ERROR] ARITH: destination must be register not literal #v(DEST)
	if ISPSEUDO(DEST)  ;preserve pseudo-registers
DEST_1 = DEST
DEST_2 = DEST
DEST_3 = DEST
	else
DEST_1 = HIBYTE(DEST)  ;make correction for non-contiguous upper byte 
		ERRIF !ISREG(DEST, destbytes), [ERROR] ARITH: invalid multi-byte destination register #v(destbytes):#v(DEST)
;NOTE: we don't want PCL to be the destination because we would need to start with LSB (for carry/borrow purposes), and that would cause a jump before PCLATH was updated
		ERRIF (DEST == PCL) && (destbytes > 1) && ((func == A'+') || (func == A'-')), [ERROR] ARITH: PCL cannot be destination for +/-
	endif
;assembler bug: getting "symbol undefined" on destbytes within WHILE loop below, so use local vbls instead
	if (destbytes) < 2
DEST_1 = LIT0
	endif
	if (destbytes) < 3
DEST_2 = LIT0
	endif
	if (destbytes) < 4
DEST_3 = LIT0
	endif

	if ISLIT(VAL)  ;unpack VAL bytes
VAL_1 = _LITERAL((UNLIT(VAL)/0x100) & 0xff)
VAL_2 = _LITERAL((UNLIT(VAL)/0x10000) & 0xff)
VAL_3 = _LITERAL((UNLIT(VAL)/0x1000000) & 0xff)
VAL_0 = _LITERAL(VAL & 0xff)  ;trim off extra bits after they are copied to VAL_1+2+3
			if (VAL_3 == _LITERAL(0x7f)) && (VAL_2 == _LITERAL(0xff))  ;KLUDGE: caller probably wanted all the bits
VAL_3 = _LITERAL(0xff)
			endif
		WARNIF (valbytes < 2) && (VAL_1 != LIT0), [CAUTION] ARITH: dropping upper bytes1 #v(VAL_1) from value #v(VAL)
		WARNIF (valbytes < 3) && (VAL_2 != LIT0), [CAUTION] ARITH: dropping upper bytes2 #v(VAL_2) from value #v(VAL)
		WARNIF (valbytes < 4) && (VAL_3 != LIT0), [CAUTION] ARITH: dropping upper bytes3 #v(VAL_3) from value #v(VAL)
	else
		if ISPSEUDO(VAL)  ;preserve pseudo-registers
VAL_1 = VAL
VAL_2 = VAL
VAL_3 = VAL
			ERRIF (VAL == WREG) && (valbytes > 1), [ERROR] ARITH: WREG will be lost during multi-byte operation
		else
VAL_1 = HIBYTE(VAL)  ;make correction for non-contiguous upper byte 
			ERRIF !ISREG(VAL, valbytes), [ERROR] ARITH: invalid multi-byte value register #v(valbytes):#v(VAL)
		endif
	endif
	if (valbytes) < 2
VAL_1 = LIT0
	endif
	if (valbytes) < 3
VAL_2 = LIT0
	endif
	if (valbytes) < 4
VAL_3 = LIT0
	endif

	ERRIF valbytes > srcbytes, [ERROR] ARITH: SRC reg bytes #v(srcbytes) must be at least as large as VAL bytes #v(valbytes) to add/sub/compare
	ERRIF (func != A'+') && (func != A'-') && (func != A'?') && (func != A'A') && (func != A'I') && (func != A'X'), [ERROR] ARITH: unknown function code func
	if ISLIT(SRC) && ISLIT(VAL)  ;constant case: add/sub/cmp/and/ior/xor literal to/from literal
SRC = UNLIT(SRC_3)*0x1000000 | UNLIT(SRC_2)*0x10000 | UNLIT(SRC_1)*0x100 | UNLIT(SRC_0)
VAL = UNLIT(VAL_3)*0x1000000 | UNLIT(VAL_2)*0x10000 | UNLIT(VAL_1)*0x100 | UNLIT(VAL_0)
		if func == A'+'
;this doesn't save any space or time if we need to set STATUS.Z and STATUS.C
			mov_ _LITERAL(SRC + VAL), destbytes, DEST, destbytes, FALSE, TRUE
			if HINTS & (EQUAL_HINT 0)
				setbit STATUS, Z, SRC+VAL  ;set STATUS.Z for caller
			endif
			if HINTS & (COMPARE_HINT 0)
				setbit STATUS, C, IIF(SRC+VAL < MIN(SRC, VAL), C_CARRY, C_NOCARRY)  ;set STATUS.C for caller
			endif
			EXPAND_POP,
			EXITM
		endif
		if func == A'-'
;this doesn't save any space or time if we need to set STATUS.Z and STATUS.C
			mov_ _LITERAL(SRC - VAL), destbytes, DEST, destbytes, FALSE, TRUE
			if HINTS & (EQUAL_HINT 0)
				setbit STATUS, Z, (SRC == VAL)  ;set STATUS.Z for caller
			endif
			if HINTS & (COMPARE_HINT 0)
				setbit STATUS, C, IIF(SRC < VAL, C_BORROW, C_NOBORROW)  ;set STATUS.C for caller
			endif
			EXPAND_POP,
			EXITM
		endif
		if func == A'?'
			if HINTS & (EQUAL_HINT 0)
				setbit STATUS, Z, (SRC == VAL)  ;set STATUS.Z for caller
			endif
			if HINTS & (COMPARE_HINT 0)
				setbit STATUS, C, IIF(SRC < VAL, C_BORROW, C_NOBORROW)  ;set STATUS.C for caller
			endif
			EXPAND_POP,
			EXITM
		endif
		if func == A'A'
			mov_ _LITERAL(SRC & VAL), destbytes, DEST, destbytes, FALSE, TRUE
			if HINTS & (EQUAL_HINT 0)
				setbit STATUS, Z, (SRC & VAL)  ;set STATUS.Z for caller
			endif
			EXPAND_POP,
			EXITM
		endif
		if func == A'I'
			mov_ _LITERAL(SRC | VAL), destbytes, DEST, destbytes, FALSE, TRUE
			if HINTS & (EQUAL_HINT 0)
				setbit STATUS, Z, (SRC | VAL)  ;set STATUS.Z for caller
			endif
			EXPAND_POP,
			EXITM
		endif
		if func == A'X'
			mov_ _LITERAL(SRC ^ VAL), destbytes, DEST, destbytes, FALSE, TRUE
			if HINTS & (EQUAL_HINT 0)
				setbit STATUS, Z, (SRC ^ VAL)  ;set STATUS.Z for caller
			endif
			EXPAND_POP,
			EXITM
		endif
		ERRIF TRUE, [ERROR] ARITH: unknown function code func
		EXPAND_POP,
		EXITM
	endif

	LOCAL byte = IIF(func == A'?', 3, 0)  ;compare must be done MSB to LSB
	while byte != IIF(func == A'?', -1, 4)  ;0, 1, 2, 3 or 3, 2, 1, 0
		if DEST_#v(byte) != LIT0
;TODO: if VAL_#v(byte) == LIT0 && (func == A'-' || func == A'+') and8 STATUS, 1, WREG  ;saves 1 word and 1 instr cycle
			loadw_ VAL_#v(byte)
			if (byte > 0) && ((func == A'-') || (func == A'+'))  ;carry/borrow from previous byte
;CAUTION: MPASM trashes params within recursive macros, so we can't use ifbit here
;				ifbit STATUS, C, IIF(func == A'+', C_CARRY, C_BORROW), addlw 1,, NOJUMP_HINT  ;incw; this will still do the add/sub, even if result is 0
				TESTBIT(STATUS, C, func != A'+')  ;skip if no carry (+) or no borrow (-)
																				EXPAND  ;show generated code
	addlw 1
				NOEXPAND
			endif
			if ISLIT(SRC_#v(byte))
				if func == A'+'
																				EXPAND  ;show generated code
	addlw UNLIT(SRC_#v(byte))
					NOEXPAND
				endif
				if (func == A'-') || (func == A'?')
																				EXPAND  ;show generated code
	sublw UNLIT(SRC_#v(byte))
					NOEXPAND
				endif
				if func == A'A'
																				EXPAND  ;show generated code
	andlw UNLIT(SRC_#v(byte))
					NOEXPAND
				endif
				if func == A'I'
																				EXPAND  ;show generated code
	iorlw UNLIT(SRC_#v(byte))
					NOEXPAND
				endif
				if func == A'X'
																				EXPAND  ;show generated code
	xorlw UNLIT(SRC_#v(byte))
					NOEXPAND
				endif
			else
				BANKCHK SRC_#v(byte), MAYBE,
				if func == A'+'
																				EXPAND  ;show generated code
	addwf SRC_#v(byte), IIF(SRC == DEST, F, W)
					NOEXPAND
				endif
				if (func == A'-') || (func == A'?')
																				EXPAND  ;show generated code
	subwf SRC_#v(byte), IIF((SRC == DEST) && (func == A'-') && !NEED_COMP, F, W)
					NOEXPAND
					if NEED_COMP && (func == A'-')
																				EXPAND  ;show generated code
	sublw 0  ;0 - W to compensate for swapped operands
						NOEXPAND
					endif
				endif
				if func == A'A'
																				EXPAND  ;show generated code
	andwf SRC_#v(byte), IIF(SRC == DEST, F, W)
					NOEXPAND
				endif
				if func == A'I'
																				EXPAND  ;show generated code
	iorwf SRC_#v(byte), IIF(SRC == DEST, F, W)
					NOEXPAND
				endif
				if func == A'X'
																				EXPAND  ;show generated code
	xorwf SRC_#v(byte), IIF(SRC == DEST, F, W)
					NOEXPAND
				endif
			endif
			if (func == A'?') && (byte != 0)  ;short-circuit compare at most significant difference
;CAUTION: MPASM trashes params within recursive macros, so we can't use ifbit here
;				ifbit STATUS, Z, FALSE, LEAVE_PLACEHOLDER1 goto_ cmpdone#v(UNIQID),, ((hints 0) & ~(EQUAL_HINT COMPARE_HINT 0)) |
																				EXPAND  ;show generated code
	btfss STATUS, Z  ;skip if zero
				NOEXPAND
				LEAVE_PLACEHOLDER1 neargoto cmpdone#v(UNIQID)
			endif
			if (SRC != DEST) || NEED_COMP
				storew_ DEST_#v(byte)
			endif
		endif
byte += IIF(func == A'?', -1, 1)
	endw
	if func == A'?'  ;fill in unresolved forward references for short-circuit compares
 jumpdest cmpdone_#v(UNIQID):,,,, FALLTHRU
byte = 1
		while byte < destbytes  ;1, 2, 3, ..., destbytes-1
			FILL_PLACEHOLDER, neargoto cmpdone_#v(UNIQID)  ;KLUDGE: help assembler resolve forward reference address
byte += 1
		endw
	endif
	EXPAND_POP,
	ENDM


;convenience wrappers for incdec_:
inc8 MACRO reg
	NOEXPAND
	inc_(reg, reg, 1)
	ENDM

;inc into different target register:
inc8to MACRO reg, dest
	NOEXPAND
	inc_(reg, dest, 1)
	ENDM

inc8_ifnz MACRO reg, non0stmt, if0stmt, hints
	NOEXPAND
	incdec_ reg, reg, 1, TRUE, non0stmt, if0stmt, hints
	ENDM

dec8 MACRO reg
	NOEXPAND
	dec_(reg, reg, 1)
	ENDM

dec8_ifnz MACRO reg, non0stmt, if0stmt, hints
	NOEXPAND
	incdec_ reg, reg, 1, FALSE, non0stmt, if0stmt, hints
	ENDM

inc16 MACRO reg
	NOEXPAND
	inc_(reg, reg, 2)
	ENDM

inc16_ifnz MACRO reg, non0stmt, if0stmt, hints
	NOEXPAND
	incdec_ reg, reg, 2, TRUE, non0stmt, if0stmt, hints
	ENDM

dec16 MACRO reg
	NOEXPAND
	dec_(reg, reg, 2)
	ENDM

dec16_ifnz MACRO reg, non0stmt, if0stmt, hints
	NOEXPAND
	incdec_ reg, reg, 2, FALSE, non0stmt, if0stmt, hints
	ENDM

inc24 MACRO reg
	NOEXPAND
	inc_(reg, reg, 3)
	ENDM

dec24 MACRO reg
	NOEXPAND
	dec_(reg, reg, 3)
	ENDM

inc32 MACRO reg
	NOEXPAND
	inc_(reg, reg, 4)
	ENDM

dec32 MACRO reg
	NOEXPAND
	dec_(reg, reg, 4)
	ENDM

;reduce macro nesting a little:
#define inc_(reg, dest, bytes)  incdec_ reg, dest, bytes, TRUE,,,
#define dec_(reg, dest, bytes)  incdec_ reg, dest, bytes, FALSE,,,

;inc_ MACRO reg, dest, bytes
;	NOEXPAND
;	incdec_  reg, dest, bytes, TRUE,,,
;	ENDM

;dec_ MACRO reg, dest, bytes
;	NOEXPAND
;	incdec_  reg, dest, bytes, FALSE,,,
;	ENDM


	VARIABLE NESTED_INCDEC = 0  ;track nested macros; MPASM trashes params, so warn about this

;increment/decrement a 8/16/24/32-bit counter (little endian):
;multi-byte reg can cross bank boundary
;NOTE: if0 check on dec means it hits 0, not -1 (not a true wrap)
;NOTE: non0stmt must be 1 word if there is no if0stmt; otherwise if0/non0 stmt can be any size
;NOTE: PAGECHK if0stmt/non0stmt needs to call PAGECHK at the end if PCLATH has been altered; this macro can only set it at the start
;params:
; reg = source register
; dest = destination register
; bytes = #bytes in src and dest
; updown = flag telling whether to increment or decrement
; non0stmt = statement(s) to execute if dest does not become 0
; if0stmt = statement(s) to execute if dest becomes 0
; hints = execution hints; zero or more of:
;          BIGSTMT_HINT ==> ifstmt is more than 1 instr and must be jumped over (not needed if elsestmt used)
;          BOTHOK_HINT ==> okay to execute elsestmt first in order to avoid jump over (reduces code size); in this case ifstmt still needs to be 1 instruction (but BIGSTMT_HINT can override this)
;          NOJUMP_HINT ==> PCLATH does not need to be valid to execute ifstmt; BIGSTMT_HINT will override this
incdec_ MACRO reg, dest, bytes, updown, non0stmt, if0stmt, hints
	NOEXPAND  ;hide clutter in LST file
	LOCAL REG = reg, DEST = dest, UPDOWN = updown, BYTES = bytes  ;NOTE: assembler trashes parameter synbols if same macro is invoked multiple times on the same line
	LOCAL HINTS = hints 0  ;optional param; must end with "|" if passed
	WARNIF2 NESTED_INCDEC && FULL_WARN, [CAUTION] Recursive usage #v(NESTED_INCDEC) of incdec, params may be trashed  ;MPASM trashes params, so warn about this
NESTED_INCDEC += 1
	LOCAL NOTMY_HINTS = HINTS & ~(NOJUMP_HINT 0)  ;line too long
	WARNIF NOTMY_HINTS, [WARNING] Ignoring unrecognized INCDEC hint #v(NOTMY_HINTS)
	if TRUE  ;MEXPAND_STACKLO & 1
																				EXPAND  ;show INCDEC params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
BYTES = bytes  ;show INCDEC params
UPDOWN = updown  ;show INCDEC params
REG = reg  ;show INCDEC params
DEST = dest  ;show INCDEC params
HINTS = hints 0  ;show INCDEC params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	LOCAL REG_0 = REG, REG_1 = (REG+1), REG_2 = (REG+2), REG_3 = (REG+3)
	ERRIF ISLIT(reg), [ERROR] INCDEC: cannot increment/decrement a literal #v(bytes):#v(reg)
	if ISPSEUDO(REG)  ;preserve pseudo-registers
REG_1 = REG
REG_2 = REG
REG_3 = REG
	else
REG_1 = HIBYTE(REG)  ;make correction for non-contiguous upper byte 
		ERRIF !ISREG(REG, bytes), [ERROR] INCDEC: invalid multi-byte source register #v(bytes):#v(reg)
	endif
	LOCAL DEST_0 = DEST, DEST_1 = (DEST+1), DEST_2 = (DEST+2), DEST_3 = (DEST+3)
	ERRIF ISLIT(dest), [ERROR] INCDEC: cannot increment/decrement a literal #v(bytes):#v(dest)
	if ISPSEUDO(DEST)  ;preserve pseudo-registers
DEST_1 = DEST
DEST_2 = DEST
DEST_3 = DEST
	else
DEST_1 = HIBYTE(DEST)  ;make correction for non-contiguous upper byte 
		ERRIF !ISREG(DEST, BYTES), [ERROR] INCDEC: invalid multi-byte destination register #v(bytes):#v(dest)
;NOTE: we don't want PCL to be the destination because we would need to start with LSB (for carry/borrow purposes), and that would cause a jump before PCLATH was updated
		ERRIF (DEST == PCL) && (BYTES > 1), [ERROR] INCDEC: PCL cannot be destination for +/-
	endif
	if HINTS & (BOTHOK_HINT 0)  ;speculative IF; execute else stmt first to cut down on instructions (assumed to be benign)
																				EXPAND  ;show generated code
	if0stmt  ;do this one first to avoid the need to jump around it later if condition is false
		NOEXPAND
	endif
	if (HINTS & (NOJUMP_HINT BIGSTMT_HINT 0) != (NOJUMP_HINT 0)) || (BYTES > 1)  ;need to jump over
;NOTE: it appears that PCLATH does NOT need to be correct when executing an BTFSC/BTFSS instruction, so only set it if conditional instruction needs it
		PAGECHK $, MAYBE,  ;allow neargoto in short-circuit path
	endif
	LOCAL byte = 0
	while byte < BYTES  ;0, 1, 2, ... bytes-1
		if byte < BYTES-1  ;short-circuit non-MSB bytes
			bumpskipz REG_#v(byte), DEST_#v(byte), UPDOWN, TRUE
			LEAVE_PLACEHOLDER1 neargoto IFSTMT  ;no-wrap short circuit path
		endif
byte += 1
	endw
byte -= 1  ;correction for partial time thru loop at end
	BANKCHK REG, MAYBE,
	LEAVE_PLACEHOLDER_ IIF(DEST == REG, 1, 2), incdec-MSB  ;leave a placeholder for inc/dec of LSB; must analyze if0/non0 stmt before we know whether we need a conditional inc/dec here or not
	LOCAL MYPH = PLACEHOLDER_STACK  ;remember top of placeholder stack in case if0stmt or non0stmt adds more
;execute next instr if value is non-0 after inc/dec:
	ifelseblock#v(NESTED_INCDEC + NESTED_IFBIT) non0stmt, if0stmt, HINTS|
	START_FILL_PLACEHOLDER MYPH&,
	bumpskipz REG_#v(byte), DEST_#v(byte), UPDOWN, IFSTMT != ELSESTMT  ;is0#v == done#v  ;1 or 2 instr, depending on whether DEST == REG
	END_FILL_PLACEHOLDER MYPH&, TRUE
byte = 0
	while byte < BYTES-1  ;0, 1, 2, ..., bytes-2
		FILL_PLACEHOLDER (MYPH - BYTES + 1)&, neargoto IFSTMT  ;non0#v(UNIQID)  ;MYPH-1, MYPH-2, ..., MYPH-(BYTES-1) are all the same, so we can just specify the first one for the loop
byte += 1
	endw
	EXPAND_POP,
NESTED_INCDEC -= 1
	ENDM
 

;convenience wrappers:

shiftl8 MACRO reg
	NOEXPAND
	shiftl_(reg, 1, 1)
	ENDM

shiftl8to MACRO reg, dest
	NOEXPAND
	shiftto_ reg, dest, 1, TRUE, 1
	ENDM

;don't clear Carry first:
shiftl8_NC MACRO reg
	NOEXPAND
	shiftl_(reg, -1, 1)
	ENDM

;don't clear Carry first:
shiftl8to_NC MACRO reg, dest
	NOEXPAND
	shiftto_ reg, dest, -1, TRUE, 1
	ENDM

;shift multiple times:
multi_shiftl8 MACRO reg, count
	shiftl_(reg, 1, count)
	ENDM

shiftr8 MACRO reg
	NOEXPAND
	shiftr_(reg, 1, 1)
	ENDM

shiftr8to MACRO reg, dest
	NOEXPAND
	shiftto_ reg, dest, 1, FALSE, 1
	ENDM

;shift left 16 into another register (leaves source intact):
shiftl16to MACRO reg, dest
	NOEXPAND
	shiftto_ reg, dest, 2, TRUE, 1
	ENDM

;shift multiple times:
multi_shiftl16 MACRO reg, count
	shiftl_(reg, 2, count)
	ENDM

shiftl16 MACRO reg
	NOEXPAND
	shiftl_(reg, 2, 1)
	ENDM

shiftr16 MACRO reg
	NOEXPAND
	shiftr_(reg, 2, 1)
	ENDM

;don't clear Carry first:
shiftr16_NC MACRO reg
	NOEXPAND
	shiftr_(reg, -2, 1)
	ENDM

shiftl24 MACRO reg
	NOEXPAND
	shiftl_(reg, 3, 1)
	ENDM

shiftr24 MACRO reg
	NOEXPAND
	shiftr_(reg, 3, 1)
	ENDM

shiftl32 MACRO reg
	NOEXPAND
	shiftl_(reg, 4, 1)
	ENDM

shiftr32 MACRO reg
	NOEXPAND
	shiftr_(reg, 4, 1)
	ENDM

;reduce macro nesting a little:
#define shiftl_(reg, bytes, count)  shiftto_ reg, reg, bytes, TRUE, count
#define shiftr_(reg, bytes, count)  shiftto_ reg, reg, bytes, FALSE, count

;shiftl_ MACRO reg, bytes, count
;	NOEXPAND
;	shiftto_ reg, reg, bytes, TRUE, count
;	ENDM

;shiftr_ MACRO reg, bytes, count
;	NOEXPAND
;	shiftto_ reg, reg, bytes, FALSE, count
;	ENDM


;shift left/right a 8/16/24/32-bit register (little endian):
;variations include clear/no clear Carry first, place result in another register, etc.
;params:
; reg = register value to be shifted
; dest = destination where to store shifted value
; bytes = #bytes to shift; < 0 => chain (do not clear C)
; enlarge = flag telling whether to shift left or right
; count = #bit positions to shift
shiftto_ MACRO reg, dest, bytes, enlarge, count
	NOEXPAND  ;hide clutter in LST file
	LOCAL REG = reg
	LOCAL DEST = dest
	LOCAL BYTES = bytes
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show SHIFT params
REG = reg  ;show SHIFT params
DEST = dest  ;show SHIFT params
BYTES = bytes  ;show SHIFT params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	LOCAL byte = 0, bit = 0
	if BYTES < 0
BYTES = -BYTES
	else
		setbit STATUS, C, FALSE  ;make sure we don't shift in junk bits
	endif

	LOCAL REG_0 = REG, REG_1 = (REG+1), REG_2 = (REG+2), REG_3 = (REG+3)
	ERRIF ISLIT(reg), [ERROR] SHIFT: cannot shift a literal #(bytes):#v(reg)
	if ISPSEUDO(REG)  ;preserve pseudo-registers
REG_1 = REG
REG_2 = REG
REG_3 = REG
	else
REG_1 = HIBYTE(REG)  ;make correction for non-contiguous upper byte 
		ERRIF !ISREG(REG, BYTES), [ERROR] SHIFT: invalid multi-byte source register #v(bytes):#v(reg)
;NOTE: we don't want PCL to be the destination because we would need to start with LSB (for carry/borrow purposes), and that would cause a jump before PCLATH was updated
		ERRIF (REG == PCL) && (bytes > 1), [ERROR] SHIFT: PCL cannot be destination for +/-
	endif
	LOCAL DEST_0 = DEST, DEST_1 = (DEST+1), DEST_2 = (DEST+2), DEST_3 = (DEST+3)
	ERRIF ISLIT(dest), [ERROR] SHIFT: destination cannot be a literal #v(bytes):#v(dest)
	if ISPSEUDO(DEST)  ;preserve pseudo-registers
DEST_1 = DEST
DEST_2 = DEST
DEST_3 = DEST
	else
DEST_1 = HIBYTE(DEST)  ;make correction for non-contiguous upper byte 
		ERRIF !ISREG(DEST, BYTES), [ERROR] SHIFT: invalid multi-byte source register #v(bytes):#v(dest)
;NOTE: we don't want PCL to be the destination because we would need to start with LSB (for carry/borrow purposes), and that would cause a jump before PCLATH was updated
		ERRIF (REG == PCL) && (bytes > 1), [ERROR] INCDEC: PCL cannot be destination for +/-
	endif

	while bit < (count)  ;0, 1, 2, ..., count-1
byte = 0
		while byte < BYTES  ;0, 1, 2, ... bytes-1
			BANKCHK REG_#v(byte), MAYBE,
			if enlarge  ;left
																				EXPAND  ;show generated code
	rlf REG_#v(byte), IIF(REG_#v(byte) == DEST_#v(byte), F, W)  ;LSB first
				NOEXPAND
				if REG_#v(byte) != DEST_#v(byte)
					storew_ DEST_#v(byte)
				endif
			else  ;right
																				EXPAND  ;show generated code
	rrf REG_#v(BYTES-byte-1), IIF(REG_#v(BYTES-byte-1) == DEST_#v(BYTES-byte-1), F, W)  ;MSB first
				NOEXPAND
				if REG_#v(BYTES - byte - 1) != DEST_#v(BYTES - byte - 1)
					storew_ DEST_#v(BYTES-byte-1)
				endif
			endif
byte += 1
		endw
bit += 1
	endw
	EXPAND_POP,
	ENDM


;#define mult8_2x(src, dest)  mult_2x 1, src, dest
;#define mult16_2x(src, dest)  mult_2x 2, src, dest

;multiply by 2:
;NOTE: result can be larger than operands
;mult8to16_2x MACRO reg
;	mult_2xto 1, reg, reg
;	ENDM

;mult_2xto MACRO bytes, src, dest
;	add#v(8*bytes) src, src, dest,  ;2*value
;	ifbit CARRY, inc8 (dest)+bytes,, NOJUMP_HINT
;	ENDM


;multiply a register by 3 (destructive):
;params:
; reg = src/dest register
mult8_3x MACRO reg
	NOEXPAND
	mult8_3xto reg, reg
	ENDM

;multiply a register by 3:
;non-destructive if dest != src
;params:
; src = source register
; dest = destination register
mult8_3xto MACRO src, dest
	NOEXPAND
	add8 src, src, WREG,  ;2*value
	add8 src, WREG, dest,  ;3*value
	ENDM


;multiply a register by 4:
;non-destructive if dest != src
;params:
; src = source register
; dest = destination register
mult8_4xto MACRO src, dest
	NOEXPAND  ;hide clutter in LST file
	if src == dest
		multi_shiftl8 src, 2  ;destructive
	else
		if dest != WREG
			add8 src, src, dest,  ;2*value
			add8 dest, dest, dest,  ;4*value
		else
			add8 src, src, WREG,  ;2*value
			add8 src, WREG, WREG,  ;3*value
			add8 src, WREG, dest,  ;4*value
		endif
	endif
	ENDM


;multiply a register by 10 (destructive):
;params:
; reg = src/dest register
mult8_10x MACRO reg
	NOEXPAND  ;hide clutter in LST file
	add8 reg, reg, reg,  ;2*value
	add8 reg, reg, WREG,  ;4*value
	add8 reg, WREG, reg,  ;6*value, W still = 4*value
	add8 reg, WREG, reg,  ;10*value
	ENDM


;multiply a register by 16 (destructive):
;params:
; reg = src/dest register
mult8_16x MACRO reg
	NOEXPAND
	mult8_16xto reg, reg
	ENDM

;multiply a register by 16:
;non-destructive if dest != src
;params:
; src = source register
; dest = destination register
mult8_16xto MACRO src, dest
	NOEXPAND
;this is faster than shifting 4x, but doesn't set Carry
	swap_ src, WREG
	and8 WREG, _LITERAL(0xf0), dest,
	ENDM


;divide a register by 16:
;non-destructive if dest != src
;params:
; src = source register
; dest = destination register
div8_16xto MACRO src, dest
	NOEXPAND
;this is faster than shifting 4x, but doesn't set Carry
	gettop4 src, dest
	ENDM


;divide by 2:
;non-destructive if dest != src
;params:
; src = source register
; dest = destination register
div8_2xto MACRO reg, dest
	NOEXPAND
	shiftr8to reg, dest
	ENDM


;multiply an unsigned 16 bit register/literal by another:
;for now, this just uses an Add loop, and assumes that the value is small and non-0
;TODO: use a real 8/16/24/32-bit function here
;params:
; src = source register
; val = multiplier value; can be register or literal value, but assumed to be small
; dest = destination register
mult16 MACRO src, val, dest
	NOEXPAND  ;hide clutter in LST file
	LOCAL mult_loop
	mov16 _LITERAL(0), dest
	if ISLIT(val)
		REPEAT UNLIT(val), ADD16_VAL2REG(src, dest)
	else
		dec8 val
 jumpdest mult_loop:,,,, FALLTHRU  ;bank already selected
		ADD16_VAL2REG(src, dest)
		dec8_ifnz val, neargoto mult_loop,,
	endif
	ENDM


;convenience wrappers:
div8 MACRO num, den, quo, rem, temp
	NOEXPAND
	div_ num, 1, den, 1, quo, rem, temp,
	ENDM

div16 MACRO num, den, quo, rem, temp
	NOEXPAND
	div_ num, 2, den, 2, quo, rem, temp,
	ENDM

div24 MACRO num, den, quo, rem, temp
	NOEXPAND
	div_ num, 3, den, 3, quo, rem, temp,
	ENDM

div32 MACRO num, den, quo, rem, temp
	NOEXPAND
	div_ num, 4, den, 4, quo, rem, temp,
	ENDM

div16_8 MACRO num, den, quo, rem, temp
	NOEXPAND
	div_ num, 2, den, 1, quo, rem, temp,
	ENDM

div32_8 MACRO num, den, quo, rem, temp
	NOEXPAND
	div_ num, 4, den, 1, quo, rem, temp,
	ENDM

div32_16 MACRO num, den, quo, rem, temp
	NOEXPAND
	div_ num, 4, den, 2, quo, rem, temp,
	ENDM


;divide an unsigned 8/16/24/32 bit register/literal by another:
;this is EXPENSIVE; use sparingly!
;division algorithm is based on http://www.piclist.com/techref/microchip/math/div/tutorial.htm, which claims worst case is 144 instr cycles (30 usec at 20 MHz) for 8 bits
;numerator and denominator can be different sizes
;numerator, quotient and remainder are assumed to be same size (even though remainder won't end up bigger than denominator, it's used as a temp copy of numerator)
;denominator and temp are assumed to be the same size
;params:
; numerator
; numbytes = #bytes in numerator
; denominator
; denbytes = #bytes in denominator
; quotient; cannot be same register as denominator (gets overwritten during)
; remainder
; divtemp = a temp variable used for intermediate results; must be same size as denominator
; hints = execution hints; (reserved for future optimization of DIV)
div_ MACRO numerator, numbytes, denominator, denbytes, quotient, remainder, divtemp, hints
	NOEXPAND  ;hide clutter in LST file
	WARNIF (hints 0) & ~(0), [WARNING] Ignoring unrecognized DIV hint #v(hints 0)
	LOCAL NUM = numerator, DEN = denominator
	LOCAL QUO = quotient, REM = remainder
	LOCAL NUMBYTES = numbytes, DENBYTES = denbytes
	LOCAL QUOBIT = divtemp, SVDEN = DEN
	LOCAL HINTS = hints 0  ;optional param; must end with "|" if passed
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show DIV params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
NUM = numerator  ;show DIV params
NUMBYTES = numbytes  ;show DIV params
DEN = denominator  ;show DIV params
DENBYTES = denbytes  ;show DIV params
QUO = quotient  ;show DIV params
REM = remainder  ;show DIV params
QUOBITS = divtemp  ;show DIV params
HINTS = hints 0  ;show DIV params
		NOEXPAND
	endif
	EXPAND_PUSH TRUE  ;show what this does
;assembler bug; save as LOCALs to force eval (else "missing operand" and "missing argument" errors)
	LOCAL shift, subloop, toobig, done, enlarge
;TODO: add 16-bit pseudo-registers
	ERRIF ISLIT(QUO), [ERROR] DIV: quotient destination #v(QUO) cannot be a literal
	ERRIF ISLIT(REM), [ERROR] DIV: remainder destination #v(REM) cannot be a literal
	if ISLIT(NUM)  ;trim off extra bits
		if (numbytes) < 4
NUM = _LITERAL(UNLIT(NUM)%0x1000000)
		endif
		if (numbytes) < 3
NUM = _LITERAL(UNLIT(NUM)%0x10000)
		endif
		if (numbytes) < 2
NUM = _LITERAL(UNLIT(NUM)%0x100)
		endif
	endif
	if ISLIT(DEN) && ISLIT(NUM)  ;constant case: divide literal by literal
		if (UNLIT(DEN)/0x1000000) !=  0xff  ;KLUDGE: caller probably didn't want top bit
DEN &= ~ASM_MSB
		endif
		if (UNLIT(NUM)/0x1000000) !=  0xff  ;KLUDGE: caller probably didn't want top bit
NUM &= ~ASM_MSB
		endif
		mov_ _LITERAL(NUM/DEN), numbytes, QUO, numbytes, FALSE, TRUE
		mov_ _LITERAL(NUM%DEN), numbytes, REM, numbytes, FALSE, TRUE
		EXPAND_POP,
		EXITM
	endif
	ERRIF (denbytes) != (numbytes), [ERROR] DIV: Numerator #v(numbytes) and denominator #v(denbytes) must be same size  ;TODO: fix this
	movnop_ NUM, numbytes, REM, numbytes, FALSE, TRUE
	ERRIF ISLIT(DEN), [ERROR] DIV: Denominator cannot be a literal #v(DEN)  ;TODO
	ERRIF QUO == DEN, [ERROR] DIV: Denominator #v(DEN) and quotient #v(QUO) cannot be the same  ;TODO
	mov_ _LITERAL(0), numbytes, QUO, numbytes, FALSE, TRUE
	mov_ _LITERAL(1), denbytes, QUOBIT, denbytes, FALSE, TRUE  ;loop control bit
	ifelse_ DEN, denbytes, EQ, _LITERAL(0), denbytes, LEAVE_PLACEHOLDER1 neargoto done,,  ;avoid junk answer and unneeded work
 jumpdest enlarge:, NUM|, DEN|,, FALLTHRU  ;find largest n-bit multiple of denominator
	shiftl_ QUOBIT, denbytes, 1
	shiftl_ DEN, denbytes, 1
	ifbit DEN+(denbytes)-1, 7, TRUE, LEAVE_PLACEHOLDER1 neargoto subloop,,  ;check if highest bit reached yet (little endian)
	ifelse_ NUM, numbytes, GE, DEN, denbytes, neargoto enlarge,,  ;no need to keep shifting if it's already larger than numerator
 jumpdest subloop:, NUM|, DEN|,, FALLTHRU
	FILL_PLACEHOLDER, neargoto subloop
	ifelse_ REM, numbytes, LT, DEN, denbytes, LEAVE_PLACEHOLDER1 neargoto toobig,,
	sub_(REM, numbytes, DEN, denbytes, REM,)
	add_(QUO, numbytes, QUOBIT, denbytes, QUO,)
 jumpdest toobig:, DEN|, REM|,, FALLTHRU  ;prepare next smaller multiple of denominator
	FILL_PLACEHOLDER, neargoto toobig
	shiftr_  QUOBIT, denbytes, 1
	ifbit STATUS, C, TRUE, LEAVE_PLACEHOLDER1 neargoto done,,
	shiftr_ DEN, denbytes, 1
	neargoto subloop
 jumpdest done:, DEN|, QUOBIT|,,
	FILL_PLACEHOLDER, neargoto done
	FILL_PLACEHOLDER, neargoto done
	EXPAND_POP,
	ENDM


;wrappers for 8/16/24/32-bit divide (cuts down on code size):

DIV8 MACRO num, den, quo, rem
	NOEXPAND
	DIVEXPANDER 1, num, den, quo, rem
	ENDM

;cast numerator + denominator parameters to 32 bits, then use 32-bit divide:
DIV8by16_as32 MACRO num, den, quo, rem
	NOEXPAND
	mov8nop num, DIV32_NUM
	mov24 _LITERAL(0), DIV32_NUM+1  ;8-to-32 up-cast (unsigned)
	mov16nop den, DIV32_DEN
	mov16 _LITERAL(0), DIV32_DEN+2  ;16-to-32 up-cast (unsigned)
	DIVEXPANDER 4, DIV32_NUM, DIV32_DEN, quo, rem
	ENDM

;cast numerator + denominator parameters to 32 bits, then use 32-bit divide:
DIV16_as32 MACRO num, den, quo, rem
	NOEXPAND
	mov16nop num, DIV32_NUM
	mov16 _LITERAL(0), DIV32_NUM+2  ;16-to-32 up-cast (unsigned)
	mov16nop den, DIV32_DEN
	mov16 _LITERAL(0), DIV32_DEN+2  ;16-to-32 up-cast (unsigned)
	DIVEXPANDER 4, DIV32_NUM, DIV32_DEN, quo, rem
	ENDM

DIV16 MACRO num, den, quo, rem
	NOEXPAND
	DIVEXPANDER 2, num, den, quo, rem
	ENDM

;cast numerator + denominator parameters to 32 bits, then use 32-bit divide:
DIV24by16_as32 MACRO num, den, quo, rem
	NOEXPAND
	mov24nop num, DIV32_NUM
	mov8 _LITERAL(0), DIV32_NUM+3  ;24-to-32 up-cast (unsigned)
	mov16nop den, DIV32_DEN
	mov16 _LITERAL(0), DIV32_DEN+2  ;16-to-32 up-cast (unsigned)
	DIVEXPANDER 4, DIV32_NUM, DIV32_DEN, quo, rem
	ENDM

DIV24 MACRO num, den, quo, rem
	NOEXPAND
	DIVEXPANDER 3, num, den, quo, rem
	ENDM

DIV32 MACRO num, den, quo, rem
	NOEXPAND
	DIVEXPANDER 4, num, den, quo, rem
	ENDM


	VARIABLE DIV_EXPANDED = 0  ;bitmap of which parts have been expanded

;allocate temps and expand code (once only):
;params:
; bytes = size of operands
; num = numerator
; den = denominator
; quo = quotient
; rem = remainder
DIVEXPANDER MACRO bytes, num, den, quo, rem
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH !(DIV_EXPANDED & (1<<bytes))
	if !(DIV_EXPANDED & (1<<bytes))  ;avoid redundant code and data expansions; expand first time only
DIV_EXPANDED |= 1<<bytes
		if DIV_EXPANDED == 1<<bytes  ;first time; define variables also
;32-bit arith temps (num == rem):
;this uses 16 bytes of temp space (12 banked, 4 non-banked)
			BDCL DIV32_BTEMPS:0  ;CAUTION: define symbol to allow this space to be reused elsewhere
			BDCL DIV32_DEN:4
			BDCL DIV32_QUO:4
			BDCL DIV32_NUM:0  ;CAUTION: NUM == REM
			BDCL DIV32_REM:4
			BDCL DIV32_BTEMPS_END:0
			REALLOC_SPACE DIV32_BTEMPS, DIV32_BTEMPS_END  ;allow this space to be reallocated; caller must ensure mutually exclusive usage
			NBDCL DIV32_NBTEMP:4  ;this one is non-banked to cut down on bank switching in putdec32 (uses EE read for tables)
		endif
		LOCAL around_div#v(8*bytes), svbank = PREVBANK
		LEAVE_PLACEHOLDER1JUMP neargoto around_div#v(8*bytes)

		UGLY_PASS12FIX 1  ;KLUDGE: assembler needs help
 jumpdest div#v(8*bytes)_:, DIV32_NUM|, DIV32_DEN|,,
		div#v(8*bytes) DIV32_NUM, DIV32_DEN, DIV32_QUO, DIV32_REM, DIV32_NBTEMP
		return_ div#v(8*bytes)_

 jumpdest around_div#v(8*bytes):, svbank|,,,
		FILL_PLACEHOLDER, neargoto around_div#v(8*bytes)
	endif
	mov#v(8*bytes)nop num, DIV32_NUM
	mov#v(8*bytes)nop den, DIV32_DEN
	call_ div#v(8*bytes)_
	mov#v(8*bytes)nop DIV32_QUO, quo
	mov#v(8*bytes)nop DIV32_REM, rem
	EXPAND_POP,
	ENDM


#if FALSE  ;test/debug
djtest:
	CBLOCK NEXT_BRAM
	REM:0, NUM:4  ;NUM+REM
	QUO:4  ;QUO
	DEN:4  ;DEN
	DIVTEMP:4
	SAFE_ENDC TRUE
	EXPAND_PUSH TRUE

;	cmp8    REM, WREG
;	nop
;	mov32  _LITERAL(0x1A345678), NUM
;	shiftl32 NUM
;	nop
;	mov32  _LITERAL(0x1A345678), NUM
;	shiftr32 NUM
;	nop
;	mov32  _LITERAL(0), NUM
;	sub32_8 NUM, _LITERAL(1), NUM
;	nop
;	add32_8 NUM, _LITERAL(2), NUM
;	nop
	mov8   _LITERAL(d'231'), NUM
	mov8   _LITERAL(d'10'), DEN
	div8    NUM, DEN, QUO, REM, DIVTEMP
	nop
	mov32    _LITERAL(d'25'), NUM
;	mov8    _LITERAL(d'25'), DEN
;	div32_8    NUM, DEN, QUO, REM, DIVTEMP
	mov32    _LITERAL(d'25'), DEN
	div32    NUM, DEN, QUO, REM, DIVTEMP
	nop
	mov32   _LITERAL(d'1234567890'), NUM
;	mov8    _LITERAL(d'2'), DEN
;	div32_8    NUM, DEN, QUO, REM, DIVTEMP
	mov32   _LITERAL(d'2'), DEN
	div32    NUM, DEN, QUO, REM, DIVTEMP
	nop
	mov32   _LITERAL(d'20'), NUM
	mov32   _LITERAL(d'1234567890'), DEN
	div32    NUM, DEN, QUO, REM, DIVTEMP
	nop
	div32_16  _LITERAL(d'1234567890'), _LITERAL(d'10000'), QUO, REM, DIVTEMP
	nop
;;	mov16 _LITERAL(d'10000'), DEN
;;	div32_16  _LITERAL(d'1234567890'), DEN, QUO, REM, DIVTEMP
;	mov32 _LITERAL(d'10000'), DEN
;	div32  _LITERAL(d'1234567890'), DEN, QUO, REM, DIVTEMP
;	nop
	mov32 _LITERAL(d'1234567890'), NUM
;	div32_8  NUM, _LITERAL(d'10000'), QUO, REM, DIVTEMP
	div32  NUM, _LITERAL(d'10000'), QUO, REM, DIVTEMP
	nop
	mov16    _LITERAL(d'65530'), NUM
;	mov8    _LITERAL(d'100'), DEN
;	div16_8    NUM, DEN, QUO, REM, DIVTEMP
	mov16   _LITERAL(d'100'), DEN
	div16    NUM, DEN, QUO, REM, DIVTEMP
	nop
;	mov32   _LITERAL(d'1234567890'), NUM
;;	mov16   _LITERAL(d'11111'), DEN
;;	div32_16    NUM, DEN, QUO, REM, DIVTEMP
;	mov32   _LITERAL(d'11111'), DEN
;	div32    NUM, DEN, QUO, REM, DIVTEMP
;	nop
	goto_ $
	EXPAND_POP,
	return
#endif
 

	VARIABLE MEMSET_EXPANDED = FALSE  ;avoid redundant code and data expansions

;fill memory with a register/literal value, can span RAM banks:
;NOTE: assumes len > 1, otherwise there would be no reason to use this
;assumes memory to be set does not cross a bank boundary
;uses stack pointer for large or variable length
;params:
; dest = destination start address
; len = #bytes to fill
; val = register or literal value to fill with
memset_ MACRO dest, len, val
;assembler bug; save as LOCALs to force eval (else "missing operand" and "missing argument" errors)
	NOEXPAND  ;hide clutter in LST file
	LOCAL DEST = dest, LEN = len, VAL = val
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show MEMSET params
DEST = dest  ;show MEMSET params
LEN = len  ;show MEMSET params
VAL = val  ;show MEMSET params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	LOCAL ofs = 0
	LOCAL loop
	if ISLIT(DEST) && ISLIT(LEN) && (UNLIT(LEN) <= d'10')  ;length known at compile time, just expand in-line
		loadw_ val
		REPEAT UNLIT(LEN), storew_ UNLIT(dest)+repeater
	else  ;use indirect addressing (space-efficient if len is large or varies at run-time)
		ERRIF VAL == WREG, [ERROR] MEMSET value in WREG will be lost  ;TODO: need a temp in order to use W
		EXPAND_PUSH TRUE  ;show what this does
		if ISLIT(DEST) || (DEST != INDF)
			mov8 _LITERAL(LOW(dest)), FSR
			BANKSEL_INDIRECT(UNLIT(DEST))  ;need to set IRP for bank 2/3
		endif
		if !MEMSET_EXPANDED
MEMSET_EXPANDED = TRUE
			NBDCL MEMSET_LEN
		endif
		mov8 len, MEMSET_LEN
		loadw_ val
 jumpdest loop:,,,, FALLTHRU
		storew_ STACK
		dec8_ifnz MEMSET_LEN, neargoto loop,
		EXPAND_POP,
	endif
	EXPAND_POP,
	ENDM


;******************************************************************************
;tty helper functions + macros section:
;provides a few stdio-like serial I/O functions
;all functions are wrapped as macros so they take no code space unless used
;the PIC's 2-byte hardware buffers are safe to use as-is for serial pass-thru only if rate of output <= rate of input
;if output rate > input rate, software output buffer would be needed in order to prevent thread blocking
;if we can't keep up with input rate, software input buffer would be needed in order to prevent receiver overruns
;interrupts are NOT enabled, but the interrupt status bits ARE used (via polling)
;these functions should NOT be called from within ISR since they use busy-waits
;caller must supply YIELD macro for cooperative multi-tasking
;******************************************************************************

#ifndef IO_STATS
 #define IO_STATS  0  ;default to Off if not specified
#endif  ;IO_STATS


;initialize tty stats:
TTY_init MACRO
	BDCL IDLE_COUNT32:4  ;32-bit idle counter; busy-waits on I/O are considered to be idle time
	mov32 _LITERAL(0), IDLE_COUNT32
#if IO_STATS
	BDCL IO_COUNT24:3  ;total I/O bytes (actually, just input bytes); will wrap after ~ 24 minutes at 115K baud constant I/O 
	BDCL OERR_COUNT  ;overrun I/O error counts (won't wrap past 255)
	BDCL LAST_OERR24:3  ;packet# of last overflow error
	BDCL FERR_COUNT  ;framing I/O error counts (won't wrap past 255)
	BDCL LAST_FERR24:3  ;packet# of last frame error
	BDCL JUNK_COUNT  ;junk char count (won't wrap past 255)
	BDCL TIMEOUT_COUNT  ;packet timed out (incomplete before time limit expired)

	mov24 _LITERAL(0), IO_COUNT24
	mov8 _LITERAL(0), OERR_COUNT
	mov24 _LITERAL(0), LAST_OERR24
	mov8 _LITERAL(0), FERR_COUNT
	mov24 _LITERAL(0), LAST_FERR24
	mov8 _LITERAL(0), JUNK_COUNT
	mov8 _LITERAL(0), TIMEOUT_COUNT
#endif
	ENDM


;check if EUSART input/output buffers are empty/full:
#define IsCharAvailable(yesno)  PIR1, RCIF, yesno  ;ifbit params
#define IsRoomAvailable(yesno)  PIR1, TXIF, yesno  ;ifbit params


;comma-less extra parameters for GETCHAR/PUTCHAR:
;KLUDGE because MPASM doesn't allow nested macro calls with params
#define NOBLOCKIO  |0x1000  ;non-blocking; check and get it if it is there; otherwise, don't wait
#define NOWAITIO  |0x2000  ;do not wait (don't even check); assumes char is already available
#define TIMEOUT_20msec_start  |0x4000  ;start a new timeout period
#define TIMEOUT_20msec_continue  |0x8000  ;continue an existing timeout period


	VARIABLE GETCHAR_EXPANDED = FALSE  ;avoid redundant code and data expansions

;get an input char from serial port (using EUSART):
;NOTE: must be called at least each 87 usec (1 char time) to keep up with incoming data stream at 115K baud
;if caller requests, this function will busy-wait if the input buffer is empty (since there is nothing else to do, anyway)
;caller must supply YIELD macro for multi-tasking
;uses EVT_20msec for timeouts
;params:
; reg = destination for received char; can use optional qualifers (listed above) to control wait behavior
GETCHAR MACRO reg
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH !GETCHAR_EXPANDED
	if !GETCHAR_EXPANDED  ;avoid redundant code and data expansions
GETCHAR_EXPANDED = TRUE
		LOCAL around_getchar, svbank = PREVBANK
		LEAVE_PLACEHOLDER1JUMP neargoto around_getchar

		BITDCL GetChar_Timeout

;main tty input function:
;set up timeout handling:
 jumpdest getch_timeout_start:,,, OFBANK(0)|,  ;return with bank 0, since that is where next data typically is
		setbit BITVAR(EVT_pre20msec), FALSE  ;initialize event notification
		setbit BITVAR(EVT_20msec), FALSE
 jumpdest getch_timeout_continue:,,, OutBankOf(getch_timeout_start)|, FALLTHRU  ;common return must use same bank affinity
		setbit BITVAR(GetChar_Timeout), TRUE  ;watch for timeout notification
		neargoto getch_wait
		UGLY_PASS12FIX -1  ;KLUDGE: assembler needs help resolving addresses
		UGLY_PASS12FIX690 -1  ;KLUDGE: assembler needs help resolving addresses
 jumpdest getch_ferr:, RCSTA|,,,  ;carry over bank select from source address
#if IO_STATS
		inc8_ifnz FERR_COUNT,, dec8 FERR_COUNT, NOJUMP_HINT  ;don't let it wrap
		mov24 IO_COUNT24, LAST_FERR24  ;gives an indication of how long ago it happened
#endif
		setbit RCSTA, SPEN, FALSE  ;turn off receive to reset FERR (page 92)
		setbit RCSTA, SPEN, TRUE  ;turn on receive again
		REPEAT 3, loadw_ RCREG  ;try to clear Framing error
;?		ifbit RCSTA, FERR, FALSE, neargoto getch  ;error cleared
 jumpdest getch_empty:, PIR1|,,, FALLTHRU  ;make bank select consistent when falling thru
		ifbit BITVAR(GetChar_Timeout), FALSE, neargoto getch_yield,,  ;not waiting for timeout
		ifbit BITVAR(EVT_20msec), FALSE, neargoto getch_yield,,
#if IO_STATS
		inc8_ifnz TIMEOUT_COUNT,, dec8 TIMEOUT_COUNT, NOJUMP_HINT  ;don't let it wrap
#endif
		return_ getch_empty
 jumpdest getch_yield:, PIR1|, BITVADRS(EVT_20msec)|,,  ;LONGJUMP_HINT|
		YIELD TRUE  ;handle other tasks; NOTE: must reset watchdog timer occassionally
 jumpdest getch_wait:,,, OutBankOf(getch_timeout_start)|, FALLTHRU  ;common return must use same bank affinity
		ifbit IsCharAvailable(FALSE), neargoto getch_empty,,  ;input buffer empty; wait for it to fill
 jumpdest getch_nocheck:, ,, OutBankOf(getch_timeout_start)|, FALLTHRU  ;common return must use same bank affinity
;?		ifbit RCSTA, SPEN, FALSE, neargoto rxoff
;?		ifbit RCSTA, CREN, FALSE, neargoto rxoff
		ifbit RCSTA, FERR, TRUE, neargoto getch_ferr,,  ;must read FERR before RCREG (page 91)
#if IO_STATS
		inc24 IO_COUNT24  ;count successful chars only
#endif
		setbit BITVAR(GetChar_Timeout), FALSE  ;turn off timeout notification (for next time)
		loadw_ RCREG  ;NOTE: don't do anything below that will trash W!
;		SET_PK1LEDS(WREG)  ;DEBUG rx
		ifbit RCSTA, OERR, FALSE, return_ getch_empty,,  ;no overrun error
#if IO_STATS
		inc8_ifnz OERR_COUNT,, dec8 OERR_COUNT, NOJUMP_HINT  ;don't let it wrap
		mov24 IO_COUNT24, LAST_OERR24  ;gives an indication of how long ago it happened; THIS TRASHES WREG
#endif
;?		setbit RCSTA, SPEN, FALSE  ;turn off receive to reset OERR (page 92)
;?		setbit RCSTA, SPEN, TRUE  ;turn on receive again
		setbit RCSTA, CREN, FALSE  ;turn off receive to reset OERR (page 92)
		setbit RCSTA, CREN, TRUE  ;turn on receive again
		return_ getch_wait

 jumpdest around_getchar:, svbank|,,,
		FILL_PLACEHOLDER, neargoto around_getchar
	endif
	WARNIF (reg) == WREG, [CAUTION] Non-wait I/O not implemented for W #v(0 NOBLOCKIO NOWAITIO TIMEOUT_20msec_start TIMEOUT_20msec_continue)  ;these bits will be "hidden" if they are ORed with WREG (-8)
	if (reg) == WREG
		call_ getch_wait  ;busy-wait
		EXPAND_POP,
		EXITM
	endif
	if (reg) & (0 NOWAITIO)  ;do not wait (don't even check); assumes char is already available
		call_ getch_nocheck
	else
		if (reg) & (0 NOBLOCKIO)  ;non-blocking; check and get it if it is there; otherwise, don't wait
			loadw_ _LITERAL(0)
			ifbit IsCharAvailable(TRUE), call_ getch_nocheck,,  ;non-blocking: only check it once, then get char if available
		else
			if (reg) & (0 TIMEOUT_20msec_start)  ;start a new timeout period
				call_ getch_timeout_start
			else
				if (reg) & (0 TIMEOUT_20msec_continue)  ;continue en existing timeout period
					call_ getch_timeout_continue
				else  ;busy-wait
					call_ getch_wait  ;busy-wait
				endif
			endif
		endif
	endif
	storew_ (reg) & ~(0 NOWAITIO NOBLOCKIO TIMEOUT_20msec_start TIMEOUT_20msec_continue)
	EXPAND_POP,
	ENDM


;convenience wrapper for literal chars:
PUTLITCHAR MACRO char
	PUTCHAR _LITERAL(char)
	ENDM


	VARIABLE PUTCHAR_EXPANDED = 0  ;bitmap of code pages where code has been expanded

;send an output char to serial port (using EUSART):
;if caller requests, this function will busy-wait if the output buffer is full (since there is nothing else to do, anyway)
;that should never happen as long as output stream is shorter than input stream, but could happen for debug/console output
;caller must supply YIELD macro for multi-tasking
;NOTE: code will be expanded once only per code page (full expansion first time only, then just thunk layers for other code pages)
;this allows a short jump no matter what PCLATH is currently set to
;params:
; char = character to send; can be literal or register
PUTCHAR MACRO char
	NOEXPAND  ;hide clutter in LST file
;	LOCAL CHAR = char
	EXPAND_PUSH !(PUTCHAR_EXPANDED & (1<<(7 - LITPAGEOF($))))
	if !(PUTCHAR_EXPANDED & (1<<(7 - LITPAGEOF($))))  ;avoid redundant code and data expansions; expand first time only for each code page
PUTCHAR_EXPANDED |= 1<<(7 - LITPAGEOF($))  ;lowest page uses highest bit so HIGHBIT will tell which page it's in
		LOCAL around_putchar, svbank = PREVBANK
		LEAVE_PLACEHOLDER1JUMP neargoto around_putchar

		if PUTCHAR_EXPANDED == 1<<(7 - LITPAGEOF($))  ;first time; full expansion
;main tty output function:
;this busy-wait should take no longer than 87 usec at 115200 baud (1 char time)
 jumpdest putch_busy:, PIR1|,,,  ;LONGJUMP_HINT|  ;carry over bank select from source address
			YIELD FALSE  ;handle other tasks
 jumpdest putch#v(7 - LITPAGEOF($))_wait:, BT_UNKNOWN|,, TXREG|, FALLTHRU
;			setbit PIE1, TXIE, TRUE  ;enable EUSART transmit interrupt now that there's something more to send
			ifbit IsRoomAvailable(FALSE), neargoto putch_busy,,  ;output buffer full; wait for it to drain
;NO			storew_ TXREG  ;can't do this here; YIELD might have trashed W
			return_ putch#v(7 - LITPAGEOF($))_wait
		else  ;not first time; thunk layer only
;put a local entry point here, then far jump to the real one:
;this avoids PCLATH code space overhead in caller for frequently-used functions
 jumpdest putch#v(7 - LITPAGEOF($))_wait:,,, OutBankOf(putch#v(HIGHBIT_#v(PUTCHAR_EXPANDED) - 1)_wait)|, LONGJUMP_HINT  ;common return must use same bank affinity, PCLATH is altered
			fargoto putch#v(HIGHBIT_#v(PUTCHAR_EXPANDED) - 1)_wait  ;busy-wait
		endif

 jumpdest around_putchar:, svbank|,,,
		FILL_PLACEHOLDER, neargoto around_putchar
	endif
	ERRIF (char) == WREG, "[ERROR] PUTCHAR overrides W, can't use it as operand"
	if (char) & (0 NOBLOCKIO)  ;non-blocking; check and send it if space is available; otherwise, don't wait
		loadw_ (char) & ~(0 NOWAITIO NOBLOCKIO)
		ifbit IsRoomAvailable(TRUE), storew_ TXREG,,  ;non-blocking: only check it once, then send char if ready
	else
		if !((char) & (0 NOWAITIO))  ;busy-wait
			if PREVPAGE == -1  ;need to be on a known page
				PAGECHK $, MAYBE,
			endif
			nearcall putch#v(7 - LITPAGEOF(PREVPAGE))_wait  ;busy-wait; call whichever page PCLATH is currently pointing to
		endif
		mov8 (char) & ~(0 NOWAITIO NOBLOCKIO), TXREG  ;do not wait (don't even check); assumes space is available in output buffer
	endif
	EXPAND_POP,
	ENDM


;******************************************************************************
;string manipulation functions:
;these provide more complex string handling, built on top of GETCHAR/PUTCHAR
;all functions are wrapped as macros so they take no code space unless used
;******************************************************************************

;convert to upper case:
;params:
; char = char to convert to upper case
TOUPPER MACRO char
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH TRUE  ;show what this does in LST file
	LOCAL done
	if8 char, LT, _LITERAL(A'a'), LEAVE_PLACEHOLDER1 neargoto done,,
	if8 char, GT, _LITERAL(A'z'), LEAVE_PLACEHOLDER1 neargoto done,,
	add8 char, _LITERAL(LOW(A'A'-A'a')), char,
 jumpdest done:,,,, FALLTHRU
	FILL_PLACEHOLDER, neargoto done  ;KLUDGE: assembler can't figure out this forward reference
	FILL_PLACEHOLDER, neargoto done  ;KLUDGE: assembler can't figure out this forward reference
	EXPAND_POP,
	ENDM


	VARIABLE TOCOLUMN_EXPANDED = FALSE  ;avoid redundant code and data expansions

;tab over to a text column (using spaces):
;params:
; column = column# to tab to (1-based)
TOCOLUMN MACRO column
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH !TOCOLUMN_EXPANDED
	if !TOCOLUMN_EXPANDED  ;avoid redundant code and data expansions; expand first time only
TOCOLUMN_EXPANDED = TRUE
		LOCAL around_tocolumn, svbank = PREVBANK
		LEAVE_PLACEHOLDER1JUMP neargoto around_tocolumn

 jumpdest tocol_loop:,,, OFBANK(0)|,  ;return with bank 0 since that is where data typically is
		dec8_ifnz INCOL,, return_ tocol_loop,
		PUTCHAR _LITERAL(A' ')
		goto_ tocol_loop

 jumpdest around_tocolumn:, svbank|,,,
		FILL_PLACEHOLDER, neargoto around_tocolumn
	endif
	mov8nop column, INCOL
	call_ tocol_loop
	EXPAND_POP,
	ENDM


	VARIABLE GETNUM_EXPANDED = FALSE  ;avoid redundant code and data expansions

;get an 8-bit number:
;if second param is true, first char is already in inchar_reg
;returns STATUS.Z if eof; also sets GotNum bit var if a number was obtained
;NOTE: reuses DIV temps
;params:
; reg = destination register for value
; want_new_char = flag telling whether to get a new char first
; getchar_func = function to get next char and input char parameter
; inchar_reg = register containing current input char (reg passed in first usage is used for all other calls)
GETNUM8_ MACRO reg, want_new_char, getchar_func, inchar_reg
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH !GETNUM_EXPANDED
	if !GETNUM_EXPANDED  ;avoid redundant code and data expansions; expand first time only
GETNUM_EXPANDED = TRUE
		LOCAL around_getnum, svbank = PREVBANK
		LEAVE_PLACEHOLDER1JUMP neargoto around_getnum

		CBLOCK DIV32_NBTEMP  ;reuse memory, since this function is mutually exclusive
		INNUM  ;incoming value
		ENDC
		BITDCL GotNum  ;got at least a partial number
		BITDCL GetHex  ;allow hex digits

 jumpdest getnum8_getchar:,,, OFBANK(0)|,  ;LONGJUMP_HINT|  ;return with bank 0, since that is where next data typically is
		getchar_func
 jumpdest getnum8:,,, OutBankOf(getnum8_getchar)|, FALLTHRU  ;common return must use same bank affinity
		mov8 _LITERAL(0), INNUM
		setbit BITVAR(GotNum), FALSE  ;haven't received anything valid yet
		setbit BITVAR(GetHex), FALSE  ;assume decimal until found otherwise
 jumpdest next_digit:,,,, FALLTHRU
		ifbit BITVAR(GetHex), TRUE, loadw_ _LITERAL(0x10), loadw_ _LITERAL(26),  ;can't accept another digit if value is already larger than this
		if8 INNUM, GE, WREG, return_ getnum8,,
		if8 inchar_reg, LE, _LITERAL(A'9'), neargoto check_dec_digit,,
		ifbit BITVAR(GetHex), FALSE, return_ getnum8,,
		if8 inchar_reg, GT, _LITERAL(A'F'), return_ getnum8,,
		if8 inchar_reg, LT, _LITERAL(A'A'), return_ getnum8,,  ;W == inchar_reg - 'A'
		add8 WREG, _LITERAL(A'0'+10), inchar_reg,  ;remove gap between dec and hex digits
 jumpdest check_dec_digit:,,,, FALLTHRU  ;LONGJUMP_HINT|:
		if8 inchar_reg, LT, _LITERAL(A'0'), return_ getnum8,, EQUAL_HINT
		ifbit NEZ, neargoto no_0x_check,,
		ifbit BITVAR(GotNum), TRUE, neargoto no_0x_check,,
		getchar_func
		if8 inchar_reg, NE, _LITERAL(A'X'), neargoto not_hexnum,,  ;0x##
		setbit BITVAR(GetHex), TRUE
		getchar_func
		PAGECHK $, MAYBE,  ;restore PCLATH
 jumpdest not_hexnum:, inchar_reg|,,, FALLTHRU
		setbit BITVAR(GotNum), TRUE
		neargoto next_digit
 jumpdest no_0x_check:, inchar_reg|,,,  ;LONGJUMP_HINT|
		setbit BITVAR(GotNum), TRUE
		ifbit BITVAR(GetHex), TRUE, mult8_16x INNUM, mult8_10x INNUM,  ;shift left 1 dec or hex digit
		sub8 inchar_reg, _LITERAL(A'0'), WREG,
		add8 INNUM, WREG, INNUM,
		ifbit CARRY, return_ getnum8,,  ;overflow; leave inchar_reg intact to force syntax error
		getchar_func
		neargoto next_digit

 jumpdest around_getnum:, svbank|,,,
		FILL_PLACEHOLDER, neargoto around_getnum
	endif
	if want_new_char  ;need to get next char first
		call_ getnum8_getchar
	else  ;already have first char
		call_ getnum8
	endif
	mov8nop INNUM, reg
;	loadw_ inchar_reg  ;this will set STATUS.Z at eof (getchar_func should set inchar_reg to 0)
	EXPAND_POP,
	ENDM


;comma-less wrappers:
PUTNUM8_DEC MACRO value
	PUTNUM8 value, 10
	ENDM

PUTNUM8_HEX MACRO value
	PUTNUM8 value, 16
	ENDM

;output a number (multi-byte values are little-endian):
PUTNUM8 MACRO val, base
	mov8nop val, OUTNUM32
	PUTNUM_ 8, 0, base
	ENDM

PUTNUM8_NN MACRO val, base
	mov8nop val, OUTNUM32
	PUTNUM_ 8, 2, base
	ENDM

PUTNUM16 MACRO val, base
	mov16nop val, OUTNUM32
	PUTNUM_ 16, 0, base
	ENDM

PUTNUM24 MACRO val, base
	mov24nop val, OUTNUM32
	PUTNUM_ 24, 0, base
	ENDM

PUTNUM32 MACRO val, base
	mov32nop val, OUTNUM32
	PUTNUM_ 32, 0, base
	ENDM


	VARIABLE PUTNUM_EXPANDED = FALSE  ;avoid redundant code and data expansions

;output a 8/16/24/32 bit value in decimal or hex:
;this is a very stripped-down version of printf(%d) or (%x)
;internally converts to 32-bits to avoid needing a separate copy of the code for 8-, 16-, 24- and 32-bit numbers
;NOTE: reuses DIV temps
;params:
; bits = #bits in value to output
; paddig = #digits to pad; if result is less than this, it will be left-padded with 0's
; base = radix; 10 => decimal, 16 => hex
PUTNUM_ MACRO bits, paddig, base
	EXPAND_PUSH !PUTNUM_EXPANDED
	if !PUTNUM_EXPANDED  ;avoid redundant code and data expansions; expand first time only
PUTNUM_EXPANDED = TRUE
		LOCAL around_putnum, svbank = PREVBANK
		LEAVE_PLACEHOLDER1JUMP neargoto around_putnum

;output a 32-bit decimal value:
;uses a count loop for each digit; kludgey, but still probably faster than multiple divides
;reuse memory, since this function is mutually exclusive with DIV32
		REDCL ThisDigit
		REDCL SkipDigits  ;IsSignificant
		REDCL OUTNUM32:4
		REDCL OUTCHAR

		CBLOCK DIV32_NBTEMP  ;reuse memory, since this function is mutually exclusive; use non-banked easier banking with EEADR
		DigitValue32:4
		ENDC

;encode bottom 28 bits of a 32-bit value into prog space (little-endian):
;encoding is as follows:
;32-bit value: (MSB) VUTS RQPO NMLK JIHG   FEDC BA98 7654 3210 (LSB)
;A78 encoding: DCBA98 76543210, FERQPO NMLKJIHG  (lower 28 bits only)
#define b32a78(val)  ((val)%0x4000), (((val)/0x10000)%0x1000) | ((val)&0xc000)/4
;		UGLY_PASS12FIX -3  ;-1  ;KLUDGE: assembler needs help resolving addresses
		UGLY_PASS12FIX 1  ;KLUDGE: assembler needs help
powers_of_10:
		da b32a78(1000000000)  ;NOTE: top 2 bits of this one are missing
		da b32a78(100000000)
		da b32a78(10000000)
		da b32a78(1000000)
		da b32a78(100000)
		da b32a78(10000)
		da b32a78(1000)
		da b32a78(100)
		da b32a78(10)
		da b32a78(1)

;here is another way to do div10: from http://www.scribd.com/doc/2364207/Jones-on-Reciprocal-Multiplication
;Q = ((A >> 1) + A) >> 1  ; Q = A*.11
;Q = ((Q >> 4) + Q)  ; Q = A*.110011
;Q = ((Q >> 8) + Q)  ; Q = A*.11001100110011
;Q = ((Q >> 16) + Q) >> 3  ; Q = A*.000110011001100110011001100110011
;TODO: check if that would be more efficient than the code below

 jumpdest putdec8:,,, OFBANK(0)|,  ;return with bank 0, since that is where next data typically is
		mov8 _LITERAL(0), OUTNUM32+1
 jumpdest putdec16:,,, OutBankOf(putdec8)|, FALLTHRU  ;common return must use same bank affinity
		mov8 _LITERAL(0), OUTNUM32+2
 jumpdest putdec24:,,, OutBankOf(putdec8)|, FALLTHRU  ;common return must use same bank affinity
		mov8 _LITERAL(0), OUTNUM32+3
 jumpdest putdec32:,,, OutBankOf(putdec8)|, FALLTHRU  ;common return must use same bank affinity
		storew_ SkipDigits
		mov16 _LITERAL(powers_of_10 - 1), EEADR
		mov8 _LITERAL(NUMDIG_DEC32 - 1), ThisDigit
 jumpdest nextdig:,,,, FALLTHRU
;extract next power of 10 from prog memory:
		PROGREAD16 EEADR, _LITERAL(1), EEDAT
		mov8 EEDAT, DigitValue32
		mov8 EEDATH, DigitValue32+1
		PROGREAD16 EEADR, _LITERAL(1), EEDAT
		ifbit EEDATH, 5, TRUE, SETBIT(DigitValue32+1, 7, TRUE),,
		ifbit EEDATH, 4, TRUE, SETBIT(DigitValue32+1, 6, TRUE),,
		mov8 EEDAT, DigitValue32+2
		and8 EEDATH, _LITERAL(0x0f), DigitValue32+3,
		if8 ThisDigit, EQ, _LITERAL(NUMDIG_DEC32 - 1), loadw_ _LITERAL(0x30),,  ;KLUDGE: fill in top 2 missing bits; correct for overflow (doesn't fill in 2 A78 words)
		ifbit EQZ, ADDWF_wrapper(DigitValue32+3, TRUE),,
		mov8 _LITERAL(A'0'), OUTCHAR
 jumpdest digloop:,,,, FALLTHRU
		sub32 OUTNUM32, DigitValue32, OUTNUM32, COMPARE_HINT
		ifbit LTZ, neargoto digdone,,
		inc8 OUTCHAR
		goto_ digloop
 jumpdest digdone:, OUTNUM32|,,,
		add32 OUTNUM32, DigitValue32, OUTNUM32,  ;undo last sub (went too far)
		if8 OUTCHAR, NE, _LITERAL(A'0'), clear_ SkipDigits,,
		if8 SkipDigits, NE, _LITERAL(0), dec8 SkipDigits, PUTCHAR OUTCHAR,  ;supress leading 0s
		dec8_ifnz ThisDigit, neargoto nextdig,,  ;always show last digit
		add8 OUTNUM32, _LITERAL(A'0'), OUTCHAR,
		PUTCHAR OUTCHAR
		return_ putdec32
;output a 32-bit hex number:
 jumpdest puthex8:,,, OFBANK(0)|,  ;return with bank 0, since that is where next data typically is
		mov8 _LITERAL(0), OUTNUM32+1
 jumpdest puthex16:,,, OutBankOf(puthex8)|, FALLTHRU  ;common return must use same bank affinity
		mov8 _LITERAL(0), OUTNUM32+2
 jumpdest puthex24:,,, OutBankOf(puthex8)|, FALLTHRU  ;common return must use same bank affinity
		mov8 _LITERAL(0), OUTNUM32+3
 jumpdest puthex32:,,, OutBankOf(puthex8)|, FALLTHRU  ;common return must use same bank affinity
		storew_ SkipDigits
		mov8 _LITERAL(8), ThisDigit
		if32 OUTNUM32, LE, _LITERAL(8), neargoto nexthex,,  ;smallhexval,,
		PUTCHAR _LITERAL(A'0')
		PUTCHAR _LITERAL(A'x')
 jumpdest nexthex:,,,, FALLTHRU
		gettop4 OUTNUM32+3, WREG
		addlw LOW -10
		ifbit GEZ, addlw A'A'-(A'9'+1),,
		add8 WREG, _LITERAL(A'0'+10), OUTCHAR,
		if8 OUTCHAR, NE, _LITERAL(A'0'), clear_ SkipDigits,,
		if8 SkipDigits, NE, _LITERAL(0), dec8 SkipDigits, PUTCHAR OUTCHAR,  ;supress leading 0s
		dec8_ifnz ThisDigit,, return_ puthex32,
		REPEAT 4, shiftl32 OUTNUM32
		goto_ nexthex

 jumpdest around_putnum:, svbank|,,,
		FILL_PLACEHOLDER, neargoto around_putnum
	endif
	if (base) == 10
		loadw_ _LITERAL(NUMDIG_DEC32 - IIF(paddig, paddig, 1))
		call_ putdec#v(bits)
	else
		loadw_ _LITERAL(8 - IIF(paddig, paddig, 1))
		call_ puthex#v(bits)
		ERRIF (base) != 16, [ERROR] PUTNUM: unhandled base #v(bits):#v(base)
	endif
	EXPAND_POP,
	ENDM


;******************************************************************************
;literal pool and string I/O helpers:
;all functions are wrapped as macros so they take no code space unless used
;A1 strings are stored as one ASCII char per word in prog space
;A2 strings are stored as 2 ASCII chars per word in prog space (top char is shifted right to fit in 14 bits)
;A1 is wasteful, but okay for shorter strings; A2 is better for larger strings
;******************************************************************************

	VARIABLE LITSTR_LEN = 0, LITSTR_COUNT = 0  ;keep track of lit string space; used to track macro perf stats

;declare lit pool:
;choosing lit pool start address is trial & error
;if there's space left at the end, increase this value to shift the lit pool up in rom
;if you get warning 220 "Address exceeds maximum range for this processor", then decrease it to shift it down in rom
;params:
; start = start address of literal pool
LITPOOL MACRO start
	CONSTANT LITPOOL_START = start
	VARIABLE LITPOOL_NEXT = LITPOOL_START  ;next address to allocate
	VARIABLE LITPOOL_ADJUST = 0  ;used for conditional lit pool size adjustments
	ENDM


;comma-less wrappers:
PUTLITSTRA1 MACRO lit
	PUTLITSTR lit, 0xA1,
	ENDM

PUTLITSTRA2 MACRO lit
	PUTLITSTR lit, 0xA2,
	ENDM


;create and output a literal string:
;this macro can embed the literal string in-place and then hop over it,
;but it's better to place the string up in the literal pool area, where it doesn't need to be jumped over
;params:
; lit = literal string to be output
; a1a2 = whether to make the string A1 or A2
; litpool = flag telling whether to put literal in lit pool or not (defaults to TRUE); if not in lit pool, literal will be in-line and must be jumped over
PUTLITSTR MACRO litstr, a1a2, litpool
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
	LOCAL UNIQID = NEXT_UNIQID 
NEXT_UNIQID += 1
	LOCAL this_str#v(UNIQID)  ;KLUDGE: force LOCALs to be unique, in case assembler doesn't generate a unique hash
	if !(litpool -1)
LITPOOL_NEXT = $  ;force literal to be in-line
	endif
	LOCAL resume_here = $, INLINE = (LITPOOL_NEXT == resume_here)
	ORG LITPOOL_NEXT  ;put lit str elsewhere in memory, so we don't need to hop over it
;first create the literal string:
	if (a1a2) == 0xA1
		MAKE_LITSTR_A1 this_str#v(UNIQID), litstr, INLINE|  ;if literal sting is in-line, need to jump over it
	else  ;A2
		MAKE_LITSTR_A2 this_str#v(UNIQID), litstr, INLINE|  ;if literal sting is in-line, need to jump over it
	endif
	if !INLINE
LITPOOL_NEXT = $  ;continue here next time
		ORG resume_here
	endif
;then send it for output:
	if (a1a2) == 0xA1
		PUTSTR_A1 this_str#v(UNIQID)
	else  ;A2
		PUTSTR_A2 this_str#v(UNIQID)
	endif
	EXPAND_POP,
	ENDM


;make a named literal A1 string:
;string is embedded within code page, 1 char per word
;caller can tell macro to jump over the string if it is within a live code segment
;params:
; name = symbolic string name
; litstr = actual string value
; jumpover = flag telling whether to jump over the string (optional, default FALSE); this is needed if the literal is embedded into a live section of code
MAKE_LITSTR_A1 MACRO name, litstr, jumpover
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
	WARNIF REGPAGEOF($) != REGPAGEOF($+4), [INFO] Literal string pre-padded to avoid 256-word page boundary #v($)..#v($+4)
	while REGPAGEOF($) != REGPAGEOF($+4)
		nop  ;pad up to next 256-word page to avoid crossing page boundary (simplifies address arithmetic)
	endw
	LOCAL lit_start, lit_end, svbank = PREVBANK
	if jumpover 0
		LEAVE_PLACEHOLDER1JUMP neargoto lit_end  ;leave a placeholder for jumping over literal
	endif
 jumpdest lit_start:,,,, IIF(jumpover 0, 0, FALLTHRU 0)|
	LEAVE_PLACEHOLDER_ 4, page_address_correction  ;leave placeholder for page-spanning address correction
;name:
	add8 PCL, WREG, PCL,  ;CAUTION: the low 5 bits of PCLATH are also used here; they must be correct for the offset in W
LITSTR_COUNT += 1
LITSTR_LEN -= $
																				EXPAND  ;show generated code
	dt litstr, A'\0'  ;retlw's (1 char per word); make sure string is null terminated
	NOEXPAND
 jumpdest lit_end:, svbank|,,,  ;end of literal string
LITSTR_LEN += $
	START_FILL_PLACEHOLDER,  ;fill in page-spanning address correction
	if REGPAGEOF(lit_end) != REGPAGEOF(lit_start+4)  ;address correction needed (literal spans 256 byte boundary)
		ERRIF LITPAGEOF(lit_end) != LITPAGEOF(lit_start), [ERROR] Literal string spans 2K word boundary #v(lit_start)..#v(lit_end)
		WARNIF LITPAGEOF(lit_end) == LITPAGEOF(lit_start), [CAUTION] Literal string spans 256-word page boundaries #v(lit_start)..#v(lit_end); double check address arithmetic here
 jumpdest lit#v(0xA1)name:, PCL|, PCLATH|,,  ;name is pre-pended to force caller to use correctly specify A1/A2
		add8 WREG, _LITERAL(LOW(lit_start + 4)), WREG,  ;try out address arithmetic before committing PCL update (PCLATH must be set first)
		SETPAGE $, FALSE  ;KLUDGE: defeat PAGECHK within ifbit (not needed here)
;		ifbit STATUS, C, C_CARRY, inc8 PCLATH,  ;address wrap, bump high byte of PCL
		ifbit CARRY, inc8 PCLATH,,  ;address wrap, bump high byte of PCL
		sub8 WREG, _LITERAL(LOW(lit_start + 4)), WREG,  ;restore W for Computed Goto
	else
		REPEAT 4, nop
 jumpdest name#v(0xA1):, PCL|, PCLATH|,,  ;put this after the nops (no need to execute them); append A1/A2 to name to force correct usage by caller
	endif
	END_FILL_PLACEHOLDER, TRUE
	if jumpover 0  ;KLUDGE: assembler isn't smart enough to resolve forward address here, so we have to help it
		FILL_PLACEHOLDER, neargoto lit_end  ;go back and fill in goto
	endif
	EXPAND_POP,
	ENDM


;make a named literal A2 string:
;string is embedded within code page, 2 chars per word
;caller can tell macro to jump over the string if it is within a live code segment
;params:
; name = symbolic string name
; litstr = actual string value
; jumpover = flag telling whether to jump over the string (optional, default FALSE); this is needed if the literal is embedded into a live section of code
MAKE_LITSTR_A2 MACRO name, litstr, jumpover
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
	LOCAL UNIQID = NEXT_UNIQID 
NEXT_UNIQID += 1
	LOCAL lit_start#v(UNIQID), lit_end#v(UNIQID), svbank = PREVBANK  ;KLUDGE: force LOCALs to be unique; sometimes assembler doesn't generate a unique hash
	LOCAL svpage = PREVPAGE
	if jumpover 0
		LEAVE_PLACEHOLDER1JUMP neargoto lit_end#v(UNIQID)  ;leave a placeholder for jumping over literal
	endif
LITSTR_COUNT += 1
LITSTR_LEN -= $
 jumpdest name#v(0xA2):,,,,  ;append A1/A2 to name to force correct usage by caller
																				EXPAND  ;show generated code
	da litstr  ;2 chars per word; caller must explicitly null-terminate if null char below is not used
	NOEXPAND
#if FALSE
	da A'\0'  ;make sure string is null terminated; this uses more space than if caller ends strings with "\0"
#endif
LITSTR_LEN += $
	if jumpover 0  ;KLUDGE: assembler isn't smart enough to resolve forward address here, so we have to help it
 jumpdest lit_end#v(UNIQID):, svbank|,,,  ;end of literal string
		FILL_PLACEHOLDER, neargoto lit_end#v(UNIQID)  ;go back and fill in goto
	else
		SETPAGE svpage, FALSE  ;non-inline literal did not change prev adrs
	endif
	EXPAND_POP,
	ENDM


	VARIABLE PUTSTRA1_EXPANDED = FALSE  ;avoid redundant code and data expansions
	VARIABLE PUTSTRA2_EXPANDED = FALSE  ;avoid redundant code and data expansions

;output a null-terminated string (max 255 char, can span page boundaries):
;string must have been created with MAKE_LITSTR_A1, which contains the character unpacking code
;params:
; name = symbolic name of string to output
PUTSTR_A1 MACRO name
	NOEXPAND  ;hide clutter in LST file
	LOCAL newline = FALSE
	EXPAND_PUSH !PUTSTRA1_EXPANDED
	if !PUTSTRA1_EXPANDED  ;avoid redundant code and data expansions; expand first time only
PUTSTRA1_EXPANDED = TRUE
		CBLOCK DIV32_NBTEMP  ;reuse memory, since this function is mutually exclusive; use non-banked easier banking with EEADR
		OUTCHAROFS  ;offset of next output char
		OUTCHAR  ;next character to output
		ENDC
	endif

	LOCAL putstra1_loop, putstra1_done
	mov8 _LITERAL(0), OUTCHAROFS
 jumpdest putstra1_loop:,,,, FALLTHRU
	GotocSetup name#v(0xA1)
	loadw_ OUTCHAROFS
	nearcall SHADOW_ADRS(name#v(0xA1))  ;puts OUTCHAR'th char into W; use nearcall because PCLATH is already set
	if8 WREG, EQ, _LITERAL(0), LEAVE_PLACEHOLDER1 neargoto putstra1_done,,  ;null char; NOTE: retlw does NOT set STATUS.Z
	storew_ OUTCHAR
	PUTCHAR OUTCHAR
	inc8 OUTCHAROFS
	goto_ putstra1_loop
 jumpdest putstra1_done:, PCL|, OUTCHAROFS|,,
	FILL_PLACEHOLDER, neargoto putstra1_done  ;KLUDGE: assembler can't figure out this forward reference
	if newline
		PUTCHAR _LITERAL(A'\r')
		PUTCHAR _LITERAL(A'\n')
	endif
	EXPAND_POP,
	ENDM


;output a null-terminated string (max 255 char, can span page boundaries):
;string must have been created with MAKE_LITSTR_A2
;uses EEADR/EEDAT to read directly from prog memory
;params:
; name = symbolic name of string to output
PUTSTR_A2 MACRO name
	NOEXPAND  ;hide clutter in LST file
	LOCAL newline = FALSE
	EXPAND_PUSH !PUTSTRA2_EXPANDED
	if !PUTSTRA2_EXPANDED  ;avoid redundant code and data expansions; expand first time only
PUTSTRA2_EXPANDED = TRUE
		LOCAL around_putstr, svbank = PREVBANK
		LEAVE_PLACEHOLDER1JUMP neargoto around_putstr

		CBLOCK DIV32_NBTEMP  ;reuse memory, since this function is mutually exclusive; use non-banked easier banking with EEADR
		OUTCHAR16:2  ;next 2 bytes of output string
		ENDC

 jumpdest putstra2ofs_loop:,,, OFBANK(0)|,  ;return with bank 0, since that is where next data typically is
		add16_8 _LITERAL(LITPOOL_START - 1), WREG, EEADR,
 jumpdest putstra2_loop:,,, OutBankOf(putstra2ofs_loop)|, FALLTHRU  ;common return must use same return bank affinity
		PROGREAD16 EEADR, _LITERAL(1), OUTCHAR16  ;get next 2 chars (packed in 14 bits)
		shiftl16 OUTCHAR16
		if8 OUTCHAR16+1, EQ, _LITERAL(0), return_ putstra2_loop,, BIGSTMT_HINT  ;end of string
		PUTCHAR OUTCHAR16+1
		if8 OUTCHAR16+1, EQ, _LITERAL(A'\n'), PUTCHAR _LITERAL(A'\r'),, BIGSTMT_HINT  ;make sure CR was sent too
		shiftr8 OUTCHAR16
		if8 OUTCHAR16+0, EQ, _LITERAL(0), return_ putstra2_loop,, BIGSTMT_HINT  ;end of string
		PUTCHAR OUTCHAR16+0
		if8 OUTCHAR16+0, EQ, _LITERAL(A'\n'), PUTCHAR _LITERAL(A'\r'),, BIGSTMT_HINT  ;make sure CR was sent too
		neargoto putstra2_loop

;output a CR+LF:
;sends LF before CR so other PICs know the text is not a command for them
 jumpdest put_crlflf:,,, OFBANK(0)|,  ;return with bank 0, since that is where data typically is
		PUTCHAR _LITERAL(A'\n')
 jumpdest put_crlf:,,, OutBankOf(put_crlflf)|, FALLTHRU  ;common return must share same out bank
		PUTCHAR _LITERAL(A'\n')
		PUTCHAR _LITERAL(A'\r')
		return_ put_crlf

 jumpdest around_putstr:, svbank|,,,
		FILL_PLACEHOLDER, neargoto around_putstr
	endif
	ERRIF name#v(0xA2) < LITPOOL_START, [ERROR] Literal string #v(name#v(0xA2)) is outside lit pool #v(LITPOOL_START)
	if (name#v(0xA2) - LITPOOL_START) < 256  ;only need 1 byte for address; reduces in-line code space (1 instr vs. 5 instr)
		EXPAND_PUSH TRUE  ;show what this does
		loadw_ _LITERAL((name#v(0xA2)) - LITPOOL_START)  ;pass start offset thru W
		call_ putstra2ofs_loop
		EXPAND_POP,
	else  ;need 2 bytes for address
		EXPAND_PUSH TRUE  ;show what this does
		mov16 _LITERAL(name#v(0xA2) - 1), EEADR  ;set whole address; compensate for loop inc
		call_ putstra2_loop
		EXPAND_POP,
	endif
	if newline
		call_ put_crlf
	endif
	EXPAND_POP,
	ENDM


;******************************************************************************
;timing functions/macros:
;******************************************************************************

	CONSTANT MAX_NOOP = d'2500'  ;uses 5 stack levels; reasonably safe for BUSYWAIT if interrupts are off


	VARIABLE BUSYWAIT_EXPANDED = FALSE  ;avoid redundant code and data expansions

;wait for specified number of instructions or usec:
;1 instr == 1 usec when running at 4 MHz (default internal clock)
;NOTE: this ties up the processor, so it can't do anything else while in the delay loop, but timing is EXACT if there are no interrupts
;NOTE: long delays require caller to supply RTC_20msec and YIELD (cooperative multi-tasking)
;params:
; numinstr = #instr/usec to wait
BUSYWAIT MACRO numinstr
	NOEXPAND  ;hide clutter in LST file
	LOCAL USEC = numinstr, delayloop
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show BUSYWAIT params
USEC = numinstr  ;show BUSYWAIT params
		NOEXPAND
	endif
	EXPAND_PUSH !BUSYWAIT_EXPANDED
	WARNIF USEC < 1, [WARNING] BUSYWAIT: No delay? #v(USEC)
;delays are exact up until MAX_NOOP; after that, they are off by a few instr (due to loop control and timer loop latency)
	ERRIF (USEC / (20 * ONE_MSEC)) > d'255', [ERROR] BUSYWAIT: too long #v(USEC / (20 * ONE_MSEC)) > 255  ;max delay with 1 byte 20msec counter
	if USEC > MAX_NOOP  ;use wait loop on timer for longer times (up to 65.536 sec)
;use timer loop to wait longer periods of time (not precise):
		if !BUSYWAIT_EXPANDED  ;avoid redundant code and data expansions; expand first time only
BUSYWAIT_EXPANDED = TRUE
			BDCL WAITMSEC  ;16:2
		endif
		EXPAND_PUSH TRUE
		mov8 _LITERAL(rdiv(numinstr, 20 * ONE_MSEC)), WAITMSEC  ;#times thru 20 msec loop
 jumpdest waitloop20:,,,, FALLTHRU
		setbit SETBIT(BITVAR(EVT_20msec), FALSE)
 jumpdest waitloop:,,,, FALLTHRU
;		call_ noop100  ;do something else here if wanted
		YIELD
		ifbit BITVAR(EVT_20msec), FALSE, neargoto waitloop,,  ;a full 20 msec has now elapsed
		dec8_ifnz WAITMSEC, neargoto waitloop20,,
		EXPAND_POP,
	else
		noop_ USEC
	endif
	EXPAND_POP,
	ENDM


	VARIABLE NOOP_EXPANDED = FALSE  ;avoid redundant code and data expansions

;noop specified number of times (up to 15 msec at 4 MHz):
;based on a nop idea seen somewhere, probably PICLIST (please add reference)
;does not alter bank or page select, although page may be altered with jump around in-place expansion
;1 instr == 1 usec at 4 Mhz
;params:
; usec = #usec to wait; actually, #instr cyces (4 MHz assumed)
noop_ MACRO usec
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH !NOOP_EXPANDED
	LOCAL USEC = usec  ;, svbank = PREVBANK
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show NOOP params
USEC = usec  ;show NOOP params
		NOEXPAND
	endif
	if !NOOP_EXPANDED  ;avoid redundant code expansions; expand code first time only
NOOP_EXPANDED = TRUE
		LOCAL around_noop, svbank = PREVBANK
		LEAVE_PLACEHOLDER1JUMP neargoto around_noop

;recursive delay loop (exact unless interrupts occur):
;delay up to 12.5 msec with stack 6 deep (dangerous depth since interrupts may also be used)
;safer is 2.499 msec, which only uses 4 stack levels; reasonably safe
		SetBankAffinity noop2500,,, ~LONGJUMP_HINT,  ;fwd declare (shim for call macro)
		SetBankAffinity noop500,,, ~LONGJUMP_HINT,  ;fwd declare (shim for call macro)
		SetBankAffinity noop100,,, ~LONGJUMP_HINT,  ;fwd declare (shim for call macro)
		SetBankAffinity noop20,,, ~LONGJUMP_HINT,  ;fwd declare (shim for call macro)
		SetBankAffinity noop4,,, ~LONGJUMP_HINT,  ;fwd declare (shim for call macro)
 jumpdest noop12500:,,,,  ;2 call + 4*2500+4*500+4*100+4*20+4*4 nested noop + 2 return = 12.5 msec; 6 levels deep
		REPEAT 4, nearcall noop2500
 jumpdest noop2500:,,, OutBankOf(noop12500)|, FWDREF|  ;2 call + 4*500+4*100+4*20+4*4 nested noop + 2 return = 2.5 msec; 5 levels deep  ;common return must use same return bank affinity
		REPEAT 4, nearcall noop500
 jumpdest noop500:,,, OutBankOf(noop12500)|, FWDREF|  ;2 call + 4*100+4*20+4*4 nested noop + 2 return = 500 usec; 4 levels deep  ;common return must use same return bank affinity
		REPEAT 4, nearcall noop100
 jumpdest noop100:,,, OutBankOf(noop12500)|, FWDREF|  ;2 call + 4*20+4*4 nested noop + 2 return = 100 usec; 3 levels deep  ;common return must use same return bank affinity
		REPEAT 4, nearcall noop20
 jumpdest noop20:,,, OutBankOf(noop12500)|, FWDREF|  ;2 call + 4*4 nested noop + 2 return = 20 usec; 2 levels deep  ;common return must use same return bank affinity
		REPEAT 4, nearcall noop4
 jumpdest noop4:,,, OutBankOf(noop12500)|, FWDREF|  ;2 call + 2 return = 4 usec; 1 level deep  ;common return must use same return bank affinity
		return_ noop4

 jumpdest around_noop:, svbank|,,,
		FILL_PLACEHOLDER, neargoto around_noop
	endif
	ERRIF USEC > d'15000', [ERROR] noop: delay #v(USEC) too long > 15000  ;max delay with 6 levels of stack
	LOCAL SAVE_BANK = PREVBANK
	REPEAT USEC / d'2500', call_ noop2500  ;max 4 times
	REPEAT (USEC % d'2500') / d'500', call_ noop500  ;max 4 times
	REPEAT (USEC % d'500') / d'100', call_ noop100  ;max 4 times
	REPEAT (USEC % d'100') / d'20', call_ noop20  ;max 4 times
	REPEAT (USEC % d'20') / 4, call_ noop4  ;max 4 times
	REPEAT USEC % 4, nop  ;kill some time (1 usec when running at 4 MHz)
	SETBANK SAVE_BANK, TRUE  ;nop tree does not use bank select, so just set it to what it was at the start
	EXPAND_POP,
	ENDM


;******************************************************************************
;pin I/O and bit-banging macros/functions:
;******************************************************************************

;port control values are encoded as follows:
;- use hex so we can express the port# (A, B, C, ...), use "2" for "Z" and "5" for "save" (don't change) place-holder
;- use one hex nibble for each port# or bit
;- first digit = port# (A, C, etc)
;- following 1-6 digits = bit values (0, 1, Z), starting with MSB if control codes > 0 or LSB if control codes < 0
;- optionally, another port# and more digits can follow, up to a max of 8 chars total (32-bit assembler limitation)
;- all unspecified bits are treated as "5" (save, don't change)

;individual pin definitions (input):
RA0_IN set 0x92A
RA1_IN set 0x925A
RA2_IN set 0x9255A
RA3_IN set 0x92555A
RA4_IN set 0x925555A
RA5_IN set 0x9255555A
RC0_IN set 0x92C
RC1_IN set 0x925C
RC2_IN set 0x9255C
RC3_IN set 0x92555C
RC4_IN set 0x925555C
RC5_IN set 0x9255555C

;invidual pin definitions (output):
RA0_LOW set 0x90A
RA1_LOW set 0x905A
RA2_LOW set 0x9055A
RA3_LOW set 0x90555A
RA4_LOW set 0x905555A
RA5_LOW set 0x9055555A
RA012345_LOW set 0xA000000
RC0_LOW set 0x90C
RC1_LOW set 0x905C
RC2_LOW set 0x9055C
RC3_LOW set 0x90555C
RC4_LOW set 0x905555C
RC5_LOW set 0x9055555C
RC012345_LOW set 0xC000000

RA0_HIGH set 0x91A
RA1_HIGH set 0x915A
RA2_HIGH set 0x9155A
RA3_HIGH set 0x91555A
RA4_HIGH set 0x915555A
RA5_HIGH set 0x9155555A
RA012345_HIGH set 0xA111111
RC0_HIGH set 0x91C
RC1_HIGH set 0x915C
RC2_HIGH set 0x9155C
RC3_HIGH set 0x91555C
RC4_HIGH set 0x915555C
RC5_HIGH set 0x9155555C
RC012345_HIGH set 0xC111111


;dummy parameter to pass to SETPINS for no wait:
#define NOWAIT  BUSYWAIT 0


;set some pins high/low/off for a certain length of time
;pin direction is set only if all pins are high or low, or there is a mix of on and off states; this helps to avoid glitches related to switching pins on/off unnecessarily
;pin direction is set only if all pins are high or all pins are low, or there is a mix of on and off states
;this helps to avoid signal glitches related to switching pins on/off unnecessarily when multiple instructions are used
;use duration 0 to return immediately
;ALTERS RP0/1, DOES NOT ALTER PCLATH
;params:
; which = bitmask of pins to set
; wantdir = flag telling whether to set direction (TRIS)
; delayfunc_duration = delay time
SETPINS MACRO which, wantdir, delayfunc_duration
	NOEXPAND  ;hide clutter in LST file
	LOCAL PINS = which
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show SETPINS params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
PINS = which  ;show SETPINS params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	LOCAL NIBBLE = 0, BIT, PORTWHICH, WATCHDOG = 0, DEBUG
PORTWHICH = 0xA
	while PORTWHICH <= 0xF  ;A, B, ..., F
		LOCAL PORTMASK_#v(PORTWHICH) = 0, PORTBITS_#v(PORTWHICH) = 0, PORTDIR_#v(PORTWHICH) = 0
PORTWHICH += 1
	endw
PORTWHICH = 0
	LOCAL SETIN, SETOUT, SETLOW, SETHIGH
	while NIBBLE != 9  ;decode next nibble
		if ((which) & 0xf) >= 9  ;control codes in LSB -> MSB order
NIBBLE = PINS & 0xf
		else  ;control codes in MSB -> LSB order
NIBBLE = (PINS/0x10000000) & 0xf
			if PINS < 0
NIBBLE -= 1  ;make correction for assembler sign-extend
			endif
		endif
		if NIBBLE >= 9  ;select a new port
			if PORTWHICH
				ERRIF !PORTMASK_#v(PORTWHICH), [ERROR] SETPINS: No port bits selected for port #v(PORTWHICH)
			endif
PORTWHICH = NIBBLE
			if ((which) & 0xf) >= 9  ;LSB first
BIT = 1
			else  ;MSB first
BIT = PORT_MSB
			endif
		else  ;set port bits
			if PORTWHICH
				if (BIT < 1) || (BIT > PORT_MSB)
					ERRIF TRUE, [ERROR] SETPINS: too many port bits #v(BIT) for port #v(PORTWHICH)
					EXPAND_POP,
					EXITM
				endif
				if NIBBLE == 0  ;set pin low
PORTMASK_#v(PORTWHICH) |= BIT  ;remember which bits were set
				else
					if NIBBLE == 1  ;set pin high
PORTMASK_#v(PORTWHICH) |= BIT  ;remember which bits were set
PORTBITS_#v(PORTWHICH) |= BIT  ;set bit high
					else
						if NIBBLE == 2  ;set pin Z (input)
PORTMASK_#v(PORTWHICH) |= BIT  ;remember which bits were set
PORTDIR_#v(PORTWHICH) |= BIT  ;turn pin off (input)
						else  ;leave pin as-is (5)
							ERRIF NIBBLE != 5, [ERROR] SETPINS: Invalid port control code #v(NIBBLE)
						endif
					endif
				endif
				if ((which) & 0xf) >= 9  ;LSB first
BIT *= 2
				else  ;MSB first
BIT /= 2
				endif
			else
				ERRIF NIBBLE != 0, [ERROR] SETPINS: must select port first #v(NIBBLE)
			endif
		endif
		if ((which) & 0xf) >= 9  ;control codes in LSB -> MSB order
PINS /= 0x10  ;right-shift to next control nibble
			if (PINS < 0) && NIBBLE
PINS -= 1  ;make correction for assembler sign-extend, but only if bottom nibble was 0????
			endif
PINS &= ~0xf0000000
PINS |= 0x90000000  ;left-pad with terminator code
		else  ;control codes in MSB -> LSB order
PINS *= 0x10  ;left-shift to next control nibble
PINS |= 9  ;right-pad with terminator code
		endif
WATCHDOG += 1  ;paranoid: avoid infinite loop if there is a bug in this macro
		if WATCHDOG > 9
			ERRIF TRUE, [ERROR] SETPINS: infinite loop #v(WATCHDOG)  ;paranoid self-check
			EXPAND_POP,
			EXITM
		endif
	endw
PORTWHICH = 0xA
	while PORTWHICH <= 0xF  ;A, B, C, D, E, F,  ;set port directions first
		if PORTMASK_#v(PORTWHICH)
;			EXPAND
;DEBUG = PORTDIR_#v(PORTWHICH)  ;0=output, 1=input
;DEBUG = PORTBITS_#v(PORTWHICH)  ;0=low, 1=high
;DEBUG = PORTMASK_#v(PORTWHICH)  ;which pins to set
			EXPAND_PUSH TRUE  ;SETPINS debug info
PINS = which
SETOUT = ~PORTDIR_#v(PORTWHICH) & PORTMASK_#v(PORTWHICH)  ;pins to set to output (0/1)
SETIN = PORTDIR_#v(PORTWHICH) & PORTMASK_#v(PORTWHICH)  ;pins to set to input (Z)
SETLOW = ~PORTDIR_#v(PORTWHICH) & ~PORTBITS_#v(PORTWHICH) & PORTMASK_#v(PORTWHICH)  ;pins to set low (0)
SETHIGH = ~PORTDIR_#v(PORTWHICH) & PORTBITS_#v(PORTWHICH) & PORTMASK_#v(PORTWHICH)  ;pins to set high (1)
			EXPAND_POP,
;NOTE: sometimes PIC seems to require that pins be set to output before setting them high/low?
;			if !PORTBITS || (PORTBITS == PORTMASK) || (PORTDIR && (PORTDIR != PORTMASK))  ;need to set pin direction; do this after pins are set (helps to avoid false signals for positive logic)
			if (wantdir) && ((SETIN == PORTMASK_#v(PORTWHICH)) || (SETLOW == PORTMASK_#v(PORTWHICH)) || (SETHIGH == PORTMASK_#v(PORTWHICH)) || (SETIN && SETOUT) || ((wantdir) == TRUE))  ;all pins set same way (probably initialization) or a mix of in + out; set pin direction; do this AFTER pins are set (helps to avoid false signals for positive logic)
				if PORTMASK_#v(PORTWHICH) != (2*PORT_MSB-1)  ;need to preserve some port bits
					bitblt8 TRIS#v(PORTWHICH), SETOUT, SETIN, TRIS#v(PORTWHICH)
				else
					mov8 _LITERAL(PORTDIR_#v(PORTWHICH)), TRIS#v(PORTWHICH)
				endif
			endif
		endif
PORTWHICH += 1
	endw
PORTWHICH = 0xA
	while PORTWHICH <= 0xF  ;A, B, C, D, E, F  ;now set port bits high/low
		if PORTMASK_#v(PORTWHICH)
;			EXPAND_PUSH TRUE
SETLOW = ~PORTDIR_#v(PORTWHICH) & ~PORTBITS_#v(PORTWHICH) & PORTMASK_#v(PORTWHICH)  ;pins to set low (0)
SETHIGH = ~PORTDIR_#v(PORTWHICH) & PORTBITS_#v(PORTWHICH) & PORTMASK_#v(PORTWHICH)  ;pins to set high (1)
;			EXPAND_POP,
;NOTE: sometimes PIC seems to require that pins be set to output before setting them high/low?
			if SETLOW || SETHIGH
				if PORTMASK_#v(PORTWHICH) != (2*PORT_MSB-1)  ;need to preserve some port bits
					bitblt8 PORT#v(PORTWHICH), SETLOW, SETHIGH, PORT#v(PORTWHICH)
				else  ;set all port bits
					mov8 _LITERAL(SETHIGH), PORT#v(PORTWHICH)
				endif
			endif
		endif
PORTWHICH += 1
	endw
	EXPAND_PUSH TRUE  ;show what this does
	delayfunc_duration
	EXPAND_POP,
	EXPAND_POP,
	ENDM


;******************************************************************************
;Circular buffer/queue macros/functions:
;(useful for serial I/O or inter-process communication)
;NOTE: there must always be at least 1 byte free in the queue (for full/empty detection)
;******************************************************************************

;define/allocate space for queue:
;one extra byte is allocated (in addtion to head + tail pointers), for empty vs. full detection
;this extra byte is used to hold overflow data; think of it as holding a null character at the end of the buffer
;addressing logic assumes that head and tail are within same bank
;params:
; name = symbolic name for queue
; maxlen = #bytes to allocate for queue contents
MSGQUE MACRO name, maxlen
	NOEXPAND  ;hide clutter in LST file
	LOCAL MAXLEN = maxlen
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show MSGQUE params
PREVBANK = PREVBANK  ;show last known bank select (shown for debug)
MAXLEN = maxlen  ;show MSGQUE params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	BDCL name:0  ;define symbol so we can refer to it by name in #v expressions
	BDCL QHEAD_#v(name)  ;head and tail pointers
	BDCL QTAIL_#v(name)
	BDCL QBSTART_#v(name):MAXLEN+1  ;allocate buffer space (includes 1 extra byte), define symbolic start + end addresses
	BDCL QBEND_#v(name):0
	ERRIF2 BANKOF(QBSTART_#v(name)) != BANKOF(QBEND_#v(name)), [ERROR] Queue starts and ends in different banks: starts #v(QBSTART_#v(name)), ends #v(QBEND_#v(name))
#ifndef USE_IRP
	ERRIF BANKOF(QBEND_#v(name)) >= 2, [ERROR] Set BANK23_INDIRECT to handle bank 2/3 addresses for queue at #v(QBSTART_#v(name))..#v(QBEND_#v(name))
#endif
	EXPAND_POP,
	ENDM


;check the space allocated for a queue:
#define QueSize(queue)  (QBEND_#v(queue) - QBSTART_#v(queue) - 1)


;comma-less extra parameters for QRESET:
;KLUDGE because MPASM doesn't allow nested macro calls with params
#define QHEAD  0x1000|
#define QTAIL  0x2000|
#define QBSTART  0x4000|
#define QBEND  0x8000|


;set head/tail pointer to start/end of buffer:
;this is used as a kludgey wrapper to avoid the need for commas so it can be used as a parameter to other macros
;params:
; queue = address of queue to reset
QRESET MACRO queue
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH TRUE  ;show what this does
	LOCAL QUEUE = (queue) & ~(QHEAD QTAIL QBSTART QBEND 0)  ;strip off comma-less parameters
	LOCAL START_OR_END = (queue) & (QBSTART QBEND 0)  ;strip off start/end parameter
	LOCAL HEAD_OR_TAIL = (queue) & (QHEAD QTAIL 0)  ;strip off head/tail parameter
	ERRIF !START_OR_END, [ERROR] Missing queue start/end parameter
	ERRIF !HEAD_OR_TAIL, [ERROR] Missing queue head/tail parameter
START_OR_END = IIF(START_OR_END == (QBSTART 0), QBSTART_#v(QUEUE), QBEND_#v(QUEUE))  ;line too long
HEAD_OR_TAIL = IIF(HEAD_OR_TAIL == (QHEAD 0), QHEAD_#v(QUEUE), QTAIL_#v(QUEUE))  ;line too long
	mov8 _LITERAL(START_OR_END), HEAD_OR_TAIL
	EXPAND_POP,
	ENDM
	

;initialize queue to empty:
;params:
; queue = address of queue to initialize
InitQue MACRO name
	NOEXPAND  ;hide clutter in LST file
	LOCAL QUEUE = name
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show InitQue params
QUEUE = name  ;show InitQue params
		NOEXPAND
	endif
	EXPAND_PUSH FALSE
	QRESET QBSTART QHEAD (name)
	QRESET QBSTART QTAIL (name)
	EXPAND_POP,
	ENDM
	

;check length of current queue contents:
;returns queue len in a register (use WREG for W)
;caller is responsible for managing interrupts; pass thread_safe == FALSE to prevent interrupts/yields
;params:
; queue = address of queue to get length for
; reg = destination register for queue length
; thread_safe = flag telling whether caller is thread-safe
GetQueLen MACRO queue, reg, thread_safe
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
	if !(thread_safe)
		intenable_ FALSE  ;turn off interrupts to prevent another thread from altering data while we're looking at it
	endif
	sub8 QHEAD_#v(queue), QTAIL_#v(queue), WREG, COMPARE_HINT
	ifbit LTZ, addlw QBEND_#v(INBUF) - QBSTART_#v(INBUF),, NOJUMP_HINT  ;make length correction for address wrap
 expand
PREVBANK = PREVBANK  ;getquelen
 noexpand
	storew_ reg
	if !(thread_safe)
		intenable_ FALSE  ;turn off interrupts to prevent another thread from altering data while we're looking at it
	endif
	EXPAND_POP,
	ENDM


;reduce macro nesting a little:
#define PutMsg(queue, msg, thread_safe)  SendMsg queue, msg, thread_safe, TRUE
#define UngetMsg(queue, msg, thread_safe)  SendMsg queue, msg, thread_safe, FALSE

;convenience wrappers:

;PutMsg MACRO queue, msg, thread_safe
;	NOEXPAND
;	SendMsg queue, msg, thread_safe, TRUE
;	ENDM

;UngetMsg MACRO queue, msg, thread_safe
;	NOEXPAND
;	SendMsg queue, msg, thread_safe, FALSE
;	ENDM


;enqueue a message into a queue:
;caller can put message at either end of queue: at head == send/put, at tail == unget
;multiple threads should not update the same end of the queue at the same time
;upon return, Z indicates whether the message could be enqueued
;NOTE: this macro alters STATUS.IRP set, so caller must save/restore it
;params:
; queue = address of queue to send to
; msg = byte value to send to queue
; thread_safe = flag telling whether caller is thread-safe
; tohead = flag telling whether to enqueue to head or tail of queue
SendMsg MACRO queue, msg, thread_safe, tohead
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH TRUE  ;FALSE  ;show what this does
	if !(thread_safe)
		intenable_ FALSE  ;turn off interrupts to prevent another thread from altering data while we're looking at it
	else
		if (thread_safe) == TRUE
			mov8 FSR, ISR_FSR  ;within ISR; save/restore caller's pointer
		endif
	endif
	LOCAL qdone
	GetQueLen queue, WREG, TRUE
	if8 WREG, EQ, _LITERAL(QBEND_#v(INBUF) - QBSTART_#v(INBUF) - 1), LEAVE_PLACEHOLDER1 neargoto qdone,,  ;queue is full (Z is set); try again later
	BANKSEL_INDIRECT(QBSTART_#v(queue))  ;need to set IRP to address bank 2 + 3 indirectly
	if tohead
		if8 QHEAD_#v(queue), EQ, _LITERAL(QBEND_#v(queue)), QRESET QBSTART QHEAD queue,, BIGSTMT_HINT
		mov8 QHEAD_#v(queue), FSR
		mov8 msg, INDF  ;put message at queue head
		inc8 QHEAD_#v(queue)
	else
		if8 QTAIL_#v(queue), EQ, _LITERAL(QBSTART_#v(queue)), QRESET QBEND QTAIL queue,, BIGSTMT_HINT
		dec8 QTAIL_#v(queue)
		mov8 QTAIL_#v(queue), FSR
		mov8 msg, INDF  ;put message at end queue tail
	endif
	setbit STATUS, Z, FALSE  ;message successfully enqueued; reset Z
 jumpdest qdone:,,,, FALLTHRU
	FILL_PLACEHOLDER, neargoto qdone
	if !(thread_safe)
		intenable_ TRUE  ;turn interrupts back on again
	else
		if (thread_safe) == TRUE
			copy_ ISR_FSR, FSR, FALSE  ;within ISR; save/restore caller's pointer (doesn't affect STATUS)
		endif
	endif
	EXPAND_POP,
	ENDM  ;!Z => sent, Z => not sent (queue full or busy)


;reduce macro nesting a little:
#define GetMsg(queue, reg, thread_safe)  RcvMsg queue, reg, thread_safe, TRUE
#define UnputMsg(queue, reg, thread_safe)  RcvMsg queue, reg, thread_safe, FALSE

;convenience wrappers:

;GetMsg MACRO queue, reg, thread_safe
;	NOEXPAND
;	RcvMsg queue, reg, thread_safe, TRUE
;	ENDM

;UnputMsg MACRO queue, reg, thread_safe
;	NOEXPAND
;	RcvMsg queue, reg, thread_safe, FALSE
;	ENDM


;dequeue a message from a queue:
;caller can get message from either end of queue: from head == rcv/get, from tail == unput
;multiple threads should not update the same end of the queue at the same time
;upon return, Z indicates whether there was a message to denqueue
;NOTE: this macro alters STATUS.IRP set, so caller must save/restore it
;params:
; queue = address of queue to send to
; reg = destination register for dequeued value
; thread_safe = flag telling whether caller is thread-safe
; fromtail = flag telling whether to dequeue from tail or head of queue
RcvMsg MACRO queue, reg, thread_safe, fromtail
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH TRUE  ;FALSE  ;show what this does
	if !(thread_safe)
		intenable_ FALSE  ;turn off interrupts to prevent another thread from altering data while we're looking at it
	else
		if (thread_safe) == TRUE
			mov8 FSR, ISR_FSR  ;within ISR; save/restore caller's pointer
		endif
	endif
	LOCAL qdone
	if8 QHEAD_#v(queue), EQ, QTAIL_#v(queue), LEAVE_PLACEHOLDER1 neargoto qdone,,  ;queue is empty (Z is set); try again later
	BANKSEL_INDIRECT(QBSTART_#v(queue))  ;need to set IRP to address bank 2 + 3 indirectly
	if fromtail
		if8 QTAIL_#v(queue), EQ, _LITERAL(QBEND_#v(queue)), QRESET QBSTART QTAIL queue,, BIGSTMT_HINT
		mov8 QTAIL_#v(queue), FSR
		mov8 INDF, reg  ;get message at queue tail
		inc8 QTAIL_#v(queue)
	else
		if8 QHEAD_#v(queue), EQ, _LITERAL(QBSTART_#v(queue)), QRESET QBEND QHEAD queue,, BIGSTMT_HINT
		dec8 QHEAD_#v(queue)
		mov8 QHEAD_#v(queue), FSR
		mov8 INDF, reg  ;get message at end queue tail
	endif
	setbit STATUS, Z, FALSE  ;message successfully dequeued; reset Z
 jumpdest qdone:,,,, FALLTHRU
	FILL_PLACEHOLDER, neargoto qdone
	if !(thread_safe)
		intenable_ TRUE  ;turn interrupts back on again
	else
		if (thread_safe) == TRUE
			copy_ ISR_FSR, FSR, FALSE  ;within ISR; save/restore caller's pointer (doesn't affect STATUS)
		endif
	endif
	EXPAND_POP,
	ENDM  ;!Z => received/non-empty, Z => not received (queue empty or busy)


;******************************************************************************
;PickKit 1 board definitions + helper functions/macros:
;******************************************************************************

#ifndef PK1DEV
 #define PK1DEV  FALSE  ;set default if caller did not specify
#endif


;PK1 LED charlieplexing timing:
;#define LEDTIME  (5*ONE_MSEC)  ;5 msec * 8 => 40 msec refresh rate (8 LEDs muxed); slight flicker
#define LEDTIME  (3*ONE_MSEC)  ;3 msec * 8 => 24 msec refresh rate (8 LEDs muxed); no flicker

#if LEDTIME > MAX_NOOP  ;too big for busy-waits; reduce it
 #undefine LEDTIME
 #define LEDTIME  MAX_NOOP  ;largest value so that BUSYWAIT will not use a timer
#endif


;#define LED1SEC_4MHz  (ONE_SECOND/(8*LEDTIME))  ;# times thru LED loop in 1 second with default int osc (4 MHz)
	CONSTANT LED1SEC_4MHz = ONE_SECOND / (8 * LEDTIME)  ;#times thru PK1LED loop in 1 second with default int osc (4 MHz)
#define LED1SEC(clock)  rdiv(LED1SEC_4MHz*InstrPerSec(clock), ONE_SECOND)  ;#times thru PK1LED loop for 1 sec at specified clock speed


	VARIABLE PK1LED_EXPANDED = FALSE  ;avoid redundant code and data expansions

;dummy definitions to avoid compile errors:
#if !PK1DEV
 #define SET_PK1LEDS(value)  ASM_NOP
#endif


;insert an easily recognizable "label" into code:
;this make it easier to find places within a disassembly listing
;params:
; value = byte value to insert into prog space
LABDCL MACRO value
	NOEXPAND  ;hide clutter in LST file
#if PK1DEV
	LOCAL around_label
	LEAVE_PLACEHOLDER1 neargoto around_label
	nop
	REPEAT 3, da (value)
 jumpdest around_label:,,,, FALLTHRU
	FILL_PLACEHOLDER neargoto around_label
#endif
	ENDM


#if PK1DEV
	messg "[CAUTION] COMPILED FOR PK1 BOARD - DEVELOPMENT"

;Pic Kit1 programmer board LED control values:
;NOTE: these are MUXed ("charlieplexed") on RA4+5+2+1, so more than 1 bit is affected and it is not possible to turn them all on at the same time
;see SETPINS for description of values
PK1_LED0 set 0xA015225  ;RA4->RA5
PK1_LED1 set 0xA105225  ;RA5->RA4
PK1_LED2 set 0xA215025  ;RA4->RA2
PK1_LED3 set 0xA205125  ;RA2->RA4
PK1_LED4 set 0xA125025  ;RA5->RA2
PK1_LED5 set 0xA025125  ;RA2->RA5
PK1_LED6 set 0xA225105  ;RA2->RA1
PK1_LED7 set 0xA225015  ;RA1->RA2
PK1_LED8 set 0xA125205  ;RA5->RA1 unpopulated
PK1_LED9 set 0xA025215  ;RA1->RA5 unpopulated
PK1_LED10 set 0xA215205  ;RA4->RA1 unpopulated
PK1_LED11 set 0xA205215  ;RA1->RA4 unpopulated
PK1_ALLOFF set 0xA225225
PK1_LED56 set 0xA02510  ;LED5+6
;other PK1 peripherals:
PK1_ADC set 0x92A  ;RA0
PK1_BUTTON set 0xA552  ;RA3


;set value to be displayed on PK1 LEDs:
#define SET_PK1LEDS(value)  SET_PK1LEDS_ value

SET_PK1LEDS_ MACRO value
	NOEXPAND
	mov8 value, PK1LEDS
	ENDM


;display a value on PK1 LEDs:
;this also generates helper functions (once only) to display the value using the PK1 LEDs
;params:
; delayfunc = delay function
SHOW_PK1LEDS MACRO delayfunc
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH !PK1LED_EXPANDED
	LOCAL pk1bit = 0
	if !PK1LED_EXPANDED  ;avoid redundant code and data expansions; expand first time only
PK1LED_EXPANDED = TRUE
		BDCL PK1LEDS  ;8-bit value to be displayed

		LOCAL around_pk1leds, svbank = PREVBANK
		LEAVE_PLACEHOLDER1JUMP neargoto around_pk1leds
		while pk1bit < 8  ;0, 1, 2, 3, 4, 5, 6, 7
 jumpdest pk1_on_#v(pk1bit):, PK1LEDS|,,,
			NOEXPAND  ;avoid too much mess in LST file
			SETPINS PK1_LED#v(pk1bit), TRUE, NOWAIT
			return_ pk1_on_#v(pk1bit)
pk1bit += 1
		endw
 jumpdest pk1_alloff:, PK1LEDs|,,,
		SETPINS PK1_ALLOFF, TRUE, NOWAIT
		return_ pk1_alloff

 jumpdest around_pk1leds:, svbank|,,,
		FILL_PLACEHOLDER, neargoto around_pk1leds
	endif

	EXPAND_PUSH FALSE
pk1bit = 0
	while bit < 8  ;0, 1, 2, 3, 4, 5, 6, 7
		ifbit PK1LEDS, bit, TRUE, nearcall pk1_on_#v(bit), nearcall pk1_alloff,
		delayfunc LEDTIME  ;for consistent brightness
bit += 1
	endw
	EXPAND_POP,
	ENDM


;single-step (for debug):
;params:
; delayfunc = delay function
WAITFORKEY MACRO delayfunc
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
	LOCAL loop1, loop2
	SETPINS PK1_BUTTON, TRUE, NOWAIT
	whilebit PORTA, 3, FALSE,,  ;RA3 = PK1 push button
	delayfunc 3*ONE_MSEC  ;debounce
	whilebit PORTA, 3, TRUE,,  ;RA3 = PK1 push button
	delayfunc 3*ONE_MSEC  ;debounce
	EXPAND_POP,
	ENDM
#endif


;******************************************************************************
;misc PIC functions:
;******************************************************************************

	CONSTANT ADC_STARTDELAY = 5  ;5*ONE_MSEC  ;5 usec delay required for acquisition (BEFORE starting conversion)
;#define ADC_RESULT  ADRESH  ;left-justified result allows a single 8-bit register to be used
;ADC input impedance should be <= 10K


;start ADC, wait for results:
;params:
; delayfunc = delay function
ADCREAD MACRO delayfunc
	NOEXPAND  ;hide clutter in LST file
;	LOCAL loop
	if MEXPAND_STACKLO & 1
																				EXPAND  ;show ADCREAD params
		NOEXPAND
	endif
	EXPAND_PUSH TRUE  ;show what this does
	delayfunc ADC_STARTDELAY  ;wait for ADC acquisition time
	setbit ADCON0, GO_DONE, TRUE  ;start ADC
;	DELAY 4  ;allow ADC a chance to get started
;loop:
;??	btfsc PIE1, ADIE
;	ifbit ADCON0, GO_DONE, TRUE, neargoto loop,,  ;bit is turned off when ADC is done
	whilebit ADCON0, GO_DONE, TRUE,,  ;bit is turned off when ADC is done
;or check ADIF in PIR1
	EXPAND_POP,
	ENDM


;read program memory:
;gets 14 bits of data
;params:
; baseadrs = start address to read (16 bits)
; offset = offset from start address to read (8 bits)
; result = destination register for result (16 bits)
PROGREAD16 MACRO baseadrs, offset, result
	NOEXPAND
	FLASHIO baseadrs, offset, TRUE, FALSE
	mov16nop EEDAT, result  ;NOTE: only lower 14 bits are valid
	ENDM


;read a byte from EEPROM memory:
;gets 8 bits of data
;params:
; baseadrs = start address to read (16 bits)
; offset = offset from start address to read (8 bits)
; result = destination register for result (8 bits)
EEREAD8 MACRO baseadrs, offset, result
	NOEXPAND
	FLASHIO baseadrs, offset, FALSE, FALSE
	mov8nop EEDAT, result
	ENDM


;read flash or EE memory:
;params:
; baseadrs = start address to read (16 bits)
; offset = offset from start address to read (8 bits)
; want_flash = flag telling whether to read flash (prog space) or eeprom
; want_write = flag telling whether to read or write
FLASHIO MACRO baseadrs, offset, want_flash, want_write
	NOEXPAND  ;hide clutter in LST file
;set up next address:
	add16_8 baseadrs, offset, EEADR,
#if !SELF_WRITE
	ERRIF (want_write) && (want_flash), [ERROR] This device does not support self-writes
#endif
	setbit EECON1, EEPGD, want_flash  ;read program or EE memory
	if want_write
		setbit EECON1, WREN, TRUE  ;enable EE/flash write
		intenable_ FALSE
		mov8 _LITERAL(0x55), EECON2  ;magic sequence
		mov8 _LITERAL(0xaa), EECON2  ;magic sequence
	endif
	setbit EECON1, IIF(want_write, WR, RD), TRUE  ;start EE/flash read/write
	if want_write
		intenable_ TRUE
		setbit EECON1, WREN, FALSE  ;paranoid; disable further EE/flash writes
		whilebit PIR2, EEIF, FALSE,  ;wait for EE write to complete
		setbit PIR2, EEIF, FALSE
	else
		if want_flash
			noop_ 2  ;DO NOT MOVE THIS!  required for flash read
		endif
	endif
	ENDM


;write a byte to EEPROM:
;CAUTION: this macro is NOT TESTED yet
;NOTE: errata says not to use the PIR1 register during an EE write
;params:
; baseadrs = start address to write (16 bits)
; offset = offset from start address to write (8 bits)
; value = register or literal value to write (8 bits)
EEWRITE8 MACRO baseadrs, offset, value
	NOEXPAND
	mov8nop value, EEDAT
	FLASHIO baseadrs, offset, FALSE, TRUE
	EXPAND_POP,
	ENDM


	CONSTANT ANSEL_SIZE = IIF(HIBYTE(ANSEL), 2, 1)

;enable/disable analog I/O:
;params:
; yesno = flag telling whether to enable analog I/O
ANALOG_PINS MACRO yesno
	NOEXPAND  ;hide clutter in LST file
	if yesno
		EXPAND_PUSH TRUE  ;show what this does
;	mov8 _LITERAL(1<<CM2 | 1<<CM1 | 1<<CM0), CMCON0  ;configure comparator inputs as digital I/O (no comparators); overrides TRISC (page 63, 44, 122)
;	SETPINS ADC_DETECT, TRUE, NOWAIT  ;set ADC pin to input
;	mov8  _LITERAL(1<<ANS0), ANSEL  ;set RA0/(AN0) to analog input, all other pins are digital
;	mov8  _LITERAL(0<<ADCS2 | 0<<ADCS1 | 1<<ADCS0), ADCON1  ;ADC clock = Fosc/8
;NOTE: left-justified result allows a single 8-bit register to be used (ADRESH)
;	mov8 _LITERAL(0<<ADFM | 0<<VCFG | 0<<CHS2 | 0<<CHS1 | 0<<CHS0 | 0<<GO_DONE | 1<<ADON), ADCON0  ; left Justified Sample, Use VDD, channel 0, do not start, turn on ADC
		error "[TODO] need some code here to set analog pins"
		EXPAND_POP,
	else
		EXPAND_PUSH TRUE  ;show what this does
		mov#v(8*ANSEL_SIZE) _LITERAL(0), ANSEL  ;configure A/D I/O as digital (no A/D converter); overrides TRISA (page 36, 122)
#ifdef CMCON0
		mov8 _LITERAL(1<<CM2 | 1<<CM1 | 1<<CM0), CMCON0  ;configure comparator inputs as digital I/O (no comparators); overrides TRISC (page 63, 44, 122)
#endif
#ifdef CM1CON0
		mov8 _LITERAL(0), CM1CON0
#endif
#ifdef CM2CON0
		mov8 _LITERAL(0), CM2CON0
#endif
		EXPAND_POP,
	endif
	ENDM


	LIST  ;turn listing back on in caller