;	LIST n=60, c=200, t=on, n=0  ;line (page) size, column size, truncate, no paging
	title  "Renard-HC - DIYC High Channel Count Renard firmware for Microchip PIC16Fxxx"
;******************************************************************************
; Renard-HC - DIYC High(er) Channel Count Renard firmware for Microchip PIC16Fxxx
; Copyright (c) 2008 - 2010 Don Julien.  djulien@thejuliens.net
; May be freely distributed and used for non-commerical purposes (non-commercial share-alike).  
; See the license agreement terms below.
;
; This software was written and is released in the hope that it will help honor and celebrate 
; the birth of Christ, who is *the* Light of the World.  In recognition and appreciation of His 
; love and sacrifice for us, and by His inspiration and enablement, I offer up this work to Him.
; If you feel inclined to support this project financially, please make a donation to a needy 
; family or to a local (Biblical) church.
; For comments, bugs, or other inquiries, please contact the author at djulien@thejuliens.net.
;
; This file contains DIYC Vixen and Renard-compatible AC and DC dimmer firmware for mid-range PICs.
; Although this firmware was initially written for the PIC16F688 in a DIYC Vixen/Renard system, 
; it should be easy to adapt to other microcontrollers or applications.  For siblings such as 
; the PIC16F87 and PIC16F726, ideally only some definitions in the .inc file will need to be changed.
; Don't let the size of this file overwhelm you; the underlying logic in here is fairly simple.
;
; HARDWARE SUMMARY:
; When using a 14-pin PIC, up to 8x7=56 dimmable channels (or 2x8x7=112 with SSR doubling) can be used in a 
; standard Renard configuration.  Up to 3 additional pins can be "freed up" for SSR control control if ext clock 
; or serial in/out are not needed (ext clock is only needed for 115k baud).  Thus, a 14-pin PIC can actually 
; control up to 11x10=110 dimmable channels (or 2x11x10=220 with SSR doubling) in a stand-alone configuration.
; A 22-pin PIC can control up to 16x15=240 dimmable channels (or 2x16x15=480 with SSR doubling).
; Multiple PICs can be chained together for larger channel counts (each acts independently).
; Any PIC16Fxxx device with EUSART, Timer0 and Timer1, and at least 12 I/O pins should work.
; At least 256 bytes RAM is needed to support 56 channels (due to double buffering).
; At least 368 bytes RAM is needed to support 90 channels (due to double buffering).
; ZC signal determines AC rate (50 or 60 Hz); must be polarized if using AC SSR doubling so the PIC 
; can distinguish + and - half cycles; no ZC => use DC rate of 50 Hz (suitable for DC LEDs or servos).
;
; The target hardware of a DIYC Renard controller is as follows:
; - RA5 is for external 18.432 MHz clock signal; OPTIONAL; only needed for reliable communication at 115k baud or SSR over-clocking
; - RA3 is for a ZC signal; OPTIONAL; needed only for AC SSR operation; must be POLARIZED for SSR doubling (use H11A1 instead of H11AA1)
; - RC5 is for Renard serial in; REQUIRED unless running the controller completely stand-alone (requires further code changes)
; - RC4 is for Renard serial out; REQUIRED unless running as the last PIC in a chain and no return loop is needed
; - the remaining I/O pins are used as chipiplexed SSR control lines:
; - for a PIC16F688: 8 remaining pins (RA0-2,4 + RC0-2,3) => up to 8x7 = 56 channels (or 112 when SSR-doubled)
; - for a PIC16F87: 12 remaining pins => up to 12x11 = 132 (264 doubled) channels (max ~ 50% on, due to RAM limits)
; - for a PIC16F726: 21 remaining pins => up to 21x20 = 420 (840 doubled) channels (max ~ 15% on, due to RAM limits)
; - for a PIC16F689/690: 14 remaining pins => up to 14x13 = 182 (364 doubled) channels (max ~ 30% on, due to RAM limits)
; - charlieplexing/chipiplexing is inherently non-PWM, but PWM *can* be achieved by using an I/O pin as if it were a
;   dedicated pin, and setting the charlieplexed/chipiplexed values for the entire row/column appropriately
;
; A standard Renard controller can be retrofitted by customizing the SSR control lines and then reflashing the PICs.
; AC SSR doubling can be used if the ZC is polarized (use an H1A11 instead of H11A11).
; ZC is not needed for DC SSRs or direct LED drive applications such as LED-Triks or RGB LED pixels.
; Standard SSRs can be used with chipiplexing by shorting or removing the SSR resistors.
; A chipiplexed PIC16F688 can handle 7x as many channels as a standard Renard, with 1/7 or 1/14 as many cat5 cables.
; This comes at a slight cost: the chipiplexing is LOSSY under SOME dimming conditions.  The firmware supports
; an adjustable control pulse width in order to allow the level of lossiness to be adjusted.
;
; SOFTWARE SUMMARY:
; - main line initializes registers, then goes into main loop to handle protocol
; - the first SYNC character is used to determine the baud rate being used (choices are 115k, 56k, 38k, or 19k baud)
; - Vixen data or debug/console commands can be sent over the serial line; the protocol state machine can handle these intermixed
; - Timer 0 is used for control pulse timing within the AC cycle, or to drive a DC cycle
; - Timer 1 is used as a RTC (seconds elapsed since startup), in order to measure ZC rate
; - interrupts are NOT used for dimming or serial I/O; ISR overhead is ~ 20 instructions when using FSR, PCLATH,
;   and RP0/RP1/IRP, which is TOO MUCH when used at 256x ZC rate; instead, "cooperative" multi-tasking is used;
;   hardware events are checked and handled during idle times when there is no other time-critical processing
; - cooperative multi-tasking results in very low timing jitter on the SSR timeslot handling (max 4 instructions < 1 usec);
;   data for the next dimming cycle is set up right AFTER the previous one has occurred, when there is still lots of time to do it
; - a dual-protocol state machine is used; Renard data packets can be interleaved with ASCII commands for easier
;   development and testing; for example, use Vixen to run a sequence, then use a terminal emulator to inquire;
;   according to the STATS command, a PIC16F688 is at most 80% busy after heavy dimming at 115k baud
; - the SSR control pulse can be "tuned" (shortened) so that more SSRs can be triggered within a given timeslot;
;   at 256 dimming levels at 60 Hz AC, standard Renard control pulses are ~ 32.6 usec; by increasing the opto
;   current, this pulse can be dropped as low as 5? usec, which allows more SSRs to be on at the same brightness
; - this firware will work with or without an external 18.432 MHz osc (it will fall back to int osc at 8 MHz if absent)
; - true PWM is not possible since the control lines are not dedicated to a channel; however, pseudo-PWM can be supported
;   by auto-repeating previous control pulses; this allows visible LEDs to be driven in a LEDTriks-like arrangement
; - if you are making changes or trying to understand this code, be sure to look at the LST file; it is helpful for seeing generated code
; - this firmware is somewhat optimized, in that it avoids redundant page and bank selects (handled via macros); the
;   savings are substantial (see the stats at the end of the LST file for details)
; - for anyone wanting to jump right to the real "magic", just look at the SSR_ISR, Chipiplex, and FBUnpack macros
;
; The RenardChipiplexer Vixen plug-in sends and this firmware expects the following data packet format:
; - sync, then a command byte of 0x80 + address (like Renard)
; - first byte of actual data = config byte:
;   - <1..0> pulse width: 1x (32.6 usec), 1.5x (20 usec), 2x (16 usec), 2.5x (12 usec)
;   - <3..2> reserved for RGB handling; channels will be treated as R,G,B triplets rather than discrete channels
;   - <4> reserved for checksum handling; useful for environments with communications problems
;   - <5> SSR doubling; channel outputs alternate between positive and negative AC half-cycles
;   - <6> pseudo-PWM; auto-repeat channels until end of half-cycle
;   - <7> macro mode select; other config bits have special meaning if this bit is on
; - remaining bytes are (brightness level, row index, row, row, row, ...), (brightness level, row index, row, row, ...), 0
; - brightness levels are in decreasing order of brightness (since the brighter channels need to be triggered first)
; - a brightness level of 0 is used as end-of-list indicator
;
; This code was compiled at tested with MPASM 5.20, which is BUGGY.
; I had to work around several MPASM bugs - these are described in the FunWithMacros-PIC16Fxxx.inc companion file.
; With the work-arounds, the generated code appears to be correct.  ALWAYS CHECK THE LST OR DISASSEMBLY FILES IF IN DOUBT.
;
; REFERENCES:
; Any page#s referenced in the comments refer to Micrchip's PIC16F688 datasheet, 
; available from Microchip at http://www.microchip.com
;
; Chipiplexing is described at http://www.edn.com/article/CA6615603.html.
; Charlieplexing is described in various places, including the Microchip tech notes at http://ww1.microchip.com/downloads/en/DeviceDoc/chapter%201.pdf
;
; REVISION HISTORY:
; version  date    who  why
; 0.3     3/2008    DJ  first partial version; wrote basic functions
; 0.4     4/2008    DJ  add threads, arithmetic helpers, etc
; 0.5     2/2009    DJ  add half-cycle detection (requires H11A1 instead of H11AA1) and charlieplexing, WDT, FSCM
; 0.61    8/2009    DJ  remove serial I/O software buffers (needless overhead), use double framebuffer to reduce flicker, pseudo-PWM compatibility, software config byte
; 0.62    8/2009    DJ  rewrite to use variable timer 0 presets (reduced overhead from 70% busy when not dimming down to 30% busy)
; 0.7     10/2009   DJ  rewrite timing logic again to avoid interrupts and support AC SSRs (ISR overhead was too high); Chipiplexing interleave assumed DC, didn't work with AC
; 0.8     10/2009   DJ  tigthen up AC SSR version for use with 2009 display (mostly reduce instruction overhead, streamline addressing, etc)
; 1.0     12/14/09  DJ  misc fixes; first actually "usable" version (56k baud, 256 dimmer levels, 2x SSR overclock at 18 MHz, 2 rows/level)
; 1.0a    12/28/09  DJ  avoid repetitive/redundant FSCM/OSF processing (only need to handle OSFIF once at start, not on each timeslot); fix baud rate adjust after FSCM
; 1.0b    1/10/10   DJ  auto-detect 56k vs. 115k baud; now only one version of firmware is needed for any given PIC; finally resolved ghosting row address problem
; 1.0c    2/25/10   DJ  fix row latching in pseudo-PWM mode; fix dump column#; misc code clean up; found LED flicker problem?
; 1.0d    3/23/10   DJ  compress console code a little to free up some more space; increase DC refresh rate from 50 to 100 Hz to cut down on DC visible LED flicker; make compatible with 16F689/690 (NOT TESTED)
; 1.0e    4/17/10   DJ  finish off readability clean-up in preparation for first source code publication
;
; SOFTWARE LICENSE AGREEMENT:
; This work is licensed under the Creative Commons Attribution-Noncommercial-Share Alike 3.0 
; Unported License.  To view a copy of this license, visit 
; http://creativecommons.org/licenses/by-nc-sa/3.0/ or send a letter to 
; Creative Commons, 171 Second Street, Suite 300, San Francisco, California, 94105, USA.
;
; Basically, this software is supplied to you solely for your own personal, non-commerical use,
; and I ask that if you make changes, adaptations, fixes, etc that you freely share those.
; THIS SOFTWARE IS PROVIDED "AS IS".  NO EXPRESS, IMPLIED, STATUTORY OR OTHER WARRANTIES 
; OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE APPLY TO THIS SOFTWARE.  AUTHOR 
; SHALL NOT BE LIABLE FOR SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, FOR ANY REASON OR
; UNDER ANY CIRCUMSTANCES WHATSOEVER.  THIS SOFTWARE MUST NOT BE USED FOR CRITICAL APPLICATIONS
; SUCH AS HEAVY MACHINERY, LIFE SUPPORT, ETC.  YOU USE THIS SOFTWARE AT YOUR OWN RISK.
; But if you have questions or problems with it, let me know and I'll try to help if I can.
;******************************************************************************

	LIST n=60, c=200, t=on, n=0  ;line (page) size, column size, truncate, no paging
;	LIST r=hex  ;radix; NOTE: this affects interpretation of literals, as well as output listing; ALWAYS use D'' with literals > 8
	LIST R=DEC
	LIST mm=on  ;memory map
	LIST st=on  ;symbol table

 
;******************************************************************************
;global compile-time options:
;definitions in this section control various behavior and timing calculations throughout the rest of the code
;if possible, it's better to add/change these definitions rather than making changes directly to the code further on down
;******************************************************************************

;NOTE: 115200 is not reliable with int osc even at 8 MHz; requires 18.432 MHz ext osc for reliable communication at that speed
;Ext clock allows PIC to be driven faster than int osc, which reduces timing errors with EUSART and allows more work to be done.
;External clock for Renard is 18.432 MHz, which results in an instruction cycle time of 217 nsec (~ 4.6 MIPS).
;Highest int osc speed for PIC16F688 is 8 MHz, for an instruction time of 500 nsec (2 MIPS).

#define BAUD_RATE  115200  ;highest baud rate to use; lower rates will be auto-detected; set to 0 for stand-alone operation to free up 2 I/O pins
#define HIGHBAUD_OK  FALSE  ;set this only if you are using 16F688 rev A4 or later silicon; see Errata for details
#define AUTO_DETECT_BAUD  TRUE  ;set this to enable auto-detect baud rate on first SYNC char
#define INT_OSC_FREQ  8000000  ;run int osc at max speed (8MHz)
#define EXT_CLK_FREQ  18432000  ;Renard ext clock frequency 18.432MHz, if present; works well with 115200 baud; set to 0 if ext clock is never present
#define WANT_ZC  TRUE  ;turn this off if not using ZC and you want to free up an I/O pin for other purposes
#define DC_REFRESH  50;100;50 ;#cycles/second for DC; 50 is suitable for servos; 100 is suitable for flicker-free visible LEDs; this also controls internal RTC rate, and is used as ZC if no external ZC is supplied
#define DIMMER_RES  256  ;#distinguishable dimming levels wanted (Vixen is hard-coded for 1-byte values, so 256 is the max permitted value for this)
#define SSR_PRECLEAR  FALSE  ;use this to try to avoid random channels turning on
#define RECLAIM_PINS  FALSE;TRUE  ;set this to relcaim unused pins for SSR control; ZC, ext clock, and serial in/out may be reclaimed if not otherwise needed
#define VIXEN_PLUGIN_BUGGY  FALSE  ;set this if Vixen plug-in is unreliable
#define ZC_TRACKING  TRUE;FALSE  ;set this to track ZC variance (useful for flicker diagnosis)
#define ZC_FILTER  4  ;use this to adjust how much filtering to apply to ZC signal; 4 timeslots seems to work well

;When chipiplexing, it is desirable to use as short an SSR control pulse as possible, in order to allow more rows to be sent within a given timeslot.
;Standard U.S. control pulses are ~ 32.6 usec (256th of 60 Hz AC half-cycle), or ~39 usec for U.K. (256th of 50 Hz AC half-cycle).
;According to the LED current vs. turn-on time curve in the MOC3023 spec sheets, the optos will turn on faster if they are driven at higher current.
;Hence, lowering the opto series resistor will allow the SSRs to be "over-clocked".
;The standard AC SSR circuit drives the optos at ~ 5 mA, but the PIC I/O pins can handle up to 25 mA, so the SSRs could actually be driven 4x as fast.
;Adjust the value below to over-clock the SSRs at 2x or 4x the normal Renard speed.
;It looks like the PIC16F688 will max out at about 2.5 x at 18 MHz ext clock and 2*60 Hz refresh:
#define SSR_OVERCLOCK  2;1  ;SSRs can be overclocked up 4x; set to 1 for standard Renard speed (32.6 usec for 60 Hz AC, 39 usec for 50 Hz AC)

;Using interrupts gives the flexibility of responding to them any time from the caller (as opposed to cooperative multi-tasking, which only handles them selectively),
;BUT this comes at a cost.  Since PCLATH and FSR also need to be saved/restored, it's about 20 instructions of overhead.
;That adds significant ISR latency and adds up to a lot of overhead, since Timer 0 is used at 1x, 2x or 4x the ZC/256 frequency.
;(20 instructions at 4x 60 Hz AC ZC/256 rate = 120*256*4=120KHz ~= 2.5 MIPS, which eats up over half the processor at 18.432 MHz or all of it at 8 MHz)
;It turns out to be more efficient to just poll for the desired interrupts cooperatively.  We can get this down to a very tight loop
;(5 instructions rather than 20 only uses about 13% processor time at 18.432 MHz or 30% at 8 MHz).
#define USE_INTS  FALSE;TRUE  ;interrupts no longer used (too much ISR overhead); cooperative multi-tasking now used instead

;#define ISR_LONGJUMPS  FALSE  ;not using long jumps within ISR (PCLATH not changed)
;#define ONEINT_ONLY  FALSE  ;service all interrupts in one ISR call
#define ONEINT_ONLY  TRUE  ;service only 1 interrupt each time thru ISR

#define IDE_DEBUG  FALSE;TRUE  ;pull in extra code to allow step-thru with IDE; MUST BE OFF FOR REAL USAGE
#define PK1DEV  FALSE  ;using the PicKit1 board; DEVELOPMENT ONLY
#define PK1SIM  TRUE;FALSE  ;test non-PK1 operation while still on PK1 dev board (requires PK1 board mod: move PK1LEDs from RA5 to RA0 so Renard ext clock can be used on RA5)
#define PERF_STATS  FALSE;TRUE  ;keep performance stats (adds some overhead, but provides very useful info)
#define IO_STATS  TRUE  ;keep I/O stats (adds a little overhead, but provides very useful info)
#define DEBUG_TTY  TRUE  ;allow console control/debug via tty; very little run-time overhead when not used; useful for dev/test
#define FULL_WARN  FALSE;TRUE  ;leave this ON if you are unfamiliar with the code; gives full warnings
#define WANT_DEBUG  FALSE  ;include additional paranoid checks (mainly useful for debug)

#define VERSION  "1.0e (4/17/10)"  ;update this whenever firmware is released!


;******************************************************************************
;helper macros:
;******************************************************************************

;direct addressing is not needed above Bank 1, so we don't ever need to set RP1 (reduces BANKSEL instructions):
#define USE_RP0  (MAX_SFR > BANK_SIZE)
#define USE_RP1  (MAX_SFR > 2*BANK_SIZE)

	INCLUDE "FunWithMacros-PIC16Fxxx.inc"  ;PIC16xxx macros, pseudo-opcodes, and helper functions
	ERRIF2 PIC16Fxxx_version != 0x10f, [WARNING] Macros are incompatible: got #v(PIC16Fxxx_version), expected #v(0x10f)  ;compatibility checking


	CONSTANT AC60HZ = 2 * 60  ;AC half-cycles per second is DOUBLE
	CONSTANT AC50HZ = 2 * 50  ;AC half-cycles per second is DOUBLE


;kludgey int-to-string macros for some values used later:
;(MPASM doesn't convert numbers to strings)
#define tostr_0  "0"
#define tostr_1  "1"
#define tostr_2  "2"
#define tostr_3  "3"
#define tostr_4  "4"


;******************************************************************************
;config section:
;this area constructs config register values
;ideally, the compile-time options above should be used to alter these register values, rather than changing the code in here
;******************************************************************************

;set missing defines:
#ifndef BAUD_RATE
 #define BAUD_RATE  0  ;defaults no serial I/O if not specified earlier
#endif  ;BAUD_RATE

#ifndef EXT_CLK_FREQ
 #define EXT_CLK_FREQ  0  ;defaults to no ext clock if not specified earlier
#endif  ;EXT_CLK_FREQ


;resolve config conflicts:
#if PK1DEV
	WARNIF EXT_CLK_FREQ, [INFO] Ext clock mhz(EXT_CLK_FREQ) cannot be used with PicKit1 dev board (due to RA5 conflict): turning off ext clock
 #if EXT_CLK_FREQ
  #undefine EXT_CLK_FREQ
  #define EXT_CLK_FREQ  0  ;get rid of ext clock; RA5 is used for the PK1 LEDs
 #endif  ;EXT_CLK_FREQ
#endif  ;PK1DEV


#if EXT_CLK_FREQ > INT_OSC_FREQ
 #define OVERCLOCK_LIMIT  3  ;max practical limit on ext 18.432 MHz clock
#else  ;EXT_CLK_FREQ
 #define OVERCLOCK_LIMIT  1  ;max practical limit on int 8 MHz osc
#endif  ;EXT_CLK_FREQ
	WARNIF SSR_OVERCLOCK > OVERCLOCK_LIMIT, [INFO] SSR overclock #v(SSR_OVERCLOCK) reduced to #v(OVERCLOCK_LIMIT): int osc mhz(INT_OSC_FREQ) not fast enough
#if SSR_OVERCLOCK > OVERCLOCK_LIMIT
 #undefine SSR_OVERCLOCK
 #define SSR_OVERCLOCK  OVERCLOCK_LIMIT
#endif  ;SSR_OVERCLOCK > 3

#if SSR_OVERCLOCK > 1
	WARNIF PERF_STATS, [INFO] Forcing stats off: too much overhead for SSR over-clocking #v(SSR_OVERCLOCK)
 #undefine PERF_STATS
 #define PERF_STATS  FALSE  ;too much overhead at this speed; turn it off
#endif  ;SSR_OVERCLOCK > 1

#if !WANT_ZC
 #undefine ZC_TRACKING
 #define ZC_TRACKING  FALSE  ;there's no point in tracking it if it's not there
#endif

#if PK1DEV && PK1SIM
 #undefine PK1SIM
 #define PK1SIM  FALSE  ;don't need this if really using PK1 board
#endif


	VARIABLE MY_CONFIG = -1  ;start with all bits on, then EXPLICITLY turn them off below
#if EXT_CLK_FREQ  ;ext clock might be present
MY_CONFIG &= _EC_OSC  ;I/O on RA4, CLKIN on RA5; external clock (18.432 MHz); if not present, int osc will be used
MY_CONFIG &= _FCMEN_ON  ;turn on fail-safe clock monitor in case external clock is not connected or fails (page 33); RA5 will still be configured as input, though
#else  ;EXT_CLK_FREQ
MY_CONFIG &= _INTRC_OSC_NOCLKOUT  ;I/O on RA4+5; internal clock (default 4 MHz, later bumped up to 8 MHz)
MY_CONFIG &= _FCMEN_OFF  ;disable fail-safe clock monitor; NOTE: this bit must explicitly be turned off since MY_CONFIG started with all bits ON
#endif  ;EXTCLK_FREQ
MY_CONFIG &= _IESO_OFF  ;internal/external switchover not needed; turn on to use optional external clock?  disabled when EC mode is on (page 31); TODO: turn on for battery-backup or RTC
MY_CONFIG &= _BOR_OFF  ;brown-out disabled; TODO: turn this on when battery-backup clock is implemented?
MY_CONFIG &= _CPD_OFF  ;data memory (EEPROM) NOT protected; TODO: CPD on or off? (EEPROM cleared)
MY_CONFIG &= _CP_OFF  ;program code memory NOT protected (maybe it should be?)
MY_CONFIG &= _MCLRE_OFF  ;use MCLR pin as INPUT pin (required for Renard); no external reset needed anyway
MY_CONFIG &= _PWRTE_ON  ;hold PIC in reset for 64 msec after power up until signals stabilize; seems like a good idea since MCLR is not used
MY_CONFIG &= _WDT_ON  ;use WDT to restart if software crashes (paranoid); WDT has 8-bit pre- (shared) and 16-bit post-scalars (page 125)

	__config MY_CONFIG


	WARNIF MY_CONFIG & ~(_INTRC_OSC_NOCLKOUT | _INTRC_OSC_CLKOUT), [CAUTION] Using ext clock mhz(EXT_CLK_FREQ): RA5 is clock INPUT only
	WARNIF !(MY_CONFIG & ~(_INTRC_OSC_NOCLKOUT | _INTRC_OSC_CLKOUT)), [INFO] Using int osc mhz(INT_OSC_FREQ) +/-: RA5 can be an OUTPUT line


;******************************************************************************
;hardware section:
;contains stuff related to pin assignments
;initially this is set up for the Renard PIC (16F688), but others can be added
;******************************************************************************

;PIC16F688 pinout:
;  Vdd 1  14 Vss
;  RA5 2  13 RA0
;  RA4 3  12 RA1
;  RA3 4  11 RA2
;  RC5 5  10 RC0
;  RC4 6   9 RC1
;  RC3 7   8 RC2

;Renard PIC16F688 8-bit standard I/O pin assignments:
;RA5 = ext clock
;RA3 = ZC (this pin can only be used for input anyway)
;RA4,RA2,RA1,RA0 = SSR control
;RC4,RC5 = serial I/O
;RC3,RC2,RC1,RC0 = SSR control
;NOTE: RA0/RA5 usage conflicts with PicKit1 board, which has LEDs on RA5 instead of RA0


;IOPIN decoders:
;example: if ZC is pin A3, then IOPORT(ZC) = 0xA, and IOBIT(ZC) = 3
#define IOPORT(iopin)  ((iopin)/0x10)
#define IOBIT(iopin)  ((iopin)%0x10)


;add an I/O pin to a list:
;params:
; purpose = symbolic name containing list of I/O pins
; iopin = which I/O pin to add to the list
ADDPIN MACRO purpose, iopin
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE  ;don't show messy macro expansion
	if iopin < 0  ;output pin
purpose#v(IOPORT(-iopin)) &= ~(1<<IOBIT(-iopin))  ;add this to list of pins for this purpose
	else  ;input pin
purpose#v(IOPORT(iopin)) |= (1<<IOBIT(iopin))  ;add this to list of pins for this purpose
	endif
	EXPAND_POP,
	ENDM


;generate I/O pin change buffers for each port that controls SSRs:
IOINIT_CONSTS MACRO
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE  ;don't show messy macro expansion
	VARIABLE port
port = FIRST_PORT;0xA
	while port <= LAST_PORT;0xF
		VARIABLE TRISINIT#v(port) = PORT#v(port)_BITS  ;default TRIS setting; all Z/input unless set otherwise
		VARIABLE SSRPINS#v(port) = 0  ;keep track of which pins are used for SSR control
port += 1
	endw  ;port <= 0xF
	EXPAND_POP,
	ENDM

	IOINIT_CONSTS

;define SSR control pins:
;see standard Renard I/O pin assignments listed above
	ADDPIN SSRPINS, 0xC3
	ADDPIN SSRPINS, 0xC2
	ADDPIN SSRPINS, 0xC1
	ADDPIN SSRPINS, 0xC0
	if (NUMBITS_#v(PORT#v(0xC)_BITS) == 8) && (IOPORT(SERIN_PIN) != 0xC) && (IOPORT(SEROUT_PIN) != 0xC) && (IOPORT(RESET_PIN) != 0xC)  ;put all SSR pins on the same port
		ADDPIN SSRPINS, 0xC4
		ADDPIN SSRPINS, 0xC5
		ADDPIN SSRPINS, 0xC6
		ADDPIN SSRPINS, 0xC7
	else  ;split across ports
		ADDPIN SSRPINS, 0xA4
		ADDPIN SSRPINS, 0xA2
		ADDPIN SSRPINS, 0xA1
		ADDPIN SSRPINS, IIF(PK1DEV, 0xA5, 0xA0)  ;RA0 <-> RA5; whoops; PK1 Board LEDs are connected to RA5 rather than RA0 in an 8-bit Renard
	endif


;other I/O port pins:
;these can be used for SSR control if not otherwise needed
#if WANT_ZC
	CONSTANT ZC = RESET_PIN  ;0xA3
	ADDPIN TRISINIT, ZC  ;set this pin to input
#else  ;WANT_ZC
;NO:	ADDPIN SSRPINS, 0xA3  ;this pin canNOT be used for SSR control
#endif  ;WANT_ZC

#if BAUD_RATE
	ADDPIN TRISINIT, SERIN_PIN  ;set this pin to input
	ADDPIN TRISINIT, -SEROUT_PIN  ;set this pin to output
#else  ;BAUD_RATE
 #if RECLAIM_PINS
	ADDPIN SSRPINS, SERIN_PIN  ;this pin can be used for SSR control
	ADDPIN SSRPINS, SEROUT_PIN  ;this pin can be used for SSR control
 #else  ;RECLAIM_PINS
  #undefine SERIN_PIN
  #undefine SEROUT_PIN
 #endif  ;RECLAIM_PINS
#endif  ;BAUD_RATE

#if EXT_CLK_FREQ
	ADDPIN TRISINIT, CLKIN_PIN  ;set this pin to input
#else  ;EXT_CLK_FREQ
 #if RECLAIM_PINS
	ADDPIN SSRPINS, CLKIN_PIN  ;this pin can be used for SSR control
 #else  ;RECLAIM_PINS
  #undefine CLKIN_PIN
 #endif  ;RECLAIM_PINS
#endif  ;EXT_CLK_FREQ

	WARNIF2 TRUE, [INFO] Port A SSR pin bitmap: #v(SSRPINS#v(0xA)), Port C SSR pin bitmap: #v(SSRPINS#v(0xC))  ;debug


;******************************************************************************
;control register section:
;this area constructs control register values
;also defines speed-dependent clock/timing values
;******************************************************************************

	VARIABLE MY_IOC#v(0xA) = 0
;no longer used (ZC is now sampled by Timer 0):
;#if WANT_ZC
;MY_IOC#v(IOPORT(ZC)) |= 1<<IOC#v(IOBIT(ZC))  ;RA3 is ZC signal; that's the only one we need to watch; BOTH EDGES are monitored
;#endif  ;WANT_ZC


	VARIABLE MY_WPU#v(0xA) = 0
;WPUA not applicable to RA3 (input-only, MCLR):
;#if WANT_ZC
;MY_WPU#v(IOPORT(ZC)) |= 1<<IOC#v(IOBIT(ZC))  ;RA3 is ZC signal; that's the only one we need to watch; BOTH EDGES are monitored
;#endif  ;WANT_ZC
;#if BAUD_RATE > 0
;MY_WPU#v(IOPORT(SERIN)) |= 1<<IOC#v(IOBIT(SERIN))
;#endif  ;BAUD_RATE > 0
#if EXT_CLK_FREQ
MY_WPU#v(IOPORT(CLKIN_PIN)) |= 1<<IOC#v(IOBIT(CLKIN_PIN))  ;pull up in case absent or fails
#endif  ;EXT_CLK_FREQ


;AC (phase angle) and DC (PWM) dimming timing window analysis:
;60Hz AC = 120 half cycles/sec ~= 8.3 msec/half cycle
;50Hz AC = 100 half cycles/sec = 10 msec/half cycle
;50Hz DC (typical servo timing) = 20 msec/cycle
;100Hz DC (POV for visible-LED applications) = 10 msec/cycle
;For 256-level AC dimming at 2*60 Hz, we need time slices of ~ 8333/256 ~= 32.6 usec.
;For 256-level AC dimming at 2*50 Hz, we need time slices of 10000/256 ~= 39 usec.
;For 256-level DC PWM for servos at 50 Hz (20 msec interval), we need time slices of ~ 20000/256 ~= 78 usec.
;For 256-level DC POV for visible LEDs at 100 Hz, we need time slices of 10000/256 ~= 39 usec.
;Hence, the target interval is ~ 32.6, 39, or 78 usec.
;If SSRs are overclocked at 2x, the interval would be half of that: ~ 16.3, 29.5, or 39 usec.

;For Renard I/O at 115200 baud, we would need bit time slices of ~9 usec (too much overhead for s/w),
;so the EUSART is used for Renard I/O; this frees up the Timer 0 for use exclusively for SSR control pulse timing.
;Timer 0 is used as timing control to turn on SSRs/LEDs.
;Timer 0 is 8 bits, and will wrap after 256 * .217 usec * 4 ~= 222 usec at 18.432 MHz or 256 * .5 usec * 4 = 512 usec at 8 MHz with no prescalar.
;Since both of these wrap-around times are > the 33/39/78 usec needed for the dimmer time slices (1/256th ZC interval), NO Timer 0 prescalar is needed.
;To get exact time intervals of ~32.6, ~39 or ~78 usec with an 18.432 MHz external clock, Timer 0 needs to wrap after ~38/45/86 ticks.
;To get exact time intervals of ~32.6, ~39 or ~78 usec with the 8 MHz int osc, Timer 0 needs to wrap after ~16/20/39 ticks.
;Possible prescalar values are 1, 2, 4, 8, 16, 32, 64, 128, or 256.
;Use as small a value as possible for better accuracy, but as large a value as possible to reduce overhead.
;For simpler timing calculations, the same prescalar value is used for all possible clock speeds.
;However, the preset value must be varied to get the correct tick rate.
	CONSTANT Timer0_Prescalar = IIF(SSR_OVERCLOCK == 1, 2, 1)  ;1 tick every .5 usec at 8MHz or .217 usec at 18.432MHz; wraps after 256 ticks (512 usec at 8MHz or ~222 usec at 18.432 MHz)
	CONSTANT Timer0_MinPrescalar = 2  ;this is the smallest configurable Timer 0 prescalar
	ERRIF Timer0_Prescalar%(Timer0_MinPrescalar/2), [ERROR] Timer 0 prescalar #v(Timer0_Prescalar) out of range >= #v(Timer0_MinPrescalar)  ;paranoid
	CONSTANT SSR_RATE = ONE_SECOND/(DIMMER_RES*SSR_OVERCLOCK)  ;ZC * rate at which each SSR will be refreshed in usec (this/ZC = actual refresh time)

#define SSR_Timeslot(clock)  (InstrPerSec(clock)/Timer0_Prescalar/DIMMER_RES/SSR_OVERCLOCK)  ;#instructions per dimmer value at specified clock speed and refresh rate for Timer 0

;max #instructions needed for timeslot handling (between SSR control timeslots):
;this is used as a compile-time check to verify there is enough PIC bandwidth for chipiplexing at the specified frequency
	CONSTANT TSIMAX = 20  ;estimate of #instructions needed for chipiplexing; Chipiplex macro takes ~10, and busy wait loop takes about 4, plus misc other overhead

;Set Timer 0 to this value to get it to wrap at the correct interval:
;2 instr are lost when updating Timer 0 according to Microchip notes
;Ideally, 256 * Timer 0 tick rate would be == desired timeslot
;Since it is not an exact multiple, we must rouind up or round down.
;Rounding down will undershoot the cycle, so that the lowest dimming levels are not achievable
;Rounding up will cause slight overshoot at the end of the cycle and start the next cycle too early;
;however, this can be compensated for by leaving the tail end of the cycle dead
;The ZC signal is used to resync the timing to the AC waveform, in case timing errors or lost interrupts cause drift during the cycle.
;This means that the inaccuracy is absolute, not cumulative across cycles.
#define Timer0_Preset(clock, refresh)  (256 - rdiv(SSR_Timeslot(clock), refresh) + 2/Timer0_Prescalar)


;OPTIONS register (page 16, 49, 122):
;this register will be set once at startup only
	VARIABLE MY_OPTIONS = 0  ;default to all bits off, then turn on as needed
MY_OPTIONS |= IIF(!MY_WPU#v(0xA), 1<<NOT_RAPU, 0)  ;disable weak pull-ups on PORT A disabled (page 36); NOTE: double negative grammar
MY_OPTIONS |= IIF(DONT_CARE, 1<<INTEDG, 0)  ;RA2 interrupt edge (not used as an external interrupt line)
MY_OPTIONS |= IIF(FALSE, 1<<T0CS, 0)  ;timer 0 clock source = chip clock (could be int osc or ext clock); timer 0 is 8 bits + configurable 8-bit prescaler
MY_OPTIONS |= IIF(DONT_CARE, 1<<T0SE, 0)  ;don't care on timer 0 edge, as long as it's consistent
MY_OPTIONS |= IIF(Timer0_Prescalar == 1, 1<<PSA, 0)  ;FALSE => pre-scalar assigned to timer 0, TRUE => WDT
;<PS2, PS1, PS0> form a 3-bit binary value:
MY_OPTIONS |= IIF(BIT_#v(Timer0_Prescalar/Timer0_MinPrescalar) & 4, 1<<PS2, 0)  ;1:32, 1:64, 1:128 or 1:256 prescalar on timer 0 (page 49)
MY_OPTIONS |= IIF(BIT_#v(Timer0_Prescalar/Timer0_MinPrescalar) & 2, 1<<PS1, 0)  ;1:8, 1:16, 1:128 or 1:256 prescalar on timer 0 (page 49)
MY_OPTIONS |= IIF(BIT_#v(Timer0_Prescalar/Timer0_MinPrescalar) & 1, 1<<PS0, 0)  ;1:4, 1:16, 1:64 or 1:256 prescalar on timer 0 (page 49)


;INTCON (interrupt control) register (page 17):
;this register will be set at startup, but individual interrupt bits will be cleared again each time their interrupt occurs
	VARIABLE MY_INTCON = 0  ;default to all bits off, then turn on as needed
;#if USE_INTS
MY_INTCON |= IIF(FALSE, 1<<GIE, 0)  ;global interrupt-enable off; this will be turned on after ISR data is initialized (if compiled for interrupts)
MY_INTCON |= IIF(TRUE, 1<<PEIE, 0)  ;peripheral interrupts enabled
MY_INTCON |= IIF(TRUE, 1<<T0IE, 0)  ;timer 0 overflow interrupt; timer 0 is 8 bits, used for 1/256th ZC timing
MY_INTCON |= IIF(FALSE, 1<<INTE, 0)  ;RA2 external interrupt disabled (not needed since we can detect changes on Port A pins via IOCA)
MY_INTCON |= IIF(MY_IOC#v(0xA), 1<<RAIE, 0)  ;RA change interrupt enabled (also needs IOCA; page 17)
;#endif  ;USE_INTS


;PIE1 (peripheral interrupt enable) register (page 18):
;this register will be set at startup; individual interrupt bits will be cleared again each time their interrupt occurs
	VARIABLE MY_PIE1 = 0  ;default to all bits off, then turn on as needed
;#if USE_INTS
MY_PIE1 |= IIF(TRUE, 1<<EEIE, 0)  ;EE write complete enabled (used for persistent data)
MY_PIE1 |= IIF(FALSE, 1<<ADIE, 0)  ;A/D converter disabled
;EUSART interrupts NOT used; handled by busy waits in main-line instead:
MY_PIE1 |= IIF(FALSE, 1<<RCIE, 0)  ;EUSART receive full interrupt disabled (non-interrupt polling is used)
MY_PIE1 |= IIF(FALSE, 1<<TXIE, 0)  ;EUSART transmit empty interrupt disabled; MUST BE OFF UNLESS TRANSMITTING (page 87); polled
MY_PIE1 |= IIF(FALSE, 1<<C1IE | 1<<C2IE, 0)  ;comparators 1 + 2 disabled
MY_PIE1 |= IIF(EXT_CLK_FREQ, 1<<OSFIE, 0)  ;ext oscillator fail enabled; switch to int osc when this happens
;NOTE: TMR1IE must be on to get TMR1IF, even if not using ints?
MY_PIE1 |= IIF(TRUE, 1<<TMR1IE, 0)  ;timer 1 overflow enabled; timer 1 is 16 bits, used for real-time clock and ZC fallback
;#endif  ;USE_INTS


;detect whether ext clock is present and working:
;NOTE: this must return TRUE if there is no ext clock
;PIR1.OSFIF and OSCCON.SCS don't appear to be reliable; Microchip notes say OSCCON.OSC might not reflect pending OSF change
;#define USING_INTOSC(yesno)  PIR1, OSFIF, yesno  ;ext oscillator failed/absent, running from int osc
;#define USING_INTOSC(yesno)  OSCCON, OSTS, yesno  ;ext oscillator failed/absent, running from int osc
;#define USING_INTOSC(yesno)  OSCCON, SCS, yesno  ;ext oscillator failed/absent, running from int osc; FSCM does not update SCS (page 31)
#define USING_INTOSC(yesno)  BITVAR(UsingExtClock), !(yesno)  ;ext oscillator failed/absent, running from int osc; FSC does not update SCS (page 31)

;rather than checking multiple bits, we use our own to keep track of it:
	BITDCL UsingExtClock  ;using ext clock; FALSE if failed or not present (using int osc); starts as 0 so it will be correct if there is no ext clock

	CONSTANT IntOscScaler = 125000/2  ;smallest configurable HF int osc frequency (shifted right 1 bit to bypass LF int osc frequency)
	ERRIF INT_OSC_FREQ%IntOscScaler, [ERROR] Int osc frequency mhz(INT_OSC_FREQ) out of range >= #v(IntOscScaler)

;OSCCON (oscillator control) register (page 24):
;set this register once at startup only
	VARIABLE MY_OSCCON = 0  ;default to all bits off, then turn on as needed
MY_OSCCON |= IIF(FALSE, 1<<SCS, 0)  ;use CONFIG clock source (ext clock)
;default internal osc freq is 4 MHz; bump it up to 8 MHz (fail-safe frequency, page 33)
;<IRCF2, IRCF1, IRCF0> form a 3-bit binary value:
MY_OSCCON |= IIF(BIT_#v(INT_OSC_FREQ/IntOscScaler) & 1, 1<<IRCF0, 0)  ;125KHz, 500KHz, 2MHz or 8MHz int osc (page 22)
MY_OSCCON |= IIF(BIT_#v(INT_OSC_FREQ/IntOscScaler) & 2, 1<<IRCF1, 0)  ;250KHz, 500KHz, 4MHz or 8Mhz int osc (page 22)
MY_OSCCON |= IIF(BIT_#v(INT_OSC_FREQ/IntOscScaler) & 4, 1<<IRCF2, 0)  ;1MHz, 2MHz, 4MHz or 8MHz int osc (page 22)

#define NOT_4MHz  OSCCON, IRCF0, TRUE  ;setbit params to check for adjusted clock; it's sufficient to just check this bit, since 8 MHz is the only other value used


;BAUDCTL (page 96):
;set this register once at startup, but adjust it after first char received (auto baud rate detection)
	VARIABLE MY_BAUDCTL = 0  ;default to all bits off, then turn on as needed
MY_BAUDCTL |= IIF(HIGHBAUD_OK, 1<<BRG16, 0)  ;16-bit baud rate generator; SEE ERRATA PAGE 2 REGARDING BRGH && BRG16; NOT needed since we are NOT using low baud rates
 messg CHECK THIS ^^
MY_BAUDCTL |= IIF(FALSE, 1<<WUE, 0)  ;no wakeup-enable
MY_BAUDCTL |= IIF(FALSE, 1<<ABDEN, 0)  ;no auto-baud rate detect (not reliable - consumes first char, which must be a Break)
MY_BAUDCTL |= IIF(FALSE, 1<<SCKP, 0)  ;synchronous clock polarity select: 1 => xmit inverted; 0 => xmit non-inverted


;Transmit Status and Control (TXSTA) register (page 87, 100):
;this register will be set at startup, but individual bits may be turned off/on again later to reset the UART
	VARIABLE MY_TXSTA = 0  ;default to all bits off, then turn on as needed
MY_TXSTA |= IIF(TRUE, 1<<BRGH, 0)  ;high baud rates
MY_TXSTA |= IIF(FALSE, 1<<SYNC, 0)  ;async mode
MY_TXSTA |= IIF(FALSE, 1<<TX9, 0)  ;don't want 9 bit data for parity
MY_TXSTA |= IIF(TRUE, 1<<TXEN, 0)  ;transmit enable (sets TXIF interrupt bit)
MY_TXSTA |= IIF(DONT_CARE, 1<<CSRC, 0)  ;clock source select; internal for sync mode (master); don't care in async mode; Errata for other devices says to set this even though it's a don't care
MY_TXSTA |= IIF(FALSE, 1<<SENDB, 0)  ;don't send break on next char (async mode only)
MY_TXSTA |= IIF(FALSE, 1<<TX9D, 0)  ;clear 9th bit


;Receive Status and Control (RCSTA) register (page 90, 100):
;this register will be set at startup, but individual bits may be turned off/on again later to reset the UART
	VARIABLE MY_RCSTA = 0  ;default to all bits off, then turn on as needed
MY_RCSTA |= IIF(TRUE, 1<<SPEN, 0)  ;serial port enabled
MY_RCSTA |= IIF(FALSE, 1<<RX9, 0)  ;9 bit receive disabled (no parity)
MY_RCSTA |= IIF(DONT_CARE, 1<<SREN, 0)  ;single receive enable; don't care in async mode
MY_RCSTA |= IIF(TRUE, 1<<CREN, 0)  ;continuous receive enabled
MY_RCSTA |= IIF(DONT_CARE, 1<<ADDEN, 0)  ;address detect disabled (not used for async mode)
MY_RCSTA |= IIF(FALSE, 1<<RX9D, 0)  ;clear 9th bit


;SPBRG register:
;Microchip baud rate formulas on page 97 dosn't quite match charts on pages following in some cases,
;but I think it is supposed to be:
;   when brgh=brg16=1   spbrg_value = (xtal_freq/(baudrate*4))-1
;   when brgh=brg16=0   spbrg_value = (xtal_freq/(baudrate*64))-1
;   when brgh!=brg16   spbrg_value = (xtal_freq/(baudrate*16))-1
	CONSTANT BRSCALE = IIF(MY_BAUDCTL & (1<<BRG16), 1, 4) * IIF(MY_TXSTA & (1<<BRGH), 1, 4)  ;part of denominator that depends on BRGH + BRG16
#define SPBRG_Preset(clock, rate)  (rdiv(InstrPerSec(clock)/BRSCALE, rate) - 1)  ;s/w programmable baud rate generator value
#if HIGHBAUD_OK  ;only set this if you are using 16F688 rev A4 or later silicon
	WARNIF (MY_BAUDCTL & (1<<BRG16)) && (MY_TXSTA & (1<<BRGH)) && FULL_WARN, [CAUTION] 16F688.A3 errata says NOT to use BRG16 and BRGH together: only safe on A4 and later silicon  ;paranoid self-check
#else
	ERRIF (MY_BAUDCTL & (1<<BRG16)) && (MY_TXSTA & (1<<BRGH)) && FULL_WARN, [ERROR] 16F688.A3 errata says NOT to use BRG16 and BRGH together: only safe on A4 and later silicon  ;paranoid self-check
#endif
	WARNIF TRUE, [INFO] Compiled for kbaud(BAUD_RATE) max baud rate with scaler #v(BRSCALE)

;18.432 MHz / 4 / 40 == 115200, which gives 0% error at high baud rates
;10 bit char time at 115200 means pumping out chars at 11520 Hz, which is every ~87 usec (== 400 instr @18.432MHz)
;8 MHz / 4 / 17 ~= 117647 ~= 2% error at 115200 baud rate (page 100)
;10 bit char time at 115200 means pumping out chars at 11520 Hz, which is every ~87 usec (== 174 instr @8MHz)

;KLUDGE: even though SPBRG charts and formula show that 115k is attainable at 8 MHz, in actual practice this doesn't appear to be true
;Therefore, the definitions below are used to reduce the baud rate to a reliable speed:
 	CONSTANT BRLIMIT = IIF((MY_BAUDCTL & (1<<BRG16)) && (MY_TXSTA & (1<<BRGH)), 35, 26)
;#define MaxBaudRate(clock)  InstrPerSec(clock)/(2*BRSCALE))  ;maximum attainable baud rate at specified clock freq
;#define MaxBaudRate(clock)  IIF((clock) <= 8000000, 57600, (InstrPerSec(clock)/(2*BRSCALE)))  ;maximum attainable baud rate at specified clock freq
;#define MaxBaudRate(clock)  1000000
#define MaxBaudRate(clock)  (InstrPerSec(clock)/BRLIMIT)  ;maximum attainable baud rate at specified clock freq

;define alternate (lower) baud rates if it's too high:
	CONSTANT BAUD_RATE_STEPDOWN#v(115200) = 57600
	CONSTANT BAUD_RATE_STEPDOWN#v(57600) = 38400
	CONSTANT BAUD_RATE_STEPDOWN#v(38400) = 19200
	CONSTANT BAUD_RATE_STEPDOWN#v(19200) = 9600


;WDT time base is 31.5 KHz
;Since there will be a RTC tick every 20 msec, WDT interval should be > 20 msec
;With a 1:1 prescalar, possible values are 1:32 (1 msec), 1:64, 1:128, 1:256, 1:512 (16 msec), 1:1024, 1:2048, 1:4096, 1:8192, 1:16384, 1:32768, or 1:65536 (~ 2 sec)
;NOTE: CONSTANT 65536 == 0 (MPASM uses 16-bit CONSTANTs)
;To make timing calculations easier, the same value is used regardless of clock; a value is chosen that will work regardless of the clock speed
;	CONSTANT WDT_Postscalar = 1024  ;the smallest WDT internval > 20 msec
	CONSTANT WDT_Postscalar = 4096/IIF(MY_OPTIONS & (1<<PSA), 2, 1)  ;if nothing has happened within 1/10 sec, something's broken (must be > 20 msec)
	CONSTANT WDT_MinPostscalar = 32  ;this is the smallest configurable WDT postscalar
	ERRIF WDT_Postscalar%WDT_MinPostscalar, [ERROR] WDT postscalar #v(WDT_Postscalar) out of range >= #v(WDT_MinPostscalar)

;WDTCON (watchdog control) register (page 126):
;set this register once at startup only
	VARIABLE MY_WDTCON = 0  ;default to all bits off, then turn on as needed
;<WDTPS3, WDTPS2, WDTPS1, WDTPS0> form a 4-bit binary value:
MY_WDTCON |= IIF(BIT_#v(WDT_Postscalar/WDT_MinPostscalar) & 1, 1<<WDTPS0, 0)  ;1:64, 1:256, 1:1K, 1:4K, 1:16K or 1:64K postscalar on WDT (page 126)
MY_WDTCON |= IIF(BIT_#v(WDT_Postscalar/WDT_MinPostscalar) & 2, 1<<WDTPS1, 0)  ;1:128, 1:256, 1:2K, 1:4K, 1:32K or 1:64K postscalar on WDT (page 126)
MY_WDTCON |= IIF(BIT_#v(WDT_Postscalar/WDT_MinPostscalar) & 4, 1<<WDTPS2, 0)  ;1:512, 1:1K, 1:2K or 1:4K postscalar on WDT (page 126)
MY_WDTCON |= IIF(BIT_#v(WDT_Postscalar/WDT_MinPostscalar) & 8, 1<<WDTPS3, 0)  ;1:8K, 1:16K, 1:32K or 1:64K postscalar on WDT (page 126)
;	if !(MY_CONFIG & ~_WDT_OFF)  ;WDT not turned on via CONFIG
MY_WDTCON |= IIF(TRUE, 1<<SWDTEN, 0)  ;s/w enable watchdog not used (WDT is enabled via CONFIG), but turn it on anyway
;	endif


;Timer 1 is used for RTC (seconds elapsed) and measuring ZC
;Timer 1 also supplies an internal ZC signal if there is no external ZC signal (for 50 Hz DC applications).
;Timer 1 will wrap after 65536 * .217 usec * 4 ~= 57 msec running at 18.432 MHz or 65536 * .5 usec * 4 ~= 131 msec running at 8 MHz.
;Since both of these wrap-around times are > the 20 msec RTC interval we want, NO Timer 1 prescalar is needed.
;To get exact time intervals of 20 msec with an 18.432 MHz external clock, Timer 1 needs to wrap after ~23041 ticks.
;To get exact time intervals of 20 msec with the 8 MHz int osc, Timer 1 needs to wrap after 10000 ticks.
;Possible prescalar values are 1, 2, 4, or 8.
;Use as small a value as possible for better accuracy, but as large a value as possible to reduce overhead.
;For simpler timing calculations, the same prescalar value is used for all possible clock speeds.
;However, the preset value must be varied to get the correct tick rate.

	CONSTANT Timer1_Prescalar = 2  ;1 tick every 2*.5 usec at 8MHz or 2*.217 usec at 18.432MHz; wraps after 65536 ticks (~65.5 msec at 8MHz or ~28.4 msec at 18.432 MHz); OK FOR 50 Hz
	CONSTANT Timer1_MinPrescalar = 1  ;this is the smallest configurable Timer 1 prescalar
	ERRIF Timer1_Prescalar%Timer1_MinPrescalar, [ERROR] Timer 1 prescalar #v(Timer1_Prescalar) out of range >= #v(Timer1_MinPrescalar)

#define RTC_Timeslot(clock)  (InstrPerSec(clock)/Timer1_Prescalar)  ;#instructions per tick at specified clock speed and refresh rate

;Set Timer 1 to this value to get it to wrap at RTC/DC ZC interval:
;5 instr are lost when updating timer 1 (turn off timer, 16-bit add, turn timer back on)
;Ideally, Timer 1 tick rate would be == desired timeslot (20 msec)
;Rounding errors will prevent this from being exact, but with a 16-bit value it can be very close.
#define Timer1_Preset(clock, refresh)  (65536 - rdiv(RTC_Timeslot(clock), refresh) + 5)

#define RTC_Preset(refresh)  (256 - refresh)  ;adding this value to RTC after wrap will generate a tick (overflow) every 1.0 seconds


;T1CON (timer1 control) register (page 51, 53):
;set this register once at startup only, except that Timer 1 is turned off/on briefly during preset (as recommended by Microchip)
	VARIABLE MY_T1CON = 0  ;default to all bits off, then turn on as needed
MY_T1CON |= IIF(FALSE, 1<<TMR1GE, 0)  ;timer 1 gate-enable off (timer 1 always on)
MY_T1CON |= IIF(FALSE, 1<<T1OSCEN, 0)  ;LP osc disabled (timer 1 source = ext osc)
MY_T1CON |= IIF(DONT_CARE, 1<<NOT_T1SYNC, 0)  ;no sync with ext clock needed
MY_T1CON |= IIF(FALSE, 1<<TMR1CS, 0)  ;use system clock always
MY_T1CON |= IIF(TRUE, 1<<TMR1ON, 0)  ;timer 1 on
;<TICKPS1, TICKPS0> form a 2-bit binary value:
MY_T1CON |= IIF(BIT_#v(Timer1_Prescalar/Timer1_MinPrescalar) & 2, 1<<T1CKPS1, 0)  ;1:4 or 1:8 prescalar on timer 1 (page 53)
MY_T1CON |= IIF(BIT_#v(Timer1_Prescalar/Timer1_MinPrescalar) & 1, 1<<T1CKPS0, 0)  ;1:2 or 1:8 prescalar on timer 1 (page 53)


	VARIABLE TIMER1TICKS2USEC_EXPANDED = FALSE  ;avoid redundant code and data expansions

;Convert #Timer 1 ticks to uSec:
;This is used only for approximate measure of time intervals for debug/console commands.
;#ticks and #usec are assumed to be small (8 bits), but intermediate results can be up to 32 bits.
;The basic ticks-to-usec conversion is: #usec = #T1 ticks * instuctions per second / T1 prescalar / 1 second, but integer arithmetic will cause severe rounding errors for 18MHz.
;For 8MHz and a prescalar of 2, usec == #Timer 1 ticks * 2 * .5 usec == #ticks.
;For 18MHz and a prescalar of 2, usec == #Timer 1 ticks * 2 * .217 == #ticks * .434 ~= #ticks * 4/9.
;This uses 1 expensive DIV function and associated temps.
Timer1Ticks2uSec MACRO t1ticks, usec
	mov8nop t1ticks, DIV32_NUM
	EXPAND_PUSH !TIMER1TICKS2USEC_EXPANDED
	if !TIMER1TICKS2USEC_EXPANDED  ;avoid redundant code and data expansions; expand first time only
TIMER1TICKS2USEC_EXPANDED = TRUE
		LOCAL around_t1ticks, svbank = PREVBANK
		LEAVE_PLACEHOLDER1JUMP neargoto around_t1ticks

		CONSTANT ticks_div_INT = RVpack(_LITERAL(RTC_Timeslot(4 * INT_OSC_FREQ)/ONE_SECOND), DIV32_DEN)  ;4 * usec/#ticks; line too long
		CONSTANT ticks_div_EXT = RVpack(_LITERAL(RTC_Timeslot(4 * EXT_CLK_FREQ)/ONE_SECOND), DIV32_DEN)  ;4 * usec/#ticks; line too long

 jumpdest t1t2usec:,,, OFBANK(0)|,  ;return with bank 0, since that is where next data is typically located
;18MHz/4 cycles per instr/2 prescalar results is a divide by 2.25, which will be very inaccurate with integer arithmetic
;For better accuracy, the code below multiplies by 4 and then divides by 9, which will give a much more accurate answer.
		multi_shiftl16 DIV32_NUM, 2  ;4* numerator
		ifbit USING_INTOSC(TRUE), MOV16RV ticks_div_INT, MOV16RV ticks_div_EXT,  ;choose correct denominator
;32-bit divide is slower, but will avoid the need for a separate copy of 16-bit divide function:
;speed is not critical since this function is only used by debug/console code anyway
		DIV8by16_as32 DIV32_NUM, DIV32_DEN, DIV32_QUO, DIV32_REM
		return_ t1t2usec

 jumpdest around_t1ticks:, svbank|,,,
		FILL_PLACEHOLDER, neargoto around_t1ticks
	endif
	call_ t1t2usec
	mov8nop DIV32_QUO, usec
	EXPAND_POP,
	ENDM


;paranoid/debug checking for timing calculations:
	VARIABLE VALIDATE
	if EXT_CLK_FREQ  ;verify timing attainability and accuracy when running at 18.432MHz (ext clock):
VALIDATE = Timer0_Preset(EXT_CLK_FREQ, AC50HZ)  ;should be -(18,432,000/4/1/(2*50)/256/OVERCLOCK-2) mod 256
		ERRIF (VALIDATE & ~0xff) || (VALIDATE > 256-TSIMAX), [ERROR] 50Hz AC is not attainable with Timer 0 prescalar #v(Timer0_Prescalar) at ext clock speed mhz(EXT_CLK_FREQ)
VALIDATE = Timer0_Preset(EXT_CLK_FREQ, AC60HZ)  ;should be -(18,432,000/4/1/(2*60)/256/OVERCLOCK-2) mod 256
		ERRIF (VALIDATE & ~0xff) || (VALIDATE > 256-TSIMAX), [ERROR] 60Hz AC is not attainable with Timer 0 prescalar #v(Timer0_Prescalar) at ext clock speed mhz(EXT_CLK_FREQ)
VALIDATE = Timer0_Preset(EXT_CLK_FREQ, DC_REFRESH)  ;should be -(18,432,000/4/1/50/256/OVERCLOCK-2) mod 256
		ERRIF (VALIDATE & ~0xff) || (VALIDATE > 256-TSIMAX), [ERROR] 50Hz DC is not attainable with Timer 0 prescalar #v(Timer0_Prescalar) at ext clock speed mhz(EXT_CLK_FREQ)
VALIDATE = Timer1_Preset(EXT_CLK_FREQ, DC_REFRESH)  ;should be -(18,432,000/4/1/50-5) mod 65536
		ERRIF VALIDATE & ~0xffff, [ERROR] 50Hz DC is not attainable with Timer 1 prescalar #v(Timer1_Prescalar) at ext clock speed mhz(EXT_CLK_FREQ)
VALIDATE = RTC_Preset(DC_REFRESH)  ;should be -50 mod 256
		ERRIF VALIDATE & ~0xff, [ERROR] RTC is not attainable with Timer 1 prescalar #v(Timer1_Prescalar) at ext clock speed mhz(EXT_CLK_FREQ)
VALIDATE = SPBRG_Preset(EXT_CLK_FREQ, BAUD_RATE)  ;see BRGH charts (page 97)
		ERRIF !VALIDATE, [ERROR] Baud rate kbaud(BAUD_RATE) not achievable with current BRGH/BRG16 settings and ext clock mhz(EXT_CLK_FREQ)
;the purpose of a faster clock is for accuracy, so these are treated as errors:
VALIDATE = err_rate(SSR_Timeslot(EXT_CLK_FREQ)/AC50HZ, rdiv(SSR_Timeslot(EXT_CLK_FREQ), AC50HZ))  ;the smaller the better
		ERRIF VALIDATE, [ERROR] 50 Hz AC SSR timeslot not exact at ext clock speed mhz(EXT_CLK_FREQ)
VALIDATE = err_rate(SSR_Timeslot(EXT_CLK_FREQ)/AC60HZ, rdiv(SSR_Timeslot(EXT_CLK_FREQ), AC60HZ))  ;the smaller the better
;		ERRIF VALIDATE, "[ERROR] 60 Hz AC SSR timeslot not exact at ext clock speed", EXT_CLK_FREQ, VALIDATE
		WARNIF VALIDATE, [ERROR] 60 Hz AC SSR timeslot not exact at ext clock speed mhz(EXT_CLK_FREQ)
VALIDATE = err_rate(SSR_Timeslot(EXT_CLK_FREQ)/DC_REFRESH, rdiv(SSR_Timeslot(EXT_CLK_FREQ), DC_REFRESH))  ;the smaller the better
		ERRIF VALIDATE, [ERROR] 50 Hz DC SSR timeslot not exact at ext clock speed mhz(EXT_CLK_FREQ)
VALIDATE = err_rate(RTC_Timeslot(EXT_CLK_FREQ)/DC_REFRESH, rdiv(RTC_Timeslot(EXT_CLK_FREQ), DC_REFRESH))  ;the smaller the better
		ERRIF VALIDATE, [ERROR] 50 Hz DC RTC timeslot not exact at ext clock speed mhz(EXT_CLK_FREQ)
VALIDATE = err_rate(InstrPerSec(EXT_CLK_FREQ)/BAUD_RATE, rdiv(InstrPerSec(EXT_CLK_FREQ), BAUD_RATE))  ;the smaller the better
;		ERRIF VALIDATE, "[ERROR] EUSART char rate not exact multiple of ext clock", EXT_CLK_FREQ, VALIDATE
		WARNIF VALIDATE, [ERROR] EUSART char rate kbaud(BAUD_RATE) not exact multiple of ext clock mhz(EXT_CLK_FREQ)
	endif
	if TRUE  ;verify timing attainability and accuracy when running at 8MHz (int osc):
VALIDATE = Timer0_Preset(INT_OSC_FREQ, AC50HZ)  ;should be -(8,000,000/4/1/(2*50)/256/OVERCLOCK-2) mod 256
		WARNIF (VALIDATE & ~0xff) || (VALIDATE > 256-TSIMAX), [WARNING] 50Hz AC is not attainable with Timer 0 prescalar #v(Timer0_Prescalar) at int osc speed mhz(INT_OSC_FREQ)
VALIDATE = Timer0_Preset(INT_OSC_FREQ, AC60HZ)  ;should be -(8,000,000/4/1/(2*60)/256/OVERCLOCK-2) mod 256
		WARNIF (VALIDATE & ~0xff) || (VALIDATE > 256-TSIMAX), [WARNING] 60Hz AC is not attainable with Timer 0 prescalar #v(Timer0_Prescalar) at int osc speed mhz(INT_OSC_FREQ)
VALIDATE = Timer0_Preset(INT_OSC_FREQ, DC_REFRESH)  ;should be -(8,000,000/4/1/50/256/OVERCLOCK-2) mod 256
		WARNIF (VALIDATE & ~0xff) || (VALIDATE > 256-TSIMAX), [WARNING] 50Hz DC is not attainable with Timer 0 prescalar #v(Timer0_Prescalar) at int osc speed mhz(INT_OSC_FREQ)
VALIDATE = Timer1_Preset(INT_OSC_FREQ, DC_REFRESH)  ;should be -(8,000,000/4/1/50-5) mod 65536
		WARNIF VALIDATE & ~0xffff, [WARNING] 50Hz DC is not attainable with Timer 1 prescalar at int osc speed mhz(INT_OSC_FREQ)
VALIDATE = RTC_Preset(DC_REFRESH)  ;should be -50 mod 256
		WARNIF VALIDATE & ~0xff, [WARNING] RTC is not attainable with Timer 1 prescalar at ext clock speed mhz(INT_OSC_FREQ)
VALIDATE = SPBRG_Preset(INT_OSC_FREQ, BAUD_RATE)  ;see BRGH charts (page 97)
		WARNIF !VALIDATE, [WARNING] Target baud rate kbaud(BAUD_RATE) not achievable with current BRGH/BRG16 settings and int osc mhz(INT_OSC_FREQ)
;the slower int osc is a fall-back plan, so these are treated as warnings, but always checked
VALIDATE = err_rate(SSR_Timeslot(INT_OSC_FREQ)/AC50HZ, rdiv(SSR_Timeslot(INT_OSC_FREQ), AC50HZ))  ;the smaller the better
		WARNIF VALIDATE, [WARNING] 50 Hz AC SSR timeslot not exact at int osc speed mhz(INT_OSC_FREQ)
VALIDATE = err_rate(SSR_Timeslot(INT_OSC_FREQ)/AC60HZ, rdiv(SSR_Timeslot(INT_OSC_FREQ), AC60HZ))  ;the smaller the better
		WARNIF VALIDATE, [WARNING] 60 Hz AC SSR timeslot not exact at int osc speed mhz(INT_OSC_FREQ)
VALIDATE = err_rate(SSR_Timeslot(INT_OSC_FREQ)/DC_REFRESH, rdiv(SSR_Timeslot(INT_OSC_FREQ), DC_REFRESH))  ;the smaller the better
		WARNIF VALIDATE, [WARNING] 50 Hz DC SSR timeslot not exact at int osc speed mhz(INT_OSC_FREQ)
VALIDATE = err_rate(RTC_Timeslot(INT_OSC_FREQ)/DC_REFRESH, rdiv(RTC_Timeslot(INT_OSC_FREQ), DC_REFRESH))  ;the smaller the better
		WARNIF VALIDATE, [WARNING] 50 Hz DC RTC timeslot not exact at int osc speed mhz(INT_OSC_FREQ)
VALIDATE = err_rate(InstrPerSec(INT_OSC_FREQ)/BAUD_RATE, rdiv(InstrPerSec(INT_OSC_FREQ), BAUD_RATE))  ;the smaller the better
;DEBUG = 100*InstrPerSec(INT_OSC_FREQ)/BAUD_RATE
;DEBUG = rdiv(100*InstrPerSec(INT_OSC_FREQ), BAUD_RATE)
		WARNIF VALIDATE, [WARNING] Target char rate kbaud(BAUD_RATE) not exact multiple of int osc mhz(INT_OSC_FREQ)
VALIDATE = err_rate(100*InstrPerSec(INT_OSC_FREQ)/MaxBaudRate(INT_OSC_FREQ), rdiv(100*InstrPerSec(INT_OSC_FREQ), MaxBaudRate(INT_OSC_FREQ)))  ;the smaller the better
		WARNIF VALIDATE, [WARNING] Fallback char rate baud(BAUD_RATE) not exact multiple of int osc mhz(INT_OSC_FREQ)
	endif


;set various registers that depend on clock speed:
;Timer 0, Timer 1 and baud rate generator depend on clock speed
;ext clock runs at 18.432 MHz, int osc runs at 8 MHz
;params:
; freq = clock frequency
SetSpeed MACRO freq
	NOEXPAND  ;hide clutter in LST file
	LOCAL FREQ = freq  ;avoid line length problems by expanding and evaluating here (this is usually a big arithmetic expression)
	EXPAND_PUSH TRUE  ;show what this does in LST file
	setbit USING_INTOSC(FREQ == INT_OSC_FREQ)  ;remember whether using ext clock or int osc
#if ZC_TRACKING
;?	mov8 _LITERAL(255), ZC_PREVWIDTH  ;forget previous ZC; none seen yet at this clock rate
	mov8 _LITERAL(0), ZC_PREVWIDTH  ;forget previous ZC; none seen yet at this clock rate
	mov8 _LITERAL(0), ZC_HIGHEST
	mov8 _LITERAL(255), ZC_LOWEST
#endif
;start off assuming no ZC (50 Hz DC); ISR will switch over to 50 or 60 Hz AC if ZC is detected:
	mov8 _LITERAL(Timer0_Preset(FREQ, DC_REFRESH)), T0_PRESET  ;set timer 0 to roll over at desired clock interval (1/256 ZC)
	setbit RCSTA, SPEN, FALSE  ;disable serial port; some notes say this is required when changing the baud rate
;NOTE: need to set SPBRG a while before enabling interrupts so SPBRG can stabilize
;http://www.pic101.com/mcgahee/picuart.zip also recommends a startup delay so MAX232 charge pump stabilizes,
;but it has probably already stabilized by the time we get to this code
	if MaxBaudRate(FREQ) < BAUD_RATE * 95/100  ;desired baud rate is not attainable (accurately), down-grade it
new_baud_rate set BAUD_RATE_STEPDOWN#v(BAUD_RATE)  ;select next lower rate (uses 32-bit arithmetic)
		WARNIF (new_baud_rate < BAUD_RATE * 95/100) && (FREQ == EXT_CLK_FREQ), [INFO] Baud rate kbaud(BAUD_RATE) reduced to kbaud(new_baud_rate) because ext clock mhz(EXT_CLK_FREQ) is not fast enough
		WARNIF (new_baud_rate < BAUD_RATE * 95/100) && (FREQ == INT_OSC_FREQ), [INFO] Baud rate kbaud(BAUD_RATE) reduced to kbaud(new_baud_rate) because int osc mhz(INT_OSC_FREQ) is not fast enough
		mov16 _LITERAL(SPBRG_Preset(FREQ, new_baud_rate)), SPBRG
	else  ;use rate with no change
		mov16 _LITERAL(SPBRG_Preset(FREQ, BAUD_RATE)), SPBRG  ;use baud rate as-is
	endif
; turn serial port off/on after baud rate change:
	mov8 _LITERAL(MY_TXSTA), TXSTA  ;set TXEN (enable xmiter) and reset SYNC (set async mode) after setting SPBRG
	mov8 _LITERAL(MY_RCSTA), RCSTA  ;set SPEN (enable UART and set TX pin for output + RX pin for input) and set CREN (continuous rcv) to enable rcv
;	setbit TXSTA, TXEN, TRUE  ;enable TX after RX is set
;http://www.pic101.com/mcgahee/picuart.zip says to do this 3x:
	REPEAT 3, loadw_ RCREG  ;flush input buffer
	EXPAND_POP,
	ENDM


;******************************************************************************
;error handling section:
;THIS NEEDS MORE WORK
;For now, fatal errors just flash an error code on some channels in an infinite loop.
;Eventually, it should probably log the errors into EEPROM and then gracefully restart or continue;
;then the debug/console commands or Vixen plug-in could be used to review the error log later.
;******************************************************************************

	VARIABLE SHOWBYTE_EXPANDED = FALSE  ;avoid redundant code and data expansions

;show a byte value on the PK1 LEDs for a while:
;params:
; value = value to be displayed
; delay = how long to display the value: 1 => 1 sec, 10 => 1/10 sec
SHOWBYTE MACRO value, delay
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH !SHOWBYTE_EXPANDED
	if !SHOWBYTE_EXPANDED  ;avoid redundant code and data expansions; expand first time only
SHOWBYTE_EXPANDED = TRUE
		LOCAL around_showbyte, svbank = PREVBANK
		LEAVE_PLACEHOLDER1JUMP neargoto around_showbyte

;since this code is only used for fatal error handling, it's okay to overwrite some other data:
		CBLOCK NONBANKED_RAM_START-2
		ByteValue  ;value to be displayed
		ShowDelay  ;delay loop control
		ENDC

		CONSTANT ONESEC_INTOSC = LED1SEC(INT_OSC_FREQ)+1  ;line too long
		CONSTANT ONESEC_EXTCLK = LED1SEC(EXT_CLK_FREQ)+1  ;line too long
;		UGLY_PASS12FIX -1  ;KLUDGE: assembler needs help to sync up addresses
;show byte for 1 sec:
 jumpdest showbyte_1sec:,,, OFBANK(0)|,  ;return with bank 0, since that is where next data is typically located
		storew_ ByteValue  ;display value was passed via W
;choose the right value that will display for 1 sec:
		ifbit NOT_4MHz, loadw_ _LITERAL(ONESEC_INTOSC), loadw_ _LITERAL(LED1SEC_4MHz + 1), BOTHOK_HINT  ;int osc speed not increased yet (didn't make it thru init?)
		if EXT_CLK_FREQ && (EXT_CLK_FREQ != INT_OSC_FREQ)  ;PIC is configured to use ext osc; check if it really is or not
			ifbit USING_INTOSC(FALSE), loadw_ _LITERAL(ONESEC_EXTCLK),, NOJUMP_HINT  ;ext oscillator failed/absent, running from int osc
		endif
		storew_ ShowDelay
		LEAVE_PLACEHOLDER1 neargoto show_loop

;		UGLY_PASS12FIX -1  ;KLUDGE: assembler needs help to sync up addresses
;show byte for 1/10th sec:
 jumpdest showbyte_10thsec:,,, OutBankOf(showbyte_1sec)|,  ;common return must share same out bank
		storew_ ByteValue  ;display value was passed via W
;choose the right value that will display for 1/10 sec:
		ifbit NOT_4MHz, loadw_ _LITERAL(ONESEC_INTOSC/10 + 1), loadw_ _LITERAL(LED1SEC_4MHz/10 + 1), BOTHOK_HINT  ;int osc speed not increased (didn't make it thru init)
		if EXT_CLK_FREQ && (EXT_CLK_FREQ != INT_OSC_FREQ)  ;PIC is configured to use ext osc; check if it really is or not
			ifbit USING_INTOSC(FALSE), loadw_ _LITERAL(ONESEC_EXTCLK/10 + 1),, NOJUMP_HINT  ;ext oscillator failed/absent, running from int osc
		endif
		storew_ ShowDelay
		LEAVE_PLACEHOLDER1 neargoto show_loop

#if PK1DEV+PK1SIM  ;using PK1 LEDs
		if !PK1LED_EXPANDED  ;show value using chipiplexing logic with charlieplexed PK1 LEDs

			BDCL PK1LEDS  ;8-bit value to be displayed

;turn all PK1 LEDs off:
 jumpdest pk1_alloff:, ByteValue|,,,
			FBUnpack _LITERAL(0), _LITERAL(0)
 jumpdest pk1_showbit:,,, OutBankOf(pk1_alloff)|, FALLTHRU  ;common return must share same out bank
			Chipiplex 1
			return_ pk1_showbit

;LED0 0x01 (RA4->RA5/RA0)  RA: 0/z 1 z z z z/0 (1 z z 0 x x x x after bit swap, RA5 = row 7) = ch# 3*7+1
 jumpdest pk1_on_#v(0):, ByteValue|,, OutBankOf(pk1_showbit)|,  ;common return must share same out bank
			FBUnpack _LITERAL(RCbit(0)), _LITERAL(RCbit(3))
			neargoto pk1_showbit  ;turn on LED and return

;LED1 0x02 (RA5/RA0->RA4)  RA: 1/z 0 z z z z/1  (0 z z 1 x x x x after bit swap, RA4 = row 0) = ch# 0*7+3
 jumpdest pk1_on_#v(1):, ByteValue|,, OutBankOf(pk1_showbit)|,  ;common return must share same out bank
			FBUnpack _LITERAL(RCbit(3)), _LITERAL(RCbit(0))
			neargoto pk1_showbit  ;turn on LED and return

;LED2 0x04 (RA4->RA2)  RA: z 1 z 0 z z  (1 0 z z x x x x after bit swap, RA2 = row 5) = ch# 1*7+1
 jumpdest pk1_on_#v(2):, ByteValue|,, OutBankOf(pk1_showbit)|,  ;common return must share same out bank
			FBUnpack _LITERAL(RCbit(0)), _LITERAL(RCbit(1))
			neargoto pk1_showbit  ;turn on LED and return

;LED3 0x08 (RA2->RA4)  RA: z 0 z 1 z z  (0 1 z z x x x x) = ch# 0*7+1
 jumpdest pk1_on_#v(3):, ByteValue|,, OutBankOf(pk1_showbit)|,  ;common return must share same out bank
			FBUnpack _LITERAL(RCbit(1)), _LITERAL(RCbit(0))
			neargoto pk1_showbit  ;turn on LED and return

;LED4 0x10 (RA5/RA0->RA2)  RA: 1/z z z 0 z z/1  (z 0 z 1 x x x x) = ch# 1*7+3
 jumpdest pk1_on_#v(4):, ByteValue|,, OutBankOf(pk1_showbit)|,  ;common return must share same out bank
			FBUnpack _LITERAL(RCbit(3)), _LITERAL(RCbit(1))
			neargoto pk1_showbit  ;turn on LED and return

;LED5 0x20 (RA2->RA5/RA0)  RA: 0/z z z 1 z z/0  (z 1 z 0 x x x x) = ch# 3*7+2
 jumpdest pk1_on_#v(5):, ByteValue|,, OutBankOf(pk1_showbit)|,  ;common return must share same out bank
			FBUnpack _LITERAL(RCbit(1)), _LITERAL(RCbit(3))
			neargoto pk1_showbit  ;turn on LED and return

;LED6 0x40 (RA2->RA1)  RA: z z z 1 0 z  (z 1 0 z x x x x, RA1 = row 6) = ch# 2*7+2
 jumpdest pk1_on_#v(6):, ByteValue|,, OutBankOf(pk1_showbit)|,  ;common return must share same out bank
			FBUnpack _LITERAL(RCbit(1)), _LITERAL(RCbit(2))
			neargoto pk1_showbit  ;turn on LED and return

;LED7 0x80 (RA1->RA2)  RA: z z z 0 1 z  (z 0 1 z x x x x) = ch# 1*7+2
 jumpdest pk1_on_#v(7):, ByteValue|,, OutBankOf(pk1_showbit)|,  ;common return must share same out bank
			FBUnpack _LITERAL(RCbit(2)), _LITERAL(RCbit(1))
			neargoto pk1_showbit  ;turn on LED and return
		endif

;		UGLY_PASS12FIX -1  ;KLUDGE: assembler needs help to sync up addresses
;delay so that LEDs are visible:
 jumpdest show_delay:, ByteValue|,, OutBankOf(pk1_alloff)|,  ;common return must share same out bank
		clrwdt  ;prevent fido from interrupting diagnostic info
		BUSYWAIT LEDTIME
		neargoto pk1_alloff  ;all-off for consistent brightness

;display byte value using charlieplexing for specified length of time:
;NOTE: this uses busy waits, so it ties up the processor
 jumpdest show_loop:, ShowDelay|,, OutBankOf(showbyte_1sec)|,  ;common return must share same out bank
		FILL_PLACEHOLDER, neargoto show_loop
		FILL_PLACEHOLDER, neargoto show_loop
		VARIABLE bit = 0
		while bit < 8  ;0, 1, 2, 3, 4, 5, 6, 7
			ifbit ByteValue, bit, TRUE, nearcall pk1_on_#v(bit),,
			nearcall show_delay
bit += 1
		endw  ;bit < 8
		dec8_ifnz ShowDelay, neargoto show_loop,,
		return_ show_loop

#else  ;PK1DEV
 jumpdest show_delay:,,, OFBANK(0)|,  ;return with bank 0, since that is where data typically is
		clrwdt  ;prevent fido from interrupting diagnostic info
		BUSYWAIT LEDTIME
		return_

;display byte value using chipiplexed channels for specified length of time:
 jumpdest show_loop:, ShowDelay|,, OutBankOf(showbyte_1sec)|,  ;common return must share same out bank
		and8 ByteValue, _LITERAL(LOW(~RCbit(0))), ByteValue,  ;drop top bit and use for row address
		FBUnpack ByteValue, _LITERAL(RCbit(0))
		Chipiplex 1
		REPEAT 8, nearcall show_delay  ;delay time calculation assumed 8 bits to display
		dec8_ifnz ShowDelay,, return_ show_loop,
		neargoto show_loop
#endif  ;PK1DEV

 jumpdest around_showbyte:, svbank|,,,
		FILL_PLACEHOLDER, neargoto around_showbyte
	endif
	loadw_ value  ;pass display value via W
	if (delay) > 1
		call_ showbyte_10thsec
	else
		call_ showbyte_1sec
#if PK1SIM  ;values shown for 1 sec are more important, so make sure the entire value is displayed
;When RA5 is config as clock input, half the PK1 LEDs (LEDs 0, 1, 4, and 5) can't be turned on.
;To compensate, these will be "reflected" into the visible LEDs (2, 3, 6, and 7) using the code below.
		if value != _LITERAL(0)
			add8 ByteValue, ByteValue, ByteValue,  ;2*previous value
			add8 ByteValue, ByteValue, WREG,  ;4*previous value
			call_ showbyte_1sec
		endif
#endif  ;PK1SIM
	endif
	EXPAND_POP,
	ENDM


	VARIABLE FATAL_EXPANDED = FALSE  ;avoid redundant code and data expansions
#define FATAL(errcode, errval, errmsg)  FATAL_ ((errcode)<<8 | (errval));, errmsg

;something bad happened; show error code and stop:
;err msg is just comment text for now
;params:
; errcodeval = error code and value to display (packed into a single param)
FATAL_ MACRO errcodeval  ;, errmsg
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH !FATAL_EXPANDED
	if !FATAL_EXPANDED  ;avoid redundant code and data expansions; expand first time only
FATAL_EXPANDED = TRUE
		LOCAL around_fatal, svbank = PREVBANK
		LEAVE_PLACEHOLDER1JUMP neargoto around_fatal

		BDCL FatalValue  ;error code
		BDCL OtherValue  ;additional info

;show fatal error condition forever:
;this turns off interrupts and uses busywaits
;		UGLY_PASS12FIX -1  ;KLUDGE: assembler needs help to sync up addresses
 jumpdest fatal:,,,,
		storew_ FatalValue  ;error code was passed via W
		intenable_ FALSE  ;since there is something seriously wrong we can't rely on timer interrupts
;show a little animated pattern to attract attention:
		SHOWBYTE _LITERAL(1), 10
		SHOWBYTE _LITERAL(2), 10
		SHOWBYTE _LITERAL(4), 10
		SHOWBYTE _LITERAL(8), 10
		SHOWBYTE _LITERAL(0x80), 10
		SHOWBYTE _LITERAL(0x40), 10
		SHOWBYTE _LITERAL(0x20), 10
		SHOWBYTE _LITERAL(0x10), 10
		SHOWBYTE _LITERAL(1), 10
		SHOWBYTE _LITERAL(2), 10
		SHOWBYTE _LITERAL(4), 10
		SHOWBYTE _LITERAL(8), 10
		SHOWBYTE _LITERAL(0x80), 10
		SHOWBYTE _LITERAL(0x40), 10
		SHOWBYTE _LITERAL(0x20), 10
		SHOWBYTE _LITERAL(0x10), 10
		SHOWBYTE _LITERAL(0), 10
;		UGLY_PASS12FIX -1  ;KLUDGE: assembler needs help to sync up addresses
 jumpdest fatal_loop:,,,, FALLTHRU
		SHOWBYTE FatalValue, 1  ;show error code 2x
		SHOWBYTE _LITERAL(0), 1
		SHOWBYTE FatalValue, 1
		SHOWBYTE _LITERAL(0), 1
		SHOWBYTE OtherValue, 1  ;then show other value 1x (a different number of times so we can tell them apart)
		SHOWBYTE _LITERAL(0), 1
		neargoto fatal_loop

 jumpdest around_fatal:, svbank|,,,
		FILL_PLACEHOLDER, neargoto around_fatal
	endif
	mov8nop LOW(errcodeval), OtherValue
	loadw_ _LITERAL(HIGH(UNLIT(errcodeval)))  ;pass error code via W
;	da errmsg
#if IDE_DEBUG
	while8 _LITERAL(1), NE, _LITERAL(0), call_ fatal,  ;allow resume without losing return address (IDE debug only)
#else  ;IDE_DEBUG
	goto_ fatal
#endif  ;IDE_DEBUG
	EXPAND_POP,
	ENDM


;******************************************************************************
;frame buffer section:
;define frame buffers (to reduce flicker during frame updates from Vixen)
;******************************************************************************

;To avoid flicker during channel updates from Vixen, double-buffering is used (active framebuffer + incoming framebuffer).
;The 2 upper 80-byte regions in the PIC16F688 are used as frame buffers; the lower region is used as working memory during processing.
;To reduce processing overhead, frame buffers need to be stored in a near-ready display format for efficiently generating SSR control pulses.
;For AC phase-angle dimming, full-on channels are turned on before dim channels, so SSR control pulses must be sent out in order from brightest to dimmest.
;For DC PWM dimming, channel order doesn't matter so much, as long as the brighter channels are on longer than the dimmer channels; BAM is probably overkill; 1 simple on/off transition per timeslot appears to be sufficient.
;The following frame buffer format appears to be a good compromise between storage space and near-usable format:
;   brightness value = 1 byte; value 0 ... 255 for channel brightness
;   row index = 1 byte; bitmap of which rows use this dimmer value; each bit represents the common cathode when chipiplexing
;   rows = n bytes (1 byte for each on bit in row index); bitmap of which columns in each row use this value; represents anodes when chipiplexing
;   (repeat above, in order of decreasing brightness)
;   0 = end of list; a brightness of 0 is useless since it's full off, so this is used as an end-of-list marker
;This allows the PIC to send out a byte at a time (channels are already grouped into rows).
;The number of channels that can be on, or that can be at a particular brightness, is variable depending on how they are spread across brightnesses and rows.
;Due to the framebuffer size limit of 80 bytes, not all possible combinations are representable.
;For example, 26 distinct brightnesses with one channel each would consume 26 * (brightness, row index, row) = 78 bytes, not leaving any space for the remaining 30 channels.
;Another example, 56 channels all at the same brightness would consume 1 * (brightness, row index, 8 rows) = 10 bytes, with plenty of empty space left over in the framebuffer.
;The more channels that share a row or share the same brightness, the more concise the repesentation will be.
;My sequences appear to use 12 - 20 distinct brightness values across the entire sequence, so this falls within the representable range of values that will fit in an 80-byte frame buffer.
;This firmware will simply truncate the frame to prevent a frame buffer overflow, which could result in important channels remaining off while lesser important channels are on correctly.
;Hence, the Vixen plug-in will coalesce distinct but "similar" brightness values as necessary to collapse the display list into the available framebuffer size.
;This results in a "lossy" protocol; since +/- 2 brightness levels are typically not noticeable with fast-changing displays, this seems to provide acceptable performance.
;The chipiplexing Renard Vixen plug-in provides 2 options for coalescing: 56 bytes or 80 bytes.
;The 80-byte limit ensures that the frame buffer will not overflow.
;The 56-byte limit ensures that the data stream size is not larger than the number of channels (for cases where the serial data stream size is a limiting factor).
;The Vixen chipiplexing plug-in will send data in the following format:
;  config byte, brightness, row index, row, row, ..., brightness, row index, row, row, ..., 0  (max of 81 bytes or 56 bytes)
;This would allow more channels to be used than the available serial data stream bandwidth can support (ie, data stream compression).
;However, this requires extra processing to reformat the list into the display-ready format shown earlier.
;This processing time is only available during the last 20% "dead zone"; at other times the PIC is potentially too busy generating SSR control pulses.

;TODO in future:
;A more concise palette-based data stream from Vixen is also possible.  This would be especially useful for RGB LED applications.
;A simple monochrome palette-based frame format would be as follows:
;  brightness values = 16 bytes; this forms a 4-bit palette of brightness values
;  56 nibbles = a 4-bit palette entry# to use for each channel
;This would provide 56 channels in 1+16+28 = 45 bytes; not that big of a savings *unless* the palette can be shared between PICs.
;On downstream PICs, a shared palette would allow 56 channels to be represented in 1+28 = 29 bytes (almost 2:1 compression), almost doubling the #channels that Renard can support.
;This could be further extended to support RGB channels, which each palette entry containing 3 bytes rather than 1.
;First packet = 1 config byte + 16 RGB palette entries + 30 channel nibbles = 47 bytes, subsequent packets = 1 config + 30 nibbles = 16 bytes.
;This would provide greater than 5x compression on the Renard data stream, making 1K RGB channels (3K monochrome channels) supportable at 115K baud and 50 msec refresh.
;Each PIC could support 30 RGB channels if ZC and ext clock are not used (2 more SSR control lines = 10, which would be 10 x 3 RGB channels with chipiplexing).


;additional data related to, but not stored directly within the frame buffers:
	BDCL FBPtr  ;pointer to next byte in currently active frame buffer; FSR may also point into frame buffer
	BDCL CfgByte  ;run-time options from Vixen plug-in; NOTE: assumed to be the same for all frame buffers

;config bit definitions:
	CONSTANT CFGBITS_PULSE_WIDTH = 0x3  ;bit mask: 0 => 1x pulse (32.6 usec), 1 => 1.5x (24 usec), 2 => 2x (16 usec), 3 => 2.5x (12 usec)
	CONSTANT CFGBIT_ECHO_PALETTE = 2  ;0x04  ;re-send palette to next PIC (gives higher Renard datastream compression)
	CONSTANT CFGBIT_RGB_PALETTE = 3  ;0x08  ;palette entries are RGB (3 bytes) vs. mono (1 byte)
	CONSTANT CFGBIT_CHECKSUM = 4  ;0x10  ;checksum is present; validate it before making frame buffer active (guards against comm errors)
	CONSTANT CFGBIT_ACSSR_DOUBLE = 5  ;0x20  ;channels doubled; alternates on each AC half cycle (requires polarized ZC)
	CONSTANT CFGBIT_PSEUDO_PWM = 6  ;0x40  ;pseudo-PWM; after turning on, channels will keep turning on until end of half-cycle
	CONSTANT CFGBIT_MACRO_MODE = 7  ;0x80 macro mode; overrides meaning of other config bits
	CONSTANT CFGMACRO_CONSOLE = (1<<CFGBIT_MACRO_MODE) + 0  ;macro function code for a debug/console command
	CONSTANT CFGMACRO_whatever = (1<<CFGBIT_MACRO_MODE) + 1  ;macro function code for (add more as needed)


;Framebuffers are placed in upper memory banks because they are relatively large, and only need indirect addressing.
;This allows the lower memory bank to be used exclusively as working storage.
;IRP must be used to supply the upper bit of addresses into the frame buffers, since FrameBuffer1 is above 0x100.
	VARIABLE fb = 1
	while fb < 3  ;1, 2
		CBLOCK UPPER_BANK_START + (fb-2)*BANK_SIZE  ;Bank 1 general RAM start = 0xA0, Bank 2 general RAM start = 0x120
		FB#v(fb)_START:UPPER_BANK_SIZE-1  ;80-1 bytes on PIC16F688
		FB#v(fb)_END  ;1 pad byte at end for end-of-list marker
		ENDC

		ERRIF FB#v(fb)_END >= MAX_RAM, [ERROR] Frame buffer #v(fb) not supported on this device: end #v(FB#v(fb)_END) > max #v(MAX_RAM)  ;paranoid
		ERRIF2 BANKOF(FB#v(fb)_START) != BANKOF(FB#v(fb)_END), [ERROR] Frame buffer #v(fb) spans banks: starts #v(FB#v(fb)_START), ends #v(FB#v(fb)_END)  ;paranoid
		ERRIF2 (BANKOF(FB#v(fb)_START) != fb) || ((FB#v(fb)_START & 0x80) != (fb&1)<<7), [ERROR] Frame buffer #v(fb) address check broken: starts #v(FB#v(fb)_START), ends #v(FB#v(fb)_END)  ;paranoid
fb += 1
	endw  ;fb < 3


;frame buffer addressing:
;Since upper banks are used, frame buffer address is more than 8 bits.
;The 9th bit can be generated based on the lower 8 bits, since those do not overlap, or top bit of lower byte can be used to check which frame buffer is active.
;Since low byte of bank 1 + bank 2 addresses do not overlap, this just checks MSB of FBPtr or FSR.
;Top bit of low byte of buffer 1 address = 1, buffer2 address = 0:
#define ISFB(reg, n)  reg, 7, (n)&1  ;run-time version (ifbit params)


	BITDCL FBSwapRequest  ;swap active/inactive frame buffers at start of next cycle
	ERRIF TRUE == 1, [ERROR] FBSELECT broken  ;verify assumptions

;prepare for indirect addressing into frame buffer:
;this sets FSR + IRP to point to the selected offset within frame buffer
;params:
;  which: 1 or 2 => select specified frame buffer, TRUE => current buffer, FALSE => inactive buffer, other => frame buffer selected by register
;  endofs = current offset from END of frame buffer
FBSELECT MACRO which, endofs
	LOCAL noswap
	if ((which) == TRUE) || ((which) == FALSE)  ;active/inactive frame buffer wanted
		if ISLIT(endofs)
			VARIABLE ENDOFS = UNLIT(endofs)  ;KLUDGE: line below is too long
			ifbit ISFB(FBPtr, IIF(which, 1, 2)), loadw_ _LITERAL(LOW(FB1_END - ENDOFS)), loadw_ _LITERAL(LOW(FB2_END - ENDOFS)), BOTHOK_HINT
			storew_ FSR
		else
			sub8 _LITERAL(LOW(FB2_END)), endofs, WREG,  ;allow endofs to be passed in W
			ifbit ISFB(FBPtr, IIF(which, 1, 2)), xorlw LOW(FB1_END ^ FB2_END),, NOJUMP_HINT
			storew_ FSR
		endif
		ifbit ISFB(FSR, 1), BANKISEL FB1_START, BANKISEL FB2_START, BOTHOK_HINT
	else
		if ((which) == 1) || ((which) == 2)  ;specified frame buffer wanted
			sub8 _LITERAL(LOW(FB#v(which)_END)), endofs, FSR,
			BANKISEL FB#v(which)_START
		else  ;select frame buffer position from register
			sub8 which, endofs, FSR,
			ifbit ISFB(FSR, 1), BANKISEL FB1_START, BANKISEL FB2_START, BOTHOK_HINT
		endif
	endif
	ENDM


;generate a row/column bitmask (msb-to-lsb order):
#define RCbit(n)  (1<<(7-(n)))


;convenience wrappers:
ChannelToRowBit MACRO channel, rowbit
	ChannelToRowCol channel, rowbit, TRUE
	ENDM

ChannelToColumnBit MACRO channel, columnbit
	ChannelToRowCol channel, columnbit, FALSE
	ENDM


	VARIABLE CHANNEL2ROWCOL_EXPANDED = FALSE  ;avoid redundant code and data expansions

;channel# -> (row, column) lookups:
;this is faster than looping through rows and cols, but takes up more space
;params:
;  optional_channel = channel# (optional param); must end with "&" if passed; uses previous channel if absent
;  rcbit = register to hold the row or column bitmask lookup results
;  want_row = flag telling whether to look up the row or column bitmask for the channel
ChannelToRowCol MACRO optional_channel, rcbit, want_row
;	EXPAND_PUSH FALSE  ;hide messy code expansion
	if !CHANNEL2ROWCOL_EXPANDED  ;avoid redundant code expansions; expand first time only
CHANNEL2ROWCOL_EXPANDED = TRUE

		LOCAL around_ch2rcbit, svbank = PREVBANK
		LEAVE_PLACEHOLDER1JUMP neargoto around_ch2rcbit

;lookup table:
;Index into the table is the channel# (0..55).
;Each entry is the row and column bitmasks for that channel (packed into 14 bits).
;Since 14 bits is not enough to hold an 8-bit row bitmask + 8-bit column bitmask, 2 bits are omitted and then reconstructed:
;  upper 7 bits = lower 7 bits of row bitmask; a value of 0 is used to represent 0x80 (since 1 bit must be on)
;  lower 7 bits = upper 7 bits of column bitmaskl a value of 0 is used to represent 0x01 (since 1 bit must be on)
ChannelToRCbits:
		REPEAT 56, da ((RCbit(repeater/7) & 0x7f)<<7) | (RCbit((repeater%7) + IIF(repeater%7 >= repeater/7, 1, 0))>>1)  ;top byte = row bit <<1 (top bit dropped), bottom byte = column bit >>1 (bottom bit dropped)

;load packed row bitmask + column bitmask into (EEDAT, EEDATH):
;upon entry, W contains channel# (1..56)
 jumpdest ch2rcbit:,,, EEDAT|,
		PROGREAD16 _LITERAL(ChannelToRCbits - 1), WREG, EEDAT  ;channel# is 1-based, but table is 0-based
		shiftl8to EEDAT, WREG  ;upper 7 bits of column bit value; move bottom bit of row into STATUS.C
		return_ ch2rcbit

 jumpdest around_ch2rcbit:, svbank|,,,
		FILL_PLACEHOLDER, neargoto around_ch2rcbit
	endif
	if (optional_channel -1) != -1  ;look up row + column bitmasks for this channel# (optional parameter was passed)
		loadw_ optional_channel -1
		call_ ch2rcbit
	endif
	if want_row
		shiftl8to_NC EEDATH, WREG  ;lower 7 bits of row bitmask value
		if8 WREG, EQ, _LITERAL(0), movlw 0x80,, NOJUMP_HINT  ;restore top bit if it was missing (0 => 0x80)
	else
		if (optional_channel -1) == -1  ;didn't look up row + column bitmasks
			shiftl8to EEDAT, WREG  ;upper 7 bits of column bit value; move bottom bit of row into STATUS.C
		endif
		if8 WREG, EQ, _LITERAL(0), movlw 1,, NOJUMP_HINT  ;restore bottom bit if it was missing
	endif
	storew_ rcbit
	ENDM


;******************************************************************************
;chipiplexing section:
;define SSR control buffers and helper functions/macros
;******************************************************************************

;for the PIC16F688, chipiplexed I/O pins will be mapped to Vixen channel#s as follows (- = Row/Cathode, + = Column/Anode):
;(this mapping is implemented in the FBUnpack macro)
;-\+   RC3  RC2  RC1  RC0  RA4  RA2  RA1  RA0/5
;RC3    --    0    1    2    3    4    5    6
;RC2     7   --    8    9   10   11   12   13
;RC1    14   15   --   16   17   18   19   20
;RC0    21   22   23   --   24   25   26   27
;RA4    28   29   30   31   --   32   33   34
;RA2    35   36   37   38   39   --   40   41
;RA1    42   43   44   45   46   47   --   48
;RA0/5  49   50   51   52   53   54   55   --

;PK1 LEDS map to the following channel#s (helpful for dev/testing):
;ch 53 = LED0 1 (RA4->RA5); RA4->RA0 for PK1 modified
;ch 34 = LED1 2 (RA5->RA4); RA0->RA4 for PK1 modified
;ch 39 = LED2 4 (RA4->RA2)
;ch 32 = LED3 8 (RA2->RA4)
;ch 41 = LED4 0x10 (RA5->RA2); RA0->RA2 for PK1 modified
;ch 54 = LED5 0x20 (RA2->RA5); RA2->RA0 for PK1 modified
;ch 47 = LED6 0x40 (RA2->RA1)
;ch 40 = LED7 0x80 (RA1->RA2)
;------
;LED0+5+(9) are on row RA5/RA0
;LED1+3+(11) are on row RA4
;LED2+4+7 are on row RA2
;LED6+(8)+(10) is on row RA1


;The top 20% of dimming range is effectively on full, and the bottom 20% is effectively off.
;For 2*60Hz, this leaves a ~ 1667 usec window to turn SSRs full-on at the start of the cycle,
;then 5000 usec to set discernable dimming levels, and then ~ 1667 usec of dead time at the end of the half cycle.
;For 2*50Hz, this leaves a 2000 usec window to turn SSRs full-on at the start of the cycle,
;then 6000 usec to set dimming levels, and then 2000 usec of dead time at the end of the half cycle.
;Dimming values from Vixen are the EARLIEST that an SSR/LED will be turned on.
;These are not guaranteed if a lot of rows come on at the same time, because the control pulses will take longer than the dimmed timeslot.
;Hence, chipiplexed dimming as implemented here is "lossy"; however, a +/-2 difference is easily achievable and not usually noticeable.

;The standard Renard on-pulse is 32 usec at 5 mA.
;According to the MOC3023 datasheet, if the pulse current is raised to 15 mA, the required turn-on pulse width is only ~ 5 usec; pad up to 6 usec for safety.
;With charlieplexing, only one output is on at a time, so this limits the number of channels that can be turned on during each dimming time slice.
;With chipiplexing, one row of outputs can be on at a time, which relaxes this limit (1 channel per pin can be on at a time).
;At 2*60Hz, the full-on window can use ~277 channels, and each mid-range Renard dimming timeslots can use ~5*7=35 channels.
;At 2*50Hz, the full-on window can use ~333 channels, and each mid-range Renard dimming timeslots can use ~6*7=42 channels.
;If we relax the dimming accuracy to +/- 2 levels, then the window is 5x as large (-2, -1, 0, +1, +2).
;At 2*60Hz, each mid-range lossy dimming timeslot can then use ~27*7=189 channels, or ~32*7=224 at 2*50Hz.
;If the channels are physically spread out, variations in dimming level will not be noticeable, as long as the channel itself gets brighter or dimmer correctly.
;For a pixel array, variations will be noticeable since the pixels are closer together, but *-Triks panels are typically only 16 pixels high, and that many pixels can easily be at the "same" brightness level.


;define I/O pin change buffers for each port that controls SSRs:
;all SSR pin changes are held in memory until start of time slice, then ports are updated as bytes (to save time)
;2 bytes are defined for each port that has SSR pins: Outmask#v contains TRIS bits, and Outbuf#v contains PORT bits
IOPIN_BUFFERS MACRO
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE  ;don't show messy macro expansion
	LOCAL numpins = 0, numports = 0
port = FIRST_PORT
	while port <= LAST_PORT
		if SSRPINS#v(port)  ;this port is used for SSR control
			NBDCL Outmask#v(port)  ;TRIS reg values for each port and subslot; non-banked to reduce bank selects when setting TRIS
			BDCL Outbuf#v(port)  ;PORT reg values for each port and subslot; banked since PORT registers are in bank 0 anyway
numports += 1
		endif
numpins += NUMBITS_#v(SSRPINS#v(port))
port += 1
	endw  ;port <= LAST_PORT
	EXPAND_POP,
	CONSTANT SSR_PIN_COUNT = numpins  ;total #SSR control pins
	CONSTANT FB_ROWSIZE = divup(numpins, 8)  ;#bytes needed to hold SSR pin bitmaps
	CONSTANT MONO_CHANNELS = numpins * (numpins - 1)  ;#monochrome channels possible (1 pair of pins per channel)
	CONSTANT RGB_CHANNELS = numpins * divup(numpins - 1, 3)  ;#RGB channels possible (3 pins and 1 common per channel)
	BDCL StickyRowIndex:FB_ROWSIZE  ;row index for next set of changes to sticky channels
	BDCL StickyRows:(SSR_PIN_COUNT*FB_ROWSIZE)  ;cumulative SSR pin values for each row; used to maintain pseudo-PWM
	ENDM

	IOPIN_BUFFERS
	WARNIF TRUE, [INFO] Using #v(SSR_PIN_COUNT) SSR control pins: #v(SSR_PIN_COUNT)x#v(SSR_PIN_COUNT-1) = #v(MONO_CHANNELS) channels (or #v(2*MONO_CHANNELS) if SSR-doubled)
	WARNIF (SSR_PIN_COUNT - 1) % 3, [CAUTION] Partial last RGB channels (#columns not a multiple of 3): only #v(RGB_CHANNELS) RGB channels available


;unpack a display list entry from frame buffer:
;NOTE: for performance tuning, this is done ahead of time using data holding registers (defined above).
;Doing this ahead of time (prior to actual I/O time) reduces actual I/O time to 10 instr: 
; 2 * bsf tris/port + 4 * (movf buf + movwf tris/port), resulting in very low dimming timing jitter.
;Otherwise, doing this at I/O time increases I/O time to 20 instr: 
; 2 * (swap/movf dlentry + and 0xf + bsf port + movwf port + comp + and + bsf tris + movwf tris) + 1 * *btfsc + xor 0x18) for port a fixup.
;20 instr at 18.432 MHz ~= 4.3 usec, which is ~ 13% of US 1x timeslot (32.6 usec) or 27% of 2x timeslot (16 usec); this will introduce a slight brightness delay and more timing jitter for AC dimming.
;params:
;  columns = bitmap of columns to be unpacked
;  rowbit = bitmask of current row being unpacked
#if !SSRPINS#v(0xA) || !SSRPINS#v(0xC)  ;SSR pins *not* split across Ports A + C (ie, 16F689/690); this is more efficient
FBUnpack MACRO columns, rowbit
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE  ;hide messy code expansion
	LOCAL ofs = 0, CommonCathode = 0, constcol = UNLIT(columns)
;unpack row + columns from display list:
port = FIRST_PORT
	while port <= LAST_PORT
		if SSRPINS#v(port)  ;this port is used for SSR control
CommonCathode = 0
constcol = UNLIT(columns)
			EXPAND_PUSH TRUE  ;show what this does in LST file
			if !ISLIT(columns)
				mov8 columns, Outbuf#v(port)
			else
				mov8 _LITERAL(constcol), Outbuf#v(port)
			endif
			EXPAND_POP,
			if ISLIT(rowbit)
CommonCathode = UNLIT(rowbit)
				EXPAND_PUSH TRUE  ;show what this does in LST file
				if !ISLIT(columns)
					xor8 WREG, _LITERAL(PORT#v(port)_BITS), WREG,  ;on channels become 0 bits, off channels become 1 bits
					if (TRISINIT#v(port) & ~CommonCathode) != PORT#v(port)_BITS  ;need to preserve some port bits
						and8 WREG, _LITERAL(TRISINIT#v(port) & ~CommonCathode), Outmask#v(port),  ;set on channels and common cathode for row to output, off channels to Z
					else
						mov8 WREG, Outmask#v(port)  ;set on channels and common cathode for row to output, off channels to Z
					endif
				else
					mov8 _LITERAL(TRISINIT#v(port) & ~CommonCathode & ~constcol), Outmask#v(port)  ;set on channels and common cathode for row to output, off channels to Z
				endif
				EXPAND_POP,
			else
				EXPAND_PUSH TRUE  ;show what this does in LST file
				if !ISLIT(columns)
					xor8 WREG, _LITERAL(PORT#v(port)_BITS), WREG,  ;on channels become 0 bits, off channels become 1 bits
					and8 WREG, _LITERAL(TRISINIT#v(port)), Outmask#v(port),  ;set on channels and common cathode for row to output, off channels to Z
				else
					mov8 _LITERAL(TRISINIT#v(port) & ~constcol), Outmask#v(port)  ;set on channels and common cathode for row to output, off channels to Z
				endif
				xor8 rowbit, _LITERAL(PORT#v(port)_BITS), WREG,  ;on channels become 0 bits, off channels become 1 bits
				and8 Outmask#v(port), WREG, Outmask#v(port),  ;turn common cathode off
				EXPAND_POP,
			endif
		endif
port += 1
	endw  ;port <= LAST_PORT
	EXPAND_POP,
	if (columns) == INDF
		REPEAT FB_ROWSIZE, inc8 FSR  ;advance to next brightness
;use this to simulate PWM by making the first row in frame buffer sticky; this allows easier debug with PK1 board:
;		ifbit CfgByte, CFGBIT_PSEUDO_PWM, FALSE, inc8 FSR,, NOJUMP_HINT  ;advance to next brightness
	endif
	ENDM

#else  ;SSR pins are split across Port A and C
 #ifdef __16F688  ;Port A + C hard-coded for 16F688
;verify hard-coded assumptions:
	LOCAL SSRPINS_a = ~IIF(EXT_CLK_FREQ, 1<<IOBIT(CLKIN_PIN), 0) & SSRPINS#v(0xA)  ;ignore ext clock bit when checking SSR pins; line too long
	ERRIF SSRPINS#v(0xC) != 0xF, [ERROR] hard-coded outbuf unpack code is wrong: C bits #v(SSRPINS#v(0xC)) vs. #v(0xf)  ;paranoid
	ERRIF SSRPINS_a != IIF(PK1DEV, 0x36, 0x17), [ERROR] hard-coded outbuf unpack code is wrong: A bits #v(SSRPINS_a) vs. #v(IIF(PK1DEV, 0x36, 0x17))  ;paranoid

FBUnpack MACRO columns, rowbit
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE  ;hide messy code expansion
	LOCAL ofs = 0, CommonCathode = 0, constcol = UNLIT(columns)
;unpack row + columns from display list:
port = FIRST_PORT
	while port <= LAST_PORT
		if SSRPINS#v(port)  ;this port is used for SSR control
CommonCathode = 0
constcol = UNLIT(columns)
;16F688 hard-coding starts here
			if port == 0xA
				EXPAND_PUSH TRUE  ;show what this does in LST file
				if !ISLIT(columns)
					gettop4 columns, WREG
					ifbit columns, 3+4, TRUE, xorlw (1<<3) | (1<<4),, NOJUMP_HINT  ;move 0x08 up to 0x10; it was packed in an unoccupied bit because the correct bit was already used by a Port C bit
#if PK1DEV  ;an additional bit needs to be moved
					ifbit columns, 0+4, TRUE, xorlw (1<<0) | (1<<5),, NOJUMP_HINT  ;move 0x01 up to 0x20 (A0 and A5 are swapped on PK1 board)
#endif
					mov8 WREG, Outbuf#v(port)  ;IIF(IsBuffered, (outbufs)+ofs, PORT#v(port))  ;data
				else
constcol = (constcol>>4) & 0xf
					if constcol & (1<<3)
constcol ^= (1<<3) | (1<<4)  ;move 0x08 up to 0x10; it was packed in an unoccupied bit because the correct bit was already used by a Port C bit
					endif
#if PK1DEV  ;an additional bit needs to be moved
					if constcol & (1<<0)
constcol ^= (1<<0) | (1<<5)  ;move 0x01 up to 0x20 (A0 and A5 are swapped on PK1 board)
					endif
#endif
					mov8 _LITERAL(constcol), Outbuf#v(port)  ;IIF(IsBuffered, (outbufs)+ofs, PORT#v(port))  ;data
				endif
				EXPAND_POP,
				if ISLIT(rowbit)
CommonCathode = (UNLIT(rowbit)>>4) & 0xf
					if UNLIT(rowbit) & (1<<(3+4))
CommonCathode ^= (1<<3) | (1<<4)  ;move 0x08 up to 0x10; it was packed in an unoccupied bit because the correct bit was already used by a Port C bit
					endif
#if PK1DEV  ;an additional bit needs to be moved
					if UNLIT(rowbit) & (1<<(0+4))
CommonCathode ^= (1<<0) | (1<<5)  ;move 0x01 up to 0x20 (A0 and A5 are swapped on PK1 board)
					endif
#endif  ;PK1DEV
				endif
			endif
			if port == 0xC  ;this port is used for SSR control
				EXPAND_PUSH TRUE  ;show what this does in LST file
				if !ISLIT(columns)
					getbottom4 columns, Outbuf#v(port)  ;IIF(IsBuffered, (outbufs)+ofs, PORT#v(port))  ;data
				else
constcol &= 0x0f
					mov8 _LITERAL(constcol), Outbuf#v(port)  ;IIF(IsBuffered, (outbufs)+ofs, PORT#v(port))  ;data
				endif
				EXPAND_POP,
				if ISLIT(rowbit)
CommonCathode = UNLIT(rowbit) & 0xf
				endif
			endif
			EXPAND_PUSH TRUE  ;show what this does in LST file
			if ISLIT(rowbit)
				if !ISLIT(columns)
					xor8 WREG, _LITERAL(PORT#v(port)_BITS), WREG,  ;on channels become 0 bits, off channels become 1 bits
					if (TRISINIT#v(port) & ~CommonCathode) != PORT#v(port)_BITS  ;need to preserve some port bits
						and8 WREG, _LITERAL(TRISINIT#v(port) & ~CommonCathode), Outmask#v(port),  ;set on channels and common cathode for row to output, off channels to Z
					else
						mov8 WREG, Outmask#v(port)  ;set on channels and common cathode for row to output, off channels to Z
					endif
				else
					mov8 _LITERAL(TRISINIT#v(port) & ~CommonCathode & ~constcol), Outmask#v(port)  ;set on channels and common cathode for row to output, off channels to Z
				endif
			else
				if !ISLIT(columns)
					xor8 WREG, _LITERAL(PORT#v(port)_BITS), WREG,  ;on channels become 0 bits, off channels become 1 bits
					and8 WREG, _LITERAL(TRISINIT#v(port)), Outmask#v(port),  ;set on channels and common cathode for row to output, off channels to Z
				else
					mov8 _LITERAL(TRISINIT#v(port) & ~constcol), Outmask#v(port)  ;set on channels and common cathode for row to output, off channels to Z
				endif
				if port == 0xA
					gettop4 rowbit, WREG
					ifbit rowbit, 3+4, TRUE, xorlw (1<<3) | (1<<4),, NOJUMP_HINT  ;move 0x08 up to 0x10; it was packed in an unoccupied bit because the correct bit was already used by a Port C bit
#if PK1DEV  ;an additional bit needs to be moved
					ifbit rowbit, 0+4, TRUE, xorlw (1<<0) | (1<<5),, NOJUMP_HINT  ;move 0x01 up to 0x20 (A0 and A5 are swapped on PK1 board)
#endif
				endif
				if port == 0xC
					getbottom4 rowbit, WREG
				endif
				xor8 WREG, _LITERAL(PORT#v(port)_BITS), WREG,  ;on channels become 0 bits, off channels become 1 bits
				and8 Outmask#v(port), WREG, Outmask#v(port),  ;turn common cathode off
			endif
			EXPAND_POP,
;16F688 hard-coding ends here
		endif
port += 1
	endw  ;port <= LAST_PORT
	EXPAND_POP,
	if (columns) == INDF
		REPEAT FB_ROWSIZE, inc8 FSR  ;advance to next brightness
;use this to simulate PWM by making the first row in frame buffer sticky; this allows easier debug with PK1 board:
;		ifbit CfgByte, CFGBIT_PSEUDO_PWM, FALSE, inc8 FSR,, NOJUMP_HINT  ;advance to next brightness
	endif
	ENDM

 #else  ;__16F688  ;Port A + C hard-coded for 16F688
FBUnpack MACRO columns, rowbit
	error TODO: need to implement FBUnpack here
	ENDM
 #endif  ;__16F688  ;Port A + C hard-coded for 16F688
#endif  ;SSR pins are split across Port A and C


	WARNIF SSR_PRECLEAR, [INFO] Using extra SSR port clear cycle

;initialize or flush pending SSR pin changes:
;For chipiplexing, the common cathode column (row) is low, but any number of other columns can be high.
;(it could also be the other way around, but I just used the polarity shown in the EDN article).
;This macro just commits I/O pin changes; the actual changes were determined prior to this code.
;Pending changes should be flushed as quickly as possible after trigger timer event to reduce timing jitter.
;This macro is expanded in-line in multiple places in order to avoid additional call/return overhead; the code is only a few instructions anyway.
;First pass sets TRIS, second pass sets PORT; this minimizes bank switching when multiple ports are used.
;TODO: maybe turn on a few pins at a time to avoid power supply noise?
;params:
;  iotype = type of I/O: <0 => init TRIS and PORT; >=0 => flush pending SSR pin changes to TRIS and PORT
Chipiplex MACRO iotype
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE  ;hide messy code expansion
#if SSR_PRECLEAR  ;use this to try to avoid random channels turning on
;pre-pass to clear pins first:
;If we just set TRISA + TRISC then PORTA + PORTC, there is a small window (8 instructions, or ~1.7 usec at 18 MHz) where a spurious signal may be generated for "incorrect" combinations of TRIS + PORT values.
;It's unclear how fast the MOC optos can actually react; published time is ~16 usec at 20 mA or 30 usec at 5 mA, but it may be lower than that.
;It's probably more than 2 usec, though, so this section of code probably is not needed.
;To avoid the chance of turning on stray channels, we reset the output pins to 0 before changing TRIS.
	if iotype >= 0  ;set initial PORT reg
port = FIRST_PORT
		while port <= LAST_PORT
			if SSRPINS#v(port)  ;this port is used for SSR control
				EXPAND_PUSH TRUE  ;show what this does in LST file
				mov8 _LITERAL(0), PORT#v(port)  ;set initial PORTx for I/O ports (some will change later)
				EXPAND_POP,
			endif
port += 1
		endw  ;port <= LAST_PORT
	endif
#endif  ;SSR_PRECLEAR
;set in/out during first pass:
port = FIRST_PORT
	while port <= LAST_PORT
		if SSRPINS#v(port)  ;this port is used for SSR control
			EXPAND_PUSH TRUE  ;show what this does in LST file
			if iotype < 0  ;set initial TRIS reg
				mov8 _LITERAL(TRISINIT#v(port)), TRIS#v(port)  ;set initial TRISx for I/O ports (some will change later)
			else  ;send pending output buffer changes
				mov8 Outmask#v(port)+0, TRIS#v(port)
			endif
			EXPAND_POP,
		endif
port += 1
	endw  ;port <= LAST_PORT
;write data during second pass
port = FIRST_PORT
	while port <= LAST_PORT
		if SSRPINS#v(port)  ;this port is used for SSR control
			EXPAND_PUSH TRUE  ;show what this does in LST file
			if iotype < 0  ;set initial PORT reg
				mov8 _LITERAL(0), PORT#v(port)  ;set initial PORTx for I/O ports (some will change later)
			else  ;send pending output buffer changes
				mov8 Outbuf#v(port)+0, PORT#v(port)
			endif
			EXPAND_POP,
		endif
port += 1
	endw  ;port <= LAST_PORT
	EXPAND_POP,
	ENDM


;******************************************************************************
;hardware event handling section:
;Helper macros/functions for channel control, chipiplexing, timers, etc.
;CAUTION: do not call nested functions unless there is enough stack space
;Most code is wrapped in macros so it can be duplicated or defined without generating code (must be after address 4).
;2 levels of stack are used in here: 1 for main ISR + 1 for peripheral-specific ISR
;4 types of interrupts are currently used: Timer 0 (1/256th ZC), Timer 1 (20 msec), fail-safe clock, and EE write.
;RA change (ZC detect), TX and RX are just polled.
;All tty I/O is handled by main-line, not in here (to simplify the state machine).
;******************************************************************************

	BDCL ThisBrightness  ;brightness of current timeslot; decreases from full on to full off as the cycle progresses
	BDCL RowIndex:FB_ROWSIZE  ;bitmap of rows for current brightness timeslot
	BDCL T0_PRESET  ;value to preload Timer 0 so it will wrap at the end of a timeslot
	BDCL RTC_20msec  ;20 msec counter (50 == 1 second)
	BDCL RTC_SECONDS16:2  ;real-time clock; 65536 seconds will wrap after ~ 18.2 hr

	NBDCL ZC_EDGE  ;used for detecting ZC transitions: 0xf0 = falling edge , 0x0f = rising edge (after 4-timeslot filter/delay)
	BDCL ZC_RATE  ;ZC refresh rate (either direction); ZC parity indicates which polarity AC half-cycle we are in; partial count except at 1 second boundaries
	BDCL ZC_PREVRATE  ;ZC count on previous half cycle  ;stable count each second
#if ZC_TRACKING
	BDCL ZC_PREVWIDTH  ;remember when ZC rising edge occurred
	BDCL ZC_WIDTH  ;this will tell how much "dead" time there is at the start of the cycle
	BDCL ZC_HIGHEST  ;remember highest ZC rate seen (helpful for flicker tracking)
	BDCL ZC_LOWEST  ;remember lowest ZC rate seen (helpful for flicker tracking)
#endif


;ALL interrupt flags that caused an interrupt must be cleared, to avoid recurring interrupts (PIC spec page 121).
;However, since there can be up to 8 reasons for interrupts, ~ 3.5 * 8 ~= 28 instruction cycles are needed just to check them all.
;In order to reduce AVERAGE overhead for the MOST COMMON interrupts, those are checked and processed first.
;If any less common interrupts are still set, then the ISR will be called again later; this gives extra overhead for the less common interrupts, but makes the most common cases more efficient.
;Most time-critical stuff is handled in hardware (EUSART, Timer 0+1), some in software (ZC).
;Timer 1 must remain accurate over long periods of time (since it is used as a real-time clock), so it must be very accurate.
;Timer 0 latency is important, since it affects brightness of lights (1/256th of ZC).
;To avoid jitter, timers are preset by ADDING a preset value to them, rather than SETTING it to a specific value; this automatically accounts for varying ISR latency.

#if ONEINT_ONLY  ;assume only 1 interrupt occurred; ISR will come back in again later if needed
;	messg "[INFO] ISR only handling one interrupt each time (shorter best/typical cases, longer worst case)"
 #define ISR_UNKNOWN  FATAL(0xe0, PIR1, "unknown interrupt")
 #if USE_INTS
  #define ISR_DONE  neargoto isr_done  ;exit after first ISR is processed (assumes that 2 interrupts rarely occur at the same time)
 #else  ;USE_INTS
  #define ISR_DONE  ;check and handle remaining interrupts
 #endif  ;USE_INTS
#else  ;ONEINT_ONLY  ;handle all interrupts that occurred
;	messg "[INFO] ISR handling all interrupts each time (long best/typical cases, shorter worst case)"
 #define ISR_UNKNOWN  ASM_NOP
 #define ISR_DONE  ;check and handle remaining interrupts
#endif


	VARIABLE PERFSTATS_EXPANDED = FALSE  ;avoid redundant code and data expansions
	VARIABLE NEXTROWSETUP_EXPANDED = FALSE  ;avoid redundant code and data expansions

;performance stats (for debug, tuning):
;params:
;  update = flag telling what to do: FALSE => record start time, TRUE => update cumulative stats, OTHER => initialize
PerfStats MACRO update
	NOEXPAND  ;hide clutter in LST file
#if PERF_STATS
	if !PERFSTATS_EXPANDED  ;avoid duplicate data definitions; expand first time only
PERFSTATS_EXPANDED = TRUE
		EXPAND_PUSH TRUE  ;show what this does in LST file
		BDCL ISR_COUNT24:3  ;#times ISR called; worst-case wrap is (16 Timer 0 or rows) + 1 Timer 1 + 2 ZC per cycle @120Hz ~= 2280/sec >~= 2 hr; best case ~= 6x = 12 hr
		BDCL SetupTime32:4  ;avg row time/count = 14
		BDCL SetupCount24:3
		BDCL ThisSetupTime  ;current non-timeslot processing time (for performance tuning)
		BDCL MaxSetupTime  ;longest non-timeslot processing time (for performance tuning)
		EXPAND_POP,
	endif
	if (update) == TRUE  ;update perf stats
		EXPAND_PUSH TRUE  ;show what this does in LST file
		sub8 TMR0, ThisSetupTime, ThisSetupTime, COMPARE_HINT  ;measure time spent doing other stuff
		ifbit LTZ, FATAL(0xe1, TMR0, "ISR timing error; missed a timeslot with Timer 0"),, BIGSTMT_HINT  ;paranoid self-check; Timer 0 should not wrap while doing other processing
		if8 ThisSetupTime, GE, MaxSetupTime, ADDWF_wrapper(MaxSetupTime, F),,  ;keep track of max time used elsewhere (for perf tuning)
		add32_8 SetupTime32, ThisSetupTime, SetupTime32,
		inc24 SetupCount24
		EXPAND_POP,
	else
		if (update) != FALSE  ;initialize
			EXPAND_PUSH TRUE  ;show what this does in LST file
			mov24 _LITERAL(0), ISR_COUNT24
			mov32 _LITERAL(0), SetupTime32
			mov24 _LITERAL(0), SetupCount24
			mov8 _LITERAL(0), MaxSetupTime
			EXPAND_POP,
		endif
		EXPAND_PUSH TRUE  ;show what this does in LST file
		mov8 TMR0, ThisSetupTime  ;remember start time of other processing
		EXPAND_POP,
	endif
#endif
	ENDM


;comma-less wrapper to update sticky columns for pseudo-PWM:
;params:
;  row = row# to be updated (0..7)
UpdateStickyColumns MACRO row
	ior8 StickyRows+row, UNSTACK, StickyRows+row,  ;merge new rows with sticky rows
	setbit StickyRowIndex, row, FALSE  ;don't need to update sticky rows again
	ENDM

;generate a gotoc entry to unpack top bit of a byte value:
;there is one of these entries for each possible byte value
;params:
;  bits = bitmask of rows; only top bit will be used
RowUnpackWrapper MACRO bits
	NOEXPAND  ;hide clutter in LST file
	LOCAL topbit = HIGHBIT_#v(bits)
	EXPAND_PUSH TRUE  ;show what this does in LST file
	neargoto UnpackRow#v(IIF(topbit, 1<<(topbit-1), 0))  ;must be near goto (PCLATH<3:4> must be correct)
	EXPAND_POP,
	ENDM


;setup for next sub timeslot (one or more display rows, depending on SSR_OVERCLOCK):
;This should be done *before* starting to wait for Timer 0 (sub timeslot), so the data is ready to go immediately at start of timeslot (for low timing jitter).
;All processing should be completed before the start of the next timeslot, but if it takes a little too long and we are late for the start of the timeslot, 
;not a big deal - SSRs will just be a little (unoticeably) dimmer than expected until the PIC "catches up" during empty timeslots; may not even be noticeable.
;The Vixen plug-in decides whether to compensate by merging brightnesses/rows or leave as-is and let them run a little behind schedule.
;Either way gives us "lossy" dimming, and much more flexible timing.
;params:
; subslot = flag telling if start of timeslot: 1 => start of 1/256th ZC timeslot, other => another sub-timeslot within there (only for SSR over-clocking)
NextRowSetup MACRO subslot
	NOEXPAND  ;hide clutter in LST file
	LOCAL UnpackCurrentRow, UnpackCurrentRowNoChanges
	EXPAND_PUSH !NEXTROWSETUP_EXPANDED  ;hide messy code expansion
	if !NEXTROWSETUP_EXPANDED  ;avoid redundant code expansions; expand first time only
NEXTROWSETUP_EXPANDED = TRUE
		LOCAL resume_here = $

;put the helper functions up higher in rom so we don't need to jump over them (saves 2 instr):
		ORG LIT_PAGE_SIZE * (LITPAGEOF($) + 1) - REG_PAGE_SIZE - 9*39  ;start a little before the last 256-word block within the current 1024-word, and put some helper functions there
#if PK1DEV
		ORG $ - 9*2 - 5  ;PK1 LED work-around needs a little more space
#endif

;helper functions to unpack frame buffer data:
;This one is used if there are no active rows (result will be all channels off).
 jumpdest UnpackRow#v(0):, PCL|,, OFBANK(0)|,  ;return with bank 0 since that is where next data typically is
		FBUnpack _LITERAL(0), _LITERAL(0)  ;4 instr
		return_ UnpackRow#v(0)

;now define a pair of functions for each active row:
;(could have used 2 jump tables, but that would take another 257 words of prog space, which is more than just defining another copy of the 8 functions)
		EXPAND_PUSH FALSE  ;hide messy code expansion
		LOCAL row
row = 0
		while row < SSR_PIN_COUNT  ;0, 1, 2, ..., 7
			EXPAND_PUSH TRUE  ;show what this does in LST file
;merge in sticky columns for this row before unpacking:
;NOTE: this uses StickyRows buffers, not FSR; this allows FSR to point to next row in frame buffer
 jumpdest UnpackStickyRow#v(1<<row):, CfgByte|,, OutBankOf(UnpackRow#v(0))|,  ;common return must share same out bank
			ifbit StickyRowIndex, row, TRUE, UpdateStickyColumns row,, BIGSTMT_HINT  ;column bits have not been merged in yet
			FBUnpack StickyRows+row, _LITERAL(1<<row)  ;refresh chipiplexing buffers from StickyRows rather than directly from frame buffer (takes ~ 14 instr)
			return_ UnpackStickyRow#v(1<<row)
;unpack next row's column bitmask from frame buffer:
 jumpdest UnpackRow#v(1<<row):, PCL|,, OutBankOf(UnpackRow#v(0))|,  ;common return must share same out bank
			setbit RowIndex, row, FALSE  ;don't do this row again for the remainder of this cycle; do next highest row next time
			ifbit CfgByte, CFGBIT_PSEUDO_PWM, TRUE, neargoto UnpackStickyRow#v(1<<row),,  ;it would be more efficient to use a separate jump table, but that would take more code space so we just make it conditional each time
			FBUnpack INDF, _LITERAL(1<<row)  ;unpack next column bitmask from frame buffer (takes ~ 14 instr)
			return_ UnpackRow#v(1<<row)
			EXPAND_POP,
row += 1
		endw  ;row < 8
		EXPAND_POP,

		ORG LIT_PAGE_SIZE * (LITPAGEOF($) + 1) - REG_PAGE_SIZE - 1  ;start on last word before last 256-word block within the current 1024-word address space, and put jump table there
;unpack the row selected by the top bit in RowIndex:
;NOTE: this code assumes that PCLATH:<3,4> does not change
 jumpdest UnpackTopRow:, RowIndex|,, OutBankOf(UnpackRow#v(0))|,  ;common return must share same out bank
		storew_ PCL  ;this causes a jump to one of the nearjumps below; PCLATH increments as part of this instruction, so all 256 nearjumps must be within the same 256-word code page
		ERRIF $%REG_PAGE_SIZE, [ERROR] Start of gotoc not at offset 0: #v($)  ;paranoid check
;jump table to unpack a row:
		EXPAND_PUSH FALSE  ;hide messy code expansion
		REPEAT 256, RowUnpackWrapper repeater  ;this generates a neargoto UnpackRow#v(top bit of RowIndex) for each byte value
		ERRIF REGPAGEOF($-1) != REGPAGEOF(UnpackTopRow+2), [ERROR] Bad end address for jump table: #v($) vs. #v(UnpackTopRow)  ;all jump table entries must be within same 256-word block in order for PCLATH to be correct
		EXPAND_POP,

		ORG resume_here  ;continue where we left off
	endif
	if subslot == 1  ;start of next timeslot; keep track of current brightness
		EXPAND_PUSH TRUE  ;show what this does in LST file
		dec8 ThisBrightness  ;keep track of current timeslot; sooner is brighter: starts at 255 at start of cycle and descreases to 0 (1, actually)
#if WANT_DEBUG
		ifbit EQZ, FATAL(0xc0, ThisBrightness, "Timeslot counter underflow; bad timing constants?"),, BIGSTMT_HINT  ;timeslots descrease in brightness as time passes; at end of cycle hold at 0 (actually, at 1 pre-decrement) until start of next cycle; this serves as a catch-all for all remaining non-0 brightness levels
#else  ;WANT_DEBUG
		ifbit EQZ, inc8 ThisBrightness,, NOJUMP_HINT  ;timeslots descrease in brightness as time passes; at end of cycle hold at 0 (actually, at 1 pre-decrement) until start of next cycle; this serves as a catch-all for all remaining non-0 brightness levels
#endif  ;WANT_DEBUG
		EXPAND_POP,
	endif
	EXPAND_PUSH TRUE  ;show what this does in LST file
	if#v(8*FB_ROWSIZE) RowIndex, NE, _LITERAL(0), LEAVE_PLACEHOLDER1 neargoto UnpackCurrentRow,,  ;finish rows left over from prev sub timeslot before getting more rows for next timeslot (causes lossy dimming until we catch up)
;ThisBrightness will never go to 0, so end-of-list marker will be handled here also:
;FSR/IRP points to next brightness level to be displayed within active frame buffer
	if8 INDF, LT, ThisBrightness, LEAVE_PLACEHOLDER1 neargoto UnpackCurrentRowNoChanges,,  ;short-cut (bypasses jump table); not yet time for next set of rows
	LOCAL MYPH = PLACEHOLDER_STACK  ;remember top of placeholder stack; need to unwind out of order
;if we fell behind (so that INDF > ThisBrightness), then we will catch up here
	inc8 FSR  ;move past brightness level; now points to row index
	mov#v(8*FB_ROWSIZE) UNSTACK, RowIndex  ;bitmap of rows to be turned on for this brightness timeslot; FSR updated to point to first column bitmap
#if WANT_DEBUG
	if#v(8*FB_ROWSIZE) RowIndex, EQ, _LITERAL(0), FATAL(0xe2, FSR, "Useless row index found in display list"),, BIGSTMT_HINT  ;paranoid; should have at least one row on
	if#v(8*FB_ROWSIZE) INDF, EQ, _LITERAL(0), FATAL(0xa3, FSR, "Useless row found in display list"),,  BIGSTMT_HINT  ;paranoid; should have at least one column on
#endif  ;WANT_DEBUG
	ifbit CfgByte, CFGBIT_PSEUDO_PWM, FALSE, LEAVE_PLACEHOLDER1 neargoto UnpackCurrentRow,,  ;go unpack column bitmap for the next row in row index
;update cumulative ("sticky") channels for pseudo-PWM before unpacking:
	mov#v(8*FB_ROWSIZE) RowIndex, StickyRowIndex  ;save bitmap of rows that are changing from frame buffer; used to update cumulative (sticky column bitmaps)
	mov#v(8*FB_ROWSIZE) _LITERAL(0), RowIndex  ;prepare to turn all bits on; dec8 below will set RowIndex to 255 (all rows on)
 jumpdest UnpackCurrentRowNoChanges:, ThisBrightness|,,, FALLTHRU
	FILL_PLACEHOLDER MYPH&, neargoto UnpackCurrentRowNoChanges
	ifbit CfgByte, CFGBIT_PSEUDO_PWM, TRUE, dec#v(8*FB_ROWSIZE) RowIndex,,  ;force refresh of all rows each time (RowIndex becomes 0xff)
 jumpdest UnpackCurrentRow:,,,, FALLTHRU
	FILL_PLACEHOLDER, neargoto UnpackCurrentRow
	FILL_PLACEHOLDER, neargoto UnpackCurrentRow
;now go set Chipiplexing pin buffers according to the next row (highest bit in RowIndex) and column index (INDF)
;PCLATH<4:3> must have been correct to get there, but PCLATH<2:0> must also be correct for the computed goto inside UnpackTopRow.
;Since PCLATH<4:3> and eventually PCLATH<2:0> need to be set, here we just set the entire PCLATH value (consolidates instructions).
	GotocSetup UnpackTopRow+1  ;set PCLATH to target jump table
	loadw_ RowIndex  ;load W (must be done after setting PCLATH, not before)
	nearcall UnpackTopRow  ;highest bit of W will be used as row (common cathode); INDF will be used as columns (anodes)
	EXPAND_POP,
	EXPAND_POP,
	ENDM


;keep track of how busy we aren't:
;each idle tick = 1+1+2+2+2/256+2/65536+2/16M instr = 6.00784313678741455078125 instr => 332898 idle ticks/sec @ 8 MHz when totally idle
#define IDLE_SCALE(clock)  rdiv(InstrPerSec(clock), 100*(1+1+2+2+2/0x100+2/0x10000+2/0x1000000))  ;#instr/100 for each tick of IDLE_COUNT

#if DEBUG_TTY  ;keep track of idle time
 #define  TrackIdleTime  inc32 IDLE_COUNT32
#else  ;DEBUG_TTY
 #define  TrackIdleTime  ASM_NOOP
#endif  ;DEBUG_TTY


;SSR timeslot handler:
;This sends buffered SSR control pulses to Port A + C.
;For 1x (no SSR over-clock), come thru here every ~ 32.6 usec at 60 Hz AC, ~ 39 usec at 50 Hz AC, or ~ 78 usec at 50 Hz DC.
;With SSR over-clocking, come through more often (Timer 0 wrap-around will be sooner in that case).
;This is wrapped as a macro so it can be expanded multiple times in-line (trades off code space vs. call/return overhead).
;Only come here after doing *all* other processing; this will busy-wait for next brightness timeslot (~32.6 usec at 60 Hz AC, or ~39 usec at 50 Hz AC).
;params:
; subslot = flag telling if start of timeslot: 1 => start of 1/256th ZC timeslot, other => another sub-timeslot within there (only for SSR over-clocking)
SSR_ISR MACRO subslot
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH TRUE  ;show what this does in LST file
	NextRowSetup subslot  ;first set up SSR I/O pin changes in memory
	PerfStats TRUE
#if !USE_INTS  ;use low latency busy-wait loop on Timer 0; reduces timing jitter
	whilebit INTCON, T0IF, FALSE, TrackIdleTime,  ;busy-wait for critical time event; this gives ~ 5 instruction latency (~ 1 usec at 18.432 MHz), compared to true ISR which would be ~ 20 instructions (due to register save/restore + return thru stack)
#endif  ;USE_INTS
;first send out time-critical SSR control pulses:
	Chipiplex subslot  ;flush next SSR pin changes; this happens as soon as possible after Timer 0 wraps, to minimize timing jitter
;now do other timeslot housekeeping:
	PresetTimer0 T0_PRESET  ;rewind Timer 0 to wrap again at the beginning of next (sub)timeslot
	setbit INTCON, T0IF, FALSE  ;clear interrupt flag
	PerfStats FALSE
	ISR_DONE
	EXPAND_POP,
	ENDM


;ZC handling:
;Since ZC is a relatively slow signal, we only need to poll it once in a while and check for changes.
;This is more efficient than checking for an IOCA interrupt flag every time an interrupt handler is called.
;For 120 VAC RMS, peak = 170V; H11AA1 circuit typically has 2 10K resistors in series.
;Assuming the H11AA1 opto LED has a 1.2V drop and requires 0.5 mA to turn on, H11AA1-based ZC detector requies AC voltage to reach ~.5 mA * 20K + 1.2V = 11.2V before turn-on.
;This would occur at asin(11.2/170) ~= 4 degrees from either end of the AC cycle, which is ~ 8.3 msec *4/180 ~= 185 usec or 5.3 timeslots at ZC/256.
;BTA08-400 triacs appear to be able to turn on ~ 1V, so they can be turned on at 8.3 msec * asin(1/170)/180 ~= 15 usec or 0.5 timeslot.
;H11AA1 LED can handle up to 60 mA max, but 30 mA is a safer limit; this would occur with ZC resistor total of (340-1.2)/30mA ~= 12K.
;So, if ZC resistors were dropped to 6K, this would give a ZC delay until ~.5 mA * 12K + 1.2V = 7.2V, or 8.3 msec * asin(7.2/170)/180 ~= 112 usec or 3.5 timeslots.
;Since ZC is a critical signal for correct AC dimming, noise on the line needs to be filtered.
;ZC is sampled each time thru here and then shifted into a tracking byte.  The tracking byte saves the last 8 samples.
;Noise filtering can then be done more efficiently by just checking the ZC tracking byte for a rising or falling edge pattern.
;Since the duration of ZC would typically be 2*185 usec = 370 usec ~= 11 timeslots, it seems reasonable to try to filter out noise shorter than ~ 100 usec or 3 timeslots.
;So if we detect the trailing edge of ZC, using 10K resistors in the ZC detector circuit, we are stuck with a delay of 185+100 usec = 285 usec or ~ 9 timeslots,
;but if we detect the leading edge of ZC instead, and then use the remaining 185 usec of the AC cycle as a noise filter, we can use a fixed delay of ~ 15 usec or 0.5 timeslot at the start of the next cycle, for more complete full-on behavior.
;Leading edge of ZC = rising edge, trailing edge = falling edge (ZC is opposite of ZC waveform voltage).
;For AC SSR doubling, H11A1 can be used instead of H11AA1 for a polarized ZC signal, in which case changes in both directions must be used.
;If ZC is non-polarized (H11AA1 circuit), then rising and falling edges are used to trigger a new half-cycle.
;If ZC is polazized (H11A1 circuit), then only the rising edge is used, and then channels can be interleaved for SSR-doubling.
;params:
;  want_sampling = flag telling whether to generate the sampling code (can be multiple times), or just the cycle start logic (one copy only)
ZC_ISR MACRO want_sampling
	NOEXPAND  ;hide clutter in LST file
	if !(want_sampling)  ;set up new cycle; this code should only be expanded once (generates global labels)
;NOTE: need to jump around this code if it is in-line
		EXPAND_PUSH TRUE  ;show what this does in LST file
#if WANT_ZC
;ZC falling edge detected:
 jumpdest zc_fall:, ZC_EDGE|,,,
#if ZC_TRACKING
		mov8 ZC_WIDTH, ZC_PREVWIDTH  ;remember width of ZC (mainly for debug)
#endif  ;ZC_TRACKING
		ifbit CfgByte, CFGBIT_ACSSR_DOUBLE, FALSE, return_ zc_fall,,  ;ZC not polarized; ignore falling edge
		neargoto start_cycle
;ZC rising edge detected:
 jumpdest zc_rise:, ZC_EDGE|,, OutBankOf(zc_fall)|, ;common return shares same bank affinity
#if ZC_TRACKING
		mov8 _LITERAL(0), ZC_WIDTH  ;start counting width on rising edge
#endif
		inc8 ZC_RATE  ;count AC cycles; used to adjust Timer 0 interval for 50 Hz vs. 60 Hz after 1 second
#endif  ;WANT_ZC
 jumpdest start_cycle:,,,, FALLTHRU
		StartCycle MAYBE  ;setup for start of new half-cycle
		return_ zc_rise
		EXPAND_POP,
	else  ;sample and detect
#if WANT_ZC
		EXPAND_PUSH TRUE  ;show what this does in LST file
;		add8 PORT#v(IOPORT(ZC)), _LITERAL LOW (-1<<IOBIT(ZC)), WREG, COMPARE_HINT  ;this is a more efficient way to move ZC bit to STATUS.C (2 instr instead of 3 required for bit test + set)
#if ZC_TRACKING
		inc8_ifnz ZC_WIDTH,, dec8 ZC_WIDTH,  ;only need to count once per brightness level; don't let it wrap
#endif
;ZC sampling:
;A ZC transition is recognized as 4 timeslots at one orientation followed by 4 timeslots at the other orientation.
;This filters out noise and automatically incorporates a delay of 4 timeslots after the actual transition.
;NOTE: should always use the *leading* edge (decaying AC voltage) so the filter/delay can be performed before entering next AC half-cycle, rather than the *trailing* edge (which would make the processing "late").
;TODO:	ifbit CfgByte, CFGBIT_SSR_DOUBLE, FALSE, use every second xition, use every xition
		bitcopy PORT#v(IOPORT(ZC)), IOBIT(ZC), STATUS, C
		shiftl8_NC ZC_EDGE  ;shift in next sample; oldest samples are in upper bits; STATUS.C is latest sample
		if8 ZC_EDGE, EQ, _LITERAL(LOW(0xff<<ZC_FILTER)), nearcall zc_fall,,  ;typically 0xf0 (4 timeslots)
		if8 ZC_EDGE, EQ, _LITERAL(LOW(~(0xff<<ZC_FILTER))), nearcall zc_rise,,  ;typically 0x0f (4 timeslots)
#endif  ;WANT_ZC
		EXPAND_POP,
	endif
	ISR_DONE
	ENDM


;setup for start of new half-cycle:
;Timing isn't too critical here; triacs can't turn on yet anyway, due to insufficient voltage.
;There is ~1667 usec (60 Hz) or 2000 usec (50 Hz) at start of AC half-cycle for start-of-cycle handling.
;Any full-on channels need to be turned on by the end of that time, but short delays are not noticeable.
;params:
;  want_fbsetup = flag telling whether to initialize new frame buffer
StartCycle MACRO want_fbsetup
	LOCAL noswap, non_pwm
	mov8 _LITERAL(255), ThisBrightness  ;first timeslot level is full-on
;?	mov8 _LITERAL 0, ThisBrightness  ;first timeslot level will be full-on (255) after first decrement
	mov#v(8*FB_ROWSIZE) _LITERAL(0), RowIndex  ;flush current rows
	PresetTimer0 T0_PRESET  ;restart Timer 0; otherwise any residual count will eventually accumulate to a full tick and cause potential mis-fire
	if want_fbsetup  ;set position within active frame buffer
		ifbit BITVAR(FBSwapRequest), FALSE, LEAVE_PLACEHOLDER1 neargoto noswap,,  ;only swap in new frame buffer at start of next SSR refresh cycle (avoids flicker during transition)
		xor8 FBPtr, _LITERAL(LOW(FB1_START^FB2_START)), FBPtr,  ;move active framebuffer ptr to other
		setbit BITVAR(FBSwapRequest), FALSE  ;clear frame buffer swap request; caller can check this to see if new frame buffer is now active
 jumpdest noswap:,,,, FALLTHRU
		FILL_PLACEHOLDER, neargoto noswap
		FBSELECT TRUE, _LITERAL(FB1_END-FB1_START)  ;set FSR/IRP to start of active (output) frame buffer
		ifbit CfgByte, CFGBIT_PSEUDO_PWM, FALSE, LEAVE_PLACEHOLDER1 neargoto non_pwm,,
;additional setup for pseudo-PWM:
		REPEAT SSR_PIN_COUNT*FB_ROWSIZE, clear_ StickyRows+repeater  ;reset all sticky channels
		mov#v(8*FB_ROWSIZE) _LITERAL(0), StickyRowIndex  ;no rows to change yet
 jumpdest non_pwm:,,,, FALLTHRU
		FILL_PLACEHOLDER, neargoto non_pwm
	endif
	ENDM


	BITDCL EVT_pre20msec  ;notification that 20 msec counter has ticked (might be partial, so this bit is only useful for getting ready for the actual 20 msec)
	BITDCL EVT_20msec  ;notification that another 20 msec has elapsed (guaranteed full; caller might have looked at EVT_20msec just before it ticked)
	BITDCL EVT_GOT1SEC  ;1 sec elapsed

T1Pre_INT set _LITERAL(Timer1_Preset(INT_OSC_FREQ, DC_REFRESH))  ;Timer 1 preset for int osc; line too long
T1Pre_EXT set _LITERAL(IIF(EXT_CLK_FREQ, Timer1_Preset(EXT_CLK_FREQ, DC_REFRESH), 1))  ;Timer 1 preset for ext clock (more than 16 bits); line too long


;check for RTC (Timer 1 overflow):
;This occurs infrequently, at DC_REFRESH rate (20 msec if 50Hz, or 10 msec if 100 Hz)
;Must do this at least 1x every 20 msec or 1 ZC time (~8.3 msec at 60 Hz AC or 10 msec at 50 Hz AC) in order to keep fido happy.
Timer1_ISR MACRO
	clrwdt  ;tell fido that we're still alive; must do this at least every 100 msec
	setbit T1CON, TMR1ON, FALSE  ;turn Timer 1 off; needed for 16-bit add
	ifbit USING_INTOSC(TRUE), PresetTimer1 T1Pre_INT, PresetTimer1 T1Pre_EXT,  ;ext oscillator failed/absent, running from int osc
	setbit T1CON, TMR1ON, TRUE  ;turn Timer 1 back on again
	setbit PIR1, TMR1IF, FALSE  ;clear interrupt
	ifbit BITVAR(EVT_pre20msec), TRUE, SETBIT(BITVAR(EVT_20msec), TRUE),,  ;a full 20 msec has now elapsed
	setbit BITVAR(EVT_pre20msec), TRUE  ;partial 20 msec has elapsed
	inc8 RTC_20msec
	ifbit EQZ, RTC_ISR,, BIGSTMT_HINT  ;keep track of msec; 1 second has elapsed
	if8 ZC_PREVRATE, EQ, _LITERAL(0), StartCycle TRUE,, BIGSTMT_HINT  ;no ext ZC; use Timer1 as internal ZC and sync 1/256th counter with start of DC cycle
	ISR_DONE
	ENDM


	CONSTANT T0_50HzAC_8MHz = Timer0_Preset(INT_OSC_FREQ, AC50HZ)  ;line too long
	CONSTANT T0_50HzAC_18432MHz = IIF(EXT_CLK_FREQ, Timer0_Preset(EXT_CLK_FREQ, AC50HZ), 0)  ;line too long
	CONSTANT T0_60HzAC_8MHz = Timer0_Preset(INT_OSC_FREQ, AC60HZ)  ;line too long
	CONSTANT T0_60HzAC_18432MHz = IIF(EXT_CLK_FREQ, Timer0_Preset(EXT_CLK_FREQ, AC60HZ), 0)  ;line too long

;RTC and 1-second tracking:
;This will also adjust Timer 0 preset based on #ZC cycles per second.
RTC_ISR MACRO
	add8 RTC_20msec, _LITERAL(RTC_Preset(DC_REFRESH)), RTC_20msec,  ;make it tick again at the right time; automatically compensates for ISR latency
	inc16 RTC_SECONDS16  ;#seconds since startup
#if WANT_ZC
	mov8 ZC_RATE, ZC_PREVRATE  ;save current ZC rate
#if ZC_TRACKING
	ifbit EQZ, neargoto nocheck,,  ;don't set lowest if no ZC seen
	if8 ZC_RATE, LT, ZC_LOWEST, ADDWF_wrapper(ZC_LOWEST, TRUE),, NOJUMP_HINT  ;keep track of minimum value seen
 jumpdest nocheck:,,,, FALLTHRU
	if8 ZC_RATE, GE, ZC_HIGHEST, ADDWF_wrapper(ZC_HIGHEST, TRUE),, NOJUMP_HINT  ;keep track of maximum value seen; NOTE: use >= rather than > for more compact code
#endif  ;ZC_TRACKING
	mov8 _LITERAL(0), ZC_RATE  ;start a fresh count for next 1 second interval
	setbit BITVAR(EVT_GOT1SEC), TRUE  ;signal 1 second elapsed
;we started off assuming 50 Hz DC (no ZC); now check if 50 or 60 Hz AC ZC is present:
;set up Timer0 preset according to ZC and clock freq:
	if8 ZC_PREVRATE, LT, _LITERAL(AC50HZ*4/5), LEAVE_PLACEHOLDER1 neargoto zcdone,,  ;20% tolerance; ignore anything less that 80% of 2*50 Hz
	if8 ZC_PREVRATE, GE, _LITERAL((AC60HZ+AC50HZ)/2), LEAVE_PLACEHOLDER1 neargoto zc_AC60Hz,,  ;17% tolerance; split between 50 Hz and 60 Hz AC
	LOCAL MYPH = PLACEHOLDER_STACK  ;override placeholder stack behavior
;AC50Hz:
	ifbit USING_INTOSC(TRUE), loadw_ _LITERAL(T0_50HzAC_8MHz), loadw_ _LITERAL(T0_50HzAC_18432MHz), BOTHOK_HINT
	storew_ T0_PRESET
	LEAVE_PLACEHOLDER1 neargoto zcdone
 jumpdest zc_AC60Hz:, ZC_PREVRATE|,,,
	FILL_PLACEHOLDER MYPH&, neargoto zc_AC60Hz
	ifbit USING_INTOSC(TRUE), loadw_ _LITERAL(T0_60HzAC_8MHz), loadw_ _LITERAL(T0_60HzAC_18432MHz), BOTHOK_HINT
	storew_ T0_PRESET
 jumpdest zcdone:,,,, FALLTHRU
	FILL_PLACEHOLDER, neargoto zcdone
	FILL_PLACEHOLDER, neargoto zcdone
#endif  ;WANT_ZC
	ENDM


	VARIABLE RENISR_EXPANDED = FALSE  ;avoid redundant code and data expansions

;handle serial I/O (Renard + console/debug):
;Renard protocol is handled by caller; must return to caller within ~ 87 usec to not drop Renard data (1 char time at 115K baud).
;This is wrapped as a function, but just returns to caller and then resumes when caller yields again by calling in to here.
Renard_ISR MACRO
	if RENISR_EXPANDED
		mov8 FSR, FBPtr  ;save current position in frame buffer
		return_ CoopISR  ;return to caller to handle Renard protocol
	else
;caller resumes here:
 jumpdest CoopISR:,,, OFBANK(0)|,  ;return with bank 0 since that is where most data is
#if PERF_STATS
		inc24 ISR_COUNT24  ;#times ISR was called; do this after ISR processing to reduce timing jitter
#endif
		FBSELECT FBPtr, _LITERAL(0)  ;set FSR/IRP to current place in active (output) frame buffer; caller will change them as frame buffer is processed
	endif
RENISR_EXPANDED = TRUE
	ENDM


;check if ext oscillator failed/not present:
;This is called maybe 1x only near start, or never
OSF_ISR MACRO
;leave interrupt flag set (so it can be checked), but disable interrupt:
;	setbit PIR1, OSFIF, FALSE  ;clear interrupt
;	setbit PIE1, OSFIE, FALSE  ;disable further OSFIF interrupts, but leave flag set so we can check it
	setbit OSCCON, SCS, TRUE  ;NOTE: FCM does not set this, so set it manually (page 31); SERIAL PORT WILL BE MESSED UP WITHOUT THIS!
;	mov8 _LITERAL MY_OSCCON, OSCCON  ;DO NOT DO THIS HERE!  resetting SCS with no ext clock will cause watchdog reset
	SetSpeed INT_OSC_FREQ  ;adjust timer presets to work with 8 MHz int osc
	ISR_DONE
	ENDM


;NOTE: cannot have any real code before this (address 0 must have startup code)
;interrupt handler must start at address 4 for PIC16Fxxx
	ERRIF $ > INTERRUPT_VECTOR, [ERROR] ISR is not in the right place #v($): some code was generated ahead of ISR #v(INTERRUPT_VECTOR)  ;paranoid self-check
	org INTERRUPT_VECTOR  ;PIC jumps here on interrupt; NOTE: although PC == 4, PCLATH might NOT be 0!


#if USE_INTS  ;NOTE: this code is no longer used, but left here in case need to revert back to using interrupts in future; probably has compile errors now
;ISR save/restore state values:
;ISR_W, ISR_STATUS and ISR_PCLATH are the only ones that really need to use non-banked RAM here (in order to preserve RP1+RP0 status).
;However, since the space is available, we might as well put other stuff in here too; helps a little (avoids extra bank selects).
	NBDCL ISR_W  ;save registers during ISR; MUST BE NON-BANKED
	NBDCL ISR_STATUS  ;save registers during ISR; MUST BE NON-BANKED
	NBDCL ISR_PCLATH  ;caller might be using PCLATH, so we need to save/restore it BEFORE any jumps within ISR use it
	NBDCL ISR_FSR  ;only need to save this if changed during ISR (mostly for PutMsg, GetMsg, FrameBuffer handling)
;	ISR_PI1  ;check which peripheral interrupts are wanted


;NOTE: save ALL registers that are set by this function, to make interrupts transparent to interruptee.
;NOTE: ISR_* save registers must be NON-BANKED; otherwise, RP0+RP1 would need to be set before saving STATUS, which would trash STATUS and interfere with interruptee's RP0+RP1.
;Actually, there is a way to save registers in BANKED RAM, by checking the current value of RP0 and RP1 first and then jumping to a different code block for each bank, but that takes more code space.
	SETPAGE -1, FALSE  ;CAUTION: PCLATH is unknown when we arrive here (int sets PCL directly, not via PCLATH)
 jumpdest realISR:,,,,
	storew_ ISR_W  ;save accumulator first (page 124)
	swap_ STATUS, ISR_STATUS  ;then grab status WITHOUT altering status bits
	mov8 PCLATH, ISR_PCLATH  ;save PCLATH register (since main uses it, ISR needs to set it before any jumps, so we need to save/restore it
	PAGECHK $, MAYBE,  ;allow near jumps (PCLATH was unknown before here)
	mov8 FSR, ISR_FSR  ;save FSR (only needs to be saved if set within ISR)
#if WANT_DEBUG
	mov8 TMR0, ISR_STARTTIME  ;save current timer to measure ISR execution time
	sub8 ISR_STARTTIME, _LITERAL(30/Timer0_Prescalar), ISR_STARTTIME,  ;add estimated ISR entry/exit overhead (approx 30 instr @ 1:2 prescalar ~= 15)
#endif  ;WANT_DEBUG
;check interrupts in decreasing order of frequency/priority; this helps to lower average ISR overhead if ONEINT_ONLY == TRUE, and latency otherwise in all cases
;NOTE: if the IE bit in the constant is on, the code below assumes that the int is wanted; if not, then use the PI1 code
	BANKCHK OFBANK(0), MAYBE  ;set to known bank; ISR sub-functions assume bank 0
	if MY_INTCON & (1<<T0IE)  ;want timer 0 interrupts
		ifbit INTCON, T0IF, TRUE, SSR_ISR,, BIGSTMT_HINT  ;1/256'th ZC (every 32.5/39/78 usec)
	endif
	if MY_INTCON & (1<<RAIE)  ;want RA change interrupts
		ifbit INTCON, RAIF, TRUE, RAX_ISR,, BIGSTMT_HINT  ;50/60 Hz during ZC
	endif
	if MY_INTCON & (1<<INTE)  ;want ext RA2 interrupts
		ifbit INTCON, INTF, TRUE, RA2_ISR,, BIGSTMT_HINT  ;ext int (RA2); unused
	endif
;check PIE1/PIR1 assumptions:
;	ERRIF (TXIF != TXIE) || (RCIF != RCIE) || (TMR1IF != TMR1IE) || (EEIF != EEIE) || (OSFIF != OSFIE) || (ADIF != ADIE) || (C1IF != C1IE) || (C2IF != C2IE), "Int flags != int enable bits", (1<<TXIF) | (1<<RCIF) | (1<<TMR1IF) | (1<<EEIF) | (1<<OSFIF) | (1<<ADIF) | (1<<C1IF) | (1<<C2IF), (1<<TXIE) | (1<<RCIE) | (1<<TMR1IE) | (1<<EEIE) | (1<<OSFIE) | (1<<ADIE) | (1<<C1IE) | (1<<C2IE)  ;the AND below assumes the bits are the same between PIE1 and PIR1
;	ERRIF TXIF != TXIE, "[ERROR] Int flags != int enable bits (TXIF vs. TXIE)", 1<<TXIF, 1<<TXIE  ;paranoid self-check
;	ERRIF RCIF != RCIE, "[ERROR] Int flags != int enable bits (RCIF vs. RCIE)", 1<<RCIF, 1<<RCIE  ;paranoid self-check
;	ERRIF TMR1IF != TMR1IE, "[ERROR] Int flags != int enable bits (TMR1IF vs. TMR1IE)", 1<<TMR1IF, 1<<TMR1IE  ;paranoid self-check
;	ERRIF EEIF != EEIE, "[ERROR] Int flags != int enable bits (EEIF vs. EEIE)", 1<<EEIF, 1<<EEIE  ;paranoid self-check
;	ERRIF OSFIF != OSFIE, "[ERROR] Int flags != int enable bits (OSFIF vs. OSFIE)", 1<<OSFIF, 1<<OSFIE  ;paranoid self-check
;	ERRIF ADIF != ADIE, "[ERROR] Int flags != int enable bits (ADIF vs. ADIE)", 1<<ADIF, 1<<ADIE  ;paranoid self-check
;	ERRIF C1IF != C1IE, "[ERROR] Int flags != int enable bits (C1IF vs. C1IE)", 1<<C1IF, 1<<C1IE  ;paranoid self-check
;	ERRIF C2IF != C2IE, "[ERROR] Int flags != int enable bits (C2IF vs. C2IE)", 1<<C2IF, 1<<C2IE  ;paranoid self-check
;	and8 PIE1, PIR1, ISR_PI1  ;mask off unwanted interrupts
;	SETBANK 0, MAYBE  ;set to known bank; ISR functions assume bank 0
	if (MY_INTCON & (1<<PEIE)) && ((MY_PIE1 & (1<<TXIE)))  ;|| (MY_RCSTA & (1<<SPEN)) || (MY_TXSTA & (1<<TXEN)))  ;MIGHT want EUSART transmit interrupts
;		ifbit ISR_PI1, TXIF, TRUE, ISR_HANDLE TX_ISR,  ;every 87 usec or longer, but only during Renard output
		ifbit PIR1, TXIF, TRUE, TX_ISR,, BIGSTMT_HINT  ;every 87 usec or longer, but only during Renard output
	endif
	if (MY_INTCON & (1<<PEIE)) && ((MY_PIE1 & (1<<RCIE)))  ;|| (MY_RCSTA & (1<<SPEN)) || (MY_RCSTA & (1<<CREN)))  ;MIGHT want EUSART receive interrupts
;		ifbit ISR_PI1, RCIF, TRUE, ISR_HANDLE RX_ISR,  ;every 87 usec or longer, but only during Renard output
		ifbit PIR1, RCIF, TRUE, RX_ISR,, BIGSTMT_HINT  ;every 87 usec or longer, but only during Renard output
	endif
	if (MY_INTCON & (1<<PEIE)) && (MY_PIE1 & (1<<TMR1IE))  ;want Timer 1 interrupts
;		ifbit ISR_PI1, TMR1IF, TRUE, ISR_HANDLE TMR1_ISR_nojitter,  ;50 Hz
		ifbit PIR1, TMR1IF, TRUE, ZCFallBack_ISR,, BIGSTMT_HINT  ;50 Hz
	endif
	if (MY_INTCON & (1<<PEIE)) && (MY_PIE1 & (1<<EEIE))  ;want EE write interrupts
;		ifbit ISR_PI1, EEIF, TRUE, ISR_HANDLE EE_ISR,  ;EE write complete
		ifbit PIR1, EEIF, TRUE, EE_ISR,, BIGSTMT_HINT  ;EE write complete
	endif
	if (MY_INTCON & (1<<PEIE)) && (MY_PIE1 & (1<<OSFIE))  ;want osc fail monitor interrupts
;		ifbit ISR_PI1, OSFIF, TRUE, ISR_HANDLE OSF_ISR,  ;infrequent; maybe 1x only near startup, or never
		ifbit PIR1, OSFIF, TRUE, OSF_ISR,, BIGSTMT_HINT  ;infrequent; maybe 1x only near startup, or never
	endif
	if (MY_INTCON & (1<<PEIE)) && (MY_PIE1 & (1<<ADIE))  ;want A/D conversion interrupts
;		ifbit ISR_PI1, ADIF, TRUE, ISR_HANDLE AD_ISR,  ;A/D conversion complete (unused)
		ifbit PIR1, ADIF, TRUE, AD_ISR,, BIGSTMT_HINT  ;A/D conversion complete (unused)
	endif
	if (MY_INTCON & (1<<PEIE)) && (MY_PIE1 & (1<<C1IE))  ;want comparator 1 interrupts
;		ifbit ISR_PI1, C1IF, TRUE, ISR_HANDLE C1_ISR,  ;Comparator 1 output changed (unused)
		ifbit PIR1, C1IF, TRUE, C1_ISR,, BIGSTMT_HINT  ;Comparator 1 output changed (unused)
	endif
	if (MY_INTCON & (1<<PEIE)) && (MY_PIE1 & (1<<C2IE))  ;want comparator 2 interrupts
;		ifbit ISR_PI1, C2IF, TRUE, ISR_HANDLE C2_ISR,  ;Comparator 2 output changed (unused)
		ifbit PIR1, C2IF, TRUE, C2_ISR,, BIGSTMT_HINT  ;Comparator 2 output changed (unused)
	endif
	ISR_UNKNOWN  ;unknown reason for interrupt
 jumpdest isr_done:,,,, FALLTHRU
#if PERF_STATS
	inc24 ISR_COUNT24  ;#times ISR was called; do this after ISR processing to reduce timing jitter
#endif
#if WANT_DEBUG
;NOTE: this arithmetic assumes that ISR will never take more than 256 ticks * 1:2 prescalar = 512 instructions
	sub8 TMR0, ISR_STARTTIME, ISR_STARTTIME,  ;length of time spent in ISR (including adjustment for entry/exit logic)
	add32_8 ISR_TOTTIME, ISR_STARTTIME, ISR_TOTTIME,  ;total time spent
	if8 ISR_STARTTIME, LT, ISR_MINTIME, ;, FALSE
	if8 ISR_STARTTIME, LT, ISR_MINTIME, ADDWF_wrapper(ISR_MINTIME, TRUE),, NOJUMP_HINT  ;keep track of minimum time taken
	if8 ISR_STARTTIME, GE, ISR_MAXTIME, ADDWF_wrapper(ISR_MAXTIME, TRUE),, NOJUMP_HINT  ;keep track of maximum value seen; NOTE: use >= rather than > for more compact code
#endif  ;WANT_DEBUG
	mov8 ISR_FSR, FSR  ;restore FSR (only needs to be saved if set within ISR)
	mov8 ISR_PCLATH, PCLATH  ;restore PCLATH register; main might have been using it
	swap_ ISR_STATUS, STATUS  ;restore status bits (page 124)
	copy_ ISR_W, WREG  ;restore accumulator WITHOUT altering status bits
	nearretfie realISR  ;return, re-enable interrupts
#else  ;USE_INTS

;Cooperative multi-tasking is now used rather than interrupts (to reduce timing jitter and avoid excessive ISR overhead).
;This function *must* be called 1x at least every 1/256th of ZC in order to provide the correct timing, *before* the time required.
;Channel control gets priority at the start of each timeslot, and then misc other tasks will occur after that during idle time (ZC, RTC, Renard, etc).
;Any extra time is burned using a busy-wait (IDLE_COUNT32 keeps track of this, for performance monitoring).
;After the main T0IF timing event occurs, lower priority tasks are performed; this allows them to be completed *before* the next critical time event, but they still get done.
;A tight busy-wait is used to poll T0IF when waiting for the next timeslot.
;This gives lower Timer 0 latency than an ISR: typically ~ 2 instr, compared to 10 - 20 instr for an ISR (lots of registers to save/restore since ISR can happen at any time).

;When over-clocking the SSRs, there is not enough spare time between fractional timeslots to do all other processing, so it's divided up and a different portion done after each fractional timeslot.
;1x timeslots ~= 32.5 usec (60 Hz AC) or ~ 39 usec (50 Hz AC), which is 150 or 180 instr at 18.432 MHz or ~65 or ~78 instr at 8 MHz.
;2x would be 75 or 90 instr at 18.432 MHz, or ~32 or ~34 instr at 8 MHz.
;4x would be ~37 or 45 instr at 18.432 MHz, or ~16 or ~17 instr at 8 MHz; not attainable at 8 MHz.
;2x is about the maximum we can get out of a PIC16F688 (splitting the SSR control pins across 2 ports costs some performance).

	FATAL(0, 0, "dummy")  ;catch jump to interrupt vector; also forces initial FATAL code expansion (too big to expand in-line as nested macro parameter - causes symbol problems in MPASM)
	ZC_ISR FALSE  ;cycle start code; placed up here to avoid needing to jump around it

;cooperative multi-tasking (YIELD) calls in to here:
;TODO: allow run-time selection of overclock freq
	Renard_ISR  ;saves Renard state before handling hardware events
#if SSR_OVERCLOCK > 2
	SSR_ISR 2  ;second fractional timeslot (3x and 4x only)
#endif  ;SSR_OVERCLOCK > 2
;not needed		ifbit PIR1, OSFIF, TRUE, OSF_ISR,, BIGSTMT_HINT  ;infrequent; maybe 1x only near startup, or never
#if SSR_OVERCLOCK == 4
	SSR_ISR 3  ;third fractional timeslot (4x only)
#endif  ;SSR_OVERCLOCK == 4
	ifbit PIR1, TMR1IF, TRUE, Timer1_ISR,, BIGSTMT_HINT  ;50 Hz; used for RTC and fall-back ZC
#if SSR_OVERCLOCK > 1
	SSR_ISR SSR_OVERCLOCK  ;last fractional timeslot (2x, 3x and 4x only)
#endif  ;SSR_OVERCLOCK > 1
	UGLY_PASS12FIX -1  ;KLUDGE: assembler needs help resolving addresses
	ZC_ISR TRUE  ;ZC sample+detect
	SSR_ISR 1  ;full or first fractional timeslot
	UGLY_PASS12FIX -1  ;KLUDGE: assembler needs help resolving addresses
	Renard_ISR  ;restores Renard state before returning to state machine
#endif  ;USE_INTS


#if PERF_STATS
	BDCL yield_count:4
#endif

;cooperative multi-tasking hook:
;Since interrupts are not used in this version, other threads must be called cooperatively thru here.
;Other tasks *must* return before another character arrives in EUSART, otherwise a saturated Renard data stream will drop data.
;At 115K baud, this would be every ~87 usec, which is over 2.5x the 1/256th ZC rate, which provides plenty of safety.
;params:
;  input = flag telling whether doing input or output
YIELD MACRO input
	NOEXPAND  ;hide clutter in LST file
	EXPAND_PUSH FALSE
#if USE_INTS  ;nothing to do but wait
	inc32 IDLE_COUNT32  ;keep track of how busy we're not
#else  ;USE_INTS
#if PERF_STATS
	inc32 yield_count
#endif
#if !IDE_DEBUG  ;real I/O
	EXPAND_PUSH TRUE
	call_ CoopISR  ;handle hardware events (cooperative)
	EXPAND_POP,
#else  ;IDE_DEBUG
	if input  ;only yield for input, skip output
		EXPAND_PUSH TRUE
		call_ CoopISR  ;handle hardware events (cooperative)
;simulate input stream:
;this can be used to simulate a hard-coded data packet from Vixen
		BDCL instream
		BITDCL instream_init
		EXPAND_POP,  ;hide clutter in LST file
		ifbit BITVAR(instream_init), TRUE,, clear_ instream,  ;initialize input stream counter
		setbit BITVAR(instream_init), TRUE
		VARIABLE gotoc_start = $
		GotocSetup $
		loadw_ instream
		inc8 instream
;PCLATH<4:3> must be correct to get there, but PCLATH<2:0> must also be correct for the computed goto inside UnpackTopRow
;since PCLATH<4:3> and eventually PCLATH<2:0> need to be set, here we just set the entire PCLATH value
		add8 PCL, WREG, PCL,  ;computed goto on instream (pre-increment)
		retlw RENARD_SYNC  ;start of packet
		retlw RENARD_CMD  ;Renard data packet
		retlw 0  ;config
		retlw 0x1a  ;brightness
#if FALSE  ;for testing buffer overflow handling
		dec8 instream  ;infinite loop on this byte
		retlw 0xff  ;non-zero value so it looks like data is still coming
#endif
		retlw 0x2a  ;data (row index)
		retlw 0x3a  ;data (columns)
		retlw 0x4a  ;data (columns)
		retlw 0x5a  ;data (columns)
#if TRUE  ;for testing console commands intermixed with Renard packets
		retlw 0  ;end of list
		retlw 1  ;junk (should be discarded)
		retlw 2  ;junk (should be discarded)
		retlw RENARD_SYNC  ;start of next packet
		retlw A'd'  ;dump command (debug/console command)
		retlw 5  ;bad char (should beeep+ignore)
		retlw A'\r'  ;end of command
		retlw 3  ;junk (should be discarded)
		retlw RENARD_SYNC  ;start of next packet
		retlw RENARD_CMD  ; Renard data packet
#endif
		retlw 0  ;config
		retlw 0x8  ;brightness
		retlw 0x2a  ;data (row index)
		retlw 0x3a  ;data (columns)
		retlw 0x4a  ;data (columns)
		retlw 0x5a  ;data (columns)
#if TRUE  ;return nulls for remainder of packet
		dec8 instream  ;prevent going past end of list
		retlw 0  ;end of list
#endif
		ERRIF2 REGPAGEOF($) != REGPAGEOF(gotoc_start), [ERROR] Computed goto spans pages: starts #v(gotoc_start), ends #v($)  ;must start and end on same page; pad if necessary
	endif
#endif  ;IDE_DEBUG
#endif  ;USE_INTS
	EXPAND_POP,
	ENDM


;initialize hardware event data and enable interrupts (if applicable):
;Various timing controls and other registers are set here.
;NOTE: interrupts are OFF at power-up (page 17)
 jumpdest ISR_init:,,,,
;set initial timer values:
;#if 1
;	if EXT_CLK_FREQ
;		mov16 _LITERAL(Timer1_Preset(EXT_CLK_FREQ, DC_REFRESH)), TMR1  ;make it tick at the right time
;	else
;		mov16 _LITERAL(Timer1_Preset(INT_OSC_FREQ, DC_REFRESH)), TMR1  ;make it tick at the right time
;	endif
;#else
;	mov16 _LITERAL(Timer1_Preset(INT_OSC_FREQ, DC_REFRESH)), TMR1  ;make it tick at the right time
	mov16 IIF(EXT_CLK_FREQ, T1Pre_EXT, T1Pre_INT), TMR1  ;make it tick at the right time
;#endif
	mov8 _LITERAL(RTC_Preset(DC_REFRESH)), RTC_20msec  ;set RTC to tick at the right time
	mov8 _LITERAL(0), TMR0  ;allow extra time before first tick for initialization
	mov16 _LITERAL(0), RTC_SECONDS16
;initialize ZC handling:
	mov8 _LITERAL(0), ZC_RATE  ;haven't seen any ZC yet
	mov8 _LITERAL(1), ZC_PREVRATE  ;pretend we saw ZC last time (prevents Timer 1 from supplying ZC while we watch for ZC)
	mov8 _LITERAL(0), ZC_EDGE  ;no ZC seen yet
;initialize chipiplexing:
	Chipiplex -1  ;initialize SSR control pins
	StartCycle FALSE  ;init cycle status; frame buffer set elsewhere
;enable interrupts (do this AFTER all interrupt control registers are set):
	mov8 _LITERAL(0), PIR1  ;NOTE: clear all interrupt flags before enabling interrupts (page 19)
	mov8 _LITERAL(MY_PIE1), PIE1  ;NOTE: set this before global interrupt enable
	mov8 _LITERAL(MY_INTCON), INTCON  ;NOTE: don't set this until other interrupt-related registers are set
	PerfStats MAYBE  ;initialize performance stats
#if USE_INTS
	retfie_ ISR_init  ;pop stack before enabling interrupts
#else  ;USE_INTS
	return_ ISR_init  ;don't enable interrupts
#endif


;******************************************************************************
;main protocol handler section:
;This implements a blended Renard and debug/console I/O handler.
;The "state" maintained by the state machine is mainly just PCL.  A few other variables are also used.
;This makes the code a lot simpler than using a state-driver ISR or hardware event handler.
;******************************************************************************

;Renard special chars:
	CONSTANT RENARD_PAD = 0x7D  ;discard this byte; used by host PC to prevent Tx overrun?
	CONSTANT RENARD_SYNC = 0x7E  ;sync byte; first byte in packet; NOTE: this can be entered in debug console using tilde ("~") character
	CONSTANT RENARD_ESC = 0x7F  ;take next char as-is (no special meaning)
	CONSTANT RENARD_CMD = 0x80  ;command/address byte; second byte in packet
	CONSTANT VIXEN_ADRS = RENARD_CMD + 0x7F  ;define a pseudo-address for Vixen so that packets can be sent to it just like a PIC


	BDCL MyAdrs  ;my position within PIC chain; 0 => last one in chain (numbered in reverse order)

;initialization:
 jumpdest Renard_init:,,,,
#if DEBUG_TTY
	MSGQUE INBUF, 3+1+1+3  ;console input command mini-buffer; longest command is ###R=###
	InitQue INBUF
#endif
#if IO_STATS
	BDCL GoodRenPackets24:3  ;count good Renard packets
	BDCL BadRenPackets  ;count bad Renard packets

	mov24 _LITERAL(0), GoodRenPackets24
	mov8 _LITERAL(0), BadRenPackets
#endif
	mov8 _LITERAL(0), MyAdrs  ;no other addresses seen yet, so assume I'm the only one
	mov8 _LITERAL(0), CfgByte  ;start with default frame buffer configuration
	FBSELECT 1, _LITERAL(LOW(FB1_END - FB1_START))  ;set FSR/IRP to start of frame buffer 1
	mov8 FSR, FBPtr  ;set ptr to start of active frame buffer
#if WANT_DEBUG
;set some initial channels (for debug only):
	mov8 _LITERAL(1<<CFGBIT_PSEUDO_PWM), CfgByte
;on PK1 board, LEDs 2+4+7 are on row RA2: RA4+RA5+RA1->RA2
	push8 _LITERAL(255)  ;full brightness
	push8 _LITERAL(RCbit(1) | RCbit(6))  ;turn on row RA2 (for PK1 board) and row RC1
	push8 _LITERAL(RCbit(0) | RCbit(3) | RCbit(2))  ;turn on LED2, 4, 7 (for PK1 board)
	push8 _LITERAL(RCbit(7))  ;RC1->RC0
	push8 _LITERAL(250)  ;near-full brightness
	push8 _LITERAL(RCbit(0) | RCbit(4))  ;turn on rows 0 and 4
	push8 _LITERAL(RCbit(2) | RCbit(6))  ;turn on columns 2 and 6 of row 0
	push8 _LITERAL(RCbit(5) | RCbit(6))  ;turn on columns 5 and 6 of row 4
	push8 _LITERAL(128)  ;half brightness
	push8 _LITERAL(RCbit(1))  ;turn on row 1
	push8 _LITERAL(RCbit(3))  ;turn on column 3 of row 1
	push8 _LITERAL(64)  ;quarter brightness
	push8 _LITERAL(RCbit(7))
	push8 _LITERAL(RCbit(6))
	push8 _LITERAL(1)  ;almost full off
	push8 _LITERAL(RCbit(1))  ;turn on row 1; should be ignored (dup)
	push8 _LITERAL(RCbit(5) | RCbit(7))  ;turn on columns 5 and 7 of row 1
#endif
	mov8 _LITERAL(0), INDF  ;end-of-list marker
 	TTY_init
	return_ Renard_init


;audible alert (used for debug console input only):
;Since beeps can be annoying, this will beep at most 1x/second.
 jumpdest beep:,,, OFBANK(0)|,
	ifbit BITVAR(EVT_GOT1SEC), TRUE, PUTCHAR _LITERAL(A'\a'),, BIGSTMT_HINT  ;bell 0x7
	setbit BITVAR(EVT_GOT1SEC), FALSE  ;don't beep too often (only once per sec)
	return_ beep


;adjust dimmer values:
;This allows the dimming to be smoothed out over the entire dimming range.
#if FALSE  ;no longer needed; Vixen 2.5 supports this
DimmerCurve:
	REPEAT 256, da repeater  ;^1  ;TODO: linearize these values; maybe also drop out top and bottom 20%?

DimmingCurve MACRO raw, adjusted
#if 0  ;read from EEPROM
	EEREAD8 _LITERAL(DimmerCurve), raw, adjusted
#else  ;read from prog memory (hard-coded)
	PROGREAD16 _LITERAL(DimmerCurve), raw, EEDAT
	mov8 EEDAT, adjusted
#endif
	ENDM
#endif


#if VIXEN_PLUGIN_BUGGY
	BDCL REN_ROWMASK  ;mask off row bit in column data from Vixen
#endif
	BDCL REN_TEMPS24:3  ;NOTE: can reuse this memory if not doing Renard I/O
	CBLOCK REN_TEMPS24
	REN_INCHAR  ;char in buf/temp
	REN_INROWS  ;row index for incoming display list
	REN_INLEN  ;in buf len
	ENDC


;	UGLY_PASS12FIX 1  ;KLUDGE: assembler needs help to sync up addresses
	GETCHAR 0  ;force initial GETCHAR code expansion; avoids extraneous jump over this code (saves 2 instr at run time)

;start of main protocol loop and state machine:
;NOTE: 2 protocols (Renard and debug console) state machines are intermixed.
;See http://www.christmasinshirley.com/wiki/index.php?title=Renard#Protocol for Renard protocol details.
;The "state variables" in this state machine are actually just PCL and a few temp variables; this simplifies the code considerably.
;Uses 2 stack levels max.
;1 char at 115K baud is ~ 87 usec, and ZC/256 is ~ 32.6 usec (60 Hz AC) or ~ 39 usec (50 Hz AC), so at least 2.5 SSR timeslots occur for *each* input char.
;This code must be executed at least 1x every 87 usec to ensure that no chars are dropped from the input stream, but that is easily done.
 jumpdest main:,,,,
	GETCHAR REN_INCHAR  ;looking for sync
#if AUTO_DETECT_BAUD
	CONSTANT REDUCE_56k_to_19k = _LITERAL(SPBRG_Preset(INT_OSC_FREQ, 19200))  ;line too long
	CONSTANT REDUCE_115k_to_38k = _LITERAL(SPBRG_Preset(EXT_CLK_FREQ, 38400))  ;line too long
	CONSTANT REDUCE_56k = _LITERAL(SPBRG_Preset(INT_OSC_FREQ, 57600))  ;line too long
;try to auto-detect baud rate the first time:
;We only need to check the commonly used baud rates: 115k, 56k, 38k, and 19k; anything slower than that is not very useful for a Vixen/Renard system.
;SYNC 0x7e is the first character that is expected; EUSART sends/receives LSB first, so this should come is as 0 (start bit), 0, 1, 1, 1, 1, 1, 1, 0, 1 (stop bit).
;SPBRG is initially set to the highest choice, and then we look at what was received to decide if that was correct or adjust the baud rate.
;For example, if SYNC is sent at half speed (ie, sent at 56k but received at 115k, or sent at 19k but received at 38k), each bit is doubled so we will get:
; 0 (first half of start bit), 0 (second half of start bit used as data), 0, 0, 1, 1, 1, 1, 1, 1 (half of data bit used as stop bit), 1, ... = 0xf8 with NO frame error.
;Initial baud rate is 56k or 115k, depending on clock speed.
;Combinations of the baud rates that we are interested in are as follows:
; condition:  SYNC sent at: received at:             will look like:                          frame error?
; 1/2 speed      56k            115k     0 (start), 0, 0, 0, 1, 1, 1, 1, 1, 1 (stop) = 0xf8       no
; 1/3 speed      38k            115k     0 (start), 0, 0, 0, 0, 0, 1, 1, 1, 1 (stop) = 0xe0       no
; 1/6 speed      19k            115k     0 (start), 0, 0, 0, 0, 0, 0, 0, 0, 0 (stop) = 0x00       yes; eventually we'll see 0 (start), 0, 0, 0, 0, 0, 1, 1, 1, 1 (stop) = 0xe0 as the SYNC finally finishes
; 2x speed      115k             56k     0 (start), 1, 1, 1, ?, ?, ?, ?, ?, ? (stop) = 0x?7/f     maybe
; 2/3 speed      38k             56k     0 (start), 0, 0, 1, 1, 1, 1, 1, 1, 1 (stop) = 0xfc       no
; 1/3 speed      19k             56k     0 (start), 0, 0, 0, 0, 0, 1, 1, 1, 1 (stop) = 0xe0       no
; 1/2 speed      19k             38k     0 (start), 0, 0, 0, 1, 1, 1, 1, 1, 1 (stop) = 0xf8       no
;This is an adaptation/simplification of a trick i saw somewhere on the web; sorry, i didn't write it down at the time; if you know, please let me know so i can add a refernce to it.
	START_CASE REN_INCHAR  ;analyze first char received
		CASE EQ, _LITERAL(RENARD_SYNC), neargoto got_sync,  ;using the correct baud rate already
		CASE EQ, _LITERAL(0xF8), neargoto rcv_speed_12,  ;got half a SYNC char
		CASE EQ, _LITERAL(0xFC), neargoto rcv_speed_23,  ;got 2/3 SYNC char
		CASE EQ, _LITERAL(0xE0), neargoto rcv_speed_13_or_16,  ;got 1/3 or 1/6 SYNC char
	END_CASE
#if IO_STATS
	inc8_ifnz JUNK_COUNT,, dec8 JUNK_COUNT, NOJUMP_HINT  ;don't let it wrap
#endif
	goto_ main  ;discard junk; wait for something recognizable
;1/2 speed => 2x SPBRG; (115k@18.4M: 9 => 56k@18.4M: 19, 3 => 8, 29 => 59, 12 => 25):
 jumpdest rcv_speed_12:, REN_INCHAR|,,,
	mov8 _LITERAL(SPBRG_Preset(EXT_CLK_FREQ, 57600)), SPBRG  ;there's only one case of interest
	goto_ flush_input
;2/3 speed => 1.5x SPBRG ( 19 => 29, 56k@8M: 8 => 38k@8M: 12):
 jumpdest rcv_speed_23:, REN_INCHAR|,,,
	mov8 _LITERAL(SPBRG_Preset(INT_OSC_FREQ, 38400)), SPBRG  ;there's only one case of interest
	goto_ flush_input
;1/3 speed => 3x SPBRG (115k@18.4M: 9 => 38k@18.4M: 29, 3 => 12, 19 => 59, 56k@8M: 8 => 19k@8M: 25):
 jumpdest rcv_speed_13_or_16:, REN_INCHAR|,,,
;if we're using a 18.432MHz clock, using a baud rate of 19.2k is overkill, so ignore this case; this removes ambiguity of SPBRG values
	if8 SPBRG, EQ, REDUCE_56k, loadw_ REDUCE_56k_to_19k, loadw_ REDUCE_115k_to_38k, BOTHOK_HINT  ;only 2 cases of interest here
	storew_ SPBRG
 jumpdest flush_input:,,,, FALLTHRU
	REPEAT 3, loadw_ RCREG  ;flush input buffer
;baud rate change with int osc seems to need to reset the serial port?? (ext clock doesn't seem to need this)
	setbit RCSTA, SPEN, FALSE  ;turn off receive to reset FERR (page 92)
	setbit RCSTA, SPEN, TRUE  ;turn on receive again
#else  ;AUTO_DETECT_BAUD
	if8 REN_INCHAR, NE, _LITERAL(RENARD_SYNC), neargoto tty_in,,  ;switch to debug console
#endif  ;AUTO_DETECT_BAUD
 jumpdest got_sync:,,,, FALLTHRU
	PUTCHAR _LITERAL(RENARD_SYNC)  ;echo sync char to down-stream controllers
	GETCHAR REN_INCHAR  ;looking for Command byte
	START_CASE REN_INCHAR
		CASE EQ, _LITERAL(RENARD_SYNC), neargoto got_sync,  ;redundant sync; still start of packet
;?		CASE EQ, _LITERAL RENARD_PAD, neargoto got_sync,
		CASE EQ, _LITERAL(VIXEN_ADRS), neargoto pass_thru,  ;packet is addressed to Vixen; don't alter it
		CASE EQ, _LITERAL(RENARD_CMD), neargoto GetRenardFrame, COMPARE_HINT  ;a Renard packet for this PIC
	END_CASE
	ifbit LTZ, neargoto tty_after_sync,,  ;getting a non-Vixen debug console command?
	if8 MyAdrs, LT, WREG, SUBWF_wrapper(MyAdrs, F),,  ;use highest command address seen to determine my position within the PIC chain
	dec8 REN_INCHAR  ;command is for a downstream PIC
#if !DEBUG_TTY  ;just Renard, no debug/console commands
 jumpdest tty_after_sync:,,, FALLTHRU
 jumpdest tty_in:,,,, FALLTHRU
#endif
;pass thru junk characters in case they mean something to somebody downstream:
 jumpdest pass_thru:,,,, FALLTHRU
	PUTCHAR REN_INCHAR
	GETCHAR REN_INCHAR
	if8 REN_INCHAR, EQ, _LITERAL(RENARD_SYNC), neargoto got_sync,,
	goto_ pass_thru

;received Renard COMMAND byte (0x80), so this packet is for me:
 jumpdest GetRenardFrame:, REN_INCHAR|,,,
;?	PUTCHAR REN_INCHAR  ;should Renard Command byte be echoed?
; messg "[TODO] should Renard command byte be echoed?"
;at 115k baud, 80 chars will take ~ 7 msec, or ~ 14 msec at 56k, so 20 msec is plenty of time to allow for the next char (or even the entire packet) to be received
	GETCHAR REN_INCHAR TIMEOUT_20msec_start  ;processing options for frame buffer; NOTE: config byte is assumed to be the same for all frame buffers; if it changes it will take effect mid-way thru a frame, which will cause frame distortion if the config change affects SSR rendering
	ifbit BITVAR(EVT_20msec), TRUE, neargoto discard_junk,,  ;reset state if packet times out
	ifbit REN_INCHAR, CFGBIT_MACRO_MODE, TRUE, neargoto macro_mode,,  ;packet is a macro function, not a display list
	mov8 REN_INCHAR, CfgByte  ;protect CfgByte until we decide that this is a data packet
;todo	ifbit CfgByte, CFGBIT_ECHO_PALETTE, TRUE, PUTCHAR CfgByte,, BIGSTMT_HINT  ;echo RGB color palette to down-stream controllers
	setbit BITVAR(FBSwapRequest), FALSE  ;paranoid: cancel any pending frame buffer swap; Renard frames can only come every 25 msec (50 msec more typically); SSR refresh cycle is every ~ 8.3 - 20 msec, so there can't be more than one in progress at a time anyway
	mov8 _LITERAL(FB1_END - FB1_START), REN_INLEN  ;set frame buffer size limit
;data stream from Vixen plug-in: brightness, row index, row, row, ..., brightness, ..., 0
 jumpdest GetNextBrightness:,,,, FALLTHRU
#if VIXEN_PLUGIN_BUGGY
	mov8 _LITERAL(0xff), REN_ROWMASK  ;don't strip off any bits
#endif
	nearcall GetDisplayListByte  ;get next brightness level, in decreasing order of brightness
	if8 REN_INCHAR, EQ, _LITERAL(0), neargoto DisplayListEnd,,  ;end of list or packet timeout
;smooth out brightness level via dimming curve:
#if FALSE  ;not needed; Vixen 2.5.0.8 has this
	DimmingCurve MACRO REN_INCHAR, INDF
#endif  ;FALSE
	nearcall GetDisplayListByte  ;row index for this brightness level
	mov8 REN_INCHAR, REN_INROWS
#if VIXEN_PLUGIN_BUGGY
	mov8 _LITERAL(RCbit(0)), REN_ROWMASK  ;keep track of which row is being processed
#endif
	UGLY_PASS12FIX 4  ;KLUDGE: assembler needs help to sync up addresses
	UGLY_PASS12FIX690 2  ;KLUDGE: assembler needs help after farcall to console forward reference above
 jumpdest GetRow:,,,, FALLTHRU
#if WANT_DEBUG  ;paranoid check (debug)
	if8 REN_INCHAR, EQ, _LITERAL(0), FATAL(0xd0, REN_INLEN, "Useless row/columns"),, BIGSTMT_HINT  ;there should be at least one bit on, otherwise there's no purpose for sending it
#endif
	if8 REN_INCHAR, NE, _LITERAL(0), neargoto SetRow,,
;premature end of list or packet timeout:
	ifbit BITVAR(EVT_20msec), TRUE, neargoto discard_junk,,  ;timeout has already been counted (if IO_STATS is on); discard this frame and move on
#if IO_STATS
	inc8_ifnz BadRenPackets,, dec8 BadRenPackets, NOJUMP_HINT  ;don't let it wrap
#endif
	goto_ discard_junk
 jumpdest SetRow:, REN_INCHAR|,,,
	if8 REN_INROWS, EQ, _LITERAL(0), neargoto GetNextBrightness,,
	shiftl8 REN_INROWS
	ifbit CARRY, nearcall GetDisplayListByte,,  ;column values for this row
#if VIXEN_PLUGIN_BUGGY
	shiftr8 REN_ROWMASK
#endif
	goto_ GetRow

 jumpdest DisplayListEnd:, REN_INCHAR|,,,  ;end of display list
	FBSELECT FALSE, REN_INLEN  ;set FSR/IRP to current place in input (non-active) frame buffer
	mov8 _LITERAL(0), INDF  ;set end-of-list indicator in case we didn't get one already; there's always room at the end of the frame buffer for this byte
;NOTE: Renard frames can only come every 25 msec (50 msec more typically); refresh cycle is every 8.3 - 20 msec, so there can't be more than one in progress at a time
	setbit BITVAR(FBSwapRequest), TRUE  ;swap in new frame buffer at start of next SSR refresh cycle
#if IO_STATS
	inc24 GoodRenPackets24
#endif
 jumpdest discard_junk:,,,, FALLTHRU
	GETCHAR REN_INCHAR
;?	ifbit BITVAR(EVT_20msec), TRUE, goto_ main,,  ;reset state if packet times out
	if8 REN_INCHAR, EQ, _LITERAL(RENARD_SYNC), neargoto got_sync,,
;?	if8 REN_INCHAR, EQ, _LITERAL(RENARD_PAD), goto_ passthru
#if IO_STATS
	inc8_ifnz JUNK_COUNT,, dec8 JUNK_COUNT, NOJUMP_HINT  ;don't let it wrap
#endif
;no:	PUTCHAR REN_INCHAR
	goto_ discard_junk

 jumpdest GetDisplayListByte:, REN_INROWS|,,,
	mov8 _LITERAL(0), REN_INCHAR  ;set char to 0 in case we don't get one
	if8 REN_INLEN, EQ, _LITERAL(0), return_ GetDisplayListByte,, NOJUMP_HINT  ;avoid buffer overrun
	GETCHAR REN_INCHAR TIMEOUT_20msec_continue
	ifbit BITVAR(EVT_20msec), TRUE, return_ GetDisplayListByte,, NOJUMP_HINT  ;receive timed out
	FBSELECT FALSE, REN_INLEN  ;set FSR/IRP to current place in input (non-active) frame buffer
#if VIXEN_PLUGIN_BUGGY
	and8 REN_INCHAR, REN_ROWMASK, INDF  ;paranoid; strip off row bit in case Vixen plug-in didn't do it
#else
	mov8 REN_INCHAR, INDF  ;take it as-is; assume Vixen plug-in is doing what it should
#endif
	dec8 REN_INLEN  ;update count of remaining chars
	return_ GetDisplayListByte

;data packet is not a display list:
 jumpdest macro_mode:, REN_INCHAR|,,,
;TODO: add other "macro" functions here
;	if8 REN_INCHAR, EQ, _LITERAL macro_function_code, goto_ handle_macro_function,,
	if8 REN_INCHAR, NE, _LITERAL(CFGMACRO_CONSOLE), neargoto discard_junk,,  ;unrecognized macro function
	PUTCHAR _LITERAL(VIXEN_ADRS)  ;address the returning packet to Vixen
	GETCHAR REN_INCHAR  ;get first char of debug command
#if DEBUG_TTY  ;console/debug command
 jumpdest tty_after_sync:,,,, FALLTHRU
 jumpdest tty_in:,,,, FALLTHRU
	if8 REN_INCHAR, GE, _LITERAL(0x7f), neargoto tty_special,,
	if8 REN_INCHAR, LT, _LITERAL(A' '), neargoto tty_special,,
 jumpdest tty_keep:,,,, FALLTHRU
	PutMsg(INBUF, REN_INCHAR, MAYBE)  ;append char to console/debug command buffer
	ifbit STATUS, Z, FALSE, PUTCHAR REN_INCHAR, call_ beep,  ;echo char to screen or signal input buffer overflow if full
 jumpdest tty_more:,,,, FALLTHRU
	GETCHAR REN_INCHAR
	if8 REN_INCHAR, NE, _LITERAL(RENARD_SYNC), neargoto tty_in,,
	farcall console  ;process console command
	goto_ got_sync  ;now switch back to Renard
 	UGLY_PASS12FIX 1  ;KLUDGE: assembler needs help after farcall to console forward reference above
	UGLY_PASS12FIX690 1  ;KLUDGE: assembler needs help after farcall to console forward reference above
 jumpdest tty_special:, REN_INCHAR|,,,
	START_CASE REN_INCHAR  ;handle special chars
		CASE EQ, _LITERAL(A'\r'), neargoto tty_CR,
		CASE EQ, _LITERAL(A'\n'), neargoto tty_LF,
		CASE EQ, _LITERAL(A'\b'), neargoto tty_BS,  ;0x8
		CASE EQ, _LITERAL(0x1b), neargoto tty_ESC,
	END_CASE
	call_ beep  ;draw attention to invalid char
	mov8 _LITERAL(A'?'), REN_INCHAR  ;echo bad char as "?"
	goto_ tty_keep
 jumpdest tty_LF:, REN_INCHAR|,,,
	setbit BITVAR(NotListening), TRUE  ;KLUDGE: it looks like another PIC already processed this command, so don't process it again
 jumpdest tty_CR:,,,, FALLTHRU
	PUTCHAR REN_INCHAR
	PUTCHAR _LITERAL(A'\n')
	farcall console  ;process console command
	goto_ tty_more
	UGLY_PASS12FIX 1  ;KLUDGE: avoid error 116 "Address label duplicated or different in second pass"; this will pad during pass 1, but be skipped during pass 2, so that following addresses are consistent between pass 1 + 2
	UGLY_PASS12FIX690 1  ;KLUDGE: assembler needs help after farcall to console forward reference above
 jumpdest tty_BS:, REN_INCHAR|,,,  ;backspace; remove prev input char
	UnputMsg(INBUF, WREG, MAYBE)  ;discard previous input char if any
	ifbit STATUS, Z, TRUE, neargoto tty_empty,,
	PUTCHAR REN_INCHAR
	PUTCHAR _LITERAL(A' ')  ;erase previous char from screen
	PUTCHAR REN_INCHAR
	goto_ tty_more
 jumpdest tty_empty:, REN_INCHAR|,,,
	call_ beep
	goto_ tty_more
 jumpdest tty_ESC:, REN_INCHAR|,,,
	InitQue INBUF  ;clear command buffer
	PUTCHAR _LITERAL(A'\r')  ;place cursor in first column
	goto_ tty_more
#endif  ;DEBUG_TTY


;******************************************************************************
;initialization section:
;StartUp initializes registers + buffers and then jumps to main loop.
;Main loop calls tty input handler for Renard and debug console.
;StartUp code is at the end so that any threads or other symbols are previously defined (avoids forward ref problems with assembler).
;******************************************************************************

;power-up test, then enter normal mode:
 jumpdest StartUp:,,,,
;If we jumped back or fell thru to 0, something's seriously wrong (probably a coding bug or assembler error).
;see STATUS.TO and STATUS.PD description (page 115, 118):
;The only valid resets here should be power-on or brown-out; report all others.
;Power-on reset: STATUS = 00011???, PCON = 0001000?
;WDT reset: STATUS = 0000????, PCON = 000?00??
;Brown-out: 00011???, PCON = 00010010
;PK1 comes up initially with STATUS = 0x1D, or 0x1F after reset or 0xB?
	and8 STATUS, _LITERAL(1<<IRP | 1<<RP1 | 1<<RP0 | 1<<NOT_TO | 1<<NOT_PD), WREG,
;CAUTION: the PIC does not appear to reset while on the PK1 board if the COM port is still connected (due to stray Vss thru RC4/RC5?)
#if PK1DEV+PK1SIM
	iorlw 1<<NOT_TO  ;KLUDGE: PK1 board reset (power off/on via MPLAB) seems to behave like wdt reset?
#endif
	CONSTANT STARTUP_BITS = _LITERAL(0<<IRP | 0<<RP1 | 0<<RP0 | 1<<NOT_TO | 1<<NOT_PD)  ;line too long
	if8 WREG, NE, STARTUP_BITS, FATAL(0xf0, STATUS, "restart: firmware bug or hardware error"),, BIGSTMT_HINT  ;check power-up reset (page 20, 117 of PIC spec)
	and8 PCON, _LITERAL(1<<ULPWUE | 1<<SBODEN), WREG,
	if8 WREG, NE, _LITERAL(0<<ULPWUE | 1<<SBODEN), FATAL(0xf1, PCON, "restart: firmware bug or hardware error"),, BIGSTMT_HINT  ;check power-up reset (page 20, 117 of PIC spec)
	setbit PCON, NOT_POR, TRUE  ;allow non-power up to be detected next time
;	ifbit PCON, NOT_BOR, FALSE, CONTINUE, FATAL(0xed, STATUS, "brownout: firmware bug or hardware error")  ;brown-out reset (page 20)
	mov8 _LITERAL(MY_OSCCON), OSCCON
#if !IDE_DEBUG  ;doesn't work in IDE/simulator
	whilebit OSCCON, HTS, FALSE,,  ;wait for int osc to stabilize
#endif
;data re/initialization:
 jumpdest reinit:,,,, FALLTHRU  ;come here for s/w restart
	SET_PK1LEDS(_LITERAL(0))  ;DEBUG
	CLEAR_BITDCLS  ;clear all BITDCL variables
;set initial control registers/options:
	clrwdt  ;16F688 errata says to clear this before setting PSA/PS0/PS1/PS2
	mov8 _LITERAL(MY_OPTIONS), OPTION_REG
	ANALOG_PINS FALSE  ;turn off A/D, comparators
;??http://www.pic101.com/mcgahee/picuart.zip says that TX must be programmed as an input
;	SETPINS RC4_LOW, TRUE, NOWAIT  ;unnecessary (SPEN sets this; ANSEL must still be set; page 91); set TX output
;	SETPINS RC5_IN, TRUE, NOWAIT  ;unnecessary (SPEN sets this; ANSEL must still be set; page 91); set RX input
;	SETPINS RA5_IN, TRUE, NOWAIT  ;set ext clock line to input; redundant with CONFIG?
	mov8 _LITERAL(MY_WPU#v(0xA)), WPUA  ;enable all weak pull-ups (RA3 only enabled when MCLRE is disabled) (page 37)
	mov8 _LITERAL(MY_IOC#v(0xA)), IOCA  ;interrupt on port A changes (page 37)
	mov8 _LITERAL(MY_T1CON), T1CON
	mov8 _LITERAL(MY_BAUDCTL), BAUDCTL
	mov8 _LITERAL(MY_WDTCON), WDTCON  ;NOTE: this seems to turn on WDT even if it shouldn't be on
;NOTE: I couldn't seem to change the baud rate after FSCM kicked in, so just set it once here at the start (since we already know by now whether or not the ext clock is working)
	ifbit PIR1, OSFIF, TRUE, OSF_ISR, SetSpeed IIF(EXT_CLK_FREQ, EXT_CLK_FREQ, INT_OSC_FREQ),  ;set timers according to whether ext clock is present or not; if ext clock is absent, Fail Safe Clock Monitor already kicked in and we are running from int osc (otherwise the PIC would not be executing this code)
	call_ Renard_init  ;set initial Renard + debug console state machines and data
	call_ ISR_init  ;preload timers, enable interrupts, etc
#if PK1DEV+PK1SIM
	goto_ main_PK1
#else
	goto_ main
#endif


;******************************************************************************
;debug/console command handler and supporting functions:
;The lower code page is reserved for frequently used code.
;Console handling is infrequent/optional, so it is placed in the upper code pages.
;******************************************************************************
#if DEBUG_TTY

;setting lit pool size is trial & error:
;If there's space left at the end of rom, increase this value.
;If you get warning 220 "Address exceeds maximum range for this processor", then decrease it.
; 	LITPOOL MAX_ROM-364+80 - LITPOOL_ADJUST  ;leave some space for literal strings
	LITPOOL 0x46c  ;seems to be a lot of empty space here

	VARIABLE empty_space = $  ;remember where we ended in case we need to come back here
	org LIT_PAGE_SIZE  ;need far jump (PCLATH<4:3>) to get to any functions in here

;force some code expansion here (assember has problems expanding some of this in-line):
;NOTE: this also declares some temp variable space that is reused below.
	DIV32 DIV32_NUM, DIV32_DEN, DIV32_QUO, DIV32_REM  ;generate DIV32 code here, then use it for all divides (slower, but saves code space)
	UGLY_PASS12FIX -3 ;-1  ;KLUDGE: assembler needs help resolving addresses
	PUTNUM8 0, 10  ;expand code once here


;convenience wrapper:
GETNUM8 MACRO reg, want_new_char
	GETNUM8_ reg, want_new_char, call_ getcmdchar, INCHAR
	ENDM


;check if we got a command char:
#define GOT_CHAR(yesno)  STATUS, Z, !(yesno)


;get next input char from command buffer:
;Also converts to upper case, keeps track of column.
 jumpdest getcmdchar:,,, OFBANK(0)|,  ;return with bank 0, since that is where data typically is
	YIELD FALSE  ;handle other hardware events first; they are more time-critical than serial input
	mov8 _LITERAL(0), INCHAR
	GetMsg(INBUF, INCHAR, MAYBE)  ;get next char from command input buffer
	ifbit GOT_CHAR(FALSE), return_ getcmdchar,, NOJUMP_HINT  ;eof
	TOUPPER INCHAR
	inc8 INCOL
	if8 INCHAR, NE, _LITERAL(A' '), return_ getcmdchar,, NOJUMP_HINT
	goto_ getcmdchar  ;skip spaces
	UGLY_PASS12FIX -2  ;KLUDGE: assembler needs help after call to getcmdchar


;try to save a little code space:
;These only save space when they are used *instead* of PUTLITSTR or PUTCHAR, not in *addition* to them.
 jumpdest put_space_lparen:,,, OFBANK(0)|,  ;return with bank 0, since that is where data typically is
	PUTCHAR _LITERAL(A' ')
	PUTCHAR _LITERAL(A'(')
	return_ put_space_lparen

 jumpdest put_rparen_comma_space:,,, OFBANK(0)|,  ;return with bank 0, since that is where data typically is
	PUTCHAR _LITERAL(A')')
 jumpdest put_comma_space:,,, OutBankOf(put_rparen_comma_space)|, FALLTHRU  ;common return must share same out bank
	PUTCHAR _LITERAL(COMMA)  ;A'\,'  ;causes MPASM syntax error unless a symbol is used
 jumpdest put_space:,,, OutBankOf(put_comma_space)|, FALLTHRU  ;common return must share same out bank
	PUTCHAR _LITERAL(A' ')
	return_ put_space


;audible + visual indicator that PK1 is ready:
;This is useful when tty is already connected to the PIC and it is reflashed.
;Also includes other more primitive tests (dev only); uncomment as needed.
 jumpdest main_PK1:,,,,
#if FALSE  ;timing test
	messg [DEV ONLY] timing test
 jumpdest timing:,,,, FALLTHRU
	intenable_ FALSE
	SHOWBYTE _LITERAL(0x3c), 1
	SHOWBYTE _LITERAL(0x5a), 1
	goto_ timing
#endif  ;timing test

#if FALSE  ;LED/timing test
	messg [DEV ONLY] LED/timing test
	mov8 _LITERAL(LOW(~3)), TRISC
	mov8 _LITERAL(LOW(~6)), TRISA
 jumpdest loop1:,,,, FALLTHRU
	revbit STATUS, IRP, ZPRESERVE_HINT
	ifbit STATUS, IRP, TRUE, loadw_ _LITERAL(1), loadw_ _LITERAL(2),
	storew_ PORTC
	ifbit STATUS, IRP, TRUE, loadw_ _LITERAL(2), loadw_ _LITERAL(4),
	storew_ PORTA
	loadw_ _LITERAL(160)
 jumpdest loop2:,,,, FALLTHRU
	clrwdt
	call_ noop12500  ;160 * 12.5 msec = 2 sec
	clrwdt
	call_ noop12500  ;160 * 12.5 msec = 2 sec
	addlw 0xff
	ifbit NEZ, goto_ loop2,,
	goto_ loop1
#endif  ;LED/timing test

#if FALSE  ;output test
	messg [DEV ONLY] output test
 jumpdest loop:,,,, FALLTHRU
	PUTCHAR _LITERAL('A')
	SHOWBYTE _LITERAL(0xcc), 1
	PUTCHAR _LITERAL('B')
	SHOWBYTE _LITERAL(0xc8), 1
	PUTCHAR _LITERAL('C')
	SHOWBYTE _LITERAL(0xc0), 1
	PUTCHAR _LITERAL('D')
	SHOWBYTE _LITERAL(0x80), 1
	PUTLITSTRA2 "abcd\0"
	SHOWBYTE _LITERAL(0), 1
	goto_ loop
#endif  ;output test

#if FALSE  ;baud rate vs. clock adjust test; useful for "finding" the correct baud rate
	messg [DEV ONLY] baud rate vs. clock test
	NBDCL baud_loop16:2
	mov16 _LITERAL(25), baud_loop16
 jumpdest baud_looper:,,,, FALLTHRU
	PUTNUM16 baud_loop16, 10
	PUTLITSTRA2 " hello\n\0"
	setbit BITVAR(EVT_GOT1SEC), FALSE
	whilebit BITVAR(EVT_GOT1SEC), FALSE, YIELD TRUE,  ;wait for 1 sec to elapse

;reset serial port and change baud rate:
	setbit TXSTA, TXEN, FALSE
	setbit RCSTA, CREN, FALSE
	mov16 baud_loop16, SPBRG
	setbit RCSTA, SPEN, TRUE  ;disable serial port
	setbit TXSTA, SYNC, FALSE
	setbit TXSTA, TXEN, TRUE
	setbit RCSTA, CREN, TRUE
	dec16_ifnz baud_loop16, neargoto baud_looper,,
#endif  ;baud rate test

#if FALSE  ;simple echo test
	messg [DEV ONLY] simple echo test
 jumpdest echo_test:,,,, FALLTHRU
	GETCHAR INCHAR
;	if8 INCHAR, NE, _LITERAL A'X',, FATAL(0xff, "test")
	TOUPPER INCHAR
	SET_PK1LEDS(INCHAR)
	PUTCHAR INCHAR
	PUTCHAR _LITERAL('=')
	PUTNUM8 INCHAR, 10
	goto_ echo_test
#endif  ;simple echo test

#if FALSE  ;echo test
	messg [DEV ONLY] echo test
 jumpdest echo_test:,,,, FALLTHRU
;	GETCHAR INCHAR
	mov8 _LITERAL(A'>'), TXREG
	whilebit IsCharAvailable(FALSE), clrwdt,  ;input buffer empty; wait for it to fill
	mov8 RCREG, INCHAR
	TOUPPER INCHAR
;	PUTCHAR INCHAR
	mov8 INCHAR, TXREG  ;do not wait (don't even check); assumes space is available in output buffer
	goto_ echo_test
;	UGLY_PASS12FIX 1  ;-1  ;KLUDGE: avoid error 116 "Address label duplicated or different in second pass"; this will pad during pass 1, but be skipped during pass 2, so that following addresses are consistent between pass 1 + 2
#endif  ;echo test

#if FALSE  ;RGB proximity/PWM test/demo
	messg [DEV ONLY] RGB proximity/PWM test
	NBDCL counter16:2
	NBDCL times
;	BITDCL onoff

	mov8 _LITERAL(0), TRISC
	mov8 _LITERAL(1), PORTC
	mov8 _LITERAL(0x39), TRISA
	mov8 _LITERAL(2), PORTA
	mov8 _LITERAL(61), times  ;60 sec
 jumpdest looper1:,,,, FALLTHRU
	mov16 _LITERAL(50*256+1), counter16  ;1 sec
;	mov16 _LITERAL(1*256+1), counter16  ;1/120 sec
 jumpdest looper2:,,,, FALLTHRU
#if 1
	whilebit INTCON, T0IF, FALSE, clrwdt,  ;busy-wait for timeslot, ~32.6 usec
	PresetTimer0 T0_PRESET  ;rewind Timer 0 to wrap again at the beginning of next timeslot
	setbit INTCON, T0IF, FALSE  ;clear interrupt flag
#else
	clrwdt
	noop_ 65  ;32.5 usec at 8 MHz
#endif
	dec16_ifnz counter16, neargoto looper2,,
	sub8 _LITERAL(3), PORTC, PORTC,
	sub8 _LITERAL(6), PORTA, PORTA,
	dec8_ifnz times, neargoto looper1,,
#endif  ;RGB proximity/PWM test/demo

#if FALSE  ;check memory; 16F688 documentation appears to be wrong regarding IRP + RP1
	messg [DEV ONLY] memory test
	intenable_ FALSE  ;turn off ints
	mov8 _LITERAL(0x20), FSR
 jumpdest bank0setloop:,,,, FALLTHRU
	sub8 FSR, _LITERAL(0x20), WREG
	push8 WREG
	if8 FSR, LT, _LITERAL(0x80), neargoto bank0setloop,

	mov8 _LITERAL(0xa0), FSR
 jumpdest bank1setloop:,,,, FALLTHRU
	sub8 FSR, _LITERAL(0x40), WREG
	push8 WREG
	if8 FSR, LT, _LITERAL(0xf0), neargoto bank1setloop,

	setbit STATUS, IRP, TRUE
	mov8 _LITERAL(0x20), FSR
 jumpdest bank2setloop:,,,, FALLTHRU
	add8 FSR, _LITERAL(0xB0), WREG
	push8 WREG
	if8 FSR, LT, _LITERAL(0x80), neargoto bank2setloop,
	setbit STATUS, IRP, FALSE

	mov8 _LITERAL(0x20), FSR
 jumpdest bank0showloop:,,,, FALLTHRU
	SHOWBYTE INDF, 1
	inc8 FSR
	if8 FSR, LT, _LITERAL(0x80), neargoto bank0showloop,

	mov8 _LITERAL(0xa0), FSR
 jumpdest bank1showloop:,,,, FALLTHRU
	SHOWBYTE INDF, 1
	inc8 FSR
	if8 FSR, NE, _LITERAL(0), neargoto bank1showloop,

	setbit STATUS, IRP, TRUE
	mov8 _LITERAL(0x20), FSR
 jumpdest bank2showloop:,,,, FALLTHRU
	SHOWBYTE INDF, 1
	inc8 FSR
	if8 FSR, LT, _LITERAL(0x80), neargoto bank2showloop,
	setbit STATUS, IRP, FALSE
	goto $-1
#endif  ;check memory; 16F688 documentation appears to be wrong regarding IRP + RP1

	PUTLITSTRA2 "ready\n\0"
	call_ beep
	goto_ main


;CAUTION: memory is reused in the code below!
;DIV32 uses 12 bytes banked + 4 bytes non-banked memory, which is a lot of space, so that space is redefined below.
;Make sure usage is mutually exclusive!
;Basically, when DIV32 is used, many of the temps (anything REDCL'ed) will get smashed.
;Any code here can safely use REN_TEMPS24, since that is only used while a command is being buffered, not executed.
	REDCL INCHAR
	REDCL INCOL  ;input column#
	REDCL SAVE_CMD  ;remember which command is being used
	REDCL CMD_TEMPS16:2  ;re-allocate a few temps for each command; CAUTION: these are overwritten by DIV32
;	BITDCL NotListening  ;flag controlling whether to respond to console commands or just pass them along


;respond to debug/console command:
;Special chars such as backspace and escape have already been handled before this code is called,
;so this code can just take the line input as-is and parse it.
 jumpdest console:,,, OutBankOf(put_crlflf)|,  ;common return must share same out bank
;	UGLY_PASS12FIX 1  ;KLUDGE: avoid error 116 "Address label duplicated or different in second pass"; this will pad during pass 1, but be skipped during pass 2, so that following addresses are consistent between pass 1 + 2
#if FALSE  ;echo test
	PUTLITSTRA2 "You entered: '\0"
 jumpdest inloop:,,,, FALLTHRU
	call_ getcmdchar
	ifbit GOT_CHAR(FALSE), neargoto eol,,
	PUTCHAR INCHAR
	goto_ inloop
 jumpdest eol:,,,,
	PUTLITSTRA2 "\'\n\n\0"
	goto_ status_cmd
	UGLY_PASS12FIX 2  ;KLUDGE: assembler needs help after call to getcmdchar
#endif
;	UGLY_PASS12FIX -1  ;KLUDGE: assembler needs help after call to getcmdchar
	mov8 _LITERAL(0), INCOL  ;reset cursor tracking
	mov8 _LITERAL(0), SAVE_CMD
	call_ getcmdchar
	ifbit GOT_CHAR(FALSE), neargoto hint_cmd,,
	START_CASE INCHAR  ;main command dispatch
		CASE EQ, _LITERAL(A'C'), neargoto clear_cmd,
		CASE EQ, _LITERAL(A'D'), neargoto dump_cmd,
		CASE EQ, _LITERAL(A'E'), neargoto error_cmd,
		CASE EQ, _LITERAL(A'G'), neargoto cfg_cmd,
		CASE EQ, _LITERAL(A'L'), neargoto listen_cmd,
		CASE EQ, _LITERAL(A'M'), neargoto mem_cmd,
		CASE EQ, _LITERAL(A'P'), neargoto playback_cmd,
		CASE EQ, _LITERAL(A'R'), neargoto reset_cmd,
		CASE EQ, _LITERAL(A'S'), neargoto status_cmd,
		CASE EQ, _LITERAL(A'W'), neargoto pwidth_cmd,
		CASE EQ, _LITERAL(A'?'), neargoto help_cmd,
		CASE LT, _LITERAL(A'0'), neargoto invalid_cmd_not_recognized,
		CASE GT, _LITERAL(A'9'), neargoto invalid_cmd_not_recognized,
	END_CASE
	GETNUM8 GETCHANNEL, FALSE
	if8 GETCHANNEL, EQ, _LITERAL(0), neargoto invalid_cmd_bad_value,,
	ifbit CfgByte, CFGBIT_RGB_PALETTE, TRUE, loadw_ _LITERAL(RGB_CHANNELS+1), loadw_ _LITERAL(MONO_CHANNELS+1), BOTHOK_HINT
	if8 GETCHANNEL, GE, WREG, neargoto invalid_cmd_bad_value,,  ;channel# out of range (1-based)
	if8 INCHAR, EQ, _LITERAL(A'='), neargoto setch_cmd,,  ;find channel, then set new value
	if8 INCHAR, EQ, _LITERAL(0), neargoto getch_cmd,,  ;end of command, no R/G/B specified
	goto_ invalid_cmd_bad_value  ;TODO: RGB commands
#if 0  ;TODO: RGB commands
;	if8 INCHAR, EQ, _LITERAL A'-', neargoto getrange_cmd,
;	if8 INCHAR, EQ, _LITERAL A',', neargoto getlist_cmd,
	ifbit CfgByte, CFGBIT_RGB_PALETTE, FALSE, neargoto invalid_cmd_not_rgb,,  ;RGB not valid
	mov8 INCHAR, SAVE_CMD  ;remember R/G/B/= character
	START_CASE INCHAR
		CASE EQ, _LITERAL(A'R'), neargoto getRGB_cmd,
		CASE EQ, _LITERAL(A'G'), neargoto getRGB_cmd,
		CASE EQ, _LITERAL(A'B'), neargoto getRGB_cmd,
	END_CASE
	UGLY_PASS12FIX 2  ;KLUDGE: assembler needs help after call to getcmdchar
;show "invalid command" message and beep:
 jumpdest invalid_cmd:,,,, FALLTHRU
#endif
	UGLY_PASS12FIX 1  ;KLUDGE: assembler needs help after call to getcmdchar
;discard the rest of the command:
 jumpdest discard_cmd:,,,,
	call_ put_crlflf
	call_ beep
 jumpdest discard_loop:,,,, FALLTHRU
	call_ getcmdchar
	ifbit GOT_CHAR(FALSE), return_ console,, NOJUMP_HINT  ;return to main protocol loop
	goto_ discard_loop
;show invalid command message and beep:
 jumpdest invalid_cmd_not_eof:,,,,
	TOCOLUMN INCOL  ;space over to first invalid char
	PUTLITSTRA2 "^ junk at end\0"
	goto_ discard_cmd
 jumpdest invalid_cmd_not_recognized:,,,,
	TOCOLUMN INCOL  ;space over to first invalid char
	PUTLITSTRA2 "^ unknown command\0"
	goto_ discard_cmd
 jumpdest invalid_cmd_not_rgb:,,,,
 jumpdest invalid_cmd_bad_value:,,,, FALLTHRU
	TOCOLUMN INCOL  ;space over to first invalid char
	PUTLITSTRA2 "^ bad value\0"
	goto_ discard_cmd
 jumpdest invalid_cmd_missing_value:,,,,
	TOCOLUMN INCOL  ;space over to first invalid char
	PUTLITSTRA2 "^ missing value\0"
	goto_ discard_cmd


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;set channel/R/G/B command:
;syntax: #[RGB]=#

	CBLOCK CMD_TEMPS16  ;reuse memory, since this function is mutually exclusive
	GETCHANNEL
	SETBRIGHT
	ENDC


;TODO: set R, G, B not complete yet; need to nail down RGB palette handling first
 jumpdest setch_cmd:, INCHAR|,,,
	GETNUM8 SETBRIGHT, TRUE
	ifbit BITVAR(GotNum), FALSE, neargoto invalid_cmd_missing_value,,
	if8 INCHAR, NE, _LITERAL(0), neargoto invalid_cmd_not_eof,,
#if 0
	if8 SAVE_CMD, EQ, _LITERAL(A'='), neargoto skip_palette_check,,
	ifbit CfgByte, CFGBIT_RGB_PALETTE, TRUE, loadw_ _LITERAL(RGB_CHANNELS), loadw_ _LITERAL(MONO_CHANNELS), BOTHOK_HINT
	if8 SETBRIGHT, GE, WREG, neargoto invalid_cmd_bad_value,,  ;channel# (0-based) is past end of frame buffer
 jumpdest skip_palette_check:,,,, FALLTHRU
	setbit BITVAR(FBSwapRequest), FALSE  ;paranoid: cancel any pending frame buffer swap; Renard frames can only come every 25 msec (50 msec more typically); SSR refresh cycle is every ~ 8.3 - 20 msec, so there can't be more than one in progress at a time anyway
#endif
;this just sets 1 channel within an empty frame buffer, rather than trying to do a messy merge with what's there already:
	FBSELECT FALSE, _LITERAL(LOW(FB1_END - FB1_START))  ;set FSR/IRP to start of inactive frame buffer
	push8 SETBRIGHT  ;brightness level
	ChannelToRowBit GETCHANNEL&, WREG
	push8 WREG  ;row index
	ChannelToColumnBit , WREG  ;use cached row/column data
	push8 WREG  ;column value
	mov8 _LITERAL(0), INDF  ;end-of-list marker
	setbit BITVAR(FBSwapRequest), TRUE  ;swap in new frame buffer at start of next SSR refresh cycle
	PUTLITSTRA2 "channel \0"
 jumpdest confirm_update:,,,, FALLTHRU
	PUTNUM8 GETCHANNEL, 10
	if8 SAVE_CMD, NE, _LITERAL(A'='), PUTCHAR SAVE_CMD,, BIGSTMT_HINT  ;R/G/B/channel indicator
	PUTLITSTRA2 " set to \0"
	PUTNUM8 SETBRIGHT, 10
	goto_ put_crlflf


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;get channel/R/G/B command:
;syntax: #[RGB]

;NOTE: also uses CMD_TEMPS16 from above
	CBLOCK REN_TEMPS24  ;reuse memory, since this function is mutually exclusive
	FindRowColBit  ;row or column bit to find
	CheckOrSkipRows  ;row index to check or skip
	CheckOrSkipColumns  ;columns to check or skip
	ENDC
	BITDCL NotTargetRow

 jumpdest getRGB_cmd:, INCHAR|,,,
	call_ getcmdchar
	if8 INCHAR, EQ, _LITERAL(A'='), neargoto setch_cmd,,
	if8 INCHAR, NE, _LITERAL(0), neargoto invalid_cmd_not_eof,,
;TODO: get R, G, B not complete yet; need to nail down RGB palette handling first
 jumpdest getch_cmd:,,,, FALLTHRU
	FBSELECT TRUE, _LITERAL(FB1_END - FB1_START)  ;set FSR/IRP to start of active frame buffer
 jumpdest findrow_loop:,,,, FALLTHRU
	unpush8 SETBRIGHT  ;remember current brightness in case this was the correct entry
	if8 SETBRIGHT, EQ, _LITERAL(0), neargoto NotFound,,  ;end of display list in frame buffer
	unpush8 CheckOrSkipRows  ;row index
	if8 CheckOrSkipRows, EQ, _LITERAL(0), neargoto NotFound,,  ;premature end of display list
	ChannelToRowBit GETCHANNEL&, FindRowColBit  ;row index bit to find
	and8 CheckOrSkipRows, FindRowColBit, WREG,
	bitcopy STATUS, Z, BITVAR(NotTargetRow)  ;remember whether this row index contains target row
 jumpdest skiprow_loop:,,,, FALLTHRU  ;skip through columns for this row index
	shiftl8 CheckOrSkipRows
	ifbit CARRY, unpush8 CheckOrSkipColumns, clear_ CheckOrSkipColumns, BOTHOK_HINT BIGSTMT_HINT  ;CAUTION: assumes else is executed first
	ifbit BITVAR(NotTargetRow), TRUE, neargoto skip_column_check,,  ;don't need to look for target column
	shiftl8 FindRowColBit
	ifbit NOCARRY, neargoto skip_column_check,,  ;column is off
	ChannelToColumnBit , FindRowColBit  ;column bit to find; use cached row/column data
	and8 CheckOrSkipColumns, FindRowColBit, WREG,
	ifbit NEZ, neargoto FoundColumn,,  ;target column is on
 jumpdest skip_column_check:,,,, FALLTHRU
	if8 CheckOrSkipRows, NE, _LITERAL(0), neargoto skiprow_loop,,
	goto_ findrow_loop
 jumpdest NotFound:, SETBRIGHT|,,,
	mov8 _LITERAL(0), SETBRIGHT  ;not found in display list; display it as off
 jumpdest FoundColumn:,,,, FALLTHRU
	PUTNUM8 GETCHANNEL, 10
	if8 SAVE_CMD, NE, _LITERAL(A'='), PUTCHAR SAVE_CMD,, BIGSTMT_HINT  ;R/G/B/channel indicator
	PUTLITSTRA2 " = \0"
	PUTNUM8 SETBRIGHT, 10
	goto_ put_crlflf


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;clear all channels command:
;syntax: c[#]
	CBLOCK CMD_TEMPS16  ;reuse memory, since this function is mutually exclusive
	INLEVEL  ;brightness level to set all channels to
	ENDC

;	UGLY_PASS12FIX 2  ;KLUDGE: assembler needs help
;TODO: auto-inc or auto-dec?
 jumpdest clear_cmd:, INCHAR|,,,
	GETNUM8 INLEVEL, TRUE
	if8 INCHAR, NE, _LITERAL(0), neargoto invalid_cmd_not_eof,,
	setbit BITVAR(FBSwapRequest), FALSE  ;paranoid: cancel any pending frame buffer swap; Renard frames can only come every 25 msec (50 msec more typically); SSR refresh cycle is every ~ 8.3 - 20 msec, so there can't be more than one in progress at a time anyway
	FBSELECT FALSE, _LITERAL(LOW(FB1_END - FB1_START))  ;set FSR/IRP to start of inactive frame buffer
#if FALSE  ;set some initial channels (for debug):
;PK1BOARD uses 2x150 ohm = (5V-.6V-.7V-1.6V)/300 ohm ~= 7 mA; 1 PIC I/O should be able to drive 3 LEDs
;LED0 0x01 (RA4->RA5/RA0)  RA: 0/z 1 z z z z/0  (1 z z 0 x x x x after bit swap, RA5 = row 7) = ch# 3*7+1
;LED1 0x02 (RA5/RA0->RA4)  RA: 1/z 0 z z z z/1  (0 z z 1 x x x x after bit swap, RA4 = row 0) = ch# 0*7+3
;LED2 0x04 (RA4->RA2)  RA: z 1 z 0 z z  (1 0 z z x x x x after bit swap, RA2 = row 5) = ch# 1*7+1
;LED3 0x08 (RA2->RA4)  RA: z 0 z 1 z z  (0 1 z z x x x x) = ch# 0*7+1
;LED4 0x10 (RA5/RA0->RA2)  RA: 1/z z z 0 z z/1  (z 0 z 1 x x x x) = ch# 1*7+3
;LED5 0x20 (RA2->RA5/RA0)  RA: 0/z z z 1 z z/0  (z 1 z 0 x x x x) = ch# 3*7+2
;LED6 0x40 (RA2->RA1)  RA: z z z 1 0 z  (z 1 0 z x x x x, RA1 = row 6) = ch# 2*7+2
;LED7 0x80 (RA1->RA2)  RA: z z z 0 1 z  (z 0 1 z x x x x) = ch# 1*7+2
;-\+   RC3  RC2  RC1  RC0  RA4  RA2  RA1  RA0/5
;RC3    --    0    1    2    3    4    5    6
;RC2     7   --    8    9   10   11   12   13
;RC1    14   15   --   16   17   18   19   20
;RC0    21   22   23   --   24   25   26   27
;RA4    28   29   30   31   --   32   33   34
;RA2    35   36   37   38   39   --   40   41
;RA1    42   43   44   45   46   47   --   48
;RA0/5  49   50   51   52   53   54   55   --
;row4:32,34,row5:39,40,41,row6:47,row7:53,54
;#if PK1DEV
;PK1 DEV board instrumentation only handles first row
;on PK1 board, LEDs 2+4+7 are on row RA2: RA4+RA5+RA1->RA2
;RA=LED4 LED2, z(LED2) row2 LED7 z(LED4)
	push8 _LITERAL(255)  ;full brightness
	push8 _LITERAL(RCbit(6))  ;turn on row RA2 (for PK1 board) and RC1 + RC0
	push8 _LITERAL(RCbit(7))
	push8 _LITERAL(128)  ;half brightness
	push8 _LITERAL(RCbit(7))
	push8 _LITERAL(RCbit(6))
	mov8 _LITERAL(0), INDF  ;end-of-list marker
#endif  ;FALSE
	if8 INLEVEL, EQ, _LITERAL(0), neargoto no_testpattern,,
;for simplicity, this display list just uses 8 rows/level; trailing rows will be slightly dimmer, but that should be okay for testing purposes
	push8 INLEVEL  ;set brightness
	push8 _LITERAL(0xff)  ;select all 8 rows; this uses 8 rows/level in order to simplify this code and also to test the timeslot rescheduling
	REPEAT 8, push8 _LITERAL(LOW(~RCbit(repeater)))  ;turn on all 7 columns for each row
 jumpdest no_testpattern:,,,, FALLTHRU
	mov8 _LITERAL(0), INDF  ;end-of-list marker
;NOTE: Renard frames can only come every 25 msec (50 msec more typically); refresh cycle is every 8.3 - 20 msec, so there can't be more than one in progress at a time
	setbit BITVAR(FBSwapRequest), TRUE  ;swap in new frame buffer at start of next SSR refresh cycle
	PUTLITSTRA2 "cleared to \0"
	PUTNUM8 INLEVEL, 10
	goto_ put_crlflf


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;dump framebuffer command:
;syntax: d[d]
	CBLOCK CMD_TEMPS16  ;reuse memory, since this function is mutually exclusive
	DUMPLEN  ;#bytes to dump from framebuffer
	DUMPROWS  ;row index currently being dumped
	ENDC

	CBLOCK REN_TEMPS24  ;reuse memory, since this function is mutually exclusive
	DUMPSTARTCOL  ;column# for first column in current row
	DUMPSKIPROW  ;row#/column# to skip for this row
	DUMPCOL  ;current column#
	ENDC

	BITDCL DUMP_FIRSTBR  ;delimiter handling
	BITDCL DUMP_DIDROW  ;delimiter handling
	BITDCL DUMP_FIRSTCOL  ;delimiter handling

;	UGLY_PASS12FIX 1  ;KLUDGE: assembler needs help with addresses
 jumpdest dump_cmd:, INCHAR|,,,
	call_ getcmdchar
	if8 INCHAR, NE, _LITERAL(A'D'), neargoto no_dump_params,,  ;optional char: selects inactive vs. active framebuffer
	mov8 _LITERAL(LOW(FB1_START ^ FB2_START)), SAVE_CMD  ;use inactive rather than active framebuffer
	call_ getcmdchar
 jumpdest no_dump_params:,,,, FALLTHRU
	if8 INCHAR, NE, _LITERAL(0), neargoto invalid_cmd_not_eof,,
#if PK1DEV+PK1SIM  ;show PK1 LED channel mapping
;LED0 0x01 (RA4->RA5/RA0)  RA: 0/z 1 z z z z/0  (1 z z 0 x x x x after bit swap, RA5 = row 7) = ch# 3*7+1
;LED1 0x02 (RA5/RA0->RA4)  RA: 1/z 0 z z z z/1  (0 z z 1 x x x x after bit swap, RA4 = row 0) = ch# 0*7+3
;LED2 0x04 (RA4->RA2)  RA: z 1 z 0 z z  (1 0 z z x x x x after bit swap, RA2 = row 5) = ch# 1*7+1
;LED3 0x08 (RA2->RA4)  RA: z 0 z 1 z z  (0 1 z z x x x x) = ch# 0*7+1
;LED4 0x10 (RA5/RA0->RA2)  RA: 1/z z z 0 z z/1  (z 0 z 1 x x x x) = ch# 1*7+3
;LED5 0x20 (RA2->RA5/RA0)  RA: 0/z z z 1 z z/0  (z 1 z 0 x x x x) = ch# 3*7+2
;LED6 0x40 (RA2->RA1)  RA: z z z 1 0 z  (z 1 0 z x x x x, RA1 = row 6) = ch# 2*7+2
;LED7 0x80 (RA1->RA2)  RA: z z z 0 1 z  (z 0 1 z x x x x) = ch# 1*7+2
;-\+   RC3  RC2  RC1  RC0  RA4  RA2  RA1  RA0/5
;RC3    --    0    1    2    3    4    5    6
;RC2     7   --    8    9   10   11   12   13
;RC1    14   15   --   16   17   18   19   20
;RC0    21   22   23   --   24   25   26   27
;RA4    28   29   30   31   --   32   33   34
;RA2    35   36   37   38   39   --   40   41
;RA1    42   43   44   45   46   47   --   48
;RA0/5  49   50   51   52   53   54   55   --
	PUTLITSTRA2 "PK1LEDs (1-based): 0=22, 1=3, 2=8, 3=1, 4=10, 5=23, 6=16, 7=9\n\0"
#endif
	mov8 _LITERAL(FB1_END - FB1_START), DUMPLEN  ;set dump ofs to start of frame buffer
	setbit BITVAR(DUMP_FIRSTBR), TRUE  ;first-time flag
 jumpdest DumpNextBrightness:,,,, FALLTHRU
	ifbit BITVAR(DUMP_FIRSTBR), FALSE, PUTLITSTRA2 "]\n\0",, BIGSTMT_HINT
	setbit BITVAR(DUMP_FIRSTBR), FALSE
	PUTLITSTRA2 "dim \0"
	nearcall GetDumpListByte  ;next brightness level
	PUTNUM16 FSR, 16
	PUTLITSTRA2 "(\'\0"
	xor8 DUMPLEN, SAVE_CMD, WREG,  ;switch over to other framebuffer address
	FBSELECT TRUE, WREG  ;set FSR/IRP to current place in active/inactive frame buffer
	ifbit ISFB(FSR, 1), loadw_ _LITERAL(LOW(FB1_START + 1)), loadw_ _LITERAL(LOW(FB2_START + 1)), BOTHOK_HINT
	sub8 FSR, WREG, WREG,  ;offset relative to start of framebuffer
	PUTNUM8 WREG, 10
	PUTLITSTRA2 "): \0"
	if8 INCHAR, EQ, _LITERAL(0), neargoto DumpListEnd,,  ;0 = end-of-list indicator
	PUTNUM8 INCHAR, 10  ;brightness level
	PUTLITSTRA2 ", rows \0"
	nearcall GetDumpListByte  ;row index for this brightness level
	if8 INCHAR, EQ, _LITERAL(0), neargoto DumpListEnd,,  ;this shouldn't happen, but if it does, treat it as end-of-list indicator
	PUTNUM8 INCHAR, 16  ;row index
	PUTLITSTRA2 ", channels \0"
	mov8 INCHAR, DUMPROWS  ;remember which rows we are looking at
	mov8 _LITERAL(LOW(1 - SSR_PIN_COUNT)), DUMPSTARTCOL
	mov8 _LITERAL(0), DUMPSKIPROW
	setbit BITVAR(DUMP_DIDROW), FALSE  ;first-time flag
	UGLY_PASS12FIX 4  ;KLUDGE: assembler needs help to sync up addresses
	UGLY_PASS12FIX690 2  ;KLUDGE: assembler needs help after farcall to console forward reference above
 jumpdest DumpNextRow:,,,, FALLTHRU
	add8 DUMPSTARTCOL, _LITERAL(SSR_PIN_COUNT - 1), DUMPSTARTCOL,  ;first column# for this row
	inc8 DUMPSKIPROW  ;keep track of which column# to skip within this row (column is used as common cathode, so we need to skip over it when dumping columns)
	if8 DUMPROWS, EQ, _LITERAL(0), neargoto DumpNextBrightness,,  ;no more rows
	ifbit BITVAR(DUMP_DIDROW), TRUE, PUTLITSTRA2 "], \0",, BIGSTMT_HINT
	setbit BITVAR(DUMP_DIDROW), FALSE  ;first-time flag
	shiftl8 DUMPROWS  ;look at next row index bit
	ifbit NOCARRY, neargoto DumpNextRow,,
	nearcall GetDumpListByte  ;columns for this row
	if8 INCHAR, EQ, _LITERAL(0), neargoto DumpListEnd,,  ;this shouldn't happen, but if it does, treat it as end-of-list indicator
	PUTNUM8 INCHAR, 16  ;columns
	setbit BITVAR(DUMP_FIRSTCOL), TRUE  ;first-time flag
	mov8 _LITERAL(LOW(-1)), DUMPCOL
	setbit BITVAR(DUMP_DIDROW), TRUE
	UGLY_PASS12FIX 2  ;KLUDGE: assembler needs help to sync up addresses
	UGLY_PASS12FIX690 1  ;KLUDGE: assembler needs help after farcall to console forward reference above
 jumpdest DumpNextColumn:,,,, FALLTHRU
	inc8 DUMPCOL
	if8 INCHAR, EQ, _LITERAL(0), neargoto DumpNextRow,,  ;no more columns
	shiftl8 INCHAR  ;look at next column bit
	ifbit NOCARRY, neargoto DumpNextColumn,,
	ifbit BITVAR(DUMP_FIRSTCOL), TRUE, PUTLITSTRA2 " [\0", PUTLITSTRA2 ", \0",
	setbit BITVAR(DUMP_FIRSTCOL), FALSE
	if8 DUMPCOL, GE, DUMPSKIPROW, loadw_ _LITERAL(0), loadw_ _LITERAL(1), BOTHOK_HINT  ;adjust for skipped column
	add8 DUMPCOL, WREG, WREG,
	add8 DUMPSTARTCOL, WREG, WREG,
	PUTNUM8 WREG, 10  ;column#
	goto_ DumpNextColumn

;get next framebuffer byte if there is still data to dump:
 jumpdest GetDumpListByte:, OFBANK(0)|,,,  ;use bank 0, since that is where data typically is
	mov8 _LITERAL(0), INCHAR
	if8 DUMPLEN, EQ, _LITERAL(0), return_ GetDumpListByte,, NOJUMP_HINT
	xor8 DUMPLEN, SAVE_CMD, WREG,  ;switch over to other framebuffer address
	FBSELECT TRUE, WREG  ;set FSR/IRP to current place in active/inactive frame buffer
	mov8 INDF, INCHAR  ;FSR will be altered during PUTCHAR, so need to save INDF here
	dec8 DUMPLEN  ;keep track of how much data remains
	return_ GetDumpListByte

 jumpdest DumpListEnd:, INCHAR|,,,  ;end of display list
	PUTLITSTRA2 "0 END\0"
	goto_ put_crlflf


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;error test:
;syntax: e[#]
;mainly used for debug, to exercise error reporting code
	CBLOCK CMD_TEMPS16  ;reuse memory, since this function is mutually exclusive
	ErrCode
	ENDC

 jumpdest error_cmd:, INCHAR|,,,
	GETNUM8 ErrCode, TRUE
	if8 INCHAR, NE, _LITERAL(0), neargoto invalid_cmd_not_eof,,
	FATAL(0x5a, ErrCode, "Testing fatal error handler")
	return_ console


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;config command:
;syntax: g[#]

 jumpdest cfg_cmd:, INCHAR|,,,
	GETNUM8 INLEVEL, TRUE  ;optional config byte value
	if8 INCHAR, NE, _LITERAL(0), neargoto invalid_cmd_not_eof,,
	ifbit BITVAR(GotNum), TRUE, MOV8RV RVpack(INNUM, CfgByte),, BIGSTMT_HINT  ;don't change CfgByte until command syntax was validated
	PUTLITSTRA2 "config = \0"
	PUTNUM8 CfgByte, 10
	goto_ put_crlflf


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;hint command (kinda useless):
 jumpdest hint_cmd:, INCHAR|,,,
	PUTLITSTRA2 "type \"?\" for help\0"
	goto_ put_crlflf


;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;listen command:
;syntax: l[#]
;use this command to tell other PICs in a multi-PIC controller to ignore console commands

	BITDCL NotListening  ;remember whether to listen to console commands

#if 0  ;TODO: not working yet
 jumpdest notlistening:,,,,
	ifbit GOT_CHAR(FALSE), return_ console,, NOJUMP_HINT
	if8 INCHAR, NE, _LITERAL(A'L'), return_ console,, NOJUMP_HINT
 jumpdest listen_cmd:,,,, FALLTHRU
	call_ getcmdchar
	ifbit GOT_CHAR(FALSE), neargoto invalid_listen_cmd,,
	GETNUM8 INNUM, FALSE
	ifbit GOT_CHAR(TRUE), neargoto invalid_listen_cmd,,
	if8 MyAdrs, EQ, INNUM, Listen TRUE, Listen FALSE, BOTHOK_HINT
	ifbit BITVAR(NotListening), TRUE, return_ console,, NOJUMP_HINT
	PUTLITSTRA2 "Listening \0"
	PUTNUM8 MyAdrs, 10
	goto_ put_crlf
 jumpdest invalid_listen_cmd:,,,,
	ifbit BITVAR(NotListening), TRUE, return_ console,, NOJUMP_HINT
	goto_ invalid_cmd
#endif

 jumpdest listen_cmd:, INCHAR|,,,
	GETNUM8 INNUM, TRUE
	if8 INCHAR, NE, _LITERAL(0), neargoto invalid_cmd_not_eof,,
	PUTNUM8 INNUM, 10
	PUTLITSTRA2 " listening\0"
	goto_ put_crlflf
#if 0  ;TODO: not working yet
	Listen FALSE
	call_ getcmdchar
	ifbit GOT_CHAR(FALSE), neargoto no_listen_param,,
	GETNUM8 INNUM, FALSE
	ifbit GOT_CHAR(TRUE), neargoto invalid_cmd,,
	if8 MyAdrs, EQ, INNUM, Listen TRUE, Listen FALSE, BOTHOK_HINT
 jumpdest no_listen_param:,,,, FALLTHRU
;	ifbit BITVAR(NotListening), TRUE, return_,, NOJUMP_HINT
	PUTLITSTRA2 "Listening \0"
	PUTNUM8 MyAdrs, 10
	call_ put_crlf
 jumpdest listen_loop:,,,,
	GETCHAR INCHAR
	PUTCHAR INCHAR
	if8 INCHAR, EQ, _LITERAL(4), return_ console,, NOJUMP_HINT  ;^D to end tunnelling
	goto_ listen_loop
#endif


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;memory/register command:
;syntax: m#=[#]

;NOTE: uses same temps as get/set channel command
 jumpdest mem_cmd:, INCHAR|,,,
	GETNUM8 GETCHANNEL, TRUE
	if MAX_RAM < 0xff
		if8 GETCHANNEL, GT, _LITERAL(MAX_RAM), neargoto invalid_cmd_bad_value,,
	endif
	if8 INCHAR, EQ, _LITERAL(A'='), neargoto update_mem,,
	if8 INCHAR, NE, _LITERAL(0), neargoto invalid_cmd_not_eof,,
	ifbit BITVAR(GotNum), FALSE, neargoto invalid_cmd_missing_value,,
	mov8 GETCHANNEL, FSR
	setbit STATUS, IRP, FALSE  ;just look at lower memory addresses; Dump command can be used to look at upper memory
	mov8 INDF, SETBRIGHT
	PUTLITSTRA2 "mem/reg \0"
	BANKCHK SETBRIGHT, MAYBE,
	goto_ FoundColumn
 jumpdest update_mem:, INCHAR|,,,
	GETNUM8 SETBRIGHT, TRUE
	if8 INCHAR, NE, _LITERAL(0), neargoto invalid_cmd_not_eof,,
	ifbit BITVAR(GotNum), FALSE, neargoto invalid_cmd_missing_value,,
	mov8 GETCHANNEL, FSR
	setbit STATUS, IRP, FALSE  ;just look at lower memory addresses; Dump command can be used to look at upper memory
	mov8 SETBRIGHT, INDF
	PUTLITSTRA2 "mem/reg \0"
	goto_ confirm_update


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;playback command:
	CBLOCK REN_TEMPS24  ;reuse memory, since this function is mutually exclusive
	chipi_row  ;currently active row
	chipi_col  ;currently active column
	start_time
	ENDC

;for now this is just a chipiplex SSR test:
;cycles through all row/col combinations one column at a time, with 1/2 second on each
 jumpdest playback_cmd:, INCHAR|,,,
	call_ getcmdchar
	ifbit GOT_CHAR(TRUE), neargoto invalid_cmd_not_eof,,
;first cycle through each row/column, one at a time:
	PUTLITSTRA2 "cols \0"
	mov8 _LITERAL(RCbit(0)), chipi_row  ;first row
 jumpdest row_loop:,,,, FALLTHRU
	mov8 _LITERAL(RCbit(0)), chipi_col  ;first column
	mov8 _LITERAL(A'0'), INCHAR  ;visual column indicator
 jumpdest col_loop:,,,, FALLTHRU
	nearcall fbsetup
	shiftr8 chipi_col  ;next column for this row
	if8 chipi_col, NE, _LITERAL(0), neargoto col_loop,,
	shiftr8 chipi_row  ;next row
	if8 chipi_row, NE, _LITERAL(0), neargoto row_loop,,
	PUTLITSTRA2 "\nrows \0"
;next cycle through each rows, turning on all columns at the same time:
	mov8 _LITERAL(RCbit(0)), chipi_row  ;first row
	mov8 _LITERAL(0xff), chipi_col  ;all columns
	mov8 _LITERAL(A'0'), INCHAR  ;visual row indicator
	UGLY_PASS12FIX 2  ;KLUDGE: assembler needs help to sync up addresses
	UGLY_PASS12FIX690 1  ;KLUDGE: assembler needs help after farcall to console forward reference above
 jumpdest rowcol_loop:,,,, FALLTHRU
	nearcall fbsetup
	shiftr8 chipi_row  ;next row
	if8 chipi_row, NE, _LITERAL(0), neargoto rowcol_loop,,
	PUTLITSTRA2 "\ndone\0"
	goto_ put_crlflf

;set up framebuffer and display it for 1/2 sec:
	UGLY_PASS12FIX 2  ;KLUDGE: assembler needs help to sync up addresses
	UGLY_PASS12FIX690 1  ;KLUDGE: assembler needs help after farcall to console forward reference above
 jumpdest fbsetup:, OFBANK(0)|,,,  ;use bank 0, since that is where data typically is
	FBSELECT FALSE, _LITERAL(LOW(FB1_END - FB1_START))  ;set FSR/IRP to start of inactive frame buffer
	push8 _LITERAL(255)  ;set brightness
	push8 chipi_row  ;row index
	xor8 chipi_row, _LITERAL(0xff), WREG,
	and8 chipi_col, WREG, WREG,  ;avoid row/column conflict
	ifbit EQZ, neargoto skip_col,,  ;no columns left to turn on
	push8 WREG  ;column for this row
	mov8 _LITERAL(0), INDF  ;end-of-list marker
	setbit BITVAR(FBSwapRequest), TRUE  ;swap in new frame buffer at start of next SSR refresh cycle
;now wait a half second:
	PUTCHAR INCHAR
	mov8 RTC_20msec, start_time
 jumpdest chipi_delay:,,,, FALLTHRU
	YIELD 0
	sub8 RTC_20msec, start_time, WREG,  ;subtract and then compare compensates for wrap around
	ifbit LTZ, addlw RTC_Preset(DC_REFRESH),, NOJUMP_HINT  ;compensate for wrap
;	if8 WREG, LT, _LITERAL 500/20, goto_ chipi_delay,,
	add8 WREG, _LITERAL(256 - 500/20), WREG,  ;add negative to avoid operand swap
	ifbit NOCARRY, neargoto chipi_delay,,
 jumpdest skip_col:,,,, FALLTHRU
	inc8 INCHAR
	return_ fbsetup


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;reset stats or processor command:
;syntax: r[r]
 jumpdest reset_cmd:, INCHAR|,,,
	call_ getcmdchar
	ifbit GOT_CHAR(FALSE), neargoto no_reset_params,,
	if8 INCHAR, NE, _LITERAL(A'R'), neargoto invalid_cmd_not_recognized,,  ;optional char to reset entire processor
	call_ getcmdchar
	ifbit GOT_CHAR(TRUE), neargoto invalid_cmd_not_eof,,
	goto_ reinit  ;full processor reset

 jumpdest no_reset_params:, OutBankOf(getcmdchar)|,,,
	PerfStats MAYBE  ;initialize performance stats
	call_ Renard_init  ;this just clears tty I/O stats
	mov16 _LITERAL(0), RTC_SECONDS16  ;reset elapsed time counter
	PUTLITSTRA2 "stats reset\0"
	goto_ put_crlflf


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;show stats/status command:
;CAUTION: DIV overwrites all temps used in console (16 bytes)
;can't use REDCL here

	CBLOCK REN_TEMPS24  ;reuse memory, since this function is mutually exclusive
	SVMINSEC16:2
	SVSEC
	ENDC

 jumpdest status_cmd:, INCHAR|,,,
	call_ getcmdchar
	ifbit GOT_CHAR(TRUE), neargoto invalid_cmd_not_eof,,
	PUTLITSTRA2 "Renard-HC " VERSION "\0"  ;NOTE: first part of string must be even length for string concatenation to work here
	PUTLITSTRA2 "\nTime: \0"
;32-bit divide is slower, but will avoid the need for a separate copy of 16-bit divide function:
	DIV16_as32 RTC_SECONDS16, _LITERAL(60*60), DIV32_QUO, DIV32_REM
	mov16 DIV32_REM, SVMINSEC16  ;remainder = minutes + seconds; save for later (DIV32 will overwrite)
	PUTNUM8 DIV32_QUO, 10  ;hours
	PUTCHAR _LITERAL(A':')
;32-bit divide is slower, but will avoid the need for a separate copy of 16-bit divide function:
	DIV16_as32 SVMINSEC16, _LITERAL(60), DIV32_QUO, DIV32_REM
	mov8 DIV32_REM, SVSEC  ;remainder = seconds
	PUTNUM8_NN DIV32_QUO, 10  ;minutes
	PUTCHAR _LITERAL(A':')
	PUTNUM8_NN SVSEC, 10
	call_ put_space_lparen  ;" ("
	PUTNUM16 RTC_SECONDS16, 10  ;show actual elapsed time counter, in case the above arithmetic is messed up
	call_ put_rparen_comma_space  ;"), "
	if EXT_CLK_FREQ && (EXT_CLK_FREQ != INT_OSC_FREQ)  ;check for ext clock
		ifbit USING_INTOSC(TRUE), PUTLITSTRA2 "8MHz int (failover)\0", PUTLITSTRA2 "18MHz ext\0",
	else
		PUTLITSTRA2 "8MHz int\0"
	endif
	PUTLITSTRA2 ", ts1 (\0"
	ifbit USING_INTOSC(TRUE), MOV16RV RVpack(T1Pre_INT, OUTNUM32), MOV16RV RVpack(T1Pre_EXT, OUTNUM32),  ;ext oscillator failed/absent, running from int osc
	PUTNUM16 OUTNUM32, 10
	call_ put_rparen_comma_space  ;"), "

#if BAUD_RATE
	CONSTANT baud_div_INT = RVpack(_LITERAL(rdiv(InstrPerSec(INT_OSC_FREQ), BRSCALE)), DIV32_NUM)  ;line to long
	CONSTANT baud_div_EXT = RVpack(_LITERAL(rdiv(InstrPerSec(EXT_CLK_FREQ), BRSCALE)), DIV32_NUM)  ;line to long

	PUTLITSTRA2 "baud \0"
	ifbit USING_INTOSC(TRUE), MOV24RV baud_div_INT, MOV24RV baud_div_EXT,  ;get clock speed
	add16 SPBRG, _LITERAL(1), DIV32_DEN,
	DIV24by16_as32 DIV32_NUM, DIV32_DEN, DIV32_QUO, DIV32_REM  ;clock/(SPBRG+1)
	PUTNUM24 DIV32_QUO, 10  ;actual baud rate
	call_ put_space_lparen  ;" ("
	PUTNUM16 SPBRG, 10  ;show SPBRG in case the above arithmetic is messed up
	call_ put_rparen_comma_space  ;"), "
#endif  ;BAUD_RATE
	PUTLITSTRA2 "adrs: \0"
	PUTNUM8 MyAdrs, 10  ;address relative to tail end of PIC chain (numbered backwards)
	ifbit BITVAR(NotListening), FALSE, PUTCHAR _LITERAL(A'L'),, BIGSTMT_HINT  ;show listen indicator

	PUTLITSTRA2 "\nZC: \0"
	loadw_ ZC_PREVRATE
	ifbit EQZ, loadw_ _LITERAL(DC_REFRESH),, NOJUMP_HINT
	PUTNUM8 WREG, 10
	if8 ZC_PREVRATE, EQ, _LITERAL(0), PUTLITSTRA2 " fallback\0",, BIGSTMT_HINT  ;0 => using DC_REFRESH rate
#if ZC_TRACKING
	PUTLITSTRA2 ", min/max \0"
	PUTNUM8 ZC_LOWEST, 10
	PUTCHAR _LITERAL(A'/')
	PUTNUM8 ZC_HIGHEST, 10
	PUTLITSTRA2 ", width \0"
	PUTNUM8 ZC_PREVWIDTH, 10  ;previous ZC pulse width seen
#endif  ;ZC_TRACKING
	PUTLITSTRA2 ", ts0 \0"
	mov8 T0_PRESET, TMR0  ;restart Timer 0
	mov8 TMR1L, SVSEC
	whilebit INTCON, T0IF, FALSE,,  ;wait for Timer 0 to wrap
	sub8 TMR1L, SVSEC, SVSEC,  ;#Timer 1 ticks it took Timer 0 to wrap; auto-corrects for wrap
	Timer1Ticks2uSec SVSEC, OUTNUM32
	PUTNUM8 OUTNUM32, 10
	call_ put_space_lparen  ;" ("
	PUTNUM8 T0_PRESET, 10
	PUTLITSTRA2 "), ovcl " tostr_#v(SSR_OVERCLOCK) "\0"  ;NOTE: first part of string must be even length for string concatenation to work here

#if IO_STATS
	PUTLITSTRA2 "\nTTY: good \0"
	PUTNUM24 GoodRenPackets24, 10
	PUTLITSTRA2 ", bad \0"
	PUTNUM8 BadRenPackets, 10
	PUTLITSTRA2 ", char \0"
	PUTNUM24 IO_COUNT24, 10
	PUTLITSTRA2 ", ovr \0"
	PUTNUM8 OERR_COUNT, 10
	PUTCHAR _LITERAL(A'@')
	PUTNUM24 LAST_OERR24, 10  ;show when last overrun occurred
	PUTLITSTRA2 ", frm \0"
	PUTNUM8 FERR_COUNT, 10
	PUTCHAR _LITERAL(A'@')
	PUTNUM24 LAST_FERR24, 10  ;show when last frame error occurred
	PUTLITSTRA2 ", timo \0"
	PUTNUM8 TIMEOUT_COUNT, 10
	PUTLITSTRA2 ", junk \0"
	PUTNUM8 JUNK_COUNT, 10
#endif

#if PERF_STATS
	PUTLITSTRA2 "\nidle \0"
	PUTNUM32 IDLE_COUNT32, 10
	call_ put_space_lparen  ;" ("
	mov16 _LITERAL(0), DIV32_DEN+2
	mov16 RTC_SECONDS16, DIV32_DEN
	DIV32 IDLE_COUNT32, DIV32_DEN, DIV32_QUO, DIV32_REM
	mov32 DIV32_QUO, DIV32_NUM  ;#idle/sec
	ifbit USING_INTOSC(TRUE), MOV32RV RVpack(_LITERAL(IDLE_SCALE(INT_OSC_FREQ)), DIV32_DEN), MOV32RV RVpack(_LITERAL(IDLE_SCALE(EXT_CLK_FREQ)), DIV32_DEN),  ;get clock speed
	DIV32 DIV32_NUM, DIV32_DEN, DIV32_QUO, DIV32_REM
	PUTNUM16 DIV32_QUO, 10
	PUTLITSTRA2 "%), isr \0"
	PUTNUM24 ISR_COUNT24, 10
	PUTLITSTRA2 "\nrow setup \0"
	PUTNUM32 SetupTime32, 10
	PUTCHAR _LITERAL(A'/')
	PUTNUM24 SetupCount24, 10
	PUTLITSTRA2 " = \0"
	mov8 _LITERAL(0), DIV32_DEN+3
	mov24 SetupCount24, DIV32_DEN
	DIV32 SetupTime32, DIV32_DEN, DIV32_QUO, DIV32_REM
	PUTNUM32 DIV32_QUO, 10
	UGLY_PASS12FIX 1  ;KLUDGE: assembler needs help
#endif

#if FALSE  ;info about current place in framebuffer
	PUTLITSTRA2 ", thisbr \0"
	PUTNUM8 ThisBrightness, 10
	PUTLITSTRA2 ", rinx \0"
	PUTNUM8 RowIndex, 16
	PUTLITSTRA2 ", outbufs \0"
	PUTNUM16 Outbuf10, 16
	PUTCHAR _LITERAL(A' ')
	PUTNUM16 Outmask10, 16
	PUTLITSTRA2 ", fbptr \0"
	PUTNUM8 FBPtr, 16
#endif
	goto_ put_crlflf


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;SSR pulse width test:
;This will fade and ramp some channels using varying timeslot sizes.
;The smallest timeslot that does not cause flicker is the optimal timeslot size.

	CBLOCK REN_TEMPS24  ;reuse memory, since this function is mutually exclusive
	PulseWidth  ;pulse width being tested (in uints of ~ .65 usec at 18MHz or 1.5 usec at 8 MHz)
	PulseDelayCounter  ;delay loop counter
	TestChannel  ;user-selected test channel#, then row bit for test channels
	ENDC

	CBLOCK CMD_TEMPS16  ;reuse memory, since this function is mutually exclusive; CAUTION: these are overwritten by DIV32
	DimTest  ;brightness value within fade/ramp loop
	pulse_time  ;try to use Timer1 to measure actual pulse width used
	ENDC

	BITDCL Pulse_UpDown  ;keep track of ramp vs. fade
	BITDCL ShowPulse  ;only show pulse width measurement 1x/loop

;need to set these for reliable Triac turn-on at both ends of AC half-cycle:
;otherwise, we can't tell if the flicker is due to opto control pulses too short vs. insufficient voltage for Triac turn-on
#define PWIDTH_MAXBRIGHT  240  ;255  ;above this may flicker (voltage too low for triacs to latch at start of AC half-cycle)
#define PWIDTH_MINBRIGHT  20  ;5  ;below this may flicker due to insufficient voltage to latch triac at end of AC half-cycle

;delay loop timing:
#define DelayLoopTime10(clock)  Instr2uSec(6*10, clock)  ;10 * delay loop time per iteration in usec (6 instr per iteration)
#define SetupTime10(clock)  Instr2uSec(16*10, clock)  ;10 * setup overhead time before getting into delay loop (10+6 instr)

	CONSTANT DelayLoopTime10_INT = DelayLoopTime10(INT_OSC_FREQ)  ; line too long
	CONSTANT DelayLoopTime10_EXT = IIF(EXT_CLK_FREQ, DelayLoopTime10(EXT_CLK_FREQ), 1)  ;KLUDGE: avoid arith error if no ext clock; line too long
	CONSTANT SetupTime10_INT = SetupTime10(INT_OSC_FREQ)  ;line too long
	CONSTANT SetupTime10_EXT = IIF(EXT_CLK_FREQ, SetupTime10(EXT_CLK_FREQ), 0)  ;line too long

	CONSTANT PWTOTAL5_INT = rdiv(5*10 - SetupTime10_INT, DelayLoopTime10_INT)  ;#delay loop iterations for 5 usec (after overhead)  ;line too long
	CONSTANT PWTOTAL5_EXT = rdiv(5*10 - SetupTime10_EXT, DelayLoopTime10_EXT)  ;#delay loop iterations for 5 usec (after overhead)  ;line too long
	CONSTANT PWTOTAL40_INT = rdiv(40*10 - SetupTime10_INT, DelayLoopTime10_INT)  ;#delay loop iterations for 40 usec (after overhead)  ;line too long
	CONSTANT PWTOTAL40_EXT = rdiv(40*10 - SetupTime10_EXT, DelayLoopTime10_EXT)  ;#delay loop iterations for 40 usec (after overhead)  ;line too long
	CONSTANT PWDELTA3_INT = rdiv(3*10, DelayLoopTime10_INT)  ;#delay loop iterations for 3 usec; line too long
	CONSTANT PWDELTA3_EXT = rdiv(3*10, DelayLoopTime10_EXT)  ;#delay loop iterations for 3 usec; line too long
	ERRIF2 !PWDELTA3_INT || !PWDELTA3_EXT, [ERROR] Pulse width change is 0: loop will never end. int delta = #v(PWDELTA3_INT), ext delta = #v(PWDELTA3_EXT)


;NOTE: if there is flicker with pulse width > 32.6 or 39 usec, SSR resistor is probably not the correct value (too high for reliable opto turn-on).
;This test starts at 40 usec to provide a little extra range of testing.
;There's no point in testing > 4x speed (down to 8 usec); PIC won't run that fast (although the optos might).
 jumpdest pwidth_cmd:, INCHAR|,,,
	GETNUM8 TestChannel, TRUE
	if8 INCHAR, NE, _LITERAL(0), neargoto invalid_cmd_not_eof,,
	ifbit CfgByte, CFGBIT_RGB_PALETTE, TRUE, loadw_ _LITERAL(RGB_CHANNELS), loadw_ _LITERAL(MONO_CHANNELS), BOTHOK_HINT
	ifbit BITVAR(GotNum), FALSE, storew_ TestChannel,,  ;default to last channel
	if8 TestChannel, GT, WREG, neargoto invalid_cmd_bad_value,,  ;channel# out of range (1-based)
	if8 TestChannel, EQ, _LITERAL(0), neargoto invalid_cmd_bad_value,,
	ChannelToRowBit TestChannel&, TestChannel  ;row index bit to set
	ifbit USING_INTOSC(TRUE), loadw_ _LITERAL(PWTOTAL40_INT), loadw_ _LITERAL(PWTOTAL40_EXT), BOTHOK_HINT  ;start delay loop at approx 40 usec
	storew_ PulseWidth
 jumpdest pwloop:,,,, FALLTHRU
	setbit BITVAR(ShowPulse), TRUE
	mov8 _LITERAL(PWIDTH_MAXBRIGHT), DimTest  ;full fade and ramp; 1 step each half-cycle ~ 2.6 sec at 50 Hz or 2.1 sec at 60 Hz
	setbit BITVAR(Pulse_UpDown), FALSE
 jumpdest dimloop:,,,, FALLTHRU  ;BITVADRS(Pulse_UpDown)
	FBSELECT FALSE, _LITERAL(LOW(FB1_END - FB1_START))  ;set FSR/IRP to start of inactive frame buffer; assumes no pending framebuffer swap
	mov8 PulseWidth, PulseDelayCounter
	push8 DimTest  ;brightness level
	push8 TestChannel  ;_LITERAL(RCbit(7))  ;select test row
	not8 TestChannel, STACK
	mov8 _LITERAL(0), INDF  ;end-of-list marker
	setbit BITVAR(FBSwapRequest), TRUE
	whilebit BITVAR(FBSwapRequest), TRUE, YIELD FALSE,  ;wait for start of AC half-cycle (when new frame buffer is activated)
	while8 DimTest, LT, ThisBrightness, YIELD FALSE,  ;wait until target timeslot; uses regular timing handlers for AC phase dimming (ZC + timeslot)
	mov8 TMR1L, pulse_time
;each step in PulseDelay loop is 3 instr; at 18 MHz ~= .65 usec, at 8 MHz = 1.5 usec; add this to YIELD return latency (~4 instr) + chipiplex (~6 instr) to determine actual timeslot width used here
;TODO: use a scope to verify the timing here; seems to work correctly.
	while8 PulseDelayCounter, NE, _LITERAL(0), dec8 PulseDelayCounter,  ;postpone next SSR pin change; precise override of timing handlers (NON-cooperative)
	Chipiplex -1  ;turn off all SSRs after precise delay (pulse on-time)
	sub8 TMR1L, pulse_time, pulse_time,  ;excludes YIELD latency, which is about the same as the arithmetic overhead on Timer1, so just use value as-is; auto-corrects for wrap
	ifbit BITVAR(ShowPulse), FALSE, LEAVE_PLACEHOLDER1 neargoto dont_show_width,,  ;BIGSTMT_HINT  ;already displayed it for this loop iteration; leave one extra word for bank select
	setbit BITVAR(ShowPulse), FALSE  ;don't show it again during this iteration
	Timer1Ticks2uSec pulse_time, DIV32_QUO
	PUTNUM8 DIV32_QUO, 10
	PUTLITSTRA2 " usec\n\0"  ;should be ~ 8 + n*3 at 8MHz or 3.5 + n*1.3 at 18MHz
	mov8 _LITERAL(PWIDTH_MAXBRIGHT), DimTest  ;KLUDGE: DIV32 overwrites DimTest, so just reinitialize it here and nobody will notice
 jumpdest dont_show_width:,,,, FALLTHRU
	FILL_PLACEHOLDER, neargoto dont_show_width
	ifbit USING_INTOSC(TRUE), loadw_ _LITERAL(2), loadw_ _LITERAL(1), BOTHOK_HINT  ;ramp/fade a little faster when running on int osc
	ifbit BITVAR(Pulse_UpDown), TRUE, ADDWF_wrapper(DimTest, TRUE), SUBWF_wrapper(DimTest, TRUE),  ;next brightness in fade/ramp test
;NOTE: ZC may kick in before ThisBrightness gets all the way to 0 (if timeslot is not quite the correct length), so stop a little short of that
	if8 DimTest, GT, _LITERAL(PWIDTH_MAXBRIGHT), neargoto doneloop,,
	if8 DimTest, GE, _LITERAL(PWIDTH_MINBRIGHT), neargoto dimloop,,
 jumpdest doneloop:,,,, FALLTHRU
	revbit BITVAR(Pulse_UpDown),  ;switch between ramp and fade
	ifbit BITVAR(Pulse_UpDown), TRUE, neargoto dimloop,,  ;now do ramp
	ifbit USING_INTOSC(TRUE), loadw_ _LITERAL(PWDELTA3_INT), loadw_ _LITERAL(PWDELTA3_EXT), BOTHOK_HINT  ;reduce delay loop by approx 3 usec
	sub8 PulseWidth, WREG, PulseWidth, COMPARE_HINT
	ifbit LTZ, neargoto pwidth_done,,
	ifbit USING_INTOSC(TRUE), loadw_ _LITERAL(PWTOTAL5_INT), loadw_ _LITERAL(PWTOTAL5_EXT), BOTHOK_HINT  ;stop delay loop at approx 5 usec
	if8 PulseWidth, GT, WREG, neargoto pwloop,,  ;PIC chipiplexing code can't run this fast anyway, so stop here
 jumpdest pwidth_done:,,,, FALLTHRU
	PUTLITSTRA2 "done\0"
	goto_ put_crlflf


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;help command:
;NOTE: put this one last; this helps with PUTLITSTR addressing optimization (saves ~ 150 words)
 jumpdest help_cmd:, INCHAR|,,,
	call_ getcmdchar
	ifbit GOT_CHAR(TRUE), neargoto invalid_cmd_not_eof,,
;output strings are concatenated to reduce overhead for calling PUTLITSTR numerous times:
;also, lit pool is nearly full, so just put these ones in-line
	PUTLITSTR "Commands:\n?\tthis list\n#[RGB]=# get/set channel/R/G/B\nc[#]\tclear\nd[d]\tdump framebuffer\ne[#]\terror test\n\0", 0xA2, FALSE&
	PUTLITSTR "g[#]\tget/set config\nl#\tlisten\nm#[=#]\tmemory/register\np\tplayback test\nr[r]\treset\ns\tstatus\nw[#]\tpulse width\0", 0xA2, FALSE&
	goto_ put_crlflf


#endif  ;DEBUG_TTY
;******************************************************************************
;start-up code segment:
;******************************************************************************

;	org EE_START
;	REPEAT 4+4, de 0  ;TODO: initialize saved stats in EEPROM


	VARIABLE resume_here = $
	SETPAGE -1, FALSE  ;PCLATH<4:3> are in an unknown state if we fall thru to here from upper code pages
	org RESET_VECTOR  ;absolute address
   	nop  ;required by in-circuit debugger?
;?	clrf	STATUS  ;don't do this; STATUS has info about restart
;NOTE: PCLATH value is unknown if execution wrapped from high address, so it should be explicitly set here
#if SELF_WRITE
	fargoto bootloader  ;TODO
#else  ;SELF_WRITE
	fargoto StartUp  ;startup logic + main line
#endif  ;SELF_WRITE
	ORG resume_here


eof:  ;this must go AFTER all executable code (MUST be a forward reference); used to detect pass 1 vs. 2 for annoying error[116] fixups
;check for correct nesting, etc:
	ERRIF2 PLACEHOLDER_STACK, [ERROR] Mismatched placeholders: final #v(PLACEHOLDER_STACK), high water #v(PLACEHOLDER_HIWATER)  ;this will be 0 unless there is a nesting mismatch
	ERRIF2 MEXPAND_CURNEST, [ERROR] Mismatched macro expanders: final #v(MEXPAND_CURNEST), high water #v(MEXPAND_MAXNEST)  ;this will be 0 unless there is a nesting mismatch
	WARNIF2 !LITPAGEOF(eof), [INFO] No need to save/restore PCLATH, code is all on same page max #v(eof) page #v(LIT_PAGE_SIZE)
;show some macro performance stats:
	WARNIF2 TRUE, [INFO] RAM usage: banked #v(NEXT_RAM#v(BOOL2INT(FALSE))) of max #v(RAM_END#v(BOOL2INT(FALSE))), non-banked #v(NEXT_RAM#v(BOOL2INT(TRUE))) of max #v(RAM_END#v(BOOL2INT(TRUE)))
	WARNIF2 TRUE, [INFO] Reallocated usage: next #v(NEXT_RAM#v(BOOL2INT(MAYBE))), max #v(RAM_END#v(BOOL2INT(MAYBE)))
	WARNIF !PLACEHOLDER_STACK, [INFO] Max placeholder nesting: #v(PLACEHOLDER_HIWATER)
	WARNIF !MEXPAND_CURNEST, [INFO] Max macro expander nesting: #v(MEXPAND_MAXNEST)
	WARNIF2 TRUE, [INFO] Bank select instructions used: #v(NUMBANKSEL), avoided: #v(NONBANKSEL)
	WARNIF2 TRUE, [INFO] Page select instructions used: #v(NUMPAGESEL), avoided: #v(NONPAGESEL)
	WARNIF2 TRUE, [INFO] Case stmt blocks: #v(CASE_COUNT/0x100), instructions saved: #v(CASE_COUNT%0x100)
	WARNIF2 TRUE, [INFO] Ugly fixups pass1: #v(PASS1_FIXUPS/0x10000):#v(PASS1_FIXUPS%0x10000), pass2: #v(PASS2_FIXUPS/0x10000):#v(PASS2_FIXUPS%0x10000)
	WARNIF TRUE, [INFO] Literal string space #v(LITSTR_COUNT):#v(LITSTR_LEN)
	WARNIF2 TRUE, [INFO] Bit variables: banked #v(BITDCL_COUNT_#v(BOOL2INT(TRUE))), non-banked #v(BITDCL_COUNT_#v(FALSE))
	WARNIF2 TRUE, [INFO] SSR control pulse width at 50 Hz AC: #v(rdiv(10*SSR_RATE, AC50HZ))/10 usec, at 60 Hz AC: #v(rdiv(10*SSR_RATE, AC60HZ))/10 usec  ;double-check timing arithmetic; should be 32.6 usec U.S., 39 usec U.K.
	WARNIF2 TRUE, [INFO] #instr with ext clock mhz(EXT_CLK_FREQ) at 50 Hz AC: #v(InstrPerSec(EXT_CLK_FREQ)/AC50HZ*SSR_RATE/_1SEC), at 60 Hz AC: #v(InstrPerSec(EXT_CLK_FREQ)/AC60HZ*SSR_RATE/_1SEC)  ;processing bandwidth
	WARNIF2 TRUE, [INFO] #instr with int osc mhz(INT_OSC_FREQ) at 50 Hz AC: #v(InstrPerSec(INT_OSC_FREQ)/AC50HZ*SSR_RATE/_1SEC), at 60 Hz AC: #v(InstrPerSec(INT_OSC_FREQ)/AC60HZ*SSR_RATE/_1SEC)  ;processing bandwidth

	END